#!/bin/sh 

EXEC_NAME=klayout
OTHER_BIN=""

CURR_DIR=`pwd`
PLATFORM="linux-32-gcc-release"
SOURCE=$CURR_DIR/src
RUN_MAKE=1
GCC_MINGW=0
WIN_CYGWIN=0

case `uname` in
MINGW32*)
  PLATFORM="win-32-mingw-release"
  ;;
CYGWIN*)
  PLATFORM="win-32-mingw-release"
  # On cygwin, provide a windows path in SOURCE 
  if ( cygpath -m $CURR_DIR >/dev/null 2>/dev/null ); then
    WIN_CYGWIN=1
    SOURCE=`cygpath -m $CURR_DIR`/src 
    if ( gcc -mno-cygwin -MM -MG -x c++ $SOURCE/tlLog.cc -I$SOURCE >/dev/null 2>/dev/null ); then
      GCC_MINGW=1
    else
      echo "*** ERROR: detected cygwin build environment but with a gcc that is a native Windows MinGW compiler"
      exit 1
    fi
  else
    echo "*** ERROR: cannot run cygpath"
    exit 1
  fi
  ;;
*)
  # Assume Linux for the remaining variants
  case `arch` in
  x86_64)
    PLATFORM="linux-64-gcc-release"
    ;;
  *)
    PLATFORM="linux-32-gcc-release"
    ;;
  esac
  ;;
esac

# Initialize the Qt paths
if [ "x$QTDIR" != "x" ]; then
  qtdir_wo_backslash=`echo "$QTDIR" | sed 's/\\\\/\\//g'`
  QTLIB="$qtdir_wo_backslash/lib"
  QTINCLUDE="$qtdir_wo_backslash/include"
  QTBIN="$qtdir_wo_backslash/bin"
else
  QTLIB="/usr/lib"
  QTINCLUDE="/usr/include/qt4"
  QTBIN="/usr/bin"
fi

RUBYINCLUDE=""
RUBYINCLUDE2="not-used"
RUBYLIBFILE=""

# Heuristics for 1.9
for ruby_version in 1.9 1.9.0 1.9.1 1.9.2 1.9.3 1.9.4 1.9.5 1.9.6 1.9.7 1.9.8 1.9.9 1.9.10; do
  for rubylib in "/usr/lib/libruby${ruby_version}.so" "/usr/lib/libruby${ruby_version}.so.${ruby_version}"; do
    if [ -e "$rubylib" ]; then
      for d in /usr/include/ruby-*; do
        if [ -e "$d/ruby.h" ]; then
          for dd in $d/*; do
            if [ -e "$dd/ruby/config.h" ]; then
              RUBYLIBFILE="$rubylib"
              RUBYINCLUDE="$d"
              RUBYINCLUDE2="$dd"
            fi
          done
        fi
      done
    fi
  done

done

# Try to localize the ruby installation
for ruby_version in 1.8 1.8.6 1.8.7 1.8.8 1.8.9 1.8.10; do
  for rubylib in "/usr/lib/libruby${ruby_version}.so" "/usr/lib/libruby${ruby_version}.so.${ruby_version}"; do
    if [ -e "$rubylib" ]; then
      for d in /usr/lib/ruby/${ruby_version}/*; do
        if [ -e "$d/ruby.h" ]; then
          RUBYLIBFILE="$rubylib"
          RUBYINCLUDE="$d"
          RUBYINCLUDE2="not-used"
        fi
      done
    fi
  done
done

BUILD=""
BIN=""

MAKE_OPT=""

while [ "$*" != "" ]; do
  a=$1
  shift
  
  case $a in 
  -qt) 
    QTLIB="$1/lib"
    QTBIN="$1/bin"
    QTINCLUDE="$1/include"
    shift
    ;;
  -qtlib) 
    QTLIB="$1"
    shift
    ;;
  -qtbin) 
    QTBIN="$1"
    shift
    ;;
  -qtinc) 
    QTINCLUDE="$1"
    shift
    ;;
  -platform) 
    PLATFORM="$1"
    shift
    ;;
  -rbinc)
    RUBYINCLUDE="$1"
    shift
    ;;
  -rbinc2)
    RUBYINCLUDE2="$1"
    shift
    ;;
  -rblib)
    RUBYLIBFILE="$1"
    shift
    ;;
  -build)
    BUILD="$1"
    shift
    ;;
  -bin)
    BIN="$1"
    shift
    ;;
  -dry-run)
    RUN_MAKE=0
    ;;
  -option) 
    MAKE_OPT="$MAKE_OPT $1"
    shift
    ;;
  -help)
    echo "usage: build [options]"
    echo ""
    echo "options:"
    echo "  -qt <dir>         Set Qt installation directory if Qt is installed in a"
    echo "                    common directory (-prefix path, default is '/usr/lib/qt4')."
    echo "                    This will set -qtbin to <dir>/bin, -qtlib to <dir>/lib"
    echo "                    and -qtinc to <dir>/include."
    echo "  -qtbin <dir>      Set Qt tools directory explicitly"
    echo "  -qtinc <dir>      Set Qt header directory explicitly"
    echo "  -qtlib <dir>      Set Qt library directory explicitly"
    echo "  -rblib <file>     Location of the .so/.dll to link for Ruby support"
    echo "  -rbinc <dir>      Location of the Ruby headers (in particular 'ruby.h')"
    echo "                    -rbinc and -rblib must be set to enable Ruby support."
    echo "  -platform <key>   Select a platform"
    echo "  -build <path>     Directory where to do the build"
    echo "  -bin <path>       Directory where to install the binary"
    echo "  -option <option>  'make' options (i.e. -j2)"
    echo ""
    echo "  -dry-run          Don't build"
    echo ""
    echo "Available platforms are:"
    for pf in `ls -1 config/*.conf.*`; do
      echo $pf | sed "s/config\/Makefile\.conf\./  /"
    done
    exit 0
    ;;
  *)
    echo "*** ERROR: unknown option $a (use -help to show options)"
    exit 1
    ;;
  esac

done

echo "Qt installation directory is $QTLIB (lib), $QTBIN (tools) and $QTINCLUDE (includes)"
echo "Ruby installation directory is $RUBYLIBFILE (lib), $RUBYINCLUDE (includes) and $RUBYINCLUDE2 (config for 1.9.x)"
echo "Platform is $PLATFORM"

gcc_path=`type gcc 2>/dev/null`
if [ "x$gcc_path" = "x" ]; then
  echo "*** ERROR: unable to locate gcc in path"
  exit 1
fi

if [ ! -r config/Makefile.conf.$PLATFORM ]; then
  echo "*** ERROR: unknown platform $PLATFORM"
  exit 1
fi

case $PLATFORM in
mac-*)
  echo "*** WARNING: no check done dof Qt-headers. -qtinc ignored"
  QTINCLUDE=dontcare
  ;;
*)
  if [ ! -r $QTINCLUDE/QtCore/Qt ]; then
    echo "*** ERROR: no Qt headers in $QTINCLUDE"
    exit 1
  fi
  ;;
esac

if [ ! -r $QTBIN/moc ]; then
  echo "*** ERROR: unable to locate Qt MOC in $QTBIN"
  exit 1
fi
if [ ! -r $QTBIN/uic ]; then
  echo "*** ERROR: unable to locate Qt UIC in $QTBIN"
  exit 1
fi

if [ "x$RUBYINCLUDE" != "x" ]; then
  HAVE_RUBY=1
  if [ "x$RUBYLIBFILE" = "x" ]; then
    echo "*** ERROR: -rbinc and -rblib must both be specified"
    exit 1
  fi
  if [ ! -f "$RUBYINCLUDE/ruby.h" ]; then
    echo "*** ERROR: wrong -rbinc path: $RUBYINCLUDE/ruby.h does not exist or is not a file"
    exit 1
  fi
  if [ ! -f "$RUBYLIBFILE" ]; then
    echo "*** ERROR: wrong -rblib path: $RUBYLIBFILE does not exist or not a file"
    exit 1
  fi
else
  HAVE_RUBY=0
fi

if [ "x$BUILD" = "x" ]; then
  BUILD=$CURR_DIR/build.$PLATFORM
fi
if [ "x$BIN" = "x" ]; then
  BIN=$CURR_DIR/bin.$PLATFORM
else
  mkdir -p $BIN
  cd $BIN
  BIN=`pwd`
  cd $CURR_DIR
fi

if ( gmake -v >/dev/null 2>/dev/null ); then
  MAKE_PRG=gmake
else
  MAKE_PRG=make
fi

mkdir -p $BUILD
echo "WIN_CYGWIN=$WIN_CYGWIN"  >$BUILD/Makefile
echo "EXEC_NAME=$EXEC_NAME"   >>$BUILD/Makefile
echo "OTHER_BIN=$OTHER_BIN"   >>$BUILD/Makefile
echo "INSTALL_BINDIR=$BIN"    >>$BUILD/Makefile
cat   src/Makefile.top        >>$BUILD/Makefile
cp -f config/Makefile.conf.$PLATFORM $BUILD/Makefile.conf

mkdir -p $BUILD/main
echo "TOP_SOURCE=$SOURCE" >$BUILD/main/Makefile
echo "SOURCE=$SOURCE" >>$BUILD/main/Makefile
echo 'default: all' >>$BUILD/main/Makefile
echo 'include ../Makefile.env' >>$BUILD/main/Makefile
echo 'include ../Makefile.conf' >>$BUILD/main/Makefile
echo 'include $(SOURCE)/Makefile.body' >>$BUILD/main/Makefile

PLUGINS=""
if [ -d src ] && [ -d src/plugins ]; then
  cd src/plugins
  PLUGINS=`echo *`
  cd ../..
  for plugin in $PLUGINS; do
    mkdir -p $BUILD/$plugin
    echo "TOP_SOURCE=$SOURCE" >$BUILD/$plugin/Makefile
    echo "SOURCE=$SOURCE/plugins/$plugin" >>$BUILD/$plugin/Makefile
    echo 'default: all' >>$BUILD/$plugin/Makefile
    echo 'include ../Makefile.env' >>$BUILD/$plugin/Makefile
    echo 'include ../Makefile.conf' >>$BUILD/$plugin/Makefile
    echo 'include $(SOURCE)/Makefile.body' >>$BUILD/$plugin/Makefile
  done
fi

echo "QTBIN=$QTBIN" >$BUILD/Makefile.env
echo "QTINCLUDE=$QTINCLUDE" >>$BUILD/Makefile.env
echo "QTLIB=$QTLIB" >>$BUILD/Makefile.env
echo "PLUGINS=$PLUGINS" >>$BUILD/Makefile.env
echo "RUBYLIBFILE=$RUBYLIBFILE" >>$BUILD/Makefile.env
echo "RUBYINCLUDE=$RUBYINCLUDE" >>$BUILD/Makefile.env
echo "RUBYINCLUDE2=$RUBYINCLUDE2" >>$BUILD/Makefile.env
echo "HAVE_RUBY=$HAVE_RUBY" >>$BUILD/Makefile.env

echo "Building plugins: $PLUGINS"
echo ""

if [ $RUN_MAKE = 0 ]; then
  exit 0
fi

# -- Running build  --

echo "Running build .."
echo "$MAKE_PRG"
cd $BUILD
$MAKE_PRG $MAKE_OPT all || exit 1
echo ""

cd $CURR_DIR

# -- Installing binaries  --

echo "Installing binaries .."
cd $BUILD
$MAKE_PRG install
cd $CURR_DIR
echo "Build done."

if [ $WIN_CYGWIN = 0 ]; then
  echo ""
  echo "Final binary is ready in $BIN/$EXEC_NAME."
  echo "To run it, you may need to add the following to your LD_LIBRARY_PATH:"
  echo "$QTLIB"
  
  LD_LIBRARY_PATH=$QTLIB:$LD_LIBRARY_PATH
  $BIN/$EXEC_NAME -v
  $BIN/$EXEC_NAME -h
fi

exit 0

