
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "dbCIFReader.h"
#include "dbStream.h"
#include "dbObjectWithProperties.h"
#include "dbArray.h"
#include "dbStatic.h"

#include "tlException.h"
#include "tlString.h"
#include "tlClassRegistry.h"

namespace db
{

// ---------------------------------------------------------------
//  CIFReader


CIFReader::CIFReader (tl::InputStreamBase &s)
  : m_stream (s),
    m_read_texts (true), m_read_properties (true), m_create_layers (true),
    m_progress (tl::translate ("Reading CIF file"), 0, 10000)
{
  m_progress.set_format (tl::translate ("%.0f MB"));
  m_progress.set_res (1.0);
  m_progress.set_unit (1024 * 1024);
}

CIFReader::~CIFReader ()
{
  //  .. nothing yet ..
}

const LayerMap &
CIFReader::read (db::Layout &layout, const db::LoadLayoutOptions &options) throw (tl::Exception)
{
  m_dbu = 0.001;
  m_wire_mode = 0;

  const db::CIFReaderOptions *specific_options = dynamic_cast<const db::CIFReaderOptions *> (options.get_specific_options ("CIF"));
  if (specific_options) {
    m_wire_mode = specific_options->wire_mode;
    m_dbu = specific_options->dbu;
  }

  m_layer_map = options.layer_map ();
  m_read_texts = options.enable_text_objects ();
  m_read_properties = options.enable_properties ();
  m_create_layers = options.create_other_layers ();
  layout.start_changes ();
  do_read (layout);
  layout.end_changes ();
  return m_layer_map;
}

const LayerMap &
CIFReader::read (db::Layout &layout) throw (tl::Exception)
{
  return read (layout, db::LoadLayoutOptions ());
}

void 
CIFReader::error (const std::string &msg) throw (tl::Exception)
{
  throw CIFReaderException (msg, m_stream.line_number (), m_cellname.c_str ());
}

void 
CIFReader::warn (const std::string &msg) 
{
  // TODO: compress
  tl::warn << msg 
           << tl::translate (" (line=") << m_stream.line_number ()
           << tl::translate (", cell=") << m_cellname
           << ")";
}

const char *
CIFReader::fetch_command ()
{

  m_cmd_buffer.clear ();

  m_stream.skip ();
  if (m_stream.at_end ()) {
    error (tl::translate ("Unexpected end of file"));
  }

  while (! m_stream.at_end () && m_stream.peek_char () != ';') {
    
    // skip over comments
    if (m_stream.peek_char () == '(') {

      m_stream.get_char ();

      char c;
      int bl = 0;
      while (! m_stream.at_end () && ((c = m_stream.get_char ()) != ')' || bl > 0)) {
        // check for nested comments (bl is the nesting level)
        if (c == '(') {
          ++bl;
        } else if (c == ')') {
          --bl;
        }
      }

    } else {
      m_cmd_buffer += m_stream.get_char ();
    }

  }

  if (m_stream.peek_char () == ';') {
    m_stream.get_char ();
  }

  return m_cmd_buffer.c_str ();
}

bool
CIFReader::read_cell (db::Layout &layout, db::Cell &cell, double sf, int level)
{
  if (fabs (sf - floor (sf + 0.5)) > 1e-6) {
    warn ("Scaling factor is not an integer - snapping errors may occur in cell '" + m_cellname + "'");
  }

  int nx = 0, ny = 0, dx = 0, dy = 0;
  int layer = -2; // no layer defined yet.
  int layer_num = 1;
  int path_mode = -1;
  size_t insts = 0;
  size_t shapes = 0;
  size_t layer_specs = 0;
  std::vector <db::Point> poly_pts;

  while (true)
  {
    tl::Extractor ex (fetch_command ());

    if (ex.at_end ()) {
      //  empty command
    } else if (ex.test ("DS")) {

      //  DS command 
      unsigned int n = 0, denom = 1, divider = 1;
      ex.read (n);
      if (ex.try_read (denom)) {
        ex.test("/"); // accept "/" as the divider
        if (ex.try_read (divider) && ! ex.at_end ()) {
          warn ("Unexpected text after 'DS' command ignored");
        }
      }

      std::string outer_cell = "C" + tl::to_string (n);
      std::swap (m_cellname, outer_cell);

      std::map <unsigned int, db::cell_index_type>::const_iterator c = m_cells_by_id.find (n);
      db::cell_index_type ci;
      if (c == m_cells_by_id.end ()) {
        ci = layout.add_cell (m_cellname.c_str ());
        m_cells_by_id.insert (std::make_pair (n, ci));
      } else {
        ci = c->second;
      } 

      db::Cell &cell = layout.cell (ci);

      read_cell (layout, cell, sf * double (denom) / double (divider), level + 1);

      std::swap (m_cellname, outer_cell);

    } else if (ex.test ("DF")) {

      if (! ex.at_end ()) {
        warn ("Unexpected text after 'DF' command ignored");
      }

      break;

    } else if (ex.test ("E")) {

      if (! ex.at_end ()) {
        warn ("Unexpected text after 'E' command ignored");
      }
      if (level > 0) {
        warn ("'E' command must be outside a cell specification");
      }

      break;

    } else if (ex.test ("C")) {

      ++insts;

      unsigned int n = 0;
      ex.read (n);
      std::map <unsigned int, db::cell_index_type>::const_iterator c = m_cells_by_id.find (n);
      if (c == m_cells_by_id.end ()) {
        std::string cn = "C" + tl::to_string (n);
        c = m_cells_by_id.insert (std::make_pair (n, layout.add_cell (cn.c_str ()))).first;
      } 

      db::CplxTrans trans;

      while (! ex.at_end ()) {

        if (ex.test ("MX")) {
          trans = db::CplxTrans (db::FTrans::m90) * trans;
        } else if (ex.test ("MY")) {
          trans = db::CplxTrans (db::FTrans::m0) * trans;
        } else if (ex.test ("T")) {

          int x = 0, y = 0;
          ex.read (x);
          ex.test (",");
          ex.read (y);
          trans = db::CplxTrans (db::Point (x * sf, y * sf)) * trans;

        } else if (ex.test ("R")) {

          int x = 1, y = 1;
          ex.read (x);
          ex.test (",");
          ex.read (y);

          if (y != 0 || x != 0) {
            double a = atan2 (double (y), double (x)) * 180.0 / M_PI;
            trans = db::CplxTrans (1.0, a, false, db::Point ()) * trans;
          }

        }

      }

      if (nx > 0 || ny > 0) {
        if (trans.is_ortho ()) {
          db::Trans simple_trans (trans);
          cell.insert (db::CellInstArray (db::CellInst (c->second), db::Trans (trans), db::Point (dx * sf, 0), db::Point (0, dy * sf), std::max (1, nx), std::max (1, ny)));
        } else {
          cell.insert (db::CellInstArray (db::CellInst (c->second), trans, db::Point (dx * sf, 0), db::Point (0, dy * sf), std::max (1, nx), std::max (1, ny)));
        }
      } else {
        if (trans.is_ortho ()) {
          db::Trans simple_trans (trans);
          cell.insert (db::CellInstArray (db::CellInst (c->second), db::Trans (trans)));
        } else {
          cell.insert (db::CellInstArray (db::CellInst (c->second), trans));
        }
      }

      if (! ex.at_end ()) {
        warn ("Unexpected text after 'C' command ignored");
      }

      nx = ny = 0;
      dx = dy = 0;

    } else if (ex.test ("L")) {

      ++layer_specs;

      std::string name;
      ex.read_word_or_quoted (name);

      std::pair<bool, unsigned int> ll = m_layer_map.logical (name);
      if (ll.first) {

        //  create the layer if it is not part of the layout yet.
        if (! layout.is_valid_layer (ll.second)) {
          layout.insert_layer (ll.second, m_layer_map.mapping (ll.second));
        }

        layer = int (ll.second);

      } else if (! m_create_layers) {

        layer = -1; // ignore geometric objects on this layer

      } else {

        //  and create the layer:
        db::LayerProperties lp;
        lp.name = name;

        unsigned int l = 0, d = 0;

        //  try to derive layer/datatype from the name ..
        //  Allowed notations are: <l>, <l>.<d>, <l>/<d>, L<l>, L<l>D<d>.
        tl::Extractor ex (name.c_str ());
        if ((ex.try_read (l) && (ex.at_end () || ((ex.test (".") || ex.test ("/")) && ex.try_read (d) && ex.at_end ()))) ||
            (ex.test ("L") && ex.try_read (l) && (ex.at_end () || (ex.test ("D") && ex.try_read (d) && ex.at_end ())))) {

          lp.layer = l;
          lp.datatype = d;

        } else {

          //  look for an unmapped layer number with datatype 0 for sequential numbering
          while (m_layer_map.logical (db::LDPair (layer_num, 0)).first) {
            ++layer_num;
          }

          lp.layer = layer_num;
          lp.datatype = 0;

          ++layer_num;

        }

        unsigned int ll = m_layer_map.next_index ();
        layout.insert_layer (ll, lp);
        m_layer_map.map (name, ll, lp);

        layer = int (ll);

      }

      if (! ex.at_end ()) {
        warn ("Unexpected text after 'L' command ignored");
      }

    } else if (ex.test ("B")) {

      ++shapes;

      if (layer < 0) {
        if (layer < -1) {
          warn ("'B' command ignored since no layer was selected");
        }
      } else {

        unsigned int w = 0, h = 0;
        int x = 0, y = 0;

        ex.read (w);
        ex.test (",");
        ex.read (h);
        ex.test (",");
        ex.read (x);
        ex.test (",");
        ex.read (y);
        ex.test (",");

        int rx = 0, ry = 0;

        if (ex.try_read (rx)) {
          ex.test (",");
          ex.read (ry);
        }

        if (rx >= 0 && ry == 0) {

          cell.shapes ((unsigned int) layer).insert (db::Box::from_double (db::DBox (sf * (x - 0.5 * w), sf * (y - 0.5 * h), sf * (x + 0.5 * w), sf * (y + 0.5 * h))));

        } else {

          double n = 1.0 / sqrt (double (rx) * double (rx) + double (ry) * double (ry));

          double xw = sf * w * 0.5 * rx * n, yw = sf * w * 0.5 * ry * n;
          double xh = -sf * h * 0.5 * ry * n, yh = sf * h * 0.5 * rx * n;

          db::Point points [4];
          points [0] = db::Point::from_double (db::DPoint (x - xw - xh, y - yw - yh));
          points [1] = db::Point::from_double (db::DPoint (x - xw + xh, y - yw + yh));
          points [2] = db::Point::from_double (db::DPoint (x + xw + xh, y + yw + yh));
          points [3] = db::Point::from_double (db::DPoint (x + xw - xh, y + yw - yh));

          db::Polygon p;
          p.assign_hull (points, points + 4);
          cell.shapes ((unsigned int) layer).insert (p);

        }

      }

      if (! ex.at_end ()) {
        warn ("Unexpected text after 'B' command ignored");
      }

    } else if (ex.test ("P")) {

      ++shapes;

      if (layer < 0) {
        if (layer < -1) {
          warn ("'P' command ignored since no layer was selected");
        }
      } else {

        poly_pts.clear ();

        while (! ex.at_end ()) {

          int rx = 0, ry = 0;
          ex.test (",");
          ex.read (rx);
          ex.test (",");
          ex.read (ry);

          poly_pts.push_back (db::Point::from_double (db::DPoint (sf * rx, sf * ry)));

        }

        db::Polygon p;
        p.assign_hull (poly_pts.begin (), poly_pts.end ());
        cell.shapes ((unsigned int) layer).insert (p);

      }

    } else if (ex.test ("R")) {

      ++shapes;

      if (layer < 0) {
        if (layer < -1) {
          warn ("'R' command ignored since no layer was selected");
        }
      } else {

        int w = 1;
        ex.read (w);

        poly_pts.clear ();

        int rx = 0, ry = 0;
        ex.test (",");
        ex.read (rx);
        ex.test (",");
        ex.read (ry);

        poly_pts.push_back (db::Point::from_double (db::DPoint (sf * rx, sf * ry)));

        db::Path p (poly_pts.begin (), poly_pts.end (), db::coord_traits <db::Coord>::rounded (sf * w), db::coord_traits <db::Coord>::rounded (sf * w / 2), db::coord_traits <db::Coord>::rounded (sf * w / 2), true);
        cell.shapes ((unsigned int) layer).insert (p);

      }

    } else if (ex.test ("W")) {

      ++shapes;

      if (layer < 0) {
        if (layer < -1) {
          warn ("'W' command ignored since no layer was selected");
        }
      } else {

        int w = 1;
        ex.read (w);

        poly_pts.clear ();

        while (! ex.at_end ()) {

          int rx = 0, ry = 0;
          ex.test (",");
          ex.read (rx);
          ex.test (",");
          ex.read (ry);

          poly_pts.push_back (db::Point::from_double (db::DPoint (sf * rx, sf * ry)));

        }

        if (path_mode == 0 || (path_mode < 0 && m_wire_mode == 1)) {
          //  Flush-ended paths 
          db::Path p (poly_pts.begin (), poly_pts.end (), db::coord_traits <db::Coord>::rounded (sf * w), 0, 0, false);
          cell.shapes ((unsigned int) layer).insert (p);
        } else if (path_mode == 1 || (path_mode < 0 && m_wire_mode == 2)) {
          //  Round-ended paths
          db::Path p (poly_pts.begin (), poly_pts.end (), db::coord_traits <db::Coord>::rounded (sf * w), db::coord_traits <db::Coord>::rounded (sf * w / 2), db::coord_traits <db::Coord>::rounded (sf * w / 2), true);
          cell.shapes ((unsigned int) layer).insert (p);
        } else {
          //  Square-ended paths
          db::Path p (poly_pts.begin (), poly_pts.end (), db::coord_traits <db::Coord>::rounded (sf * w), db::coord_traits <db::Coord>::rounded (sf * w / 2), db::coord_traits <db::Coord>::rounded (sf * w / 2), false);
          cell.shapes ((unsigned int) layer).insert (p);
        }

      }

    } else if (ex.test ("0")) {
      // ignore extensions 0 ..
    } else if (ex.test ("1")) {
      // ignore extensions 1 ..
    } else if (ex.test ("2")) {
      // ignore extensions 2 ..
    } else if (ex.test ("3")) {
      // ignore extensions 3 ..
    } else if (ex.test ("4")) {
      // ignore extensions 4 ..
    } else if (ex.test ("5")) {
      // ignore extensions 5 ..
    } else if (ex.test ("6")) {
      // ignore extensions 6 ..
    } else if (ex.test ("7")) {
      // ignore extensions 7 ..
    } else if (ex.test ("8")) {
      // ignore extensions 8 ..
    } else if (ex.test ("93")) {

        nx = ny = 0;
        dx = dy = 0;

        ex.test (",");
        ex.read (nx);
        ex.test (",");
        ex.read (dx);

        ex.test (",");
        ex.read (ny);
        ex.test (",");
        ex.read (dy);

    } else if (ex.test ("94")) {

      // label at location 
      ++shapes;

      if (layer < 0) {
        if (layer < -1) {
          warn ("'94' command ignored since no layer was selected");
        }
      } else {

        std::string text;
        ex.skip ();
        if (*ex == '\'' || *ex == '"') {
          ex.read_quoted (text);
        } else {
          ex.read (text);
        }

        int rx = 0, ry = 0;
        ex.test (",");
        ex.read (rx);
        ex.test (",");
        ex.read (ry);

        double h = 0.0;
        ex.try_read (h);

        db::Text t (text.c_str (), db::Trans (db::Point::from_double (db::DPoint (sf * rx, sf * ry))), db::coord_traits <db::Coord>::rounded (h / m_dbu));
        cell.shapes ((unsigned int) layer).insert (t);

      }

    } else if (ex.test ("95")) {

      // label in box 
      ++shapes;

      if (layer < 0) {
        if (layer < -1) {
          warn ("'95' command ignored since no layer was selected");
        }
      } else {

        std::string text;
        ex.read (text);

        int w = 0, h = 0;
        ex.test (",");
        ex.read (w);
        ex.test (",");
        ex.read (h);

        int rx = 0, ry = 0;
        ex.test (",");
        ex.read (rx);
        ex.test (",");
        ex.read (ry);

        db::Text t (text.c_str (), db::Trans (db::Point::from_double (db::DPoint (sf * rx, sf * ry))));
        cell.shapes ((unsigned int) layer).insert (t);

      }

    } else if (ex.test ("98")) {

      // Path type (0: flush, 1: round, 2: square)
      ex.read (path_mode);

    } else if (ex.test ("9 ")) {

      ex.read_word_or_quoted (m_cellname);
      m_cellname = layout.uniquify_cell_name (m_cellname.c_str ());
      layout.rename_cell (cell.cell_index (), m_cellname.c_str ());

    } 
    
  }

  //  The cell is considered non-empty if it contains more than one instance, at least one shape or
  //  has at least one "L" command.
  return insts > 1 || shapes > 0 || layer_specs > 0;
}

void 
CIFReader::do_read (db::Layout &layout) throw (tl::Exception)
{
  try {
  
    double sf = 0.01 / m_dbu;
    layout.dbu (m_dbu);

    m_cellname = "{CIF top level}";

    db::Cell &cell = layout.cell (layout.add_cell ());

    if (! read_cell (layout, cell, sf, 0)) {
      // The top cell is empty or contains a single instance: discard it.
      layout.delete_cell (cell.cell_index ());
    } else {
      layout.rename_cell (cell.cell_index (), layout.uniquify_cell_name ("CIF_TOP").c_str ());
    }

    m_cellname = std::string ();

    m_stream.skip ();
    if (! m_stream.at_end ()) {
      warn ("E command is followed by more text");
    }

  } catch (db::CIFReaderException &ex) {
    throw ex;
  } catch (tl::Exception &ex) {
    error (ex.msg ());
  }
}

}

