
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtInstPropertiesPage.h"
#include "edtDialogs.h"
#include "layObjectInstPath.h"
#include "layLayoutView.h"
#include "tlString.h"

#include <string>

namespace edt
{


// -------------------------------------------------------------------------
//  helper function to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

static db::Coord
coord_from_dcoord (double d, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DCoord dc = t.ctrans (d / dbu);
  db::Coord c = db::Coord (floor (dc + 0.5));
  if (fabs (dc - c) > 1e-6) { 
    if (! du) {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit (%g)").c_str (), d, dbu);
    } else {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit").c_str (), d);
    }
  }

  return c;
}

static db::Point
coord_from_dpoint (const db::DPoint &dp, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DPoint dpt = t * (dp * (1.0 / dbu));

  db::Coord cx = coord_from_dcoord (dpt.x (), 1.0, false, db::CplxTrans ());
  db::Coord cy = coord_from_dcoord (dpt.y (), 1.0, false, db::CplxTrans ());

  return db::Point (cx, cy);
}

// -------------------------------------------------------------------------
//  InstPropertiesPage implementation

InstPropertiesPage::InstPropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), mp_service (service), m_enable_cb_callback (true)
{
  m_selection_ptrs.reserve (service->selection ().size ());
  for (edt::Service::obj_iterator s = service->selection ().begin (); s != service->selection ().end (); ++s) {
    m_selection_ptrs.push_back (s);
  }
  m_index = 0;
  m_prop_id = 0;
  mp_service->clear_highlights ();

  setupUi (this);
  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (sel_pb, SIGNAL (clicked ()), this, SLOT (show_cell ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
  connect (dbu_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_cb, SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
}

InstPropertiesPage::~InstPropertiesPage ()
{
  mp_service->restore_highlights ();
}

void
InstPropertiesPage::show_props ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  UserPropertiesForm props_form (this);
  if (props_form.show (mp_service->view (), pos->cv_index (), m_prop_id)) {
    if (! readonly () && m_prop_id != pos->back ().inst_ptr.prop_id ()) {

      const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
      db::Cell &cell = cv->layout ().cell (pos->cell_index ());

      db::Transaction transaction (cv->layout ().manager (), tl::translate ("Change instance properties"));

      db::Instance new_inst = cell.replace_prop_id (pos->back ().inst_ptr, m_prop_id);

      if (new_inst != pos->back ().inst_ptr) {

        //  change selection to new shape
        lay::ObjectInstPath new_sel (*pos);
        new_sel.back ().inst_ptr = new_inst;

        mp_service->select (*pos, lay::Editable::Reset);
        mp_service->select (new_sel, lay::Editable::Add);

        // locate the new selection
        for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
          ;
        tl_assert (pos != mp_service->selection ().end ());
        m_selection_ptrs [m_index] = pos;

        mp_service->selection_to_view ();

      }

      update ();

    }
  }
}

void
InstPropertiesPage::display_mode_changed (bool)
{
  if (! m_enable_cb_callback) {
    return;
  }

  mp_service->view ()->dbu_coordinates (dbu_cb->isChecked ());
  mp_service->view ()->absolute_coordinates (abs_cb->isChecked ());

  update ();
}

void 
InstPropertiesPage::back ()
{
  m_index = m_selection_ptrs.size ();
}

void 
InstPropertiesPage::front ()
{
  m_index = 0;
}

bool 
InstPropertiesPage::at_begin () const
{
  return (m_index == 0);
}

bool 
InstPropertiesPage::at_end () const
{
  return (m_index == m_selection_ptrs.size ());
}

void 
InstPropertiesPage::operator-- ()
{
  --m_index;
}

void 
InstPropertiesPage::operator++ ()
{
  ++m_index;
}

void 
InstPropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

void 
InstPropertiesPage::update ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];
  tl_assert (pos->is_cell_inst ());

  mp_service->highlight (m_index);

  m_enable_cb_callback = false;
  dbu_cb->setChecked (mp_service->view ()->dbu_coordinates ());
  abs_cb->setChecked (mp_service->view ()->absolute_coordinates ());
  m_enable_cb_callback = true;

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
  double dbu = cv->layout ().dbu ();

  std::string info (tl::translate ("Cell "));
  info += cv->layout ().cell_name (pos->cell_index ());
  info_lbl->setText (info.c_str ());

  cell_name_le->setText (cv->layout ().cell_name (pos->back ().inst_ptr.cell_index ()));

  db::Point rowv, columnv;
  unsigned long rows, columns;

  db::CplxTrans gt;
  if (abs_cb->isChecked ()) {
    gt = pos->trans ();
  }

  bool du = dbu_cb->isChecked ();

  db::Trans t (pos->back ().inst_ptr.front ());

  if (pos->back ().inst_ptr.is_regular_array (rowv, columnv, rows, columns)) {

    array_grp->setChecked (true);
    rows_le->setText (tl::to_string (rows).c_str ());
    columns_le->setText (tl::to_string (columns).c_str ());
    row_x_le->setText (coord_to_string ((gt * rowv).x (), dbu, du).c_str ());
    row_y_le->setText (coord_to_string ((gt * rowv).y (), dbu, du).c_str ());
    column_x_le->setText (coord_to_string ((gt * columnv).x (), dbu, du).c_str ());
    column_y_le->setText (coord_to_string ((gt * columnv).y (), dbu, du).c_str ());

    if (! pos->back ().array_inst.at_end ()) {

      //  show the array indices
      long row = pos->back ().array_inst.index_a ();
      long column = pos->back ().array_inst.index_b ();
      inst_lbl->setText (tl::sprintf (tl::translate ("This is instance [%ld,%ld] of array with").c_str (), row, column).c_str ());

    }

  } else {

    array_grp->setChecked (false);
    rows_le->setText ("");
    columns_le->setText ("");
    row_x_le->setText ("");
    row_y_le->setText ("");
    column_x_le->setText ("");
    column_y_le->setText ("");
    inst_lbl->setText ("");

  }

  pos_x_le->setText (coord_to_string ((gt * t.disp ()).x (), dbu, du).c_str ());
  pos_y_le->setText (coord_to_string ((gt * t.disp ()).y (), dbu, du).c_str ());
  angle_le->setText (tl::to_string (pos->back ().inst_ptr.complex_trans ().angle ()).c_str ());
  mirror_cbx->setChecked (t.is_mirror ());
  mag_le->setText (tl::to_string (pos->back ().inst_ptr.complex_trans ().mag ()).c_str ());

  m_prop_id = pos->back ().inst_ptr.prop_id ();
}

void 
InstPropertiesPage::show_cell ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  lay::CellView::unspecific_cell_path_type path (mp_service->view ()->cellview (pos->cv_index ()).combined_unspecific_path ());
  for (lay::ObjectInstPath::iterator p = pos->begin (); p != pos->end (); ++p) {
    path.push_back (p->inst_ptr.cell_index ());
  }

  mp_service->view ()->set_current_cell_path (pos->cv_index (), path);
}

void
InstPropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), *m_selection_ptrs [m_index]);
}

bool 
InstPropertiesPage::readonly ()
{
  return ! mp_service->view ()->is_editable ();
}

void 
InstPropertiesPage::apply ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
  bool du = dbu_cb->isChecked ();
  double dbu = cv->layout ().dbu ();

  std::pair<bool, db::cell_index_type> ci = cv->layout ().cell_by_name (cell_name_le->text ().ascii ());
  if (! ci.first) {
    throw tl::Exception (tl::translate ("Not a valid cell name: %s").c_str (), cell_name_le->text ().ascii ());
  }

  db::Cell &cell = cv->layout ().cell (pos->cell_index ());

  //  detect recursions in the hierarchy
  std::set<db::cell_index_type> called;
  cv->layout ().cell (ci.second).collect_called_cells (called);
  if (ci.second == cell.cell_index () || called.find (cell.cell_index ()) != called.end ()) {
    throw tl::Exception (tl::translate ("Trying to build a recursive hierarchy").c_str ());
  }

  double x = 0.0, y = 0.0;
  tl::from_string (pos_x_le->text ().ascii (), x);
  tl::from_string (pos_y_le->text ().ascii (), y);

  db::CplxTrans t;
  if (abs_cb->isChecked ()) {
    t = pos->trans ().inverted ();
  }

  db::Point disp (coord_from_dpoint (db::DPoint (x, y), dbu, du, t));

  bool mirror = mirror_cbx->isChecked ();
  double angle_deg = 0.0;
  tl::from_string (angle_le->text ().ascii (), angle_deg);

  double mag = 0.0;
  tl::from_string (mag_le->text ().ascii (), mag);

  angle_deg -= (floor (angle_deg / 360.0) + 1.0) * 360.0;
  while (angle_deg < -1e-6) {
    angle_deg += 360.0;
  }

  unsigned int angle = (unsigned int) floor (angle_deg / 90.0 + 1e-6);
  bool is_complex = ((fabs (mag - 1.0) > 1e-6 || fabs (angle_deg - angle * 90.0) > 1e-6));

  db::CellInstArray inst;

  if (array_grp->isChecked ()) {

    double cx = 0.0, cy = 0.0;
    double rx = 0.0, ry = 0.0;
    unsigned long rows = 0, cols = 0;

    tl::from_string (column_x_le->text ().ascii (), cx);
    tl::from_string (column_y_le->text ().ascii (), cy);
    tl::from_string (row_x_le->text ().ascii (), rx);
    tl::from_string (row_y_le->text ().ascii (), ry);
    tl::from_string (rows_le->text ().ascii (), rows);
    tl::from_string (columns_le->text ().ascii (), cols);

    db::Point rv (coord_from_dpoint (db::DPoint (rx, ry), dbu, du, t));
    db::Point cv (coord_from_dpoint (db::DPoint (cx, cy), dbu, du, t));

    if (is_complex) {
      inst = db::CellInstArray (db::CellInst (ci.second), 
                                db::CplxTrans (mag, angle_deg, mirror, disp), rv, cv, rows, cols);
    } else {
      inst = db::CellInstArray (db::CellInst (ci.second), 
                                db::Trans (angle, mirror, disp), rv, cv, rows, cols);
    }

  } else {
    if (is_complex) {
      inst = db::CellInstArray (db::CellInst (ci.second), db::CplxTrans (mag, angle_deg, mirror, disp));
    } else {
      inst = db::CellInstArray (db::CellInst (ci.second), db::Trans (angle, mirror, disp));
    }
  }

  db::Instance new_inst = cell.replace (pos->back ().inst_ptr, inst);
  if (new_inst != pos->back ().inst_ptr) {

    //  change selection to new shape
    lay::ObjectInstPath new_sel (*pos);
    new_sel.back ().inst_ptr = new_inst;

    mp_service->select (*pos, lay::Editable::Reset);
    mp_service->select (new_sel, lay::Editable::Add);

    // locate the new selection
    for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
      ;
    tl_assert (pos != mp_service->selection ().end ());
    m_selection_ptrs [m_index] = pos;

    mp_service->selection_to_view ();

  }

  update ();
}

}

