
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "extNetTracer.h"

#include "dbLayoutUtils.h"
#include "dbPolygonTools.h"
#include "layViewObject.h"
#include "layPlugin.h"
#include "layLayoutView.h"
#include "tlLog.h"

namespace ext
{

// -----------------------------------------------------------------------------------
//  NetTracerData implementation

NetTracerData::NetTracerData ()
{
  // .. nothing yet ..
}

void 
NetTracerData::add_connection (const NetTracerConnection &connection)
{
  if (connection.layer_a () >= 0 && connection.layer_b () >= 0 && (! connection.has_via_layer () || connection.via_layer () >= 0)) {
    m_connections.push_back (connection);
  }
  if (connection.layer_a () >= 0) {
    m_connection_graph.insert (std::make_pair (connection.layer_a (), std::vector <NetTracerConnection> ())).first->second.push_back (connection);
  }
  if (connection.layer_b () >= 0) {
    m_connection_graph.insert (std::make_pair (connection.layer_b (), std::vector <NetTracerConnection> ())).first->second.push_back (connection);
  }
}

const std::vector<NetTracerConnection> &
NetTracerData::connections (unsigned int from_layer) const
{
  std::map <unsigned int, std::vector <NetTracerConnection> >::const_iterator g = m_connection_graph.find (from_layer);
  if (g != m_connection_graph.end ()) {
    return g->second;
  } else {
    static std::vector<NetTracerConnection> empty;
    return empty;
  }
}

// -----------------------------------------------------------------------------------
//  NetTracer implementation

NetTracer::NetTracer ()
  : mp_progress (0), m_name_hier_depth (-1), m_incomplete (false)
{
  //  .. nothing yet ..
}

void 
NetTracer::clear ()
{
  m_shapes_found.clear ();
}

std::string
NetTracer::name () const
{
  return m_name;
}

void 
NetTracer::set_name (const std::string &n)
{
  m_name = n;
}

void
NetTracer::trace (const lay::CellView &cellview, int cellview_index, const NetTracerShape &start, const NetTracerData &data)
{
  trace (cellview, cellview_index, start, NetTracerShape (), data);
}

void
NetTracer::trace (const lay::CellView &cellview, int cellview_index, const NetTracerShape &start, const NetTracerShape &stop, const NetTracerData &data)
{
  m_cv = cellview;
  m_cv_index = cellview_index;

  m_shapes_graph.clear ();
  m_shapes_found.clear ();

  try {

    tl::Progress progress (tl::translate ("Tracing Net"), 0, 1000);
    progress.set_format (tl::translate ("%.0f000 shapes"));
    progress.set_unit (1000);

    mp_progress = &progress;

    tl::SelfTimer timer (tl::verbosity () >= 11, tl::translate ("Net Tracing"));

    m_hit_test_queue.clear ();
    m_stop_shape = stop;

    m_start_shape = start;
    deliver_shape (start, 0, fixed_vector<unsigned int, 2> (start.layer), data);

    while (! m_hit_test_queue.empty ()) {

      std::list <NetTracerSeed>::iterator cand = m_hit_test_queue.end ();
      --cand;

      db::Box combined_box = cand->seed_is_box () ? cand->box () : cand->polygon ().box ();
      double asum = cand->seed_is_box () ? cand->box ().area () : cand->polygon ().area ();

      std::vector <NetTracerSeed> candidates;
      std::vector <std::list <NetTracerSeed>::iterator> to_delete;

      candidates.push_back (*cand);
      to_delete.push_back (cand);

      for (std::list <NetTracerSeed>::iterator c = cand; c != m_hit_test_queue.begin (); ) {

        --c;
        if (c->layer () == cand->layer ()) {

          db::Box b = c->seed_is_box () ? c->box () : c->polygon ().box ();
          double a = c->seed_is_box () ? c->box ().area () : c->polygon ().area ();

          if ((combined_box + b).area () > (asum +a) * 10.0) {
            break;
          }

          combined_box += b;
          asum += a;

          candidates.push_back (*c);
          to_delete.push_back (c);

        }

      }

      for (std::vector <std::list <NetTracerSeed>::iterator>::const_iterator td = to_delete.begin (); td != to_delete.end (); ++td) {
        m_hit_test_queue.erase (*td);
      }

      if (candidates.size () == 1) {
        if (candidates.back ().seed_is_box ()) {
          determine_interactions (candidates.back ().box (), candidates.back ().shape (), fixed_vector<unsigned int, 2> (candidates.back ().layer ()), data);
        } else if (candidates.back ().seed_is_polygon ()) {
          determine_interactions (candidates.back ().polygon (), candidates.back ().shape (), fixed_vector<unsigned int, 2> (candidates.back ().layer ()), data);
        }
      } else {
        determine_interactions (candidates, combined_box, data);
      }

    }

    m_hit_test_queue.clear ();
    m_incomplete = false;
    mp_progress = 0;

  } catch (...) {

    m_shapes_graph.clear ();

    m_hit_test_queue.clear ();
    m_incomplete = true;
    mp_progress = 0;

    throw;

  }

  try {

    if (m_stop_shape.is_valid ()) {

      std::map <NetTracerShape, std::vector<const NetTracerShape *> >::const_iterator st = m_shapes_graph.find (m_stop_shape);
      if (st == m_shapes_graph.end ()) {
        m_shapes_graph.clear ();
        throw tl::Exception (tl::translate ("Nets are not connected"));
      }

      tl::Progress search_progress (tl::translate ("Finding Path"), 0, 100);
      search_progress.set_format (tl::translate ("Iteration %.0f00"));
      search_progress.set_unit (100);

      const NetTracerShape *stop = &st->first;
      const NetTracerShape *start = &m_shapes_graph.find(m_start_shape)->first;

      //  find the shortest path with Dijkstras algorithm

      std::map<const NetTracerShape *, const NetTracerShape *> previous;
      std::map<const NetTracerShape *, size_t> cost;
      cost.insert (std::make_pair (stop, 0));
      std::set<const NetTracerShape *> visited;

      while (! cost.empty ()) {
        
        ++search_progress;

        size_t min_cost = std::numeric_limits <size_t>::max ();
        for (std::map<const NetTracerShape *, size_t>::const_iterator ac = cost.begin (); ac != cost.end (); ++ac) {
          if (ac->second < min_cost) {
            min_cost = ac->second;
          }
        }

        const NetTracerShape *current = 0;
        for (std::map<const NetTracerShape *, size_t>::iterator ac = cost.begin (); ac != cost.end (); ++ac) {
          if (ac->second == min_cost) {
            current = ac->first;
            visited.insert (current);
            cost.erase (ac);
            break;
          }
        }

        if (! current) {
          break; 
        }

        const std::vector <const NetTracerShape *> &adj = m_shapes_graph[*current];
        for (std::vector <const NetTracerShape *>::const_iterator a = adj.begin (); a != adj.end (); ++a) {
          if (visited.find (*a) == visited.end ()) {
            std::map<const NetTracerShape *, size_t>::iterator ac = cost.find (*a);
            if (ac == cost.end ()) {
              cost.insert (std::make_pair (*a, min_cost + 1));
              previous.insert (std::make_pair (*a, current));
            } else if (min_cost + 1 < ac->second) {
              ac->second = min_cost + 1;
              previous.insert (std::make_pair (*a, current));
            }
          }
        }

        if (previous.find (start) != previous.end ()) {
          break;
        }

      }

      m_shapes_found.clear ();

      const NetTracerShape *s = start;
      while (s) {
        m_shapes_found.insert (*s);
        std::map<const NetTracerShape *, const NetTracerShape *>::const_iterator p = previous.find (s);
        if (p == previous.end ()) {
          s = 0;
        } else {
          s = p->second;
        }
      }

      m_shapes_graph.clear ();

    }

  } catch (...) {
    m_shapes_found.clear ();
    m_shapes_graph.clear ();
    throw;
  }
}

void 
NetTracer::evaluate_text (const db::RecursiveShapeIterator &iter)
{
  if (iter.shape ().is_text ()) {
    if (m_name.empty () || m_name_hier_depth < 0 || m_name_hier_depth > int (iter.depth ())) {
      m_name = iter.shape ().text_string ();
      m_name_hier_depth = int (iter.depth ());
    }
  }
}

const NetTracerShape *
NetTracer::deliver_shape (const NetTracerShape &net_shape, const NetTracerShape *adjacent, const fixed_vector<unsigned int, 2> &layers, const NetTracerData &data)
{
  const NetTracerShape *shape_ptr = 0;

  if (! m_stop_shape.is_valid ()) {

    shape_ptr = &*(m_shapes_found.insert (net_shape).first);

    if (mp_progress) {
      ++(*mp_progress);
    }

  } else {
    
    std::map <NetTracerShape, std::vector<const NetTracerShape *> >::iterator n = m_shapes_graph.find (net_shape);
    if (n == m_shapes_graph.end ()) {

      n = m_shapes_graph.insert (std::make_pair (net_shape, std::vector<const NetTracerShape *> ())).first;

      //  shape_ptr != 0 indicates that we need to derive new seeds from that shape 
      shape_ptr = &n->first;

      if (mp_progress) {
        ++(*mp_progress);
      }

    }

    if (adjacent) {

      n->second.push_back (adjacent);

      //  Record the reverse interaction
      std::map <NetTracerShape, std::vector<const NetTracerShape *> >::iterator m = m_shapes_graph.find (*adjacent);
      m->second.push_back (&n->first);

    }

  }

  if (shape_ptr) {

    bool seed_added = false;

    if (shape_ptr->shape.is_box ()) {

      if (shape_ptr->trans.is_ortho ()) {

        //  Create new seed on bottom layer (or self if already at the bottom)
        db::Box box (shape_ptr->trans * shape_ptr->shape.box ());
        m_hit_test_queue.push_back (NetTracerSeed (box, layers[0], shape_ptr));

      } else {

        //  Create new seed on bottom layer (or self if already at the bottom)
        db::Polygon box_poly (shape_ptr->shape.box ());
        box_poly.transform (db::ICplxTrans (shape_ptr->trans));
        m_hit_test_queue.push_back (NetTracerSeed (box_poly, layers[0], shape_ptr));

      }

      seed_added = true;

    } else if (shape_ptr->shape.is_path () || shape_ptr->shape.is_polygon ()) {

      //  Create new seed on bottom layer (or self if already at the bottom)
      db::Polygon polygon;
      net_shape.shape.polygon (polygon);
      polygon.transform (db::ICplxTrans (shape_ptr->trans));
      m_hit_test_queue.push_back (NetTracerSeed (polygon, layers[0], shape_ptr));

      seed_added = true;

    }

    //  Look up and down through via connections
    if (seed_added && layers.size () == 1) {

      const NetTracerSeed &new_seed = m_hit_test_queue.back ();

      unsigned int layer = layers.back ();

      const std::vector <NetTracerConnection> &connections = data.connections (layer);
      for (std::vector <NetTracerConnection>::const_iterator c = connections.begin (); c != connections.end (); ++c) {

        fixed_vector<unsigned int, 2> connected_layers;

        if (c->layer_a () != int (layer)) {
          connected_layers.push_back (c->layer_a ());
        } else if (c->layer_b () != int (layer)) {
          connected_layers.push_back (c->layer_b ());
        }

        if (c->has_via_layer ()) {
          connected_layers.push_back (c->via_layer ());
        }

        if (! connected_layers.empty ()) {
          if (new_seed.seed_is_box ()) {
            determine_interactions (new_seed.box (), shape_ptr, connected_layers, data);
          } else if (new_seed.seed_is_polygon ()) {
            determine_interactions (new_seed.polygon (), shape_ptr, connected_layers, data);
          }
        }

      }

    }

  } 
  
  return shape_ptr;
}

bool
NetTracer::interacts (const db::Box &box, const NetTracerShape &net_shape)
{
  if (net_shape.shape.is_text ()) {

    return true;

  } else if (net_shape.shape.is_box ()) {

    if (net_shape.trans.is_ortho ()) {

      return true;

    } else {

      db::Polygon box_poly (net_shape.shape.box ());
      box_poly.transform (db::ICplxTrans (net_shape.trans));

      return (db::interact (box_poly, box));

    }

  } else if (net_shape.shape.is_polygon () || net_shape.shape.is_path ()) {

    db::Polygon polygon;
    net_shape.shape.polygon (polygon);
    polygon.transform (db::ICplxTrans (net_shape.trans));

    return db::interact (polygon, box);

  } else {

    return false;

  }
}

bool
NetTracer::interacts (const db::Polygon &polygon, const NetTracerShape &net_shape)
{
  if (net_shape.shape.is_text ()) {

    db::Box trans_box (net_shape.trans * net_shape.shape.bbox ());
    return db::interact (polygon, trans_box);

  } else if (net_shape.shape.is_box ()) {

    if (net_shape.trans.is_ortho ()) {

      db::Box trans_box (net_shape.trans * net_shape.shape.box ());
      return db::interact (polygon, trans_box);

    } else {

      db::Polygon box_poly (net_shape.shape.box ());
      box_poly.transform (db::ICplxTrans (net_shape.trans));

      return db::interact (polygon, box_poly);

    }

  } else if (net_shape.shape.is_polygon () || net_shape.shape.is_path ()) {

    db::Polygon p;
    net_shape.shape.polygon (p);
    p.transform (db::ICplxTrans (net_shape.trans));

    return db::interact (p, polygon);

  } else {

    return false;

  }
}

struct HitTestDataBoxConverter
{
  db::Box operator() (const NetTracerSeed &d) const
  {
    return d.seed_is_box () ? d.box () : d.polygon ().box ();
  }

  typedef db::simple_bbox_tag complexity;
};

void
NetTracer::determine_interactions (const std::vector<NetTracerSeed> &seeds, const db::Box &combined_box, const NetTracerData &data)
{
  bool extract_full_graph = m_stop_shape.is_valid ();

  typedef db::box_tree<db::Box, NetTracerSeed, HitTestDataBoxConverter, 1> HitTestDataBoxTree;

  HitTestDataBoxTree seed_tree;
  for (std::vector<NetTracerSeed>::const_iterator seed = seeds.begin (); seed != seeds.end (); ++seed) {
    seed_tree.insert (*seed);
  }
  seed_tree.sort (HitTestDataBoxConverter ());

  unsigned int layer = seeds.front ().layer ();
  fixed_vector<unsigned int, 2> layers (seeds.front ().layer ());

  db::RecursiveShapeIterator net_shapes (m_cv->layout (), *m_cv.cell (), layer, combined_box);
  while (! net_shapes.at_end ()) {

    NetTracerShape net_shape (net_shapes.trans (), net_shapes.shape (), layer, net_shapes.cell_index ());
    if (m_shapes_found.find (net_shape) == m_shapes_found.end ()) {

      db::Box shape_bbox (net_shapes.itrans () * net_shapes.shape ().bbox ());
      for (HitTestDataBoxTree::touching_iterator seed = seed_tree.begin_touching (shape_bbox, HitTestDataBoxConverter ()); ! seed.at_end (); ++seed) {

        if (tl::verbosity () >= 40) {
          tl::info << "Net net candidate: " << net_shapes.shape ().to_string () << " on " << m_cv->layout ().get_properties (layer).to_string ();
        }

        evaluate_text (net_shapes);

        bool interact = false;
        if (seed->seed_is_box ()) {
          if (interacts (seed->box (), net_shape)) {
            interact = true;
          }
        } else if (seed->seed_is_polygon ()) {
          if (interacts (seed->polygon (), net_shape)) {
            interact = true;
          }
        }

        if (interact) {
          deliver_shape (net_shape, seed->shape (), layers, data);
          if (! extract_full_graph) {
            break;
          }
        }

      }

    }

    ++net_shapes;

  }
}

void
NetTracer::determine_interactions (const db::Box &seed, const NetTracerShape *shape, const fixed_vector<unsigned int, 2> &layers, const NetTracerData &data)
{
  if (tl::verbosity () >= 40) {
    tl::info << "Continue trace net with " << seed.to_string () << " on " << m_cv->layout ().get_properties (layers.back ()).to_string ();
  }

  db::RecursiveShapeIterator net_shapes (m_cv->layout (), *m_cv.cell (), layers.back (), seed);
  while (! net_shapes.at_end ()) {

    NetTracerShape net_shape (net_shapes.trans (), net_shapes.shape (), layers.back (), net_shapes.cell_index ());
    if (m_shapes_found.find (net_shape) == m_shapes_found.end ()) {

      if (tl::verbosity () >= 40) {
        tl::info << "Net net candidate: " << net_shapes.shape ().to_string () << " on " << m_cv->layout ().get_properties (layers.back ()).to_string ();
      }

      evaluate_text (net_shapes);

      if (interacts (seed, net_shape)) {
        deliver_shape (net_shape, shape, layers, data);
      }

    }

    ++net_shapes;

  }
}

void
NetTracer::determine_interactions (const db::Polygon &seed, const NetTracerShape *shape, const fixed_vector<unsigned int, 2> &layers, const NetTracerData &data)
{
  if (tl::verbosity () >= 40) {
    tl::info << "Continue trace net with " << seed.to_string () << " on " << m_cv->layout ().get_properties (layers.back ()).to_string ();
  }

  int area_ratio = 2;

  db::Polygon::area_type poly_area = seed.area ();
  db::Polygon::area_type box_area = seed.box ().area ();

  if (poly_area == box_area && seed.vertices () == 4) {

    //  The polygon is a box
    determine_interactions (seed.box (), shape, layers, data);

  } else if (poly_area + 1 >= box_area / area_ratio) {

    //  The polygon is sufficiently "dense", so it can be used as it is.
    db::RecursiveShapeIterator net_shapes (m_cv->layout (), *m_cv.cell (), layers.back (), seed.box ());
    while (! net_shapes.at_end ()) {

      NetTracerShape net_shape (net_shapes.trans (), net_shapes.shape (), layers.back (), net_shapes.cell_index ());
      if (m_shapes_found.find (net_shape) == m_shapes_found.end ()) {

        if (tl::verbosity () >= 40) {
          tl::info << "Net net candidate: " << net_shapes.shape ().to_string () << " on " << m_cv->layout ().get_properties (layers.back ()).to_string ();
        }

        evaluate_text (net_shapes);

        if (interacts (seed, net_shape)) {
          deliver_shape (net_shape, shape, layers, data);
        }

      }

      ++net_shapes;

    }

  } else {

    //  otherwise split polygon and recursively treat these parts ...
    std::vector <db::Polygon> polygons;
    db::split_polygon (seed, polygons);

    for (std::vector<db::Polygon>::const_iterator p = polygons.begin (); p != polygons.end (); ++p) {
      determine_interactions (*p, shape, layers, data);
    }

  }
}

}

