
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_extNetTracer
#define HDR_extNetTracer

#include "config.h"

#include <vector>
#include <map>
#include <list>

#include "layCellView.h"
#include "tlProgress.h"

namespace db
{
  class RecursiveShapeIterator;
}

namespace ext
{

/**
 *  @brief Describes one connection between two conductive layers
 *
 *  This class has three members: the index of the first conductive layer, the
 *  index of the via layer and the index of the second conductive layer.
 */
class KLAYOUT_DLL NetTracerConnection
{
public:
  /**
   *  @brief Creates a connection between layer la and lb without an intermediate via layer
   *
   *  A connection described by this constructor defines a connection at all places where la overlaps lb.
   */
  NetTracerConnection (int la, int lb)
    : m_layer_a (la), m_via_layer (-1), m_has_via_layer (false), m_layer_b (lb)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Creates a connection between layer la and lb with a via layer
   *
   *  A connection described by this constructor defines a connection at all places where la overlaps the via 
   *  shape and the same via shape overlaps lb.
   */
  NetTracerConnection (int la, int via, int lb)
    : m_layer_a (la), m_via_layer (via), m_has_via_layer (true), m_layer_b (lb)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Gets the index of the first conductive layer
   *
   *  Returns a negative value if the via layer is not present (and should be considered
   *  empty).
   */
  int layer_a () const
  {
    return m_layer_a;
  }

  /**
   *  @brief Gets the index of the second conductive layer
   *
   *  Returns a negative value if the via layer is not present (and should be considered
   *  empty).
   */
  int layer_b () const
  {
    return m_layer_b;
  }

  /**
   *  @brief Returns true if this connection has a via layer
   */
  bool has_via_layer () const
  {
    return m_has_via_layer;
  }

  /**
   *  @brief Returns the index of the via layer. 
   *
   *  Returns a negative value if the via layer is not present (and should be considered
   *  empty).
   */
  int via_layer () const
  {
    return m_via_layer;
  }

private:
  int m_layer_a;
  int m_via_layer;
  bool m_has_via_layer;
  int m_layer_b;
};

/**
 *  @brief Wraps the data for a net tracing 
 */
class KLAYOUT_DLL NetTracerData
{
public:
  /**
   *  @brief Default constructor
   */
  NetTracerData ();

  /**
   *  @brief Add a connection to the connection graph
   */
  void add_connection (const NetTracerConnection &connection);

  /**
   *  @brief Returns all connectors starting or ending at the given layer.
   *
   *  If the layer does not participate in the net tracing, no connections are
   *  returned.
   */
  const std::vector<NetTracerConnection> &connections (unsigned int from_layer) const;

  /**
   *  @brief Returns true, if no connection is defined.
   */
  bool is_empty () const
  {
    return m_connections.empty ();
  }

private:
  std::vector <NetTracerConnection> m_connections;
  std::map <unsigned int, std::vector <NetTracerConnection> > m_connection_graph;
};

struct KLAYOUT_DLL NetTracerShape
{
  db::CplxTrans trans;
  db::Shape shape;
  unsigned int layer;
  db::cell_index_type cell_index;

  NetTracerShape ()
    : trans (), shape (), layer (0), cell_index (0)
  {
    // .. nothing yet ..
  }

  NetTracerShape (const db::CplxTrans &t, const db::Shape &s, unsigned int l, db::cell_index_type c)
    : trans (t), shape (s), layer (l), cell_index (c)
  {
    // .. nothing yet ..
  }

  bool is_valid () const
  {
    return !shape.is_null ();
  }

  bool operator== (const NetTracerShape &other) const
  {
    if (layer != other.layer) {
      return false;
    }
    if (cell_index != other.cell_index) {
      return false;
    }
    if (shape != other.shape) {
      return false;
    }
    return trans.equal (other.trans);
  }

  bool operator< (const NetTracerShape &other) const
  {
    if (layer != other.layer) {
      return layer < other.layer;
    }
    if (cell_index != other.cell_index) {
      return cell_index < other.cell_index;
    }
    if (shape != other.shape) {
      return shape < other.shape;
    }
    return trans.less (other.trans);
  }
};

template <class T, size_t N>
class fixed_vector 
{
public:
  typedef T value_type;
  typedef T *iterator;
  typedef const T *const_iterator;

  fixed_vector ()
    : m_size (0)
  {
    // .. nothing yet ..
  }

  fixed_vector (const T &l)
    : m_size (1)
  {
    m_values [0] = l;
  }

  void clear () 
  {
    m_size = 0;
  }

  T &back ()
  {
    tl_assert (m_size > 0);
    return m_values [m_size - 1];
  }

  const T &back () const
  {
    tl_assert (m_size > 0);
    return m_values [m_size - 1];
  }

  T &operator[] (size_t i) 
  {
    return m_values [i];
  }

  const T &operator[] (size_t i) const
  {
    return m_values [i];
  }

  void push_back (const T &l)
  {
    tl_assert (m_size < N);
    m_values [m_size++] = l;
  }

  void pop_back ()
  {
    --m_size;
  }

  size_t size () const
  {
    return m_size;
  }

  T *begin ()
  {
    return m_values;
  }

  T *end ()
  {
    return m_values + m_size;
  }

  const T *begin () const
  {
    return m_values;
  }

  const T *end () const
  {
    return m_values + m_size;
  }

  bool empty () const
  {
    return m_size == 0;
  }

private:
  size_t m_size;
  unsigned int m_values [N];
};

class KLAYOUT_DLL NetTracerSeed
{
public:
  enum SeedType { STNone, STBox, STPolygon };

  NetTracerSeed ()
    : m_seed_type (STNone), mp_shape (0)
  {
    // .. nothing yet ..
  }

  NetTracerSeed (const db::Box &seed, unsigned int layer, const NetTracerShape *shape)
    : m_seed_type (STBox), m_layer (layer), mp_shape (shape)
  {
    new (&m_seed.box) db::Box (seed);
  }

  NetTracerSeed (const db::Polygon &seed, unsigned int layer, const NetTracerShape *shape)
    : m_seed_type (STPolygon), m_layer (layer), mp_shape (shape)
  {
    new (&m_seed.polygon) db::Polygon (seed);
  }

  NetTracerSeed (const NetTracerSeed &tn)
    : m_seed_type (tn.m_seed_type), m_layer (tn.m_layer), mp_shape (tn.mp_shape)
  {
    if (m_seed_type == STBox) {
      new (&m_seed.box) db::Box (*(db::Box *)(&tn.m_seed.box));
    } else if (m_seed_type == STPolygon) {
      new (&m_seed.polygon) db::Polygon (*(db::Polygon *)(&tn.m_seed.polygon));
    }
  }

  NetTracerSeed &operator= (const NetTracerSeed &tn)
  {
    if (this != &tn) {

      release ();

      m_seed_type = tn.m_seed_type;
      m_layer = tn.m_layer;
      mp_shape = tn.mp_shape;

      if (m_seed_type == STBox) {
        new (&m_seed.box) db::Box (*(db::Box *)(&tn.m_seed.box));
      } else if (m_seed_type == STPolygon) {
        new (&m_seed.polygon) db::Polygon (*(db::Polygon *)(&tn.m_seed.polygon));
      }

    }

    return *this;
  }

  ~NetTracerSeed ()
  {
    release ();
  }

  bool seed_is_invalid () const
  {
    return m_seed_type == STNone;
  }

  bool seed_is_box () const
  {
    return m_seed_type == STBox;
  }

  bool seed_is_polygon () const
  {
    return m_seed_type == STPolygon;
  }

  const db::Box &box () const 
  {
    return *((db::Box *) (&m_seed.box));
  }

  db::Box &box () 
  {
    return *((db::Box *) (&m_seed.box));
  }

  const db::Polygon &polygon () const 
  {
    return *((db::Polygon *) (&m_seed.polygon));
  }

  db::Polygon &polygon () 
  {
    return *((db::Polygon *) (&m_seed.polygon));
  }

  unsigned int layer () const
  {
    return m_layer;
  }

  std::string to_string () const
  {
    if (seed_is_box ()) {
      return box ().to_string ();
    } else if (seed_is_polygon ()) {
      return polygon ().to_string ();
    } else {
      return std::string ();
    }
  }

  const NetTracerShape *shape () const
  {
    return mp_shape;
  }

private:
  SeedType m_seed_type;
  union {
    char box [sizeof (db::Box)];
    char polygon [sizeof (db::Polygon)];
  } m_seed;
  unsigned int m_layer;
  const NetTracerShape *mp_shape;

  void release ()
  {
    if (m_seed_type == STBox) {
      ((db::Box *) (&m_seed.box))->~box ();
    } else if (m_seed_type == STPolygon) {
      ((db::Polygon *) (&m_seed.polygon))->~polygon ();
    }
  }
};

/**
 *  @brief The net tracer
 *
 *  This object will provide a net tracing on a given cell view.
 *  Net tracing can be performed with a given seed point and given tracing data.
 *  The tracing is initiated with the "trace" method.
 */
class KLAYOUT_DLL NetTracer 
{
public:
  typedef std::set <NetTracerShape>::const_iterator iterator;

  /**
   *  @brief Construct a net tracer on the given cellview.
   */
  NetTracer ();

  /**
   *  @brief Trace the net starting from the given seed with the given data.
   */
  void trace (const lay::CellView &cellview, int cellview_index, const NetTracerShape &start, const NetTracerData &data);

  /**
   *  @brief Trace the path starting from the given seed and ending at the given shape with the given data.
   */
  void trace (const lay::CellView &cellview, int cellview_index, const NetTracerShape &start, const NetTracerShape &stop, const NetTracerData &data);

  /**
   *  @brief Begin operator for the shapes found
   */
  iterator begin () const
  {
    return m_shapes_found.begin ();
  }

  /**
   *  @brief End operator for the shapes found
   */
  iterator end () const
  {
    return m_shapes_found.end ();
  }

  /**
   *  @brief Returns the number of shapes found
   */
  size_t size () const
  {
    return m_shapes_found.size ();
  }

  /**
   *  @brief Returns true, if the net is incomplete
   *
   *  This flag is true if the extractor was aborted
   *  for example by the user.
   *  The shapes do not fully cover the net.
   */
  bool incomplete () const
  {
    return m_incomplete;
  }

  /**
   *  @brief Clear the data found so far.
   */
  void clear ();

  /**
   *  @brief Get a name for the net
   *
   *  The name can be empty if not specific net name could be determined (i.e. from a label or property)
   */
  std::string name () const;

  /**
   *  @brief Set a name for the net
   */
  void set_name (const std::string &n);

  /**
   *  @brief Get the cellview from which this net was taken
   */
  const lay::CellView &cellview () const
  {
    return m_cv;
  }

  /**
   *  @brief Get the cellview index from which this net was taken
   */
  int cellview_index () const
  {
    return m_cv_index;
  }

private:
  lay::CellView m_cv;
  int m_cv_index;
  std::set <NetTracerShape> m_shapes_found;
  std::map <NetTracerShape, std::vector<const NetTracerShape *> > m_shapes_graph;
  tl::Progress *mp_progress;
  std::list <NetTracerSeed> m_hit_test_queue;
  std::string m_name;
  int m_name_hier_depth;
  bool m_incomplete;
  NetTracerShape m_stop_shape; 
  NetTracerShape m_start_shape;

  void determine_interactions (const db::Box &seed, const NetTracerShape *shape, const fixed_vector<unsigned int, 2> &layers, const NetTracerData &data);
  void determine_interactions (const db::Polygon &seed, const NetTracerShape *shape, const fixed_vector<unsigned int, 2> &layers, const NetTracerData &data);
  void determine_interactions (const std::vector<NetTracerSeed> &seeds, const db::Box &combined_box, const NetTracerData &data);
  void evaluate_text (const db::RecursiveShapeIterator &iter);
  const NetTracerShape *deliver_shape (const NetTracerShape &shape, const NetTracerShape *adjacent, const fixed_vector<unsigned int, 2> &layers, const NetTracerData &data);
  bool interacts (const db::Box &box, const NetTracerShape &shape);
  bool interacts (const db::Polygon &polygon, const NetTracerShape &shape);
};

}

#endif

