
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "extNetTracerDialog.h"
#include "extNetTracerConfig.h"

#include "layConfigurationDialog.h"
#include "layConfig.h"
#include "layConverters.h"
#include "layFinder.h"
#include "layLayoutView.h"
#include "layTechSetupDialog.h"
#include "layMainWindow.h"
#include "layExceptions.h"
#include "layFileDialog.h"
#include "tlXMLWriter.h"

#include <QtGui/QMessageBox>
#include <QtGui/QInputDialog>
#include <QtGui/QItemDelegate>
#include <QtGui/QHeaderView>

#include <fstream>
#include <sstream>

namespace ext
{

// -----------------------------------------------------------------------------------
//  NetTracerConnectionInfo implementation

NetTracerConnectionInfo::NetTracerConnectionInfo ()
{
  // .. nothing yet ..
}

NetTracerConnectionInfo::NetTracerConnectionInfo (const db::LayerProperties &la, const db::LayerProperties &lb)
  : m_la (la), m_via (), m_lb (lb)
{
  // .. nothing yet ..
}

NetTracerConnectionInfo::NetTracerConnectionInfo (const db::LayerProperties &la, const db::LayerProperties &via, const db::LayerProperties &lb)
  : m_la (la), m_via (via), m_lb (lb)
{
  // .. nothing yet ..
}

NetTracerConnection 
NetTracerConnectionInfo::get (const lay::CellView &cv) const
{
  int la = -1, lb = -1, via = -1;

  for (db::Layout::layer_iterator l = cv->layout ().begin_layers (); l != cv->layout ().end_layers (); ++l) {
    if ((*l).second->log_equal (m_la)) {
      la = (*l).first;
      break;
    }
  }

  for (db::Layout::layer_iterator l = cv->layout ().begin_layers (); l != cv->layout ().end_layers (); ++l) {
    if ((*l).second->log_equal (m_lb)) {
      lb = (*l).first;
      break;
    }
  }

  if (! (m_via == db::LayerProperties ())) {
    for (db::Layout::layer_iterator l = cv->layout ().begin_layers (); l != cv->layout ().end_layers (); ++l) {
      if ((*l).second->log_equal (m_via)) {
        via = (*l).first;
        break;
      }
    }
    return NetTracerConnection (la, via, lb);
  } else {
    return NetTracerConnection (la, lb);
  }
}

std::string 
NetTracerConnectionInfo::to_string () const
{
  std::string res;
  res += m_la.to_string ();
  res += ",";
  res += m_via.to_string ();
  res += ",";
  res += m_lb.to_string ();

  return res;
}

void 
NetTracerConnectionInfo::parse (tl::Extractor &ex)
{
  m_la.read (ex);
  ex.expect (",");
  m_via.read (ex);
  ex.expect (",");
  m_lb.read (ex);
}

// -----------------------------------------------------------------------------------
//  NetTracerTechnologyComponent implementation

NetTracerTechnologyComponent::NetTracerTechnologyComponent ()
  : lay::TechnologyComponent ("connectivity", tl::translate ("Connectivity"))
{
  // .. nothing yet ..
}

NetTracerTechnologyComponent::NetTracerTechnologyComponent (const NetTracerTechnologyComponent &d)
  : lay::TechnologyComponent ("connectivity", tl::translate ("Connectivity"))
{
  m_connections = d.m_connections;
}

std::string 
NetTracerTechnologyComponent::to_string () const
{
  std::string res;
  res += "(";
  for (const_iterator c = begin (); c != end (); ++c) {
    res += c->to_string ();
    res += ";";
  }
  res += ")";

  return res;
}

void 
NetTracerTechnologyComponent::parse (tl::Extractor &ex)
{
  m_connections.clear ();

  ex.test ("(");

  while (! ex.at_end () && ! ex.test (")")) {
    NetTracerConnectionInfo info;
    info.parse (ex);
    ex.test (";");
    m_connections.push_back (info);
  }
}

lay::TechnologyComponentEditor * 
NetTracerTechnologyComponent::create_editor (QWidget *parent) 
{
  return new NetTracerTechComponentEditor (parent);
}

// -----------------------------------------------------------------------------------------
//  NetTracerConnectivityColumnDelegate definition and implementation

class NetTracerConnectivityColumnDelegate
  : public QItemDelegate
{
public:
  NetTracerConnectivityColumnDelegate (QWidget *parent, NetTracerTechnologyComponent *data)
    : QItemDelegate (parent), mp_data (data)
  {
    //  .. nothing yet ..
  }

  QWidget *createEditor (QWidget *parent, const QStyleOptionViewItem &option, const QModelIndex &index) const 
  {
    return new QLineEdit (parent);
  }

  void updateEditorGeometry(QWidget *editor, const QStyleOptionViewItem &option, const QModelIndex &index) const
  {
    editor->setGeometry(option.rect);
  }

  void setEditorData (QWidget *widget, const QModelIndex &index) const
  {
    QLineEdit *editor = dynamic_cast<QLineEdit *> (widget);
    if (editor) {
      int n = index.model ()->data (index, Qt::UserRole).asInt ();
      if (mp_data->size () > size_t (n)) { 
        if (index.column () == 0) {
          editor->setText (mp_data->begin () [n].layer_a ().to_string ().c_str ());
        } else if (index.column () == 1) {
          editor->setText (mp_data->begin () [n].via_layer ().to_string ().c_str ());
        } else if (index.column () == 2) {
          editor->setText (mp_data->begin () [n].layer_b ().to_string ().c_str ());
        }
      }
    }
  }

  void setModelData (QWidget *widget, QAbstractItemModel *model, const QModelIndex &index) const
  {
    QLineEdit *editor = dynamic_cast<QLineEdit *> (widget);
    if (editor) {

      int n = model->data (index, Qt::UserRole).asInt ();
      if (mp_data->size () > size_t (n)) { 

        db::LayerProperties lp;

        std::string text = editor->text ().ascii ();
        tl::Extractor ex (text.c_str ());
        lp.read (ex);

        if ((index.column () == 0 || index.column () == 2) && lp.log_equal (db::LayerProperties ())) {
          model->setData (index, QVariant (tl::translate ("Enter layer").c_str ()), Qt::DisplayRole);
          model->setData (index, QVariant (QColor (255, 0, 0)), Qt::ForegroundRole);
        } else if (index.column () == 1 && lp.log_equal (db::LayerProperties ())) {
          model->setData (index, QVariant (tl::translate ("None").c_str ()), Qt::DisplayRole);
          model->setData (index, QVariant (), Qt::ForegroundRole);
        } else {
          model->setData (index, QVariant (QString (lp.to_string ().c_str ())), Qt::DisplayRole);
          model->setData (index, QVariant (), Qt::ForegroundRole);
        }

        if (index.column () == 0) {
          mp_data->begin () [n].set_layer_a (lp);
        } else if (index.column () == 1) {
          mp_data->begin () [n].set_via_layer (lp);
        } else if (index.column () == 2) {
          mp_data->begin () [n].set_layer_b (lp);
        }

      }

    }
  }

  QSize sizeHint (const QStyleOptionViewItem &option, const QModelIndex &index) const 
  {
    QWidget *editor = createEditor (0, option, index);
    QSize size = editor->sizeHint ();
    delete editor;
    return size - QSize (2, 2);
  }

private:
  NetTracerTechnologyComponent *mp_data;
};

// -----------------------------------------------------------------------------------
//  NetTracerTechComponentEditor implementation

NetTracerTechComponentEditor::NetTracerTechComponentEditor (QWidget *parent)
  : TechnologyComponentEditor (parent)
{
  Ui::NetTracerTechComponentEditor::setupUi (this);

  connect (add_conductor_pb, SIGNAL (clicked ()), this, SLOT (add_clicked ()));
  connect (del_conductor_pb, SIGNAL (clicked ()), this, SLOT (del_clicked ()));
  connect (move_conductor_up_pb, SIGNAL (clicked ()), this, SLOT (move_up_clicked ()));
  connect (move_conductor_down_pb, SIGNAL (clicked ()), this, SLOT (move_down_clicked ()));

  connectivity_table->horizontalHeader ()->setHighlightSections (false);
  connectivity_table->verticalHeader ()->hide ();
}

void 
NetTracerTechComponentEditor::commit ()
{
  NetTracerTechnologyComponent *data = dynamic_cast <NetTracerTechnologyComponent *> (tech_component ());
  if (! data) {
    return;
  }

  *data = m_data;
}

void 
NetTracerTechComponentEditor::setup ()
{
  NetTracerTechnologyComponent *data = dynamic_cast <NetTracerTechnologyComponent *> (tech_component ());
  if (! data) {
    return;
  }

  m_data = *data;

  for (int c = 0; c < 3; ++c) {
    if (connectivity_table->itemDelegateForColumn (c) != 0) {
      delete connectivity_table->itemDelegateForColumn (c);
    }
    connectivity_table->setItemDelegateForColumn (c, new NetTracerConnectivityColumnDelegate (connectivity_table, &m_data));
  }

  update ();
}

void 
NetTracerTechComponentEditor::add_clicked ()
{
  int row = connectivity_table->currentItem () ? connectivity_table->row (connectivity_table->currentItem ()) : -1;
  if (row < 0) {
    m_data.add (NetTracerConnectionInfo ());
    row = int (m_data.size () - 1);
  } else {
    row += 1;
    m_data.insert (m_data.begin () + row, NetTracerConnectionInfo ());
  }

  update ();
  connectivity_table->setCurrentItem (connectivity_table->item (row, 0));
}

void 
NetTracerTechComponentEditor::del_clicked ()
{
  std::set<int> selected_rows;
  QModelIndexList selected_indices = connectivity_table->selectionModel ()->selectedIndexes ();
  for (QModelIndexList::const_iterator i = selected_indices.begin (); i != selected_indices.end (); ++i) {
    selected_rows.insert (i->row ());
  }

  connectivity_table->setCurrentIndex (QModelIndex ());

  int offset = 0;
  for (std::set<int>::const_iterator r = selected_rows.begin (); r != selected_rows.end (); ++r) {
    m_data.erase (m_data.begin () + (*r - offset));
    ++offset;
  }

  update ();
}

void 
NetTracerTechComponentEditor::move_up_clicked ()
{
  std::set<int> selected_rows;
  QModelIndexList selected_indices = connectivity_table->selectionModel ()->selectedIndexes ();
  for (QModelIndexList::const_iterator i = selected_indices.begin (); i != selected_indices.end (); ++i) {
    selected_rows.insert (i->row ());
  }

  QTableWidgetItem *current = connectivity_table->currentItem ();
  int n_current = current ? current->data (Qt::UserRole).asInt () : -1;

  connectivity_table->setCurrentIndex (QModelIndex ());

  int n = 0;
  for (NetTracerTechnologyComponent::iterator l = m_data.begin (); l != m_data.end (); ++l, ++n) {
    if (selected_rows.find (n + 1) != selected_rows.end () && selected_rows.find (n) == selected_rows.end ()) {
      std::swap (m_data.begin () [n + 1], m_data.begin () [n]);
      selected_rows.erase (n + 1);
      selected_rows.insert (n);
      if (n_current == n + 1) {
        n_current = n;
      }
    }
  }

  update ();

  // select the new items
  for (std::set <int>::const_iterator s = selected_rows.begin (); s != selected_rows.end (); ++s) {
    connectivity_table->selectionModel ()->select (connectivity_table->model ()->index (*s, 0), QItemSelectionModel::Select | QItemSelectionModel::Rows);
  }

  if (n_current >= 0) {
    connectivity_table->selectionModel ()->select (connectivity_table->model ()->index (n_current, 0), QItemSelectionModel::Current | QItemSelectionModel::Rows);
  }
}

void 
NetTracerTechComponentEditor::move_down_clicked ()
{
  std::set<int> selected_rows;
  QModelIndexList selected_indices = connectivity_table->selectionModel ()->selectedIndexes ();
  for (QModelIndexList::const_iterator i = selected_indices.begin (); i != selected_indices.end (); ++i) {
    selected_rows.insert (i->row ());
  }

  QTableWidgetItem *current = connectivity_table->currentItem ();
  int n_current = current ? current->data (Qt::UserRole).asInt () : -1;

  connectivity_table->setCurrentIndex (QModelIndex ());

  int n = m_data.size ();
  for (NetTracerTechnologyComponent::iterator l = m_data.end (); l != m_data.begin (); ) {
    --l;
    --n;
    if (selected_rows.find (n - 1) != selected_rows.end () && selected_rows.find (n) == selected_rows.end ()) {
      std::swap (m_data.begin () [n - 1], m_data.begin () [n]);
      selected_rows.erase (n - 1);
      selected_rows.insert (n);
      if (n_current == n - 1) {
        n_current = n;
      }
    }
  }

  update ();

  // select the new items
  for (std::set <int>::const_iterator s = selected_rows.begin (); s != selected_rows.end (); ++s) {
    connectivity_table->selectionModel ()->select (connectivity_table->model ()->index (*s, 0), QItemSelectionModel::Select | QItemSelectionModel::Rows);
  }

  if (n_current >= 0) {
    connectivity_table->selectionModel ()->select (connectivity_table->model ()->index (n_current, 0), QItemSelectionModel::Current | QItemSelectionModel::Rows);
  }
}

void
NetTracerTechComponentEditor::update ()
{
  connectivity_table->clear ();
  connectivity_table->setRowCount (m_data.size ());

  connectivity_table->setColumnCount (3);
  QStringList labels;
  labels << QString (tl::translate ("Conductor 1").c_str ());
  labels << QString (tl::translate ("Via (optional)").c_str ());
  labels << QString (tl::translate ("Conductor 2").c_str ());
  connectivity_table->setHorizontalHeaderLabels (labels);

  int n = 0;
  for (NetTracerTechnologyComponent::iterator l = m_data.begin (); l != m_data.end (); ++l, ++n) {

    for (int c = 0; c < 3; ++c) {

      QTableWidgetItem *item = new QTableWidgetItem ();
      item->setFlags (item->flags () | Qt::ItemIsEditable);
      connectivity_table->setItem (n, c, item);

      item->setData (Qt::ForegroundRole, QVariant ());

      if (c == 0) {
        if (l->layer_a ().log_equal (db::LayerProperties ())) {
          item->setData (Qt::DisplayRole, QVariant (tl::translate ("Enter Layer").c_str ()));
          item->setData (Qt::ForegroundRole, QVariant (QColor (255, 0, 0)));
        } else {
          item->setData (Qt::DisplayRole, QVariant (l->layer_a ().to_string ().c_str ()));
        }
      } else if (c == 1) {
        if (l->via_layer ().log_equal (db::LayerProperties ())) {
          item->setData (Qt::DisplayRole, QVariant (tl::translate ("None").c_str ()));
        } else {
          item->setData (Qt::DisplayRole, QVariant (l->via_layer ().to_string ().c_str ()));
        }
      } else if (c == 2) {
        if (l->layer_b ().log_equal (db::LayerProperties ())) {
          item->setData (Qt::DisplayRole, QVariant (tl::translate ("Enter Layer").c_str ()));
          item->setData (Qt::ForegroundRole, QVariant (QColor (255, 0, 0)));
        } else {
          item->setData (Qt::DisplayRole, QVariant (l->layer_b ().to_string ().c_str ()));
        }
      }

      item->setData (Qt::UserRole, QVariant (n));

    }

  }

  connectivity_table->clearSelection ();
}

// -----------------------------------------------------------------------------------
//  Net implementation

Net::Net(const NetTracer &tracer)
  : m_cv (tracer.cellview ()), m_cv_index (tracer.cellview_index ()), m_name (tracer.name ()), m_incomplete (tracer.incomplete ())
{
  size_t n = 0;
  for (NetTracer::iterator s = tracer.begin (); s != tracer.end (); ++s) {
    ++n;
  }
  m_net_shapes.reserve (n);

  for (NetTracer::iterator s = tracer.begin (); s != tracer.end (); ++s) {
    db::Shape new_shape = m_shapes.insert (s->shape);
    m_net_shapes.push_back (*s);
    m_net_shapes.back ().shape = new_shape;
  }
}

// -----------------------------------------------------------------------------------
//  NetTracerDialog implementation

NetTracerDialog::NetTracerDialog (lay::MainWindow *main_window, lay::LayoutView *view)
  : lay::Browser (main_window, view, "net_tracer_dialog"),
    lay::ViewService (view->view_object_widget ()), 
    m_net_index (1),
    m_window (ext::NTFitNet),
    m_window_dim (0.0),
    m_max_marker_count (0),
    m_marker_line_width (-1),
    m_marker_vertex_size (-1),
    m_marker_halo (-1),
    m_marker_dither_pattern (-1),
    m_marker_intensity (0),
    m_mouse_state (0)
{
  mp_export_file_dialog = new lay::FileDialog (this, tl::translate ("Export Net"), tl::translate ("KLayout net files (*.lyn);;All files (*)"));

  Ui::NetTracerDialog::setupUi (this);

  connect (add_pb, SIGNAL (clicked ()), this, SLOT (trace_net_button_clicked ()));
  connect (add2_pb, SIGNAL (clicked ()), this, SLOT (trace_path_button_clicked ()));
  connect (del_pb, SIGNAL (clicked ()), this, SLOT (delete_button_clicked ()));
  connect (clear_all_pb, SIGNAL (clicked ()), this, SLOT (clear_all_button_clicked ()));
  connect (detailed_cb, SIGNAL (clicked ()), this, SLOT (detailed_mode_clicked ()));
  connect (export_pb, SIGNAL (clicked ()), this, SLOT (export_clicked ()));
  connect (export_text_pb, SIGNAL (clicked ()), this, SLOT (export_text_clicked ()));
  connect (configure_pb, SIGNAL (clicked ()), this, SLOT (configure_clicked ()));
  connect (stack_pb, SIGNAL (clicked ()), this, SLOT (layer_stack_clicked ()));
  connect (net_list, SIGNAL (itemSelectionChanged ()), this, SLOT (item_selection_changed ()));
  connect (net_list, SIGNAL (itemDoubleClicked (QListWidgetItem *)), this, SLOT (item_double_clicked (QListWidgetItem *)));
  connect (sticky_cbx, SIGNAL (clicked ()), this, SLOT (sticky_mode_clicked ()));

  view->add_layer_list_observer (*this);
}

NetTracerDialog::~NetTracerDialog ()
{
  view ()->remove_layer_list_observer (*this);

  clear_markers ();
  clear_nets ();
}

void
NetTracerDialog::clear_nets ()
{
  for (std::vector <Net *>::iterator n = mp_nets.begin (); n != mp_nets.end (); ++n) {
    delete *n;
  }
  mp_nets.clear ();
}

void 
NetTracerDialog::item_double_clicked (QListWidgetItem *item)
{
  int item_index = net_list->row (item);
  if (item_index >= 0 && item_index < int (mp_nets.size ())) {

    QString name = mp_nets [item_index]->name ().c_str ();

    bool ok = false;
    name = QInputDialog::getText (this,
                                  tl::translate ("Net Name").c_str (),
                                  tl::translate ("Enter new net name").c_str (),
                                  QLineEdit::Normal, name,
                                  &ok);

    if (ok) {

      mp_nets [item_index]->set_name (name.ascii ());

      m_enable_selection_handler = false;
      update_list ();
      item_selection_changed ();
      m_enable_selection_handler = true;

    }

  }
}

bool 
NetTracerDialog::mouse_move_event (const db::DPoint &p, unsigned int buttons, bool prio) 
{ 
  if (prio && m_mouse_state != 0) {
    set_cursor (lay::Cursor::cross);
  }

  return false;
}

bool 
NetTracerDialog::mouse_click_event (const db::DPoint &p, unsigned int buttons, bool prio) 
{ 
  if (prio && (buttons & lay::LeftButton) != 0 && m_mouse_state != 0) {

    if (m_mouse_state == 2) {

      m_mouse_first_point = p;
      m_mouse_state = 3;

      view ()->message (tl::translate ("Click on the second point in the net"));

    } else {

      bool trace_path = (m_mouse_state == 3);

      if (trace_path || ! sticky_cbx->isChecked ()) {
        release_mouse ();
      }

      //  prepare for the net tracing 
      clear_markers ();

      NetTracerShape start;

      //  locate the seed
      {
        //  compute search box
        double l = double (lay::search_range) / widget ()->mouse_event_trans ().mag ();
        db::DBox search_box = db::DBox (p, p).enlarged (db::DPoint (l, l));

        lay::ShapeFinder finder (true /*point mode*/, false /*all levels*/, db::ShapeIterator::All);

        //  go through all visible layers of all cellviews and find a seed shape
        for (lay::LayerPropertiesConstIterator lprop = view ()->begin_layers (); ! lprop.at_end (); ++lprop) {
          if (lprop->is_visual ()) {
            finder.find (view (), *lprop, search_box);
          }
        }

        //  return, if no shape was found
        lay::ShapeFinder::iterator r = finder.begin ();
        if (r == finder.end ()) {
          return true;
        }

        m_cv_index = r->cv_index ();

        start = NetTracerShape (r->trans (), r->shape (), r->layer (), r->cell_index ());
      } 

      NetTracerShape stop;

      lay::CellView cv = view ()->cellview (m_cv_index);
      if (! cv.is_valid ()) {
        return true;
      }

      // @@@ TODO: use actual technology name from cv's layout
      std::string tech_name;

      //  fetch the net tracer data from the technology and apply to the current layout
      lay::Technology tech = *lay::MainWindow::instance ()->technology_by_name (tech_name);
      NetTracerTechnologyComponent *tech_component = dynamic_cast <NetTracerTechnologyComponent *> (tech.component_by_name ("connectivity"));

      //  if the technology component defines a layer stack, only proceed if the found seed shape is
      //  part of the layer stack:
      if (tech_component && tech_component->begin () != tech_component->end ()) {
        db::LayerProperties lp = cv->layout ().get_properties (start.layer);
        bool is_valid_layer = false;
        for (NetTracerTechnologyComponent::const_iterator c = tech_component->begin (); c != tech_component->end () && ! is_valid_layer; ++c) {
          if (c->layer_a ().log_equal (lp) || c->layer_b ().log_equal (lp) || c->via_layer ().log_equal (lp)) {
            is_valid_layer = true;
          }
        }
        if (! is_valid_layer) {
          return true;
        }
      }

      //  locate the stop shape (the first mouse click)
      if (trace_path) {

        //  compute search box
        double l = double (lay::search_range) / widget ()->mouse_event_trans ().mag ();
        db::DBox search_box = db::DBox (m_mouse_first_point, m_mouse_first_point).enlarged (db::DPoint (l, l));

        lay::ShapeFinder finder (true /*point mode*/, false /*all levels*/, db::ShapeIterator::All);

        //  go through all visible layers of all cellviews and find a stop shape
        for (lay::LayerPropertiesConstIterator lprop = view ()->begin_layers (); ! lprop.at_end (); ++lprop) {

          //  Only evaluate valid layers
          if (lprop->visible (true) && lprop->is_shape_layer () && lprop->cellview_index () == int (m_cv_index)) {

            //  if a connectivity is defined, use only those layers as seed which are 
            //  part of the table.
            bool is_valid_layer = true;
            db::LayerProperties lp = cv->layout ().get_properties (lprop->layer_index ());
            if (tech_component && tech_component->begin () != tech_component->end ()) {
              is_valid_layer = false;
              for (NetTracerTechnologyComponent::const_iterator c = tech_component->begin (); c != tech_component->end () && ! is_valid_layer; ++c) {
                if (c->layer_a ().log_equal (lp) || c->layer_b ().log_equal (lp) || c->via_layer ().log_equal (lp)) {
                  is_valid_layer = true;
                }
              }
            }

            if (is_valid_layer) {
              finder.find (view (), *lprop, search_box);
            }

          }

        }

        //  return, if no shape was found
        lay::ShapeFinder::iterator r = finder.begin ();
        if (r == finder.end ()) {
          return true;
        }

        if (r->cv_index () != m_cv_index) {
          throw tl::Exception (tl::translate ("Both shapes for path tracing must come from the same layout"));
        }

        stop = NetTracerShape (r->trans (), r->shape (), r->layer (), r->cell_index ());

      } 

      //  disable highlighting while the net is looked up.
      m_enable_selection_handler = false;

      try {

        NetTracer net_tracer;

        NetTracerData tracer_data;
        if (tech_component) {
          for (NetTracerTechnologyComponent::const_iterator c = tech_component->begin (); c != tech_component->end (); ++c) {
            tracer_data.add_connection (c->get (cv));
          }
        }

        //  and trace
        try {
          net_tracer.trace (cv, m_cv_index, start, stop, tracer_data);
        } catch (tl::BreakException &) {
          //  just keep the found shapes on break (user abort)
        }

        if (net_tracer.begin () == net_tracer.end ()) {

          net_list->setCurrentItem (0);

        } else {

          //  create a new net taking the shapes from the tracer
          mp_nets.push_back (new Net (net_tracer));

          std::string n = mp_nets.back ()->name ();
          if (n.empty ()) {
            mp_nets.back ()->set_name (tl::sprintf (tl::translate ("Net%d").c_str (), m_net_index++));
          }

          update_list ();
          item_selection_changed ();
          net_list->setCurrentItem (net_list->item (mp_nets.size () - 1));

        }

        m_enable_selection_handler = true;

      } catch (...) {
        m_enable_selection_handler = true;
        throw;
      }

    }

    return true;

  } else {
    return false; 
  }
}

bool 
NetTracerDialog::configure (const std::string &name, const std::string &value)
{
  bool need_update = false;
  bool taken = true;

  if (name == lay::cfg_background_color) {

    need_update = true;

  } else if (name == cfg_nt_window_mode) {

    nt_window_type window = m_window;
    NetTracerWindowModeConverter ().from_string (value, window);
    need_update = lay::test_and_set (m_window, window);

  } else if (name == cfg_nt_window_dim) {

    double wdim = m_window_dim;
    tl::from_string (value, wdim);
    if (fabs (wdim - m_window_dim) > 1e-6) {
      m_window_dim = wdim;
      need_update = true;
    }

  } else if (name == cfg_nt_max_shapes_highlighted) {

    unsigned int mc = 0;
    tl::from_string (value, mc);
    need_update = lay::test_and_set (m_max_marker_count, mc);

  } else if (name == cfg_nt_marker_color) {

    QColor color;
    if (! value.empty ()) {
      lay::ColorConverter ().from_string (value, color);
    }

    if (color != m_marker_color) {
      m_marker_color = color;
      need_update = true;
    }

  } else if (name == cfg_nt_marker_line_width) {

    int lw = 0;
    tl::from_string (value, lw);

    if (lw != m_marker_line_width) {
      m_marker_line_width = lw;
      need_update = true;
    }

  } else if (name == cfg_nt_marker_vertex_size) {

    int vs = 0;
    tl::from_string (value, vs);

    if (vs != m_marker_vertex_size) {
      m_marker_vertex_size = vs;
      need_update = true;
    }

  } else if (name == cfg_nt_marker_halo) {

    int halo = 0;
    tl::from_string (value, halo);

    if (halo != m_marker_halo) {
      m_marker_halo = halo;
      need_update = true;
    }

  } else if (name == cfg_nt_marker_dither_pattern) {

    int dp = 0;
    tl::from_string (value, dp);

    if (dp != m_marker_dither_pattern) {
      m_marker_dither_pattern = dp;
      need_update = true;
    }

  } else if (name == cfg_nt_marker_intensity) {

    int bo = 0;
    tl::from_string (value, bo);

    if (bo != m_marker_intensity) {
      m_marker_intensity = bo;
      need_update = true;
    }

  } else {
    taken = false;
  }

  if (active () && need_update) {
    update_highlights ();
    update_info ();
  }

  return taken;
}

void  
NetTracerDialog::menu_activated (const std::string &symbol)
{
  if (symbol == "ext::net_trace") {

    lay::CellView cv = view ()->cellview (view ()->active_cellview_index ());
    if (cv.is_valid ()) {
      show ();
      activate ();
    }

  } else {
    lay::Browser::menu_activated (symbol);
  }
}

void 
NetTracerDialog::item_selection_changed ()
{
  update_highlights ();
  update_info ();
}

void
NetTracerDialog::detailed_mode_clicked ()
{
  update_info ();
}

void
NetTracerDialog::update_info ()
{
  bool detailed = detailed_cb->isChecked ();

  std::ostrstream info_stream;
  tl::XMLWriter info (info_stream); 
  
  info.start_document ("");
  info.start_element ("html");
  info.start_element ("body");

  // @@@ TODO: use actual technology name
  std::string tech_name;

  lay::Technology tech = *lay::MainWindow::instance ()->technology_by_name (tech_name);
  NetTracerTechnologyComponent *tech_component = dynamic_cast <NetTracerTechnologyComponent *> (tech.component_by_name ("connectivity"));
  if (!tech_component || tech_component->begin () == tech_component->end ()) {
    info.start_element ("p");
    info.write_attribute ("style", "color:red; font-weight: bold");
    info.cdata (tl::translate ("No layer stack defined - use \"Layer Stack\" button to open the dialog"));
    info.end_element ("p");
  }

  QList<QListWidgetItem *> selected_items = net_list->selectedItems ();
  if (selected_items.size () == 0) {

    info.start_element ("p");
    info.cdata (tl::translate ("No net selected"));
    info.end_element ("p");

  } else {

    size_t ntot = 0;

    info.start_element ("p");
    bool first = true;

    for (QList<QListWidgetItem *>::const_iterator item = selected_items.begin (); item != selected_items.end (); ++item) {

      int item_index = net_list->row (*item);
      if (item_index >= 0 && item_index < int (mp_nets.size ())) {

        if (! first) {
          info.start_element ("br");
          info.end_element ("br");
        }
        first = false;

        info.cdata (tl::sprintf (tl::translate ("%ld Shapes").c_str (), mp_nets[item_index]->size ()));
        if (selected_items.size () > 1) {
          info.cdata (" (" + mp_nets[item_index]->name () + ")");
        }
        if (mp_nets[item_index]->incomplete ()) {
          info.start_element ("span");
          info.write_attribute ("style", "color:red; font-weight: bold");
          info.cdata(" (" + tl::translate ("Net is incomplete") + ") ");
          info.end_element ("span");
        }

        ntot += mp_nets[item_index]->size ();

      }
    }

    info.end_element ("p");

    if (selected_items.size () > 1) {
      info.start_element ("p");
      info.cdata (tl::sprintf (tl::translate ("%ld Shapes (total)").c_str (), ntot));
      info.end_element ("p");
    }

    if (ntot > m_max_marker_count) {
      info.start_element ("p");
      info.write_attribute ("style", "color:red; font-weight: bold");
      info.cdata (tl::translate ("Not all shapes are highlighted"));
      info.end_element ("p");
    }

    if (selected_items.size () == 1) {

      int item_index = net_list->row (selected_items [0]);
      if (item_index >= 0 && item_index < int (mp_nets.size ())) {

        const lay::CellView &cv = mp_nets[item_index]->cellview ();
        if (cv.is_valid ()) {

          db::CplxTrans dbu (cv->layout ().dbu ());
          db::CplxTrans dbuinv (1.0 / cv->layout ().dbu ());

          size_t max_labels = 1000;
          size_t max_cells = 1000;
          size_t max_shapes = 2000;

          if (detailed) {

            info.start_element ("h3");
            info.cdata (tl::translate ("General:"));
            info.end_element ("h3");

            info.start_element ("p");
            info.start_element ("b");
            info.cdata (tl::translate ("Net name: "));
            info.end_element ("b");
            info.cdata (mp_nets[item_index]->name ());
            info.start_element ("br");
            info.end_element ("br");
            info.start_element ("b");
            info.cdata (tl::translate ("Top cell: "));
            info.end_element ("b");
            info.cdata (cv->layout ().cell_name (cv.cell_index ()));
            info.start_element ("br");
            info.end_element ("br");
            info.start_element ("b");
            info.cdata (tl::translate ("Layout: "));
            info.end_element ("b");
            info.cdata (cv->name ());
            info.start_element ("br");
            info.end_element ("br");
            info.start_element ("b");
            info.cdata (tl::translate ("Layout file: "));
            info.end_element ("b");
            info.cdata (cv->filename ());
            info.end_element ("p");

            bool incomplete = false;
            std::map<std::string, std::set<std::string> > shapes;

            size_t tot_shapes = 0;
            for (Net::iterator net_shape = mp_nets [item_index]->begin (); net_shape != mp_nets [item_index]->end (); ++net_shape) {

              if (tot_shapes++ >= max_shapes) {
                incomplete = true;
                break;
              }

              std::string c (std::string (cv->layout ().cell_name (net_shape->cell_index)));
              c += " (with ";
              c += (dbu * net_shape->trans * dbuinv).to_string ();
              c += ")";

              std::string t;
              std::string l (cv->layout ().get_properties (net_shape->layer).to_string ());

              if (net_shape->shape.is_text ()) {
                db::Text text;
                net_shape->shape.text (text);
                t = tl::translate ("text on ") + l + ": " + (dbu * text).to_string ();
              } else if (net_shape->shape.is_box ()) {
                db::Box box;
                net_shape->shape.box (box);
                t = tl::translate ("box on ") + l + ": " + (dbu * box).to_string ();
              } else if (net_shape->shape.is_path ()) {
                db::Path path;
                net_shape->shape.path (path);
                t = tl::translate ("path on ") + l + ": " + (dbu * path).to_string ();
              } else if (net_shape->shape.is_polygon ()) {
                db::Polygon polygon;
                net_shape->shape.polygon (polygon);
                t = tl::translate ("polygon on ") + l + ": " + (dbu * polygon).to_string ();
              }

              if (! t.empty ()) {
                shapes.insert (std::make_pair (c, std::set<std::string> ())).first->second.insert (t);
              }

            }

            if (! shapes.empty ()) {

              info.start_element ("h3");
              info.cdata (tl::translate ("Shapes:"));
              info.end_element ("h3");

              for (std::map<std::string, std::set<std::string> >::const_iterator s = shapes.begin (); s != shapes.end (); ++s) {

                info.start_element ("p");

                info.start_element ("b");
                info.cdata (tl::translate ("Cell "));
                info.cdata (s->first);
                info.cdata (":");
                info.end_element ("b");

                for (std::set <std::string>::const_iterator l = s->second.begin (); l != s->second.end (); ++l) {
                  info.start_element ("br");
                  info.end_element ("br");
                  info.cdata (*l);
                }

                info.end_element ("p");

              }

              if (incomplete) {
                info.start_element ("p");
                info.cdata ("...");
                info.end_element ("p");
              }

            }

          } else {

            bool incomplete = false;
            std::set<std::string> labels;

            for (Net::iterator net_shape = mp_nets [item_index]->begin (); net_shape != mp_nets [item_index]->end (); ++net_shape) {

              if (net_shape->shape.is_text ()) {

                if (labels.size () >= max_labels) {
                  incomplete = true;
                  break;
                }

                std::string t (std::string (cv->layout ().cell_name (net_shape->cell_index)) + "." + net_shape->shape.text_string ());
                if (detailed) {
                  t += tl::translate (" (at ");
                  t += (dbu * net_shape->trans * db::CplxTrans (net_shape->shape.text ().trans ())).trans (db::DPoint (0, 0)).to_string ();
                  t += ")";
                }
                labels.insert (t);

              }

            }

            if (! labels.empty ()) {

              info.start_element ("h3");
              info.cdata (tl::translate ("Labels:"));
              info.end_element ("h3");

              info.start_element ("p");

              for (std::set <std::string>::const_iterator l = labels.begin (); l != labels.end (); ++l) {
                if (l != labels.begin ()) {
                  info.start_element ("br");
                  info.end_element ("br");
                }
                info.cdata (*l);
              }

              if (incomplete) {
                info.start_element ("br");
                info.end_element ("br");
                info.cdata ("...");
              }

              info.end_element ("p");

            }

            incomplete = false;
            std::set<std::string> cells;

            for (Net::iterator net_shape = mp_nets [item_index]->begin (); net_shape != mp_nets [item_index]->end (); ++net_shape) {

              if (cells.size () >= max_cells) {
                incomplete = true;
                break;
              }

              std::string t (cv->layout ().cell_name (net_shape->cell_index));
              if (detailed) {
                t += tl::translate (" (with ");
                t += (dbu * net_shape->trans * dbuinv).to_string ();
                t += ")";
              }
              cells.insert (t);
            }

            if (! cells.empty ()) {

              info.start_element ("h3");
              info.cdata (tl::translate ("Cells on net:"));
              info.end_element ("h3");

              info.start_element ("p");

              for (std::set <std::string>::const_iterator c = cells.begin (); c != cells.end (); ++c) {
                if (c != cells.begin ()) {
                  info.start_element ("br");
                  info.end_element ("br");
                }
                info.cdata (*c);
              }

              if (incomplete) {
                info.start_element ("br");
                info.end_element ("br");
                info.cdata ("...");
              }

              info.end_element ("p");

            }

          }

        }

      }

    }

  }

  info.end_element ("body");
  info.end_element ("html");

  info_stream << std::ends;
  net_info_text->setHtml (info_stream.str ());
}

void
NetTracerDialog::update_list ()
{
  net_list->clear ();

  for (size_t i = 0; i < mp_nets.size (); ++i) {

    QListWidgetItem *item = new QListWidgetItem (net_list);
    item->setData (Qt::DisplayRole, mp_nets [i]->name ().c_str ());

    net_list->addItem (item);

  }
}

void 
NetTracerDialog::trace_path_button_clicked ()
{
BEGIN_PROTECTED
  net_list->setCurrentItem (0);
  m_mouse_state = 2;
  view ()->message (tl::translate ("Click on the first point in the net"));
  widget ()->grab_mouse (this, false);
END_PROTECTED
}

void 
NetTracerDialog::trace_net_button_clicked ()
{
BEGIN_PROTECTED
  net_list->setCurrentItem (0);
  m_mouse_state = 1;
  view ()->message (tl::translate ("Click on a point in the net"));
  widget ()->grab_mouse (this, false);
END_PROTECTED
}

void
NetTracerDialog::sticky_mode_clicked ()
{
BEGIN_PROTECTED
  if (! sticky_cbx->isChecked ()) {
    release_mouse ();
  } else {
    trace_net_button_clicked ();
  }
END_PROTECTED
}

void 
NetTracerDialog::release_mouse ()
{
  add_pb->setChecked (false);
  add2_pb->setChecked (false);
  m_mouse_state = 0;
  view ()->message ();
  widget ()->ungrab_mouse (this);
}

void
NetTracerDialog::clear_all_button_clicked ()
{
BEGIN_PROTECTED
  release_mouse ();

  if (QMessageBox::question (this, tl::translate ("Clear All Nets").c_str (),
                             tl::translate ("Are you sure to delete all nets?\nThis operation cannot be undone.").c_str (),
                             QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes) {

    m_enable_selection_handler = false;

    clear_markers ();
    clear_nets ();
    update_list ();
    
    m_enable_selection_handler = true;

  }
END_PROTECTED
}

void 
NetTracerDialog::delete_button_clicked ()
{
BEGIN_PROTECTED
  release_mouse ();

  m_enable_selection_handler = false;

  std::vector<int> to_delete;

  QList<QListWidgetItem *> selected_items = net_list->selectedItems ();
  for (QList<QListWidgetItem *>::const_iterator item = selected_items.begin (); item != selected_items.end (); ++item) {
    int item_index = net_list->row (*item);
    if (item_index >= 0 && item_index < int (mp_nets.size ())) {
      to_delete.push_back (item_index);
    }
  }

  std::sort (to_delete.begin (), to_delete.end ());

  for (std::vector<int>::const_iterator d = to_delete.end (); d != to_delete.begin (); ) {
    --d;
    delete mp_nets [*d];
    mp_nets.erase (mp_nets.begin () + *d);
  }

  clear_markers ();
  update_list ();
    
  m_enable_selection_handler = true;
END_PROTECTED
}

void 
NetTracerDialog::layer_stack_clicked ()
{
BEGIN_PROTECTED

  release_mouse ();

  // @@@ TODO: use actual technology name
  std::string tech_name;

  //  create a temporary copy
  lay::Technology tech = *lay::MainWindow::instance ()->technology_by_name (tech_name);

  //  call the dialog and if successful, install the new technology
  lay::TechComponentSetupDialog dialog (this, &tech, "connectivity");
  if (dialog.exec ()) {
    lay::MainWindow::instance ()->set_technology (tech_name, tech);
  }

END_PROTECTED
}

struct cv2cellname 
{
  std::string to_string (const lay::CellView &cv)
  {
    if (!cv.is_valid ()) { 
      return std::string ();
    } else {
      return cv->layout ().cell_name (cv.cell_index ());
    }
  }

  lay::CellView from_string (const std::string &s) { return lay::CellView (); }
};


void 
NetTracerDialog::export_text_clicked ()
{
BEGIN_PROTECTED

  release_mouse ();

  int cv_index = view ()->active_cellview_index ();
  lay::CellView cv = view ()->cellview (cv_index);
  if (cv.is_valid ()) {

    QList<QListWidgetItem *> selected_items = net_list->selectedItems ();
    if (selected_items.size () == 0) {
      throw tl::Exception (tl::translate ("No net selected to export"));
    }

    if (mp_export_file_dialog->get_save (m_export_file_name)) {

      std::ofstream os (m_export_file_name.c_str ());
      tl::XMLWriter w (os);

      w.start_document ();

      w.start_element ("nets");

      for (QList<QListWidgetItem *>::const_iterator item = selected_items.begin (); item != selected_items.end (); ++item) {

        int item_index = net_list->row (*item);
        if (item_index >= 0 && item_index < int (mp_nets.size ())) {

          w.start_element ("net");

          const Net *net = mp_nets[item_index];

          w.start_element ("name");
          w.cdata (net->name ());
          w.end_element ("name");

          w.start_element ("top_cell");
          w.cdata (net->cellview ()->layout ().cell_name (net->cellview ().cell_index ()));
          w.end_element ("top_cell");

          w.start_element ("layout");
          w.cdata (net->cellview ()->filename ());
          w.end_element ("layout");

          w.start_element ("dbu");
          w.cdata (tl::to_string (net->cellview ()->layout ().dbu ()));
          w.end_element ("dbu");

          w.start_element ("complete");
          w.cdata (tl::to_string (! net->incomplete ()));
          w.end_element ("complete");

          w.start_element ("shapes");

          for (Net::iterator net_shape = net->begin (); net_shape != net->end (); ++net_shape) {

            w.start_element ("element");

            w.start_element ("layer");
            w.cdata (net->cellview ()->layout ().get_properties (net_shape->layer).to_string ());
            w.end_element ("layer");

            w.start_element ("cell");
            w.cdata (net->cellview ()->layout ().cell_name (net_shape->cell_index));
            w.end_element ("cell");

            w.start_element ("trans");
            w.cdata (net_shape->trans.to_string ());
            w.end_element ("trans");

            w.start_element ("shape");
            w.cdata (net_shape->shape.to_string ());
            w.end_element ("shape");

            w.end_element ("element");

          }

          w.end_element ("shapes");

          w.end_element ("net");

        }

      }

      w.end_element ("nets");

      w.end_document ();

    }

  }

END_PROTECTED
}

void 
NetTracerDialog::export_clicked ()
{
BEGIN_PROTECTED

  release_mouse ();

  int cv_index = view ()->active_cellview_index ();
  lay::CellView cv = view ()->cellview (cv_index);
  if (cv.is_valid ()) {

    QList<QListWidgetItem *> selected_items = net_list->selectedItems ();
    if (selected_items.size () == 0) {
      throw tl::Exception (tl::translate ("No net selected to export"));
    }

    bool ok;
    QString cell_name = m_export_cell_name.c_str ();
    cell_name = QInputDialog::getText (this, tl::translate ("Export Net").c_str (), tl::translate ("Export net to cell named").c_str (), QLineEdit::Normal, cell_name, &ok);
    if (ok) {

      if (cell_name.isEmpty ()) {
        throw tl::Exception (tl::translate ("No cell was specified"));
      }

      //  Clear undo buffers if layout is created.
      view ()->manager ()->clear ();

      m_export_cell_name = cell_name.ascii ();
      std::pair<bool, db::cell_index_type> cbn = cv->layout ().cell_by_name (m_export_cell_name.c_str ());
      if (! cbn.first) {
        cbn.second = cv->layout ().add_cell (m_export_cell_name.c_str ());
      } 

      db::Cell &export_cell = cv->layout ().cell (cbn.second);

      for (QList<QListWidgetItem *>::const_iterator item = selected_items.begin (); item != selected_items.end (); ++item) {

        int item_index = net_list->row (*item);
        if (item_index >= 0 && item_index < int (mp_nets.size ())) {

          const lay::CellView &cv_src = mp_nets[item_index]->cellview ();

          std::map<unsigned int, unsigned int> layer_map;

          for (Net::iterator net_shape = mp_nets[item_index]->begin (); net_shape != mp_nets[item_index]->end (); ++net_shape) {

            std::map<unsigned int, unsigned int>::const_iterator lm = layer_map.find (net_shape->layer);
            if (lm == layer_map.end ()) {

              int layer_index = -1;
              for (db::Layout::layer_iterator l = cv->layout ().begin_layers (); l != cv->layout ().end_layers (); ++l) {
                if (cv->layout ().get_properties ((*l).first).log_equal (cv_src->layout ().get_properties (net_shape->layer))) {
                  layer_index = int ((*l).first);
                  break;
                }
              }

              if (layer_index < 0) {

                layer_index = int (cv->layout ().insert_layer (cv_src->layout ().get_properties (net_shape->layer)));

                //  Add a new entry in the layer list
                lay::LayerProperties props;
                props.set_source (lay::ParsedLayerSource (cv->layout ().get_properties (layer_index), cv_index));
                view ()->init_layer_properties (props);
                view ()->insert_layer (view ()->end_layers (), props);

              }

              lm = layer_map.insert (std::make_pair (net_shape->layer, (unsigned int)layer_index)).first;

            }

            db::Shape new_shape = export_cell.shapes (lm->second).insert (net_shape->shape);
            export_cell.shapes (lm->second).transform (new_shape, db::ICplxTrans (net_shape->trans));

          }

        }

      }

      view ()->select_cell (export_cell.cell_index (), view ()->active_cellview_index ());

    }

  }

END_PROTECTED
}

void 
NetTracerDialog::configure_clicked ()
{
BEGIN_PROTECTED
  lay::ConfigurationDialog config_dialog (this, main_window (), "NetTracerPlugin");
  config_dialog.exec ();
END_PROTECTED
}

void  
NetTracerDialog::deactivated ()
{
  clear_markers ();
  release_mouse ();
}

void  
NetTracerDialog::activated ()
{
  // .. nothing yet ..
}

lay::ViewService * 
NetTracerDialog::view_service_interface ()
{
  return this;
}

void
NetTracerDialog::signal_observed ()
{
  update_highlights ();
}

void
NetTracerDialog::update_highlights ()
{
  clear_markers ();

  size_t n_marker = 0;
  db::DBox bbox;

  QList<QListWidgetItem *> selected_items = net_list->selectedItems ();
  for (QList<QListWidgetItem *>::const_iterator item = selected_items.begin (); item != selected_items.end (); ++item) {

    int item_index = net_list->row (*item);

    if (item_index >= 0 && item_index < int (mp_nets.size ())) {

      int cv_index = mp_nets[item_index]->cellview_index ();

      std::map<unsigned int, std::vector<db::CplxTrans> > tv_by_layer = view ()->cv_transform_variants_by_layer (cv_index);
      std::map<unsigned int, lay::LayerPropertiesConstIterator> layer_props;

      const lay::CellView &cv = mp_nets[item_index]->cellview ();
      if (cv.is_valid () && &cv->layout () == &view ()->cellview (cv_index)->layout ()) {

        db::DBox cv_bbox;

        //  Create markers for the shapes 
        for (Net::iterator net_shape = mp_nets [item_index]->begin (); net_shape != mp_nets [item_index]->end () && n_marker < m_max_marker_count; ++net_shape) {

          std::map<unsigned int, std::vector<db::CplxTrans> >::const_iterator tv = tv_by_layer.find (net_shape->layer);
          if (tv != tv_by_layer.end ()) {

            lay::LayerPropertiesConstIterator original;

            std::map<unsigned int, lay::LayerPropertiesConstIterator>::const_iterator lp_cache = layer_props.find (net_shape->layer);
            if (lp_cache != layer_props.end ()) {
              original = lp_cache->second;
            } else {
              for (lay::LayerPropertiesConstIterator lp = view ()->begin_layers (); !lp.at_end (); ++lp) {
                if (!lp->has_children () && lp->cellview_index () == int (cv_index) && lp->layer_index () == int (net_shape->layer)) {
                  layer_props.insert (std::make_pair (net_shape->layer, lp));
                  original = lp;
                  break;
                }
              }
            }

            mp_markers.push_back (new lay::ShapeMarker (view (), cv_index));
            mp_markers.back ()->set (net_shape->shape, net_shape->trans, tv->second);

            if (! original.at_end ()) {
              mp_markers.back ()->set_line_width (original->width (true));
              mp_markers.back ()->set_vertex_size (1);
              mp_markers.back ()->set_dither_pattern (original->dither_pattern (true));
              if (view ()->background_color ().green () < 128) {
                mp_markers.back ()->set_color (original->eff_fill_color_brighter (true, (m_marker_intensity * 255) / 100));
                mp_markers.back ()->set_frame_color (original->eff_frame_color_brighter (true, (m_marker_intensity * 255) / 100));
              } else {
                mp_markers.back ()->set_color (original->eff_fill_color_brighter (true, (-m_marker_intensity * 255) / 100));
                mp_markers.back ()->set_frame_color (original->eff_frame_color_brighter (true, (-m_marker_intensity * 255) / 100));
              }
            }

            if (m_marker_color.isValid ()) {
              mp_markers.back ()->set_color (m_marker_color);
              mp_markers.back ()->set_frame_color (m_marker_color);
            }

            if (m_marker_line_width >= 0) {
              mp_markers.back ()->set_line_width (m_marker_line_width);
            }

            if (m_marker_vertex_size >= 0) {
              mp_markers.back ()->set_vertex_size (m_marker_vertex_size);
            }

            if (m_marker_halo >= 0) {
              mp_markers.back ()->set_halo (m_marker_halo);
            }

            if (m_marker_dither_pattern >= 0) {
              mp_markers.back ()->set_dither_pattern (m_marker_dither_pattern);
            }

            db::Box shape_box = net_shape->shape.bbox ();
            for (std::vector<db::CplxTrans>::const_iterator t = tv->second.begin (); t != tv->second.end (); ++t) {
              cv_bbox += *t * net_shape->trans * shape_box;
            }

            ++n_marker;

          }

        }

        bbox += cv_bbox * cv->layout ().dbu ();

      }

    }

  }

  if (! bbox.empty ()) {
    if (m_window == NTFitNet) {

      view ()->zoom_box (bbox.enlarged (db::DPoint (m_window_dim, m_window_dim)));

    } else if (m_window == NTCenter) {

      view ()->pan_center ((bbox.p1 () + bbox.p2 ()) * 0.5);

    } else if (m_window == NTCenterSize) {

      double w = std::max (bbox.width (), m_window_dim);
      double h = std::max (bbox.height (), m_window_dim);
      db::DPoint center ((bbox.p1 () + bbox.p2 ()) * 0.5);
      db::DPoint d (w * 0.5, h * 0.5);
      view ()->zoom_box (db::DBox (center - d, center + d));

    }
  }

}

void
NetTracerDialog::clear_markers ()
{
  for (std::vector <lay::ShapeMarker *>::iterator m = mp_markers.begin (); m != mp_markers.end (); ++m) {
    delete *m;
  }

  mp_markers.clear ();
}

// -----------------------------------------------------------------------------------
//  NetTracerPlugin definition and implementation

class NetTracerPluginDeclaration
  : public lay::PluginDeclaration,
    public lay::TechnologyComponentProvider
{
public:
  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const 
  {
    options.push_back (std::pair<std::string, std::string> (cfg_nt_window_mode, "fit-net"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_window_dim, "1.0"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_max_shapes_highlighted, "10000"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_marker_color, lay::ColorConverter ().to_string (QColor ())));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_marker_line_width, "-1"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_marker_vertex_size, "-1"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_marker_halo, "-1"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_marker_dither_pattern, "-1"));
    options.push_back (std::pair<std::string, std::string> (cfg_nt_marker_intensity, "50"));
  }

  virtual std::vector<std::pair <std::string, lay::ConfigPage *> > config_pages (QWidget *parent) const 
  {
    std::vector<std::pair <std::string, lay::ConfigPage *> > pages;
    pages.push_back (std::make_pair (tl::translate ("Other Tools|Net Tracer"), new NetTracerConfigPage (parent)));
    return pages;
  }

  virtual void get_menu_entries (std::vector<lay::MenuEntry> &menu_entries) const
  {
    // @@@ where should that go?
    lay::PluginDeclaration::get_menu_entries (menu_entries);
    menu_entries.push_back (lay::MenuEntry ("net_trace_group", "tools_menu.end"));
    menu_entries.push_back (lay::MenuEntry ("ext::net_trace", "net_trace", "tools_menu.end", tl::translate ("Trace Net")));
  }
 
  virtual lay::Plugin *create_plugin (db::Manager *manager, lay::MainWindow *main, lay::LayoutView *view) const
  {
    return new NetTracerDialog (main, view);
  }

  virtual const lay::TechnologyComponentProvider *technology_component_provider () const 
  {
    return this;
  }

  virtual lay::TechnologyComponent *create_component () const
  {
    return new NetTracerTechnologyComponent ();
  }
};

static tl::RegisteredClass<lay::PluginDeclaration> config_decl (new NetTracerPluginDeclaration (), 13000, "NetTracerPlugin");

}

