
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_extNetTracerDialog
#define HDR_extNetTracerDialog

#include "NetTracerDialog.h"
#include "NetTracerTechComponentEditor.h"

#include "extNetTracer.h"
#include "extNetTracerConfig.h"

#include "layBrowser.h"
#include "layPlugin.h"
#include "layViewObject.h"
#include "layMarker.h"
#include "layTechnology.h"

#include "tlObserver.h"

namespace lay
{
  class FileDialog;
}

namespace ext
{

class NetTracerConnectionInfo
{
public:
  NetTracerConnectionInfo ();
  NetTracerConnectionInfo (const db::LayerProperties &la, const db::LayerProperties &lb);
  NetTracerConnectionInfo (const db::LayerProperties &la, const db::LayerProperties &via, const db::LayerProperties &lb);

  NetTracerConnection get (const lay::CellView &cv) const;

  std::string to_string () const;
  void parse (tl::Extractor &ex);

  const db::LayerProperties &layer_a () const 
  {
    return m_la;
  }

  void set_layer_a (const db::LayerProperties &l) 
  {
    m_la = l;
  }

  const db::LayerProperties &via_layer () const 
  {
    return m_via;
  }

  void set_via_layer (const db::LayerProperties &l) 
  {
    m_via = l;
  }

  const db::LayerProperties &layer_b () const 
  {
    return m_lb;
  }

  void set_layer_b (const db::LayerProperties &l) 
  {
    m_lb = l;
  }

private:
  db::LayerProperties m_la, m_via, m_lb;
};

class Net
{
public:
  typedef std::vector <NetTracerShape>::const_iterator iterator;

  Net (const NetTracer &tracer);

  iterator begin () const
  {
    return m_net_shapes.begin ();
  }

  iterator end () const
  {
    return m_net_shapes.end ();
  }

  size_t size () const
  {
    return m_net_shapes.size ();
  }

  /**
   *  @brief Get a name for the net
   *
   *  The name can be empty if not specific net name could be determined (i.e. from a label or property)
   */
  std::string name () const
  {
    return m_name;
  }

  /**
   *  @brief Set a name for the net
   */
  void set_name (const std::string &n)
  {
    m_name = n;
  }

  /**
   *  @brief Get the cellview from which this net was taken
   */
  const lay::CellView &cellview () const
  {
    return m_cv;
  }

  /**
   *  @brief Get the cellview index from which this net was taken
   */
  int cellview_index () const
  {
    return m_cv_index;
  }

  /**
   *  @brief Returns true, if the net is incomplete
   *
   *  This flag is true if the extractor was aborted
   *  for example by the user.
   *  The shapes do not fully cover the net.
   */
  bool incomplete () const
  {
    return m_incomplete;
  }

private:
  lay::CellView m_cv;
  int m_cv_index;
  std::string m_name;
  bool m_incomplete;
  std::vector <NetTracerShape> m_net_shapes;
  db::Shapes m_shapes;
};

class NetTracerDialog
  : public lay::Browser,
    public lay::ViewService, 
    public Ui::NetTracerDialog,
    public tl::Observer
{
Q_OBJECT

public:
  NetTracerDialog (lay::MainWindow *main_window, lay::LayoutView *view);
  virtual ~NetTracerDialog ();

  virtual bool mouse_move_event (const db::DPoint &p, unsigned int buttons, bool prio);
  virtual bool mouse_click_event (const db::DPoint &p, unsigned int buttons, bool prio);
  virtual void menu_activated (const std::string &symbol);
  virtual lay::ViewService *view_service_interface ();
  virtual void deactivated ();
  virtual void activated ();
  virtual bool configure (const std::string &name, const std::string &value);

protected slots:
  void trace_net_button_clicked ();
  void trace_path_button_clicked ();
  void delete_button_clicked ();
  void clear_all_button_clicked ();
  void layer_stack_clicked ();
  void export_clicked ();
  void export_text_clicked ();
  void detailed_mode_clicked ();
  void sticky_mode_clicked ();
  void configure_clicked ();
  void item_selection_changed ();
  void item_double_clicked (QListWidgetItem *item);

private:
  std::vector <Net *> mp_nets;
  std::vector <lay::ShapeMarker *> mp_markers;
  unsigned int m_cv_index;
  int m_net_index;
  bool m_enable_selection_handler;
  nt_window_type m_window;
  double m_window_dim;
  unsigned int m_max_marker_count;
  QColor m_marker_color;
  int m_marker_line_width;
  int m_marker_vertex_size;
  int m_marker_halo;
  int m_marker_dither_pattern;
  int m_marker_intensity;
  db::DPoint m_mouse_first_point;
  int m_mouse_state;
  std::string m_export_cell_name;
  lay::FileDialog *mp_export_file_dialog;
  std::string m_export_file_name;

  void update_highlights ();
  void clear_markers ();
  void clear_nets ();
  void update_list ();
  void update_info ();
  void signal_observed ();
  void release_mouse ();
};

class NetTracerTechnologyComponent
  : public lay::TechnologyComponent
{
public:
  typedef std::vector<NetTracerConnectionInfo>::const_iterator const_iterator;
  typedef std::vector<NetTracerConnectionInfo>::iterator iterator;

  NetTracerTechnologyComponent ();
  NetTracerTechnologyComponent (const NetTracerTechnologyComponent &d);

  const_iterator begin () const
  {
    return m_connections.begin ();
  }

  iterator begin ()
  {
    return m_connections.begin ();
  }

  const_iterator end () const
  {
    return m_connections.end ();
  }

  iterator end ()
  {
    return m_connections.end ();
  }

  void clear ()
  {
    m_connections.clear ();
  }

  void erase (iterator p)
  {
    m_connections.erase (p);
  }

  void insert (iterator p, const NetTracerConnectionInfo &info)
  {
    m_connections.insert (p, info);
  }

  void add (const NetTracerConnectionInfo &info)
  {
    m_connections.push_back (info);
  }

  size_t size () const
  {
    return m_connections.size ();
  }

  virtual std::string to_string () const;
  virtual void parse (tl::Extractor &ex);
  virtual lay::TechnologyComponentEditor *create_editor (QWidget *parent);

  lay::TechnologyComponent *clone () const
  {
    return new NetTracerTechnologyComponent (*this);
  }

private:
  std::vector<NetTracerConnectionInfo> m_connections;
};

class NetTracerTechComponentEditor
  : public lay::TechnologyComponentEditor,
    public Ui::NetTracerTechComponentEditor
{
Q_OBJECT

public:
  NetTracerTechComponentEditor (QWidget *parent);

  void commit ();
  void setup ();

public slots:
  void add_clicked ();
  void del_clicked ();
  void move_up_clicked ();
  void move_down_clicked ();

private:
  NetTracerTechnologyComponent m_data;

  void update ();
};

}

#endif

