
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsi
#define _HDR_gsi

/**
 *  @brief GSI - the generic scripting interface
 *
 *  The generic scripting interface provides a way to export C++ classes to a scripting 
 *  language in some generic way. The main target scripting language is Ruby, but the concept
 *  is not restricted to this. 
 *
 *  Terms used herein:
 *  - scripting language: the language to which the C++ objects shall be exposed
 *  - scripting client: the interpreter that provides an implementation of the scripting language
 *  - exposure: the process of making a class available to the scripting client.
 *  - binding: the process and fact of providing extensions to a certain scripting language, 
 *      availability of the extension to a certain scripting languate
 *  - object: a C++ object being exposed to the scripting language
 *  - class: a C++ class declaration being exposed to the scripting language
 *  - ownership: bascially responsibility to free resources for an object. 
 *      An object cannot have multiple owners. If an object is not owned, there is no distinct
 *      responsibility for freeing the resources of an object. In a good design, this state should
 *      be avoided in general.
 *
 *  The basic design targets for this implementation were:
 *  - dynamically extendable: This means, a class and method set should be extendable on runtime,
 *    i.e. when loading a plugin library. This way, a plugin may provide it's own set of classes
 *    without knowing about the application or other plugin's classes.
 *  - the method of exporting a class or method should not be dependend on the choice of
 *    the target scripting language. This way, multiple target language bindings may be possible
 *    without changing the export declarations.
 *  - ideally, the export mechanism should be based purely on C++ methods. This simplifies the
 *    build process.
 *
 *  All files declaring classes which use GSI features (gsi::Object, callbacks, events) must include
 *  "gsi.h", all files declaring GSI interfaces must include "gsiDecl.h".
 *
 *  This implementation fulfils these requirements by being based on a abstract interface 
 *  and a declaration mechanism that uses statically instantiated C++ objects and heavily makes use
 *  of template mechanism to bind methods to abstract declarations.
 *
 *
 *  1.) Basic declaration for exposure
 *
 *  A class declaration basically consists of the instantiation of a certain template and
 *  in particular of proving a static member of this template. The following code piece
 *  shows the declaration of a class "A" with two methods to the scripting language:
 *
 *  //  C++ declaration of the class to expose (usually somewhere else)
 *  class A {
 *  public:
 *    //  class to expose to the scripting interface
 *    void meth1 (int);            //  one example method
 *    std::string meth2 () const;  //  second example method
 *  };
 *
 *  //  Declaration of the class for exposure to the scripting client
 *  template <>
 *  gsi::Class<A> gsi::Class<A>::decl ("A", 
 *    gsi::method ("meth1", &A::meth1) +
 *    gsi::method ("meth2", &A::meth2) 
 *  );
 *
 *  Note the '+' operators that form the list of methods. If a class does not exhibit any
 *  methods to expose, no list must be given or it must be specified as "gsi::Methods()", i.e:
 *
 *  template <> gsi::Class<A> gsi::Class<A>::decl ("A");
 *
 *  or
 *
 *  template <> gsi::Class<A> gsi::Class<A>::decl ("A", gsi::Methods ());
 *
 *  Being useful but not required, the class exposed may be derived from the class
 *  gsi::Object<X> where X is the class being exposed:
 *
 *    class A : public gsi::Object<X> { ... };
 *
 *  If a class that is exposed is derived from gsi::Object<X>, additional features are
 *  enabled for the class. For Ruby for example, the objects created this way can be
 *  extended by Ruby member in a consistent way. In addition, the Ruby objects this
 *  way receive notification when the C++ object is destroyed. This allows the Ruby
 *  object to act gracefully in this case.
 *
 *  2.) Static and external methods
 *
 *  The declaration may expose static methods in a very simple way. Instead of specifying the 
 *  method pointer as the second argument to "method", simply specify the static method:
 *  
 *  //  C++ declaration of the class to expose (usually somewhere else)
 *  class A {
 *  public:
 *    //  class to expose to the scripting interface
 *    static void meth1 (int);     //  one example method
 *    std::string meth2 () const;  //  second example method
 *  };
 *
 *  //  Declaration of the class for exposure to the scripting client
 *  template <>
 *  gsi::Class<A> gsi::Class<A>::decl ("A", 
 *    gsi::method ("meth1", &A::meth1) +
 *    gsi::method ("meth2", &A::meth2) 
 *  );
 *
 *  External methods are methods that appear as methods on the client
 *  side but are implemented as static functions on the C++ side.
 *  Such methods can be declared with the gsi::method_ext and gsi::iterator_ext
 *  declarations:
 *
 *  void f(A *a, int i);  // external method
 *
 *  //  Declaration of the class for exposure of f as a method
 *  template <>
 *  gsi::Class<A> gsi::Class<A>::decl ("A", 
 *    gsi::method_ext ("f", &f) 
 *  );
 *
 *  Constructors are special static methods that "new" an object and
 *  return a reference to that object which then is owned by the scripting client.
 *  Such a constructor can be declared with the "constructor" method declaration:
 *  
 *  A *ctor (int i);
 *
 *  //  Declaration of the class for exposure of f as a method
 *  template <>
 *  gsi::Class<A> gsi::Class<A>::decl ("A", 
 *    gsi::constructor ("new", &ctor) 
 *  );
 *
 *  Using this declaration is the only way to provide constructors that take arguments
 *  for classes that cannot be copied.
 *
 *  3.) Specifying one variant from a set of polymorphic methods
 *
 *  C++ allows to declare multiple methods with different signatures but same name.
 *  A certain version may be selected by casting the method pointer to the desired
 *  variant:
 *
 *  //  C++ declaration of the class to expose (usually somewhere else)
 *  class A {
 *  public:
 *    //  class to expose to the scripting interface
 *    void meth1 (int);            //  one example method
 *    int meth1 ();                //  this one is not supposed to be used
 *    std::string meth2 () const;  //  second example method
 *  };
 *
 *  //  Declaration of the class for exposure to the scripting client
 *  template <>
 *  gsi::Class<A> gsi::Class<A>::decl ("A", 
 *    gsi::method ("meth1", (void (A::*)(int)) &A::meth1) +  //  explicitly select the first version
 *    gsi::method ("meth2", &A::meth2) 
 *  );
 *
 *
 *  4.) Changing the assumptions made for a class
 *
 *  By default, it is assumed that a class can be copied and created with an default constructor.
 *  For some classes however, this may not be given. To change the assumptions made for a class,
 *  insert a specialization of "gsi::ClassTag<X>" specifying two typedef's, "copy_semantics" and
 *  "create_semantics". These can be set to "gsi::CantCopySemanticsTag" and "gsi::CantDefaultCreateSemanticsTag"
 *  to specify that objects of the given class cannot be copied or created with a default constructor
 *  respectively.
 *
 *  The following declaration will tell the declaration processor that the class "A_NC" does not provide
 *  copy constructors or assignment operators and no default constructor.
 *
 *  namespace gsi
 *  {
 *    template <> struct ClassTag<A_NC> {
 *    typedef gsi::CantCopySemanticsTag copy_semantics;
 *    typedef gsi::CantDefaultCreateSemanticsTag create_semantics;
 *  };
 *
 *  template <>
 *  gsi::Class<A_NC> gsi::Class<A_NC>::decl ("A_NC",
 *    ...
 *  );
 *
 *
 *  5.) Inheritance:
 *
 *  A class can be declared to be derived from one base class by providing the respective declaration
 *  member as the first object to the "decl" constructor:
 *
 *  template <>
 *  gsi::Class<B> gsi::Class<B>::decl (gsi::Class<A>::decl, "B",
 *    ...
 *  );
 *
 *  This applies only if the base class is exposed as well. The effect of this declaration is that
 *  "B" will inherit all methods of "A" without having to specify them again. On scripting languages
 *  that provide a inheritance concept as well, "B" will be seen as being derived from "A" as well.
 *  The above specification requires that "B" is derived from "A" in the C++ sense as well.
 *  Only derivation from one base class is supported.
 *
 *
 *  6.) Iterators:
 *
 *  For scripting languages that provide iterator support (through "yield" on ruby for example), the
 *  declaration may declare "begin..end" iterator delivery method pairs in a special way that enables
 *  the scripting language iterator feature:
 *
 *  //  C++ declaration of the class to expose (usually somewhere else)
 *  class A {
 *  public:
 *    typedef ... iterator;
 *    iterator begin_strings ();   //  one example method
 *    iterator end_strings ();     //  one example method
 *    ...
 *  };
 *
 *  //  Declaration of the class for exposure to the scripting client
 *  template <>
 *  gsi::Class<A> gsi::Class<A>::decl ("A", 
 *    gsi::iterator ("strings", &A::begin_strings, &A::end_strings) +
 *    ...
 *  );
 *
 *  On ruby for this example, a single "strings" method is provided that iterates
 *  over all strings and yields the associated block for each string.
 *
 *  Iterator methods may also take arguments.
 *
 *
 *  7.) Implementing virtual methods in the scripting client
 *
 *  The GSI framework provides a way to allow virtual functions of C++ objects being
 *  implemented on the scripting client's side by providing a "callback" mechanism.
 *
 *  The basic concept is this: on C++ side, a "stub" provides an implementation for the
 *  virtual methods in C++ but redirects the call to a callback connector object that 
 *  passes the call further onto the scripting side. There, a suitable implementation is
 *  looked for an the method is called. The return value is then passed back to the C++
 *  side. 
 *
 *  This approach requires two measures:
 *  - A C++ stub class that provides the C++ implementation and certain connector objects
 *    of type "gsi::Callback" as public members.
 *  - A method declaration that declares the reimplemented method and the connector
 *    object.
 *
 *  The following example illustrates this and the recommended way of implementing the stub
 *  method:
 *
 *  //  C++ declaration of the class to expose (usually somewhere else)
 *  class C
 *  {
 *  public:
 *    virtual int f (const std::string &s);
 *  };
 *
 *  //  The stub class providing a potential override for "f":
 *  class C_Stub : public C
 *  {
 *  public:
 *    virtual int f (const std::string &s)
 *    {
 *      return gsi::can_issue (f_cb) ? gsi::issue (f_cb, &C_Stub::f, s) : C::f (s);
 *    }
 *  
 *    gsi::Callback f_cb;
 *  };
 *
 *  //  The declaration of the stub class to GSI
 *  template <>
 *  gsi::Class<C_Stub> gsi::Class<C_Stub>::decl ("C", 
 *    gsi::callback ("f", &C_Stub::f, &C_Stub::f_cb) 
 *  );
 *
 *  gsi::issue provides the mechanism of passing control over to the scripting client.
 *  It receives the callback connector object, the method pointer of the method that is exposed
 *  and an argument list of the arguments to pass to the scripting client. The above reimplementation
 *  of "f" first tries to find out if there is actually a reimplementation available using 
 *  "gsi::can_issue". If not, it spawns to the base classe's implementation.
 *  A reimplementation may not be available in some cases, i.e if no adequate method is found.
 *
 *  For the above example, method "f" can be reimplemented in ruby this way:
 *
 *  class C_IMPL < C
 *    def f(s)
 *      s.size    # example reimplementation delivering the size
 *    end
 *  end
 *
 *  8.) Extending a declaration
 *
 *  The standard declaration provided by Class<X>::decl can be extended by 
 *  additional declarations provided by other static Class<X> objects. These 
 *  additional declarations are merged with the standard declarations in a
 *  preparation step on the client side. This way, external compilation units
 *  may extend classes of the main application, preferably with external
 *  method declarations.
 */

#include "tlAssert.h"
#include "tlInternational.h"
#include "tlException.h"

#include "gsiObject.h"
#include "gsiTypes.h"
#include "gsiSerialisation.h"
#include "gsiCallback.h"
#include "gsiEvents.h"

#endif

