
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsiDecl.h"
#include "tlAssert.h"

namespace gsi
{

ClassBase::class_collection *ClassBase::mp_class_collection = 0;
unsigned int ClassBase::m_class_count = 0;

ClassBase::ClassBase (const std::string &name, const ClassBase *base, const std::string &doc, const Methods &mm)
  : m_name (name), mp_base (base), m_doc (doc)
{ 
  for (method_iterator m = mm.begin (); m != mm.end (); ++m) {
    add_method ((*m)->clone ());
  }
  if (base) {
    for (method_iterator m = base->begin_methods (); m != base->end_methods (); ++m) {
      add_method ((*m)->clone (), true /*from base class*/);
    }
  }
  if (! mp_class_collection) {
    mp_class_collection = new class_collection ();
  }
  mp_class_collection->insert (this, false);
  ++m_class_count;
}

ClassBase::~ClassBase ()
{
  tl_assert (m_class_count > 0);
  if (--m_class_count == 0) {
    delete mp_class_collection;
    mp_class_collection = 0;
  }
}

bool 
ClassBase::is_derived_from (const ClassBase *base) const
{
  if (base == this) {
    return true;
  } else if (mp_base == 0) {
    return false;
  } else {
    return mp_base->is_derived_from (base);
  }
}

const ClassBase::class_collection &
ClassBase::collection () 
{
  if (mp_class_collection) {
    return *mp_class_collection;
  } else {
    static class_collection empty;
    return empty;
  }
}

void
ClassBase::merge_declarations ()
{
  if (! mp_class_collection) {
    return;
  }

  //  merge the extensions to the main declaration
  static bool merged = false;
  if (! merged) {

    //  HINT: this code block must be called exactly ONCE!
    merged = true;

    std::vector <const gsi::ClassBase *> ext_declarations;

    for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {
      //  consider only extensions
      if (c->declaration () != &*c) {
        for (gsi::ClassBase::method_iterator m = c->begin_methods (); m != c->end_methods (); ++m) {
          c->declaration ()->add_method ((*m)->clone ());
        }
        ext_declarations.push_back (&*c);
      }
    }

    //  removed auxiliary classes that have been merged with the "actual" one
    for (std::vector <const gsi::ClassBase *>::const_iterator ed = ext_declarations.begin (); ed != ext_declarations.end (); ++ed) {
      mp_class_collection->remove (const_cast<gsi::ClassBase *> (*ed));
    }

  }
}

void
ClassBase::add_method (MethodBase *method, bool base_class)
{
  //  take apart the name by synonyms
  //  The name may be composed of several synonym names like "<name1>|<name2>|...".
  //  Each synonym name may start with a "#" indicating a deprecated version.
  tl::Extractor ex (method->name ().c_str ());
  while (! ex.at_end ()) {

    ex.test ("#");  //  deprecated flag
    std::string name;
    ex.read (name, "|");
    ex.test ("|");
  
    //  insert the method into the overloaded name map
    NameMap::iterator nm = m_name_map.find (name);
    if (nm != m_name_map.end ()) {
      //  The base classes methods are completely overwritten by the derived class - since the derived class
      //  is entered first, we simply do not extend the overloaded base class method list for base class methods.
      if (! base_class) {
        m_overloaded_methods [nm->second].push_back (m_methods.size ());
      }
    } else {
      m_name_map.insert (std::make_pair (name, (unsigned int) m_overloaded_methods.size ()));
      m_overloaded_names.push_back (name);
      m_overloaded_methods.push_back (std::vector<unsigned int> ());
      m_overloaded_methods.back ().push_back (m_methods.size ());
    }
    
  }

  m_methods.add_method (method);
}

}

