
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiClassBase
#define _HDR_gsiClassBase

#include "tlCollection.h"
#include "tlException.h"

#include <list>
#include <map>
#include <vector>

//  For a comprehensive documentation see gsi.h

namespace gsi
{

/**
 *  @brief The basic object to declare a class
 */
class KLAYOUT_DLL ClassBase
  : public tl::Collectable
{
public:
  typedef tl::Collection<ClassBase> class_collection; 
  typedef class_collection::const_iterator class_iterator;
  typedef Methods::iterator method_iterator;
  typedef std::map<std::string, unsigned int> NameMap;
  typedef NameMap::const_iterator name_map_iterator;
  typedef std::vector< std::vector<unsigned int> > OverloadedMethodsList;
  typedef std::vector<unsigned int>::const_iterator overloaded_methods_iterator;
  
  ClassBase (const std::string &name, const ClassBase *base, const std::string &doc, const Methods &mm);

  virtual ~ClassBase ();

  virtual ClassBase *declaration () const = 0;

  void add_method (MethodBase *method, bool base_class = false);

  const std::string &name () const
  {
    return m_name;
  }

  const std::string &doc () const
  {
    return m_doc;
  }

  /**
   *  @brief Merge base and extension declarations
   *
   *  This method must be called at least once!
   */
  static void merge_declarations ();

  static class_iterator begin_classes () 
  {
    return collection ().begin ();
  }
 
  static class_iterator end_classes () 
  {
    return collection ().end ();
  }
 
  method_iterator begin_methods () const
  {
    return m_methods.begin ();
  }

  method_iterator end_methods () const
  {
    return m_methods.end ();
  }

  name_map_iterator begin_name_map () const
  {
    return m_name_map.begin ();
  }

  name_map_iterator end_name_map () const
  {
    return m_name_map.end ();
  }

  overloaded_methods_iterator begin_overloaded_methods (unsigned int index) const
  {
    return m_overloaded_methods [index].begin ();
  }

  overloaded_methods_iterator end_overloaded_methods (unsigned int index) const
  {
    return m_overloaded_methods [index].end ();
  }

  const std::string &overloaded_name (unsigned int index) const
  {
    return m_overloaded_names [index];
  }

  bool is_derived_from (const ClassBase *base) const;

  virtual void destroy (void *obj) const = 0;
  virtual void *create () const = 0;
  virtual void assign (void *, void *) const = 0;
  virtual void *take_and_create (SerialArgs &from) const = 0;
  virtual void get_vector_of (SerialArgs &from, bool ref, void *data, void (*cb) (void *data, void *obj)) const = 0;
  virtual void push_vector_of (SerialArgs &to, bool ref, std::list<TempObject> &temp_objects, const std::vector<void *> &objects) const = 0;
  virtual void copy_to (SerialArgs &at, void *from) const = 0;
  virtual void *create_obj_iterator (SerialArgs &from) const = 0;
  virtual bool obj_iterator_at_end (void *iter) const = 0;
  virtual void *take_iterator_obj (void *iter) const = 0;
  virtual const void *get_iterator_obj (void *iter) const = 0;
  virtual void inc_obj_iterator (void *iter) const = 0;
  virtual void delete_obj_iterator (void *iter) const = 0;
  virtual void *client_data (void *p) const = 0;
  virtual bool can_copy () const = 0;
  virtual bool can_default_create () const = 0;
  virtual void register_client_handler (ClientHandler *handler) const = 0;

protected:
  static const class_collection &collection ();

private:
  std::string m_name;
  const ClassBase *mp_base;
  std::string m_doc;
  static class_collection *mp_class_collection;
  static unsigned int m_class_count;
  Methods m_methods;
  NameMap m_name_map;
  OverloadedMethodsList m_overloaded_methods;
  std::vector<std::string> m_overloaded_names;
};

}

#endif


