
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "gsiDecl.h"
#include "antObject.h"
#include "antService.h"
#include "layLayoutView.h"

namespace gsi
{

static ant::Object *new_ruler ()
{
  return new ant::Object ();
}

static int style_ruler ()       { return int (ant::Object::STY_ruler); }
static int style_arrow_end ()   { return int (ant::Object::STY_arrow_end); }
static int style_arrow_start () { return int (ant::Object::STY_arrow_start); }
static int style_arrow_both ()  { return int (ant::Object::STY_arrow_both); }
static int style_line ()        { return int (ant::Object::STY_line); }

static int outline_diag ()      { return int (ant::Object::OL_diag); }
static int outline_xy ()        { return int (ant::Object::OL_xy); }
static int outline_diag_xy ()   { return int (ant::Object::OL_diag_xy); }
static int outline_yx ()        { return int (ant::Object::OL_yx); }
static int outline_diag_yx ()   { return int (ant::Object::OL_diag_yx); }
static int outline_box ()       { return int (ant::Object::OL_box); }

static int angle_any ()         { return int (lay::AC_Any); }
static int angle_diagonal ()    { return int (lay::AC_Diagonal); }
static int angle_ortho ()       { return int (lay::AC_Ortho); }
static int angle_horizontal ()  { return int (lay::AC_Horizontal); }
static int angle_vertical ()    { return int (lay::AC_Vertical); }
static int angle_global ()      { return int (lay::AC_Global); }

static int get_style (const ant::Object *obj)
{
  return int (obj->style ());
}

static void set_style (ant::Object *obj, int style)
{
  obj->style (ant::Object::style_type (style));
}

static int get_outline (const ant::Object *obj)
{
  return int (obj->outline ());
}

static void set_outline (ant::Object *obj, int outline)
{
  obj->outline (ant::Object::outline_type (outline));
}

static int get_angle_constraint (const ant::Object *obj)
{
  return int (obj->angle_constraint ());
}

static void set_angle_constraint (ant::Object *obj, int angle_constraint)
{
  obj->angle_constraint (lay::angle_constraint_type (angle_constraint));
}


template <>
gsi::Class<ant::Object> gsi::Class<ant::Object>::decl ("Annotation", 
  gsi::constructor ("new", &gsi::new_ruler,
    "@brief Create a new ruler or marker with the default attributes"
  ) +
  gsi::method ("style_ruler", &gsi::style_ruler,
    "@brief style_ruler code used by the \\style method"
  ) +
  gsi::method ("style_arrow_end", &gsi::style_arrow_end,
    "@brief style_arrow_end code used by the \\style method"
  ) +
  gsi::method ("style_arrow_start", &gsi::style_arrow_start,
    "@brief style_arrow_start code used by the \\style method"
  ) +
  gsi::method ("style_arrow_both", &gsi::style_arrow_both,
    "@brief style_arrow_both code used by the \\style method"
  ) +
  gsi::method ("style_line", &gsi::style_line,
    "@brief style_line code used by the \\style method"
  ) +
  gsi::method ("outline_diag", &gsi::outline_diag,
    "@brief outline_diag code used by the \\outline method"
  ) +
  gsi::method ("outline_xy", &gsi::outline_xy,
    "@brief outline_xy code used by the \\outline method"
  ) +
  gsi::method ("outline_diag_xy", &gsi::outline_diag_xy,
    "@brief outline_diag_xy code used by the \\outline method"
  ) +
  gsi::method ("outline_yx ", &gsi::outline_yx ,
    "@brief outline_yx  code used by the \\outline method"
  ) +
  gsi::method ("outline_diag_yx ", &gsi::outline_diag_yx ,
    "@brief outline_diag_yx  code used by the \\outline method"
  ) +
  gsi::method ("outline_box", &gsi::outline_box,
    "@brief outline_box code used by the \\outline method"
  ) +
  gsi::method ("angle_any", &gsi::angle_any,
    "@brief angle_any code used by the \\angle_constraint method"
  ) +
  gsi::method ("angle_diagonal", &gsi::angle_diagonal,
    "@brief angle_diagonal code used by the \\angle_constraint method"
  ) +
  gsi::method ("angle_ortho", &gsi::angle_ortho,
    "@brief angle_ortho code used by the \\angle_constraint method"
  ) +
  gsi::method ("angle_horizontal", &gsi::angle_horizontal,
    "@brief angle_horizontal code used by the \\angle_constraint method"
  ) +
  gsi::method ("angle_vertical", &gsi::angle_vertical,
    "@brief angle_vertical code used by the \\angle_constraint method"
  ) +
  gsi::method ("angle_global", &gsi::angle_global,
    "@brief angle_global code used by the \\angle_constraint method.\n"
    "This code will tell the ruler or marker to use the angle constraint defined globally."
  ) +
  gsi::method ("p1", (const db::DPoint & (ant::Object::*) () const) &ant::Object::p1,
    "@brief Get the first point of the ruler or marker\n"
    "The points of the ruler or marker are always given in micron units in floating-point "
    "coordinates.\n"
    "@return The first point\n"
  ) +
  gsi::method ("p2", (const db::DPoint & (ant::Object::*) () const) &ant::Object::p2,
    "@brief Get the second point of the ruler or marker\n"
    "The points of the ruler or marker are always given in micron units in floating-point "
    "coordinates.\n"
    "@return The second point\n"
  ) +
  gsi::method ("p1=", (void (ant::Object::*) (const db::DPoint &)) &ant::Object::p1,
    "@brief Set the first point of the ruler or marker\n"
    "The points of the ruler or marker are always given in micron units in floating-point "
    "coordinates.\n"
    "@args point\n"
  ) +
  gsi::method ("p2=", (void (ant::Object::*) (const db::DPoint &)) &ant::Object::p2,
    "@brief Set the second point of the ruler or marker\n"
    "The points of the ruler or marker are always given in micron units in floating-point "
    "coordinates.\n"
    "@args point\n"
  ) +
  gsi::method ("box", &ant::Object::box,
    "@brief Get the bounding box of the object (not including text)\n"
    "@return The bounding box\n"
  ) +
  gsi::method ("transformed", &ant::Object::transformed<db::DTrans>,
    "@brief Transform the ruler or marker with the given simple transformation\n"
    "@args t\n"
    "@param t The transformation to apply\n"
    "@return The transformed object\n"
  ) +
  gsi::method ("transformed_cplx", &ant::Object::transformed<db::DCplxTrans>,
    "@brief Transform the ruler or marker with the given complex transformation\n"
    "@args t\n"
    "@param t The magnifying transformation to apply\n"
    "@return The transformed object\n"
  ) +
  gsi::method ("transformed_cplx", &ant::Object::transformed<db::ICplxTrans>,
    "@brief Transform the ruler or marker with the given complex transformation\n"
    "@args t\n"
    "@param t The magnifying transformation to apply\n"
    "@return The transformed object (in this case an integer coordinate object)\n"
    "\n"
    "This method has been introduced in version 0.18.\n"
  ) +
  gsi::method ("fmt=", (void (ant::Object::*) (const std::string &)) &ant::Object::fmt,
    "@brief Set the format used for the label\n"
    "@args format\n"
    "@param format The format string\n"
  ) +
  gsi::method ("fmt", (const std::string & (ant::Object::*) () const) &ant::Object::fmt,
    "@brief Returns the format used for the label\n"
    "@return The format string\n"
  ) +
  gsi::method ("fmt_x=", (void (ant::Object::*) (const std::string &)) &ant::Object::fmt_x,
    "@brief Set the format used for the x-axis label\n"
    "@args format\n"
    "X-axis labels are only used for styles that have a horizontal component. "
    "@param format The format string\n"
  ) +
  gsi::method ("fmt_x", (const std::string & (ant::Object::*) () const) &ant::Object::fmt_x,
    "@brief Returns the format used for the x-axis label\n"
    "@return The format string\n"
  ) +
  gsi::method ("fmt_y=", (void (ant::Object::*) (const std::string &)) &ant::Object::fmt_y,
    "@brief Set the format used for the y-axis label\n"
    "@args format\n"
    "Y-axis labels are only used for styles that have a vertical component. "
    "@param format The format string\n"
  ) +
  gsi::method ("fmt_y", (const std::string & (ant::Object::*) () const) &ant::Object::fmt_y,
    "@brief Returns the format used for the y-axis label\n"
    "@return The format string\n"
  ) +
  gsi::method_ext ("style=", &gsi::set_style, 
    "@brief Set the style used for drawing the annotation object\n"
    "@args style\n"
    "The style_... values can be used for defining the annotation object's style. The style determines "
    "if ticks or arrows are drawn."
  ) +
  gsi::method_ext ("style", &gsi::get_style,
    "@brief Returns the style of the annotation object\n"
  ) +
  gsi::method_ext ("outline=", &gsi::set_outline,
    "@brief Set the outline style used for drawing the annotation object\n"
    "@args outline\n"
    "The outline_... values can be used for defining the annotation object's outline. The "
    "outline style determines what components are drawn. "
  ) +
  gsi::method_ext ("outline", &gsi::get_outline,
    "@brief Returns the outline style of the annotation object\n"
  ) +
  gsi::method ("snap=", (void (ant::Object::*) (bool)) &ant::Object::snap,
    "@brief Set the 'snap to objects' attribute.\n"
    "@args flag\n"
    "If this attribute is set to true, the ruler or marker snaps to other objects when moved. "
  ) +
  gsi::method ("snap?", (bool (ant::Object::*) () const) &ant::Object::snap,
    "@brief Return the 'snap to objects' attribute.\n"
  ) +
  gsi::method_ext ("angle_constraint=", &gsi::set_angle_constraint,
    "@brief Set the angle constraint attribute.\n"
    "@args flag\n"
    "This attribute controls if an angle constraint is applied when moving one of the ruler's "
    "points. The angle_... values can be used for this purpose." 
  ) +
  gsi::method_ext ("angle_constraint", &gsi::get_angle_constraint,
    "@brief Return the angle constraint attribute.\n"
    "See \\angle_constraint= for a more detailed description."
  ) +
  gsi::method ("text_x", &ant::Object::text_x,
    "@brief Return the formatted text for the x-axis label."
  ) +
  gsi::method ("text_y", &ant::Object::text_y,
    "@brief Return the formatted text for the y-axis label."
  ) +
  gsi::method ("text", &ant::Object::text,
    "@brief Return the formatted text for the main label."
  ) +
  gsi::method ("to_s", &ant::Object::to_string,
    "@brief Returns the string representation of the ruler."
    "\n"
    "This method was introduced in version 0.19."
  ) +
  gsi::method ("==", &ant::Object::operator==,
    "@brief Equality operator"
  ) +
  gsi::method ("!=", &ant::Object::operator!=,
    "@brief Inequality operator"
  ),
  "@brief This class implements an \"annotation object\"\n"
  "\n"
  "Annotation objects provide a way to attach measurements or descriptive informations to a layout view. "
  "Annotation objects can appear as rulers for example. Annotation objects can be configured in "
  "different ways using the styles provided. By configuring an annotation object properly, it can appear "
  "as a rectangle or a plain line for example."
);

static void clear_annotations (lay::LayoutView *view)
{
  ant::Service *ant_service = view->get_plugin <ant::Service> ();
  if (ant_service) {
    ant_service->clear_rulers ();
  }
}

static void insert_annotation (lay::LayoutView *view, const ant::Object &obj)
{
  ant::Service *ant_service = view->get_plugin <ant::Service> ();
  if (ant_service) {
    ant_service->insert_ruler (obj, false /*do not observe the ruler count limit*/);
  }
}

static ant::AnnotationIterator begin_annotations (lay::LayoutView *view)
{
  ant::Service *ant_service = view->get_plugin <ant::Service> ();
  if (ant_service) {
    return ant_service->begin_annotations ();
  } else {
    return ant::AnnotationIterator ();
  }
}

static
gsi::Class<lay::LayoutView> layout_view_decl ("LayoutView", 
  gsi::method_ext ("clear_annotations", &gsi::clear_annotations, 
    "@brief Clear all annotations on this view"
  ) +
  gsi::method_ext ("insert_annotation", &gsi::insert_annotation, 
    "@brief Insert an annotation object into the given view\n"
    "@args obj\n"
    "Insert the annotation object given by obj into the view. "
  ) +
  gsi::iterator_ext ("each_annotation", &gsi::begin_annotations, 
    "@brief Iterate over all annotations attached to this view"
  ),
  ""
);

class AnnotationSelectionIterator 
{
public:
  typedef ant::Object value_type;
  typedef std::map<ant::Service::obj_iterator, unsigned int>::const_iterator iterator_type;
  typedef void pointer; 
  typedef const value_type &reference;
  typedef std::forward_iterator_tag iterator_category;
  typedef void difference_type;

  AnnotationSelectionIterator (const std::vector<ant::Service *> &services) 
    : m_services (services), m_service (0)
  {
    if (! m_services.empty ()) {
      m_iter = m_services [m_service]->selection ().begin ();
      next ();
    }
  }

  bool at_end () const
  {
    return (m_service >= m_services.size ());
  }

  AnnotationSelectionIterator &operator++ ()
  {
    ++m_iter;
    next ();
    return *this;
  }

  const value_type &operator* () const
  {
    return *(dynamic_cast<const ant::Object *> (m_iter->first->ptr ()));
  }

private:
  std::vector<ant::Service *> m_services;
  unsigned int m_service;
  iterator_type m_iter;

  void next ()
  {
    while (m_iter == m_services [m_service]->selection ().end ()) {
      ++m_service;
      if (m_service < m_services.size ()) {
        m_iter = m_services [m_service]->selection ().begin ();
      } else {
        break;
      }
    }
  }
};

//  extend the layout view by "edtService" specific methods 

static bool has_annotation_selection (const lay::LayoutView *view)
{
  std::vector<ant::Service *> ant_services = view->get_plugins <ant::Service> ();
  for (std::vector<ant::Service *>::const_iterator s = ant_services.begin (); s != ant_services.end (); ++s) {
    if ((*s)->has_selection ()) {
      return true;
    }
  }
  return false;
}

static AnnotationSelectionIterator begin_annotations_selected (const lay::LayoutView *view)
{
  return AnnotationSelectionIterator (view->get_plugins <ant::Service> ());
}


static
gsi::Class<lay::LayoutView> layout_view_decl2 ("LayoutView", 
  gsi::method_ext ("has_annotation_selection?", &has_annotation_selection, 
    "@brief Returns true, if annotations (rulers) are selected in this view"
    "\n"
    "This method was introduced in version 0.19."
  ) +
  gsi::iterator_ext ("each_annotation_selected", &begin_annotations_selected,
    "@brief Iterate over each selected annotation objects, yielding a \\Annotation object for each of them"
    "\n"
    "This method was introduced in version 0.19."
  ),
  ""
);

}

