
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsiDecl.h"
#include "dbLayout.h"
#include "dbClip.h"
#include "dbLayoutUtils.h"
#include "dbWriter.h"
#include "dbSaveLayoutOptions.h"
#include "tlStream.h"

namespace gsi
{

// ---------------------------------------------------------------
//  db::LayerProperties binding

static const std::string &lp_get_name (const db::LayerProperties *lp)
{
  return lp->name;
}

static void lp_set_name (db::LayerProperties *lp, const std::string &s)
{
  lp->name = s;
}

static int lp_get_layer (const db::LayerProperties *lp)
{
  return lp->layer;
}

static void lp_set_layer (db::LayerProperties *lp, int n)
{
  lp->layer = n;
}

static int lp_get_datatype (const db::LayerProperties *lp)
{
  return lp->datatype;
}

static void lp_set_datatype (db::LayerProperties *lp, int n)
{
  lp->datatype = n;
}

static 
db::LayerProperties *ctor_layer_info_default ()
{
  return new db::LayerProperties ();
}

static 
db::LayerProperties *ctor_layer_info_ld (int layer, int datatype)
{
   db::LayerProperties *ret = new db::LayerProperties ();
   ret->layer = layer;
   ret->datatype = datatype;
   return ret;
}

static 
db::LayerProperties *ctor_layer_info_ldn (int layer, int datatype, const std::string &name)
{
   db::LayerProperties *ret = new db::LayerProperties ();
   ret->layer = layer;
   ret->datatype = datatype;
   ret->name = name;
   return ret;
}

static 
db::LayerProperties *ctor_layer_info_name (const std::string &name)
{
   db::LayerProperties *ret = new db::LayerProperties ();
   ret->name = name;
   return ret;
}

//  since there already exists a "LayerProperties" object, we call this one "LayerInfo"
template <>
Class<db::LayerProperties> Class<db::LayerProperties>::decl ("LayerInfo", 
  gsi::constructor ("new", &ctor_layer_info_default, 
    "@brief The default constructor.\n"
    "Creates a default \\LayerInfo object.\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::constructor ("new", &ctor_layer_info_ld, 
    "@brief The constructor for a layer/datatype pair.\n"
    "@args layer,datatype\n"
    "Creates a \\LayerInfo object representing a layer and datatype.\n"
    "@param layer The layer number\n"
    "@param datatype The datatype number\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::constructor ("new", &ctor_layer_info_name, 
    "@brief The constructor for a named layer.\n"
    "@args name\n"
    "Creates a \\LayerInfo object representing a named layer.\n"
    "@param name The name\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::constructor ("new", &ctor_layer_info_ldn, 
    "@brief The constructor for a named layer with layer and datatype.\n"
    "@args layer,datatype,name\n"
    "Creates a \\LayerInfo object representing a named layer with layer and datatype.\n"
    "@param layer The layer number\n"
    "@param datatype The datatype number\n"
    "@param name The name\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::method ("to_s", &db::LayerProperties::to_string, 
    "@brief Convert the layer info object to a string\n"
    "@return The string\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::method ("==", &db::LayerProperties::operator==, 
    "@brief Compare two layer info objects\n"
    "@return True, if both are equal\n"
    "@args b\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::method ("!=", &db::LayerProperties::operator!=, 
    "@brief Compare two layer info objects\n"
    "@return True, if both are not equal\n"
    "@args b\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::method ("is_equivalent?", &db::LayerProperties::log_equal, 
    "@brief Equivalence of two layer info objects\n"
    "@return True, if both are equivalent\n"
    "@args b\n"
    "\n"
    "First, layer and datatype are compared. The name is of second order and used only if no layer or datatype is given.\n"
    "This is basically a weak comparison that reflects the search preferences.\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::method ("is_named?", &db::LayerProperties::is_named, 
    "@brief Returns true, if the layer is purely specified by name.\n"
    "@return True, if no layer or datatype is given.\n"
    "\n"
    "This method was added in version 0.18.\n"
  ) +
  gsi::method_ext ("name=", &lp_set_name, 
    "@brief Set the layer name\n"
    "The name is set on OASIS input for example, if the layer has a name.\n"
    "@args name"
  ) +
  gsi::method_ext ("name", &lp_get_name, 
    "@brief Gets the layer name\n"
  ) +
  gsi::method_ext ("layer=", &lp_set_layer, 
    "@brief Sets the layer number\n"
    "@args layer\n"
  ) +
  gsi::method_ext ("layer", &lp_get_layer, 
    "@brief Gets the layer number\n"
  ) +
  gsi::method_ext ("datatype=", &lp_set_datatype, 
    "@brief Set the datatype\n"
    "@args datatype"
  ) +
  gsi::method_ext ("datatype", &lp_get_datatype, 
    "@brief Gets the datatype\n"
  ),
  "@brief A structure encapsulating the layer properties\n"
  "\n"
  "The layer properties describe how a layer is stored in a GDSII or OASIS file for example."
);

// ---------------------------------------------------------------
//  db::Layout binding

static db::cell_index_type cell_by_name (db::Layout *l, const char *name)
{
  std::pair<bool, db::cell_index_type> c = l->cell_by_name (name);
  if (! c.first) {
    throw tl::Exception (tl::sprintf (tl::translate ("No such cell: '%s'").c_str (), name));
  }
  return c.second;
}

static db::cell_index_type clip (db::Layout *l, db::cell_index_type c, const db::Box &box)
{
  std::vector <db::Box> boxes;
  boxes.push_back (box);
  std::vector <db::cell_index_type> cc = db::clip_layout(*l, *l, c, boxes);
  tl_assert (! cc.empty ());
  return cc [0];
}

static db::cell_index_type clip_into (const db::Layout *l, db::cell_index_type c, db::Layout *t, const db::Box &box)
{
  std::vector <db::Box> boxes;
  boxes.push_back (box);
  std::vector <db::cell_index_type> cc = db::clip_layout(*l, *t, c, boxes);
  tl_assert (! cc.empty ());
  return cc [0];
}

static std::vector<db::cell_index_type> multi_clip (db::Layout *l, db::cell_index_type c, const std::vector<db::Box> &boxes)
{
  return db::clip_layout(*l, *l, c, boxes);
}

static std::vector<db::cell_index_type> multi_clip_into (db::Layout *l, db::cell_index_type c, db::Layout *t, const std::vector<db::Box> &boxes)
{
  return db::clip_layout(*l, *t, c, boxes);
}

static std::vector<unsigned int> layer_indices (const db::Layout *l)
{
  std::vector<unsigned int> layers;
  for (unsigned int i = 0; i < l->layers (); ++i) {
    if (l->is_valid_layer (i)) {
      layers.push_back (i);
    }
  }
  return layers;
}

static db::properties_id_type properties_id (db::Layout *layout, const std::vector<tl::Variant> &properties)
{
  db::PropertiesRepository::properties_set props;

  for (std::vector<tl::Variant>::const_iterator v = properties.begin (); v != properties.end (); ++v) {
    if (! v->is_list () || v->get_list ().size () != 2) {
      throw tl::Exception (tl::translate ("Expected a list of pairs of variants (found at least one that is not a pair)"));
    }
    db::property_names_id_type name_id = layout->properties_repository ().prop_name_id (v->get_list ()[0]);
    props.insert (std::make_pair (name_id, v->get_list () [1]));
  }

  return layout->properties_repository ().properties_id (props);
}

static std::vector<tl::Variant> properties (const db::Layout *layout, db::properties_id_type id)
{
  std::vector<tl::Variant> ret;

  if (layout->properties_repository ().is_valid_properties_id (id)) {

    const db::PropertiesRepository::properties_set &props = layout->properties_repository ().properties (id);
    ret.reserve (props.size ());

    for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
      ret.push_back (tl::Variant::empty_list ());
      ret.back ().get_list ().reserve (2);
      ret.back ().get_list ().push_back (layout->properties_repository ().prop_name (p->first));
      ret.back ().get_list ().push_back (p->second);
    }

  }

  return ret;
}

static void 
delete_cells (db::Layout *layout, const std::vector<db::cell_index_type> &cell_indices)
{
  layout->delete_cells (cell_indices.begin (), cell_indices.end ());
}

static void 
prune_cell (db::Layout *layout, db::cell_index_type cell_index, int levels)
{
  layout->prune_cell (cell_index, levels);
}

static void 
prune_subcells (db::Layout *layout, db::cell_index_type cell_index, int levels)
{
  layout->prune_subcells (cell_index, levels);
}

static void 
flatten (db::Layout *layout, db::cell_index_type cell_index, int levels, bool prune)
{
  layout->flatten (layout->cell (cell_index), levels, prune);
}

static void 
write_simple (const db::Layout *layout, const std::string &filename)
{
  bool gzip = false;
  db::SaveLayoutOptions options;
  options.set_format_from_filename (filename, gzip);

  db::Writer writer (options);
  if (gzip) {
    tl::OutputZLibFile file (filename);
    writer.write (*layout, file);
  } else {
    tl::OutputFile file (filename);
    writer.write (*layout, file);
  }
}

static void 
write_options (const db::Layout *layout, const std::string &filename, bool gzip, const db::SaveLayoutOptions &options)
{
  db::Writer writer (options);
  if (gzip) {
    tl::OutputZLibFile file (filename);
    writer.write (*layout, file);
  } else {
    tl::OutputFile file (filename);
    writer.write (*layout, file);
  }
}

db::RecursiveShapeIterator 
begin_shapes (const db::Layout *layout, db::cell_index_type starting_cell, unsigned int layer)
{
  if (! layout->is_valid_layer (layer)) {
    throw tl::Exception (tl::translate ("Invalid layer index"));
  }
  if (! layout->is_valid_cell_index (starting_cell)) {
    throw tl::Exception (tl::translate ("Invalid cell index"));
  }
  return db::RecursiveShapeIterator (*layout, layout->cell (starting_cell), layer);
}

db::RecursiveShapeIterator 
begin_shapes_touching (const db::Layout *layout, db::cell_index_type starting_cell, unsigned int layer, db::Box region)
{
  if (! layout->is_valid_layer (layer)) {
    throw tl::Exception (tl::translate ("Invalid layer index"));
  }
  if (! layout->is_valid_cell_index (starting_cell)) {
    throw tl::Exception (tl::translate ("Invalid cell index"));
  }
  return db::RecursiveShapeIterator (*layout, layout->cell (starting_cell), layer, region, false);
}

db::RecursiveShapeIterator 
begin_shapes_overlapping (const db::Layout *layout, db::cell_index_type starting_cell, unsigned int layer, db::Box region)
{
  if (! layout->is_valid_layer (layer)) {
    throw tl::Exception (tl::translate ("Invalid layer index"));
  }
  if (! layout->is_valid_cell_index (starting_cell)) {
    throw tl::Exception (tl::translate ("Invalid cell index"));
  }
  return db::RecursiveShapeIterator (*layout, layout->cell (starting_cell), layer, region, true);
}

db::Layout *layout_ctor_with_manager(db::Manager &manager)
{
  return new db::Layout (&manager);
}

db::Layout *layout_default_ctor()
{
  return new db::Layout ();
}

template <>
Class<db::Layout> Class<db::Layout>::decl ("Layout", 
  gsi::constructor ("new", &layout_ctor_with_manager,
    "@brief Create a layout object attached to a manager\n"
    "\n"
    "This method was introduced in version 0.19.\n"
  ) +
  gsi::constructor ("new", &layout_default_ctor,
    "@brief Create a layout object\n"
  ) +
  gsi::method ("clear", &db::Layout::clear, 
    "@brief Clears the layout\n"
    "\n"
    "Clears the layout completely."
  ) +
  gsi::method_ext ("properties_id", &properties_id, 
    "@brief Get the properties ID for a given properties set\n"
    "@args properties\n"
    "\n"
    "Before a set of properties can be attached to a shape, it must be converted into an ID that "
    "is unique for that set. The properties set must be given as a list of pairs of variants, "
    "each pair describing a name and a value. The name acts as the key for the property and does not need to be a string (it can be an integer or double value as well).\n"
    "The backward conversion can be performed with the 'properties' method.\n"
    "\n"
    "@param properties The array of pairs of variants (both elements can be integer, double or string)\n"
    "@return The unique properties ID for that set"
  ) +
  gsi::method_ext ("properties", &properties, 
    "@brief Get the properties set for a given properties ID\n"
    "@args properties_id\n"
    "\n"
    "Basically performs the backward conversion of the 'properties_id' method. "
    "Given a properties ID, returns the properties set as an array of pairs of "
    "variants. In this array, each key and the value are stored as pairs (arrays with two elements).\n"
    "If the properties ID is not valid, an empty array is returned.\n"
    "\n"
    "@param properties_id The properties ID to get the properties for\n"
    "@return The array of variants (see \\properties_id)\n"
  ) +
  gsi::method ("has_cell?", &db::Layout::has_cell, 
    "@brief Tell, if the cell with a given name exists\n"
    "@args name\n"
    "Returns true, if the layout has a cell with the given name"
  ) +
  gsi::method_ext ("cell_by_name", &cell_by_name, 
    "@brief Get the cell index for a given name\n"
    "@args name\n"
    "Returns the cell index for the cell with the given name. If no cell with this "
    "name exists, an exception is thrown."
  ) +
  gsi::method ("cell_name", &db::Layout::cell_name, 
    "@brief Get the name for a cell with the given index\n"
    "@args index\n"
  ) +
  gsi::method ("add_cell", &db::Layout::add_cell, 
    "@brief Add a cell with the given name\n"
    "@args name\n"
    "@return The index of the newly created cell."
  ) +
  gsi::method ("rename_cell", &db::Layout::rename_cell, 
    "@brief name\n"
    "@args index, name\n"
  ) +
  gsi::method ("delete_cell", &db::Layout::delete_cell,
    "@brief Delete a cell \n"
    "@args cell_index\n"
    "\n"
    "This deletes a cell but not the sub cells of the cell.\n"
    "These subcells will likely become new top cells unless they are used\n"
    "otherwise.\n"
    "All instances of this cell are deleted as well.\n"
    "Hint: to delete multiple cells, use \"delete_cells\" which is \n"
    "far more efficient in this case.\n"
    "\n"
    "@param cell_index The index of the cell to delete\n"
    "\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method_ext ("delete_cells", &delete_cells,
    "@brief Delete multiple cells\n"
    "@args cell_index_list\n"
    "\n"
    "This deletes the cells but not the sub cells of these cells.\n"
    "These subcells will likely become new top cells unless they are used\n"
    "otherwise.\n"
    "All instances of these cells are deleted as well.\n"
    "\n"
    "@param cell_index_list An array of cell indices of the cells to delete\n"
    "\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method_ext ("prune_subcells", &prune_subcells,
    "@brief Delete all sub cells of the cell which are not used otherwise down to the specified level of hierarchy\n"
    "@args cell_index, levels\n"
    "\n"
    "This deletes all sub cells of the cell which are not used otherwise.\n"
    "All instances of the deleted cells are deleted as well.\n"
    "It is possible to specify how many levels of hierarchy below the given root cell are considered.\n"
    "\n"
    "@param cell_index The root cell from which to delete a sub cells\n"
    "@param levels The number of hierarchy levels to consider (-1: all, 0: none, 1: one level etc.)\n"
    "\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method_ext ("prune_cell", &prune_cell,
    "@brief Delete a cell plus subcells not used otherwise\n"
    "@args cell_index, levels\n"
    "\n"
    "This deletes a cell and also all sub cells of the cell which are not used otherwise.\n"
    "The number of hierarchy levels to consider can be specified as well. One level of hierarchy means that "
    "only the direct children of the cell are deleted with the cell itself.\n"
    "All instances of this cell are deleted as well.\n"
    "\n"
    "@param cell_index The index of the cell to delete\n"
    "@param levels The number of hierarchy levels to consider (-1: all, 0: none, 1: one level etc.)\n"
    "\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method ("delete_cell_rec", &db::Layout::delete_cell_rec,
    "@brief Delete a cell plus all subcells\n"
    "@args cell_index\n"
    "\n"
    "This deletes a cell and also all sub cells of the cell.\n"
    "In contrast to \\prune_cell, all cells are deleted together with their instances even if they are used otherwise.\n"
    "\n"
    "@param cell_index The index of the cell to delete\n"
    "\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method_ext ("flatten", &flatten,
    "@brief Flatten the given cell\n"
    "@args cell_index, levels, prune\n"
    "\n"
    "This method propagates all shapes from the specified number of hierarchy levels below into the given cell.\n"
    "It also removes the instances of the cells from which the shapes came from, but does not remove the cells themselves if prune is set to false.\n"
    "If prune is set to true, these cells are removed if not used otherwise.\n"
    "\n"
    "@param cell_index The cell which should be flattened\n"
    "@param levels The number of hierarchy levels to flatten (-1: all, 0: none, 1: one level etc.)\n"
    "@param prune Set to true to remove orphan cells.\n"
    "\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method ("start_changes", &db::Layout::start_changes,
    "@brief Signal the start of an operation bringing the layout into invalid state\n"
    "\n"
    "This method should be called whenever the layout is\n"
    "about to be brought into an invalid state. After calling\n"
    "this method, \\under_construction returns false which \n"
    "tells foreign code (such as \\update which might be called\n"
    "asynchronously for example because of a repaint event)\n"
    "not to use this layout object.\n\n"
    "This state is cancelled by the \\end_changes method.\n"
    "The start_changes method can be called multiple times\n"
    "and must be cancelled the same number of times.\n\n"
    "Using this method is only required currently if a repaint event may happen while\n"
    "the layout object is in an invalid state."
  ) +
  gsi::method ("end_changes", &db::Layout::end_changes,
    "@brief Cancel the \"in changes\" state (see \"start_changes\")\n"
  ) +
  gsi::method ("under_construction", &db::Layout::under_construction,
    "@brief Tell if the layout object is under construction\n"
    "\n"
    "A layout object is either under construction if a transaction\n"
    "is ongoing or the layout is brought into invalid state by\n"
    "\"start_changes\".\n"
  ) +
  gsi::method ("update", (void (db::Layout::*) ()) &db::Layout::update,
    "@brief Update the internals of the layout\n"
    "This method updates the internal state of the layout. Usually this is done automatically\n"
    "This method is provided to ensure this state explicitly."
  ) +
  gsi::method ("dbu=", (void (db::Layout::*) (double)) &db::Layout::dbu,
    "@brief Database unit write accessor\n"
    "@args dbu\n"
  ) +
  gsi::method ("dbu", (double (db::Layout::*) () const) &db::Layout::dbu,
    "@brief Database unit read accessor\n"
  ) +
  gsi::method ("insert_layer", (unsigned int (db::Layout::*) (const db::LayerProperties &)) &db::Layout::insert_layer,
    "@brief Insert a new layer with the given properties\n"
    "@args props\n"
    "@return The index of the newly created layer\n"
  ) +
  gsi::method ("insert_layer_at", (void (db::Layout::*) (unsigned int, const db::LayerProperties &)) &db::Layout::insert_layer,
    "@brief Insert a new layer with the given properties at the given index\n"
    "@args index, props\n"
  ) +
  gsi::method ("insert_special_layer", (unsigned int (db::Layout::*) (const db::LayerProperties &)) &db::Layout::insert_special_layer,
    "@brief Insert a new special layer with the given properties\n"
    "@args props\n"
    "\n"
    "Special layers can be used to represent objects that should not participate in normal viewing or other "
    "related operations. Special layers are not reported as valid layers.\n"
    "\n"
    "@return The index of the newly created layer\n"
  ) +
  gsi::method ("insert_special_layer_at", (void (db::Layout::*) (unsigned int, const db::LayerProperties &)) &db::Layout::insert_special_layer,
    "@brief Insert a new special layer with the given properties at the given index\n"
    "@args index, props\n"
    "\n"
    "See \\insert_special_layer for a description of special layers."
  ) +
  gsi::method ("set_info", &db::Layout::set_properties,
    "@brief Set the info structure for a specified layer\n"
    "@args index, props\n"
  ) +
  gsi::method ("get_info", &db::Layout::get_properties,
    "@brief Get the info structure for a specified layer\n"
    "@args index\n"
  ) +
  gsi::method ("cells", &db::Layout::cells,
    "@brief Return the number of cells\n"
    "\n"
    "@return The number of cells (the maximum cell index)\n"
  ) +
  gsi::method ("cell", (db::Layout::cell_type &(db::Layout::*) (db::cell_index_type)) &db::Layout::cell,
    "@brief Address a cell by index\n"
    "@args i\n"
    "\n"
    "@param i The cell index\n"
    "@return A reference to the cell\n"
  ) +
  gsi::iterator ("each_cell", (db::Layout::iterator (db::Layout::*) ()) &db::Layout::begin, (db::Layout::iterator (db::Layout::*) ()) &db::Layout::end,
    "@brief Iterate the unsorted cell list\n"
  ) +
  gsi::iterator ("each_cell_bottom_up", (db::Layout::bottom_up_iterator (db::Layout::*) ()) &db::Layout::begin_bottom_up, (db::Layout::bottom_up_iterator (db::Layout::*) ()) &db::Layout::end_bottom_up,
    "@brief Iterate the bottom-up sorted cell list\n"
    "\n"
    "In bottom-up traversal a cell is not delivered before\n"
    "the last child cell of this cell has been delivered.\n"
    "The bottom-up iterator does not deliver cells but cell\n"
    "indices actually.\n"
  ) +
  gsi::iterator ("each_cell_top_down", (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::begin_top_down, (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::end_top_down,
    "@brief Iterate of the top-down sorted cell list\n"
    "\n"
    "The top-down cell list has the property of delivering all\n"
    "cells before they are instantiated. In addition the first\n"
    "cells are all top cells. There is at least one top cell.\n"
    "The top-down iterator does not deliver cells but cell\n"
    "indices actually.\n"
    "@brief begin iterator of the top-down sorted cell list\n"
  ) +
  gsi::iterator ("each_top_cell", (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::begin_top_down, (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::end_top_cells,
    "@brief Iterate the top cells\n"
    "A layout may have an arbitrary number of top cells. The usual case however is that there is one "
    "top cell."
  ) +
  gsi::method ("swap_layers", &db::Layout::swap_layers,
    "@brief Swap layers\n"
    "@args a, b\n"
    "\n"
    "Swaps the shapes of both layers.\n"
    "\n"
    "This method was introduced in version 0.19.\n"
    "\n"
    "@param a The first of the layers to swap.\n"
    "@param b The second of the layers to swap.\n"
  ) +
  gsi::method ("move_layer", &db::Layout::move_layer,
    "@brief Move a layer\n"
    "@args src, dest\n"
    "\n"
    "This method was introduced in version 0.19.\n"
    "\n"
    "Move a layer from the source to the target. The target is not cleared before, so that this method \n"
    "merges shapes from the source with the target layer. The source layer is empty after that operation.\n"
    "\n"
    "@param src The layer index of the source layer.\n"
    "@param dest The layer index of the destination layer.\n"
  ) +
  gsi::method ("copy_layer", &db::Layout::copy_layer,
    "@brief Copy a layer\n"
    "@args src, dest\n"
    "\n"
    "This method was introduced in version 0.19.\n"
    "\n"
    "Copy a layer from the source to the target. The target is not cleared before, so that this method \n"
    "merges shapes from the source with the target layer.\n"
    "\n"
    "@param src The layer index of the source layer.\n"
    "@param dest The layer index of the destination layer.\n"
  ) +
  gsi::method ("clear_layer", &db::Layout::clear_layer,
    "@brief Clear a layer\n"
    "@args layer_index\n"
    "\n"
    "Clears the layer: removes all shapes.\n"
    "\n"
    "This method was introduced in version 0.19.\n"
    "\n"
    "@param layer_index The index of the layer to delete.\n"
  ) +
  gsi::method ("delete_layer", &db::Layout::delete_layer,
    "@brief Delete a layer\n"
    "@args layer_index\n"
    "\n"
    "This does free the shapes of the cells and remembers the\n"
    "layer's index for recycling.\n"
    "\n"
    "@param layer_index The index of the layer to delete.\n"
  ) +
  gsi::method_ext ("layer_indices", &layer_indices,
    "@brief Return a list of valid layer indices\n"
    "\n"
    "This method returns an array with layer indices representing valid layers.\n"
    "\n"
    "This method was introduced in version 0.19.\n"
  ) +
  gsi::method ("layers", &db::Layout::layers,
    "@brief Return the number of layers\n"
    "The number of layers reports the maximum (plus 1) layer index used so far. Not all of the layers with "
    "an index in the range of 0 to layers-1 needs to be a valid layer. These layers can be either valid, "
    "special or unused. Use \\is_valid_layer? and \\is_special_layer? to test for the first two states.\n"
  ) +
  gsi::method ("is_valid_cell_index?", &db::Layout::is_valid_cell_index,
    "@brief Tell, if a cell index is a valid index\n"
    "@args cell_index\n"
    "\n"
    "@return true, if this is the case\n"
    "This method has been added in version 0.20.\n"
  ) +
  gsi::method ("is_valid_layer?", &db::Layout::is_valid_layer,
    "@brief Tell, if a layer index is a valid index\n"
    "@args layer_index\n"
    "\n"
    "@return true, if this is the case\n"
  ) +
  gsi::method ("is_special_layer?", &db::Layout::is_special_layer,
    "@brief Tell, if a layer index is a special layer index\n"
    "@args layer_index\n"
    "\n"
    "@return true, if this is the case\n"
  ) +
  gsi::method_ext ("begin_shapes", &begin_shapes, 
    "@brief Delivers a recursive shape iterator for the shapes below the given cell on the given layer\n"
    "@args cell_index,layer\n"
    "@param cell_index The index of the starting cell\n"
    "@param layer The layer from which to get the shapes\n"
    "@return A suitable iterator\n"
    "\n"
    "For details see the description of the \\RecursiveShapeIterator class.\n"
    "\n"
    "This method has been added in version 0.18.\n"
  ) +
  gsi::method_ext ("begin_shapes_touching", &begin_shapes_touching, 
    "@brief Delivers a recursive shape iterator for the shapes below the given cell on the given layer using a region search\n"
    "@args cell_index,layer,region\n"
    "@param cell_index The index of the starting cell\n"
    "@param layer The layer from which to get the shapes\n"
    "@param region The search region\n"
    "@return A suitable iterator\n"
    "\n"
    "For details see the description of the \\RecursiveShapeIterator class.\n"
    "This version gives an iterator delivering shapes whose bounding box touches the given region.\n"
    "\n"
    "This method has been added in version 0.18.\n"
  ) +
  gsi::method_ext ("begin_shapes_overlapping", &begin_shapes_overlapping, 
    "@brief Delivers a recursive shape iterator for the shapes below the given cell on the given layer using a region search\n"
    "@args cell_index,layer,region\n"
    "@param cell_index The index of the starting cell\n"
    "@param layer The layer from which to get the shapes\n"
    "@param region The search region\n"
    "@return A suitable iterator\n"
    "\n"
    "For details see the description of the \\RecursiveShapeIterator class.\n"
    "This version gives an iterator delivering shapes whose bounding box overlaps the given region.\n"
    "\n"
    "This method has been added in version 0.18.\n"
  ) +
  gsi::method_ext ("write", &write_options,
    "@brief Write the layout to a stream file\n"
    "@args filename, gzip, options\n"
    "@param filename The file to which to write the layout\n"
    "@param gzip True, if the file should be gzipped\n"
    "@param options The option set to use for writing. See \\SaveLayoutOptions for details\n"
  ) +
  gsi::method_ext ("write", &write_simple,
    "@brief Write the layout to a stream file\n"
    "@args filename\n"
    "@param filename The file to which to write the layout\n"
  ) + 
  gsi::method_ext ("clip", &clip,
    "@brief Clips the given cell by the given rectangle and produce a new cell with the clip\n"
    "@args cell, box\n"
    "@param cell The cell index of the cell to clip\n"
    "@param box The clip box in database units\n"
    "@return The index of the new cell\n"
    "\n"
    "This method will cut a rectangular region given by the box from the given cell. The clip will "
    "be stored in a new cell whose index is returned. The clip will be performed hierarchically. The "
    "resulting cell will hold a hierarchy of child cells, which are potentially clipped versions of "
    "child cells of the original cell." 
    "\n"
    "This method has been added in version 0.21.\n"
  ) + 
  gsi::method_ext ("clip_into", &clip_into,
    "@brief Clips the given cell by the given rectangle and produce a new cell with the clip\n"
    "@args cell, box, target\n"
    "@param cell The cell index of the cell to clip\n"
    "@param box The clip box in database units\n"
    "@param target The target layout\n"
    "@return The index of the new cell in the target layout\n"
    "\n"
    "This method will cut a rectangular region given by the box from the given cell. The clip will "
    "be stored in a new cell in the target layout. The clip will be performed hierarchically. The "
    "resulting cell will hold a hierarchy of child cells, which are potentially clipped versions of "
    "child cells of the original cell.\n"
    "\n"
    "Please note that it is important that the database unit of the " 
    "target layout is identical to the database unit of the source layout to achieve the desired results."
    "This method also assumes that the target layout holds the same layers than the source layout. It will "
    "copy shapes to the same layers than they have been on the original layout. "
    "\n"
    "This method has been added in version 0.21.\n"
  ) + 
  gsi::method_ext ("multi_clip", &multi_clip,
    "@brief Clips the given cell by the given rectangles and produce new cells with the clips, one for each rectangle.\n"
    "@args cell, boxes\n"
    "@param cell The cell index of the cell to clip\n"
    "@param boxes The clip boxes in database units\n"
    "@return The indexes of the new cells\n"
    "\n"
    "This method will cut rectangular regions given by the boxes from the given cell. The clips will "
    "be stored in a new cells whose indexed are returned. The clips will be performed hierarchically. The "
    "resulting cells will hold a hierarchy of child cells, which are potentially clipped versions of "
    "child cells of the original cell. This version is somewhat more efficient than doing individual clips "
    "because the clip cells may share clipped versions of child cells." 
    "\n"
    "This method has been added in version 0.21.\n"
  ) + 
  gsi::method_ext ("multi_clip_into", &multi_clip_into,
    "@brief Clips the given cell by the given rectangles and produce new cells with the clips, one for each rectangle.\n"
    "@args cell, boxes, target\n"
    "@param cell The cell index of the cell to clip\n"
    "@param boxes The clip boxes in database units\n"
    "@param target The target layout\n"
    "@return The indexes of the new cells\n"
    "\n"
    "This method will cut rectangular regions given by the boxes from the given cell. The clips will "
    "be stored in a new cells in the given target layout. The clips will be performed hierarchically. The "
    "resulting cells will hold a hierarchy of child cells, which are potentially clipped versions of "
    "child cells of the original cell. This version is somewhat more efficient than doing individual clips "
    "because the clip cells may share clipped versions of child cells.\n" 
    "\n"
    "Please note that it is important that the database unit of the " 
    "target layout is identical to the database unit of the source layout to achieve the desired results. "
    "This method also assumes that the target layout holds the same layers than the source layout. It will "
    "copy shapes to the same layers than they have been on the original layout. "
    "\n"
    "This method has been added in version 0.21.\n"
  ),
  "@brief The layout object\n"
  "\n"
  "The layout object basically wraps the cell graphs and\n"
  "adds functionality for managing cell names and layer names.\n"
  "The cell graph is a container for the cells and their hierarchical arrangement.\n"
  "The cell graph is constructed by creating cells and adding child instances to it.\n"
);

db::SaveLayoutOptions new_v ()
{
  return db::SaveLayoutOptions ();
}

template <>
Class<db::SaveLayoutOptions> Class<db::SaveLayoutOptions>::decl ("SaveLayoutOptions", 
  gsi::method ("new", &new_v,
    "@brief Default constructor\n"
    "\n"
    "By default, the scale factor will be 1.0, the database unit is set to\n"
    "\"same as original\" and all layers are selected as well as all cells.\n"
    "The default format is GDS2. "
  ) +
  gsi::method ("format=", &db::SaveLayoutOptions::set_format,
    "@brief Select a format\n"
    "@args format\n"
    "The format string can be either \"GDS2\" or \"OASIS\". Other formats may be available if\n"
    "a suitable plugin is installed."
  ) +
  gsi::method ("format", &db::SaveLayoutOptions::format,
    "@brief Get the format name \n"
  ) + 
  gsi::method ("add_layer", &db::SaveLayoutOptions::add_layer,
    "@brief Add a layer to be saved \n"
    "\n"
    "@args layer_index, properties\n"
    "\n"
    "Adds the layer with the given index to the layer list that will be written.\n"
    "If all layers have been selected previously, this state will \n"
    "be cleared.\n"
    "\n"
    "The 'properties' argument can be used to assign different layer properties than the ones\n"
    "present in the layout. Pass a default \\LayerInfo object to this argument to use the\n"
    "properties from the layout object. Construct a valid \\LayerInfo object with explicit layer,\n"
    "datatype and possibly a name to override the properties stored in the layout.\n"
  ) + 
  gsi::method ("select_all_layers", &db::SaveLayoutOptions::select_all_layers,
    "@brief Select all layers to be saved\n"
    "\n"
    "This method will clear all layers selected with \\add_layer so far and set the 'select all layers' flag.\n"
    "This is the default.\n"
  ) + 
  gsi::method ("deselect_all_layers", &db::SaveLayoutOptions::deselect_all_layers,
    "@brief Unselect all layers: no layer will be saved\n"
    "\n"
    "This method will clear all layers selected with \\add_layer so far and clear the 'select all layers' flag.\n"
    "Using this method is the only way to save a layout without any layers."
  ) + 
  gsi::method ("add_cell", &db::SaveLayoutOptions::add_cell,
    "@brief Add a cell (plus hierarchy) to be saved\n"
    "\n"
    "@args cell_index\n"
    "\n"
    "The index of the cell must be a valid index in the context of the layout that will be saved.\n"
    "This method clears the 'select all cells' flag.\n"
  ) + 
  gsi::method ("select_all_cells", &db::SaveLayoutOptions::select_all_cells,
    "@brief Select all cells to save\n"
    "\n"
    "This method will clear all cells specified with \\add_cells so far and set the 'select all cells' flag.\n"
    "This is the default.\n"
  ) + 
  gsi::method ("dbu=", &db::SaveLayoutOptions::set_dbu,
    "@brief Set the database unit to be used in the stream file\n"
    "@args dbu\n"
    "\n"
    "By default, the database unit of the layout is used. This method allows to explicitly use a different\n"
    "database unit. This effectively scales the layout.\n"
  ) + 
  gsi::method ("dbu", &db::SaveLayoutOptions::dbu,
    "@brief Get the explicit database unit if one is set\n"
  ) + 
  gsi::method ("no_empty_cells=", &db::SaveLayoutOptions::set_dont_write_empty_cells,
    "@brief Don't write empty cells if this flag is set\n"
    "@args flag\n"
    "\n"
    "By default, all cells are written (no_empty_cells is false).\n"
    "This applies to empty cells which do not contain shapes for the specified layers "
    "as well as cells which are empty because they reference empty cells only.\n"
  ) + 
  gsi::method ("no_empty_cells", &db::SaveLayoutOptions::dont_write_empty_cells,
    "@brief Returns a flag indicating whether empty cells are not written.\n"
  ) + 
  gsi::method ("scale_factor=", &db::SaveLayoutOptions::set_scale_factor,
    "@brief Set the scaling factor for the saving \n"
    "@args scale_factor\n"
    "\n"
    "Using a scaling factor will scale all objects accordingly. Using a scaling factor can compensate\n"
    "implicit scaling by an explicit database unit specification. Setting and scale factor plus an explicit\n"
    "database unit thus allows to transcribe a layout to a different database unit without changing the\n"
    "layout's physical dimensions (beside potential grid snapping effects).\n"
    "\n"
    "Be aware that rounding effects may occur if fractional scaling factors are used which are not compliant\n"
    "with any implicit layout grid.\n"
    "\n"
    "By default, no scaling is applied."
  ) + 
  gsi::method ("scale_factor", &db::SaveLayoutOptions::scale_factor,
    "@brief Get the scaling factor currently set\n"
  ),
  "@brief Options for saving layouts\n"
  "\n"
  "This class describes the various options for saving a layout to a stream file (GDS2, OASIS).\n"
  "There are: layers to be saved, cell or cells to be saved, scale factor, format, database unit\n"
  "and format specific options.\n"
  "Usually the default constructor provides a suitable object."
  "\n"
  "The layers are specified by either selecting all layers or by defining layer by layer using the\n"
  "\\add_layer method. \\select_all_layers will explicitly select all layers for saving, \\deselect_all_layers will explicitly clear the list of layers.\n"
  "\n"
  "Cells are selected in a similar fashion: by default, all cells are selected. Using \\add_cell, specific\n"
  "cells can be selected for saving. All these cells plus their hierarchy will then be written to the stream file.\n"
  "\n"
);

}

