
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsiDecl.h"
#include "dbShapes.h"
#include "dbShape.h"
#include "dbLayout.h"

namespace gsi
{

// ---------------------------------------------------------------
//  db::Shapes binding

template<class Sh>
static db::Shape insert (db::Shapes *s, const Sh &p)
{
  return s->insert (p);
}

template<class Sh>
static db::Shape replace (db::Shapes *s, const db::Shape &sh, const Sh &p)
{
  return s->replace (sh, p);
}

template<class Sh>
static db::Shape insert_with_properties (db::Shapes *s, const Sh &p, db::properties_id_type id)
{
  return s->insert (db::object_with_properties<Sh> (p, id));
}

static db::Shapes::shape_iterator begin (const db::Shapes *s, unsigned int flags)
{
  return s->begin (flags);
}

static db::Shapes::shape_iterator begin_all (const db::Shapes *s)
{
  return s->begin (db::ShapeIterator::All);
}

static db::Shapes::shape_iterator begin_overlapping (const db::Shapes *s, unsigned int flags, const db::Box &region)
{
  return s->begin_overlapping (region, flags);
}

static db::Shapes::shape_iterator begin_overlapping_all (const db::Shapes *s, const db::Box &region)
{
  return s->begin_overlapping (region, db::ShapeIterator::All);
}

static db::Shapes::shape_iterator begin_touching (const db::Shapes *s, unsigned int flags, const db::Box &region)
{
  return s->begin_touching (region, flags);
}

static db::Shapes::shape_iterator begin_touching_all (const db::Shapes *s, const db::Box &region)
{
  return s->begin_touching (region, db::ShapeIterator::All);
}

static db::Shape transform_shape_cplx (db::Shapes *s, const db::Shape &shape, const db::CplxTrans &trans)
{
  return s->transform (shape, db::ICplxTrans (trans));
}

static unsigned int s_all ()                 { return db::ShapeIterator::All; }
static unsigned int s_all_with_properties () { return db::ShapeIterator::AllWithProperties; }
static unsigned int s_properties ()          { return db::ShapeIterator::Properties; }
static unsigned int s_polygons ()            { return db::ShapeIterator::Polygons; }
static unsigned int s_boxes ()               { return db::ShapeIterator::Boxes; }
static unsigned int s_edges ()               { return db::ShapeIterator::Edges; }
static unsigned int s_paths ()               { return db::ShapeIterator::Paths; }
static unsigned int s_texts ()               { return db::ShapeIterator::Texts; }
static unsigned int s_user_objects ()        { return db::ShapeIterator::UserObjects; }

template <>
Class<db::Shapes> Class<db::Shapes>::decl ("Shapes", 
  gsi::method ("insert", (db::Shape (db::Shapes::*)(const db::Shape &)) &db::Shapes::insert,
    "@brief Insert a shape from a shape reference into the shapes list\n"
    "@args shape\n"
    "@return A reference (a \\Shape object) to the newly created shape\n"
    "This method has been introduced in version 0.16.\n"
  ) +
  gsi::method ("transform", (db::Shape (db::Shapes::*)(const db::Shape &, const db::Trans &)) &db::Shapes::transform,
    "@brief Transform the shape given by the reference with the given transformation\n"
    "@args shape, trans\n"
    "@return A reference (a \\Shape object) to the new shape\n"
    "This method has been introduced in version 0.16.\n"
    "The original shape may be deleted and re-inserted by this method. Therefore, a new reference is returned.\n"
    "It is permitted in editable mode only."
  ) +
  gsi::method_ext ("transform", &transform_shape_cplx,
    "@brief Transform the shape given by the reference with the given complex transformation\n"
    "@args shape, trans\n"
    "@return A reference (a \\Shape object) to the new shape\n"
    "This method has been introduced in version 0.16.\n"
    "The original shape may be deleted and re-inserted by this method. Therefore, a new reference is returned.\n"
    "It is permitted in editable mode only."
  ) +
  gsi::method_ext ("replace", &replace<db::Box>,
    "@brief Replace the given shape with a box\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args shape,box\n"
    "\n"
    "This method has been introduced with version 0.16. It replaces the given shape with the "
    "object specified. It does not change the property Id. To change the property Id, "
    "use the \\replace_prop_id method. To replace a shape and discard the property Id, erase the "
    "shape and insert a new shape."
    "\n"
    "This method is permitted in editable mode only."
  ) +
  gsi::method_ext ("replace", &replace<db::Path>,
    "@brief Replace the given shape with a path\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args shape,path\n"
    "\n"
    "This method has been introduced with version 0.16. It replaces the given shape with the "
    "object specified. It does not change the property Id. To change the property Id, "
    "use the \\replace_prop_id method. To replace a shape and discard the property Id, erase the "
    "shape and insert a new shape."
    "\n"
    "This method is permitted in editable mode only."
  ) +
  gsi::method_ext ("replace", &replace<db::Edge>,
    "@brief Replace the given shape with an edge object\n"
    "@args shape,edge\n"
    "\n"
    "This method has been introduced with version 0.16. It replaces the given shape with the "
    "object specified. It does not change the property Id. To change the property Id, "
    "use the \\replace_prop_id method. To replace a shape and discard the property Id, erase the "
    "shape and insert a new shape."
    "\n"
    "This method is permitted in editable mode only."
  ) +
  gsi::method_ext ("replace", &replace<db::Text>,
    "@brief Replace the given shape with a text object\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args shape,text\n"
    "\n"
    "This method has been introduced with version 0.16. It replaces the given shape with the "
    "object specified. It does not change the property Id. To change the property Id, "
    "use the \\replace_prop_id method. To replace a shape and discard the property Id, erase the "
    "shape and insert a new shape."
    "\n"
    "This method is permitted in editable mode only."
  ) +
  gsi::method_ext ("replace", &replace<db::SimplePolygon>,
    "@brief Replace the given shape with a simple polygon\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args shape,simple_polygon\n"
    "\n"
    "This method has been introduced with version 0.16. It replaces the given shape with the "
    "object specified. It does not change the property Id. To change the property Id, "
    "use the \\replace_prop_id method. To replace a shape and discard the property Id, erase the "
    "shape and insert a new shape."
    "\n"
    "This method is permitted in editable mode only."
  ) +
  gsi::method_ext ("replace", &replace<db::Polygon>,
    "@brief Replace the given shape with a polygon\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args shape,polygon\n"
    "\n"
    "This method has been introduced with version 0.16. It replaces the given shape with the "
    "object specified. It does not change the property Id. To change the property Id, "
    "use the \\replace_prop_id method. To replace a shape and discard the property Id, erase the "
    "shape and insert a new shape."
    "\n"
    "This method is permitted in editable mode only."
  ) +
  gsi::method_ext ("insert|#insert_box", &insert<db::Box>,
    "@brief Insert a box into the shapes list\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args box\n"
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_path", &insert<db::Path>,
    "@brief Insert a path into the shapes list\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args path\n"
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_edge", &insert<db::Edge>,
    "@brief Insert a edge into the shapes list\n"
    "@args edge\n"
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_text", &insert<db::Text>,
    "@brief Insert a text into the shapes list\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args text\n"
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_simple_polygon", &insert<db::SimplePolygon>,
    "@brief Insert a simple polygon into the shapes list\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args simple_polygon\n"
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_polygon", &insert<db::Polygon>,
    "@brief Insert a polygon into the shapes list\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "@args polygon\n"
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_box_with_properties", &insert_with_properties<db::Box>,
    "@brief Insert a box with properties into the shapes list\n"
    "@args box, property_id\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_path_with_properties", &insert_with_properties<db::Path>,
    "@brief Insert a path with properties into the shapes list\n"
    "@args path, property_id\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_edge_with_properties", &insert_with_properties<db::Edge>,
    "@brief Insert a edge with properties into the shapes list\n"
    "@args edge, property_id\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_text_with_properties", &insert_with_properties<db::Text>,
    "@brief Insert a text with properties into the shapes list\n"
    "@args text, property_id\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_simple_polygon_with_properties", &insert_with_properties<db::SimplePolygon>,
    "@brief Insert a simple polygon with properties into the shapes list\n"
    "@args simple_polygon, property_id\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::method_ext ("insert|#insert_polygon_with_properties", &insert_with_properties<db::Polygon>,
    "@brief Insert a polygon with properties into the shapes list\n"
    "@args polygon, property_id\n"
    "@return A reference to the new shape (a \\Shape object)\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
    "\n"
    "Starting with version 0.16, this method returns a reference to the newly created shape\n"
  ) +
  gsi::iterator_ext ("each", &begin, 
    "@brief Get all shapes\n"
    "@args flags\n"
    "\n"
    "@param flags An \"or\"-ed combination of the s_... constants\n"
  ) +
  gsi::iterator_ext ("each", &begin_all, 
    "@brief Get all shapes\n"
    "\n"
    "This call is equivalent to each(s_all). This convenience method has been introduced in version 0.16\n"
  ) +
  gsi::iterator_ext ("each_touching", &begin_touching, 
    "@brief Get all shapes that touch the search box (region)\n"
    "@args flags,region\n"
    "This method was introduced in version 0.16\n"
    "\n"
    "@param flags An \"or\"-ed combination of the s_... constants\n"
    "@param region The rectangular search region\n"
  ) +
  gsi::iterator_ext ("each_touching", &begin_touching_all, 
    "@brief Get all shapes that touch the search box (region)\n"
    "@args region\n"
    "@param region The rectangular search region\n"
    "\n"
    "This call is equivalent to each_touching(s_all,region). This convenience method has been introduced in version 0.16\n"
  ) +
  gsi::iterator_ext ("each_overlapping", &begin_overlapping, 
    "@brief Get all shapes that overlap the search box (region)\n"
    "@args flags,region\n"
    "This method was introduced in version 0.16\n"
    "\n"
    "@param flags An \"or\"-ed combination of the s_... constants\n"
    "@param region The rectangular search region\n"
  ) +
  gsi::iterator_ext ("each_overlapping", &begin_overlapping_all, 
    "@brief Get all shapes that overlap the search box (region)\n"
    "@args region\n"
    "@param region The rectangular search region\n"
    "\n"
    "This call is equivalent to each_overlapping(s_all,region). This convenience method has been introduced in version 0.16\n"
  ) +
  gsi::method ("erase", &db::Shapes::erase_shape, 
    "@brief Erase the shape pointed to by the given \\Shape object\n"
    "@args shape\n"
    "This method has been introduced in version 0.16. It can only be used in editable mode.\n"
    "Erasing a shape will invalidate the shape reference. Access to this reference may then render invalid results.\n"
    "\n"
    "@param shape The shape which to destroy"
  ) +
  gsi::method ("is_valid?", &db::Shapes::is_valid, 
    "@brief Test if the given \\Shape object is still pointing to a valid object\n"
    "@args shape\n"
    "This method has been introduced in version 0.16.\n"
    "If the shape represented by the given reference has been deleted, this method returns false. "
    "If however, another shape has been inserted already that occupies the original shape's position, "
    "this method will return true again.\n"
  ) +
  gsi::method ("is_empty?", &db::Shapes::empty, 
    "@brief Returns a value indicating whether the shapes container is empty\n"
    "This method has been introduced in version 0.20.\n"
  ) +
  gsi::method ("clear", &db::Shapes::clear, 
    "@brief Clear the shape container\n"
    "This method has been introduced in version 0.16. It can only be used in editable mode."
  ) +
  gsi::method ("size", (size_t (db::Shapes::*)() const) &db::Shapes::size, 
    "@brief Report the number of shapes in this container\n"
    "This method was introduced in version 0.16\n"
    "@return The number of shapes in this container\n"
  ) +
  gsi::method ("replace_prop_id", (db::Shape (db::Shapes::*) (const db::Shape &, db::properties_id_type)) &db::Shapes::replace_prop_id,
    "@brief Replace (or install) the properties of a shape\n"
    "@args shape,property_id\n"
    "@return A \\Shape object representing the new shape\n"
    "This method has been introduced in version 0.16. It can only be used in editable mode.\n"
    "Changes the properties Id of the given shape or install a properties Id on that shape if it does not have one yet.\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id.\n"
    "This method will potentially invalidate the shape reference passed to it. Use the reference "
    "returned for future references."
  ) +
  gsi::method ("s_all", &s_all) +
  gsi::method ("s_all_with_properties", &s_all_with_properties) +
  gsi::method ("s_polygons", &s_polygons) +
  gsi::method ("s_boxes", &s_boxes) +
  gsi::method ("s_edges", &s_edges) +
  gsi::method ("s_paths", &s_paths) +
  gsi::method ("s_texts", &s_texts) +
  gsi::method ("s_user_objects", &s_user_objects) +
  gsi::method ("s_properties", &s_properties),
  "@brief A collection of shapes\n"
  "\n"
  "A shapes collection is a collection of geometrical objects, such as polygons, boxes, paths, edges or text objects.\n"
);

}

