
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsiDecl.h"

#include "layBrowserDialog.h"
#include <QtGui/QMessageBox>
#include <QtGui/QInputDialog>
#include <QtGui/QFileDialog>

namespace gsi
{

// ---------------------------------------------------------------------------------
//  Value classes with "not set" capabilities

struct DoubleValue
{
  DoubleValue ()          : v (0), h (false) { }
  DoubleValue (double _v) : v (_v), h (true) { }

  double v;
  bool h;
  double value ()   const { return v; }
  bool has_value () const { return h; }
};

template <>
Class<DoubleValue> Class<DoubleValue>::decl ("DoubleValue", 
  gsi::method ("has_value?", &DoubleValue::has_value,
    "@brief True, if a value is present"
  ) +
  gsi::method ("to_f", &DoubleValue::value,
    "@brief Get the actual value (a synonym for \\value)"
  ) +
  gsi::method ("value", &DoubleValue::value,
    "@brief Get the actual value"
  ),
  "@brief Encapsulate a floating point value\n"
  "This class is provided as a return value of \\InputDialog::get_double.\n"
  "By using an object rather than a pure value, an object with \\has_value? = false can be returned indicating that\n"
  "the \"Cancel\" button was pressed."
);

struct IntValue
{
  IntValue ()       : v (0), h (false) { }
  IntValue (int _v) : v (_v), h (true) { }

  int v;
  bool h;
  int value ()      const { return v; }
  bool has_value () const { return h; }
};

template <>
Class<IntValue> Class<IntValue>::decl ("IntValue", 
  gsi::method ("has_value?", &IntValue::has_value,
    "@brief True, if a value is present"
  ) +
  gsi::method ("to_i", &IntValue::value,
    "@brief Get the actual value (a synonym for \\value)"
  ) +
  gsi::method ("value", &IntValue::value,
    "@brief Get the actual value"
  ),
  "@brief Encapsulate an integer value\n"
  "This class is provided as a return value of \\InputDialog::get_int.\n"
  "By using an object rather than a pure value, an object with \\has_value? = false can be returned indicating that\n"
  "the \"Cancel\" button was pressed."
);

struct StringValue
{
  StringValue ()                      : v (), h (false) { }
  StringValue (const std::string &_v) : v (_v), h (true) { }

  std::string v;
  bool h;
  const std::string &value () const { return v; }
  bool has_value ()           const { return h; }
};

template <>
Class<StringValue> Class<StringValue>::decl ("StringValue", 
  gsi::method ("has_value?", &StringValue::has_value,
    "@brief True, if a value is present"
  ) +
  gsi::method ("to_s", &StringValue::value,
    "@brief Get the actual value (a synonym for \\value)"
  ) +
  gsi::method ("value", &StringValue::value,
    "@brief Get the actual value"
  ),
  "@brief Encapsulate a string value\n"
  "This class is provided as a return value of \\InputDialog::get_string, \\InputDialog::get_item and \\FileDialog.\n"
  "By using an object rather than a pure value, an object with \\has_value? = false can be returned indicating that\n"
  "the \"Cancel\" button was pressed."
);

struct StringListValue
{
  StringListValue ()                                   : v (), h (false) { }
  StringListValue (const std::vector<std::string> &_v) : v (_v), h (true) { }

  std::vector<std::string> v;
  bool h;
  const std::vector<std::string> &value () const { return v; }
  bool has_value () const                        { return h; }
};

template <>
Class<StringListValue> Class<StringListValue>::decl ("StringListValue", 
  gsi::method ("has_value?", &StringListValue::has_value,
    "@brief True, if a value is present"
  ) +
  gsi::method ("value", &StringListValue::value,
    "@brief Get the actual value (a list of strings)"
  ),
  "@brief Encapsulate a string list\n"
  "This class is provided as a return value of \\FileDialog.\n"
  "By using an object rather than a pure string list, an object with \\has_value? = false can be returned indicating that\n"
  "the \"Cancel\" button was pressed."
);

// ---------------------------------------------------------------------------------
//  HTML browser

//  the stub required to provide all interface logic for the virtual methods
//  (this enables reimplementation of the virtual function on the client side)
class BrowserDialog_Stub 
  : public lay::BrowserDialog,
    public gsi::Object<BrowserDialog_Stub>
{
public:
  BrowserDialog_Stub () : lay::BrowserDialog () { }

  virtual void closed ()
  {
    return gsi::can_issue (closed_cb) ? gsi::issue (closed_cb, &BrowserDialog_Stub::closed) : lay::BrowserDialog::closed ();
  }

  gsi::Callback closed_cb;
};

template <> struct ClassTag<BrowserDialog_Stub> {
  typedef CantCopySemanticsTag copy_semantics;
  typedef CanDefaultCreateSemanticsTag create_semantics;
};

//  the stub required to provide all interface logic for the virtual methods
//  (this enables reimplementation of the virtual function on the client side)
class BrowserSource_Stub : public lay::BrowserSource
{
public:
  BrowserSource_Stub () : lay::BrowserSource () { }
  BrowserSource_Stub (const std::string &html) : lay::BrowserSource (html) { }

  virtual std::string get (const std::string &url)
  {
    return gsi::can_issue (get_cb) ? gsi::issue (get_cb, &BrowserSource_Stub::get, url) : lay::BrowserSource::get (url);
  }

  gsi::Callback get_cb;
};

//  specialize the "set_source" method to the stub class
void set_source (BrowserDialog_Stub *s, BrowserSource_Stub *src)
{
  s->set_source (src);
}

void set_size (BrowserDialog_Stub *s, int width, int height)
{
  s->resize (QSize (width, height));
}

void set_caption (BrowserDialog_Stub *s, const std::string &caption)
{
  s->setCaption (caption.c_str ());
}

template <>
Class<BrowserDialog_Stub> Class<BrowserDialog_Stub>::decl ("BrowserDialog", 
  gsi::method ("hide", &BrowserDialog_Stub::hide, 
    "@brief Hide the HTML browser window"
  ) +
  gsi::method ("show", &BrowserDialog_Stub::show, 
    "@brief Show the HTML browser window in a non-modal way"
  ) +
  gsi::method ("exec", &BrowserDialog_Stub::exec, 
    "@brief Execute the HTML browser dialog as a modal window\n"
  ) +
  gsi::method ("load", &BrowserDialog_Stub::load, 
    "@brief Load the given URL into the browser dialog\n"
  ) +
  gsi::method_ext ("set_source", &set_source, 
    "@brief Connect to a source object\n"
    "@args source\n"
    "\n"
    "Caution: this will use the object as the source but not hold a reference to that object.\n"
    "In order not to loose the source object (i.e. in RBA), a separate reference is required.\n"
  ) +
  gsi::method_ext ("set_size", &set_size, 
    "@brief Set the size of the dialog window\n"
    "@args width, height\n"
  ) +
  gsi::method_ext ("set_caption", &set_caption, 
    "@brief Set the caption of the window\n"
    "@args caption\n"
  ) +
  gsi::method ("reload", &BrowserDialog_Stub::reload, 
    "@brief Reload the current page"
  ) +
  gsi::method ("set_home", &BrowserDialog_Stub::set_home, 
    "@brief Set the browser's initial and current URL which is selected if the \"home\" location is choosen\n"
    "@args home_url\n"
  ) +
  gsi::callback ("closed", &BrowserDialog_Stub::closed, &BrowserDialog_Stub::closed_cb, 
    "@brief Callback when the dialog is closed"
    "\n"
    "This callback can be reimplemented to implement cleanup functionality when the "
    "dialog is closed."
  ),
  "@brief The HTML browser dialog\n"
  "\n"
  "The HTML browser displays HTML code in a browser panel. It receives the code\n"
  "by retrieving it from a given URL.\n"
  "\n"
  "URL's with the special scheme \"int\" are retrieved from a BrowserSource object.\n"
  "This will act as a kind of server for these URL's.\n"
);

static BrowserSource_Stub *new_html (const std::string &html)
{
  return new BrowserSource_Stub (html);
}

template <>
Class<BrowserSource_Stub> Class<BrowserSource_Stub>::decl ("BrowserSource", 
  gsi::constructor ("new_html", &new_html,
    "@brief construct a BrowserSource object with a default HTML string\n"
    "\n"
    "The default HTML string is sent when no specific implementation is provided.\n"
  ) +
  gsi::callback ("get", &BrowserSource_Stub::get, &BrowserSource_Stub::get_cb, 
    "@brief Get the HTML code for a given \"int\" URL.\n"
    "@args url\n"
    "\n"
    "If this method returns an empty string, the browser will not be set to \n"
    "a new location. This allows to implement any functionality behind such links.\n"
    "If the method returns a string, the content of this string is displayed in the HTML\n"
    "browser page."
  ),
  "@brief The \\BrowserDialog source for \"int\" URL's\n"
  "\n"
  "The source object basically acts as a \"server\" for special URL's using \"int\" as the scheme.\n"
  "Classes that want to implement such functionality must derive from BrowserSource and reimplement\n"
  "the \\get method. This method is supposed to deliver a HTML page for the given URL.\n\n"
  "Alternatively to implementing this functionality, a source object may be instantiated using the\n"
  "\\new_html constructor. This will create a source object that simply displays the given string\n"
  "as the initial and only page."
);

// ---------------------------------------------------------------------------------
//  Input dialogs

static StringValue get_string (const std::string &title, const std::string &label, const std::string &value)
{
  bool ok = false;
  QString s = QInputDialog::getText (lay::MainWindow::instance (), title.c_str (), label.c_str (), QLineEdit::Normal, value.c_str (), &ok);
  if (!ok) {
    return StringValue ();
  } else {
    return StringValue (s.ascii ());
  }
}

static StringValue get_string_password (const std::string &title, const std::string &label, const std::string &value)
{
  bool ok = false;
  QString s = QInputDialog::getText (lay::MainWindow::instance (), title.c_str (), label.c_str (), QLineEdit::Password, value.c_str (), &ok);
  if (!ok) {
    return StringValue ();
  } else {
    return StringValue (s.ascii ());
  }
}

static DoubleValue get_double (const std::string &title, const std::string &label, double value, int digits)
{
  bool ok = false;
  double s = QInputDialog::getDouble (lay::MainWindow::instance (), title.c_str (), label.c_str (), value, std::numeric_limits<double>::min (), std::numeric_limits<double>::max (), digits, &ok);
  if (!ok) {
    return DoubleValue ();
  } else {
    return DoubleValue (s);
  }
}

static DoubleValue get_double_ex (const std::string &title, const std::string &label, double value, double dmin, double dmax, int decimals)
{
  bool ok = false;
  double s = QInputDialog::getDouble (lay::MainWindow::instance (), title.c_str (), label.c_str (), value, dmin, dmax, decimals, &ok);
  if (!ok) {
    return DoubleValue ();
  } else {
    return DoubleValue (s);
  }
}

static IntValue get_int (const std::string &title, const std::string &label, int value)
{
  bool ok = false;
  int s = QInputDialog::getInteger (lay::MainWindow::instance (), title.c_str (), label.c_str (), value, std::numeric_limits<int>::min (), std::numeric_limits<int>::max (), 1, &ok);
  if (!ok) {
    return IntValue ();
  } else {
    return IntValue (s);
  }
}

static IntValue get_int_ex (const std::string &title, const std::string &label, int value, int dmin, int dmax, int step)
{
  bool ok = false;
  int s = QInputDialog::getInteger (lay::MainWindow::instance (), title.c_str (), label.c_str (), value, dmin, dmax, step, &ok);
  if (!ok) {
    return IntValue ();
  } else {
    return IntValue (s);
  }
}

static StringValue get_item (const std::string &title, const std::string &label, const std::vector<std::string> &items, int selected)
{
  bool ok = false;
  QStringList ilist;
  for (std::vector<std::string>::const_iterator i = items.begin (); i != items.end (); ++i) {
    ilist << i->c_str ();
  }
  QString s = QInputDialog::getItem (lay::MainWindow::instance (), title.c_str (), label.c_str (), ilist, selected, false, &ok);
  if (!ok) {
    return StringValue ();
  } else {
    return StringValue (s.ascii ());
  }
}

struct InputDialog { };

template <>
Class<InputDialog> Class<InputDialog>::decl ("InputDialog", 
  gsi::method ("get_string", &get_string,
    "@brief Open an input dialog requesting a string\n"
    "@args title, label, value\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param value The initial value for the input field\n"
    "@return A \\StringValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ) +
  gsi::method ("get_item", &get_item,
    "@brief Open an input dialog requesting an item from a list\n"
    "@args title, label, items, value\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param items The list of items to show in the selection element\n"
    "@param selection The initial selection (index of the element selected intially)\n"
    "@return A \\StringValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ) +
  gsi::method ("get_string_password", &get_string_password,
    "@brief Open an input dialog requesting a string without showing the actual characters entered\n"
    "@args title, label, value\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param value The initial value for the input field\n"
    "@return A \\StringValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ) +
  gsi::method ("get_double", &get_double,
    "@brief Open an input dialog requesting a floating-point value\n"
    "@args title, label, value\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param value The initial value for the input field\n"
    "@param digits The number of digits allowed\n"
    "@return A \\DoubleValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ) +
  gsi::method ("get_double_ex", &get_double_ex,
    "@brief Open an input dialog requesting a floating-point value with enhanced capabilities\n"
    "@args title, label, value, min, max, digits\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param value The initial value for the input field\n"
    "@param min The minimum value allowed\n"
    "@param max The maximum value allowed\n"
    "@param digits The number of digits allowed\n"
    "@return A \\DoubleValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ) +
  gsi::method ("get_int", &get_int,
    "@brief Open an input dialog requesting an integer value\n"
    "@args title, label, value\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param value The initial value for the input field\n"
    "@return A \\IntValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ) +
  gsi::method ("get_int_ex", &get_int_ex,
    "@brief Open an input dialog requesting an integer value with enhanced capabilities\n"
    "@args title, label, value, min, max, step\n"
    "@param title The title to display for the dialog\n"
    "@param label The label text to display for the dialog\n"
    "@param value The initial value for the input field\n"
    "@param min The minimum value allowed\n"
    "@param max The maximum value allowed\n"
    "@param step The step size for the spin buttons\n"
    "@return A \\IntValue object with has_value? set to true, if \"Ok\" was pressed and the value given in it's value attribute\n"
  ),
  "@brief Various methods to open a dialog requesting data entry"
);

// ---------------------------------------------------------------------------------
//  FileDialog

struct FileDialog { };

static StringValue get_existing_dir (const std::string &title, const std::string &dir)
{
  QString f = QFileDialog::getExistingDirectory (lay::MainWindow::instance (), title.c_str (), dir.c_str ());
  if (f.isEmpty ()) {
    return StringValue ();
  } else {
    return StringValue (f.ascii ());
  }
}

static StringListValue get_open_file_names (const std::string &title, const std::string &dir, const std::string &filter)
{
  QStringList f = QFileDialog::getOpenFileNames (lay::MainWindow::instance (), title.c_str (), dir.c_str (), filter.c_str ());
  if (f.isEmpty ()) {
    return StringListValue ();
  } else {
    std::vector <std::string> l;
    for (QStringList::const_iterator s = f.begin (); s != f.end (); ++s) {
      l.push_back (std::string (s->ascii ()));
    }
    return StringListValue (l);
  }
}

static StringValue get_open_file_name (const std::string &title, const std::string &dir, const std::string &filter)
{
  QString f = QFileDialog::getOpenFileName (lay::MainWindow::instance (), title.c_str (), dir.c_str (), filter.c_str ());
  if (f.isEmpty ()) {
    return StringValue ();
  } else {
    return StringValue (f.ascii ());
  }
}

static StringValue get_save_file_name (const std::string &title, const std::string &dir, const std::string &filter)
{
  QString f = QFileDialog::getSaveFileName (lay::MainWindow::instance (), title.c_str (), dir.c_str (), filter.c_str ());
  if (f.isEmpty ()) {
    return StringValue ();
  } else {
    return StringValue (f.ascii ());
  }
}

template <>
Class<FileDialog> Class<FileDialog>::decl ("FileDialog", 
  gsi::method ("get_existing_dir", &get_existing_dir,
    "@brief Open a dialog to select a directory\n"
    "@args title, dir\n"
    "\n"
    "@param title The title of the dialog\n"
    "@param dir The directory selected initially\n"
    "@return A \\StringValue object that contains the directory path selected or with has_value? = false if \"Cancel\" was pressed\n"
  ) +
  gsi::method ("get_open_file_names", &get_open_file_names,
    "@brief Select one or multiple files for opening\n"
    "@args title, dir, filter\n"
    "\n"
    "@param title The title of the dialog\n"
    "@param dir The directory selected initially\n"
    "@param filter The filters available, for example \"Images (*.png *.xpm *.jpg);;Text files (*.txt);;XML files (*.xml)\"\n"
    "@return A \\StringListValue object that contains the files selected or with has_value? = false if \"Cancel\" was pressed\n"
  ) +
  gsi::method ("get_open_file_name", &get_open_file_name,
    "@brief Select one file for opening\n"
    "@args title, dir, filter\n"
    "\n"
    "@param title The title of the dialog\n"
    "@param dir The directory selected initially\n"
    "@param filter The filters available, for example \"Images (*.png *.xpm *.jpg);;Text files (*.txt);;XML files (*.xml)\"\n"
    "@return A \\StringValue object that contains the files selected or with has_value? = false if \"Cancel\" was pressed\n"
  ) +
  gsi::method ("get_save_file_name", &get_save_file_name,
    "@brief Select one file for writing\n"
    "@args title, dir, filter\n"
    "\n"
    "@param title The title of the dialog\n"
    "@param dir The directory selected initially\n"
    "@param filter The filters available, for example \"Images (*.png *.xpm *.jpg);;Text files (*.txt);;XML files (*.xml)\"\n"
    "@return A \\StringValue object that contains the files selected or with has_value? = false if \"Cancel\" was pressed\n"
  ),
  "@brief Various methods to request a file name"
);


// ---------------------------------------------------------------------------------
//  MessageBox

static int b_ok ()     { return 1 << 0; }
static int b_yes ()    { return 1 << 1; }
static int b_no ()     { return 1 << 2; }
static int b_abort ()  { return 1 << 3; }
static int b_retry ()  { return 1 << 4; }
static int b_ignore () { return 1 << 5; }
static int b_cancel () { return 1 << 6; }

static int qbuttons [] = {
    int (QMessageBox::Ok),
    int (QMessageBox::Yes),
    int (QMessageBox::No),
    int (QMessageBox::Abort),
    int (QMessageBox::Retry),
    int (QMessageBox::Ignore),
    int (QMessageBox::Cancel)
};

#if QT_VERSION >= 0x040200

typedef QMessageBox::StandardButton (*msg_func) (QWidget *, const QString &, const QString &, QMessageBox::StandardButtons, QMessageBox::StandardButton);

static int show_msg_box (msg_func mf, const std::string &title, const std::string &text, int buttons)
{
  QMessageBox::StandardButton b = QMessageBox::StandardButton (0);
  for (unsigned int i = 0; i < sizeof (qbuttons) / sizeof (qbuttons [0]); ++i) {
    if ((buttons & (1 << i)) != 0) {
      buttons &= ~(1 << i);
      b = QMessageBox::StandardButton (int (b) | qbuttons [i]);
    }
  }

  int res = (*mf) (lay::MainWindow::instance (), title.c_str (), text.c_str (), QMessageBox::StandardButton (b), QMessageBox::NoButton);

  for (unsigned int i = 0; i < sizeof (qbuttons) / sizeof (qbuttons [0]); ++i) {
    if (res == qbuttons [i]) {
      return (1 << i);
    }
  }
  return 0;
}

#else

typedef int (*msg_func) (QWidget *, const QString &, const QString &, int, int, int);

static int show_msg_box (msg_func mf, const std::string &title, const std::string &text, int buttons)
{
  std::vector<int> b;

  for (unsigned int i = 0; i < sizeof (qbuttons) / sizeof (qbuttons [0]); ++i) {
    if ((buttons & (1 << i)) != 0) {
      buttons &= ~(1 << i);
      b.push_back (qbuttons [i]);
    }
  }

  while (b.size () < 3) {
    b.push_back (0);
  }

  int res = (*mf) (lay::MainWindow::instance (), title.c_str (), text.c_str (), b[0], b[1], b[2]);

  for (unsigned int i = 0; i < sizeof (qbuttons) / sizeof (qbuttons [0]); ++i) {
    if (res == qbuttons [i]) {
      return (1 << i);
    }
  }
  return 0;
}

#endif

static int critical (const std::string &title, const std::string &text, int buttons)
{
  return show_msg_box (&QMessageBox::critical, title, text, buttons);
}

static int info (const std::string &title, const std::string &text, int buttons)
{
  return show_msg_box (&QMessageBox::information, title, text, buttons);
}

static int question (const std::string &title, const std::string &text, int buttons)
{
  return show_msg_box (&QMessageBox::question, title, text, buttons);
}

static int warning (const std::string &title, const std::string &text, int buttons)
{
  return show_msg_box (&QMessageBox::warning, title, text, buttons);
}

struct MessageBox { };

template <>
Class<MessageBox> Class<MessageBox>::decl ("MessageBox", 
  gsi::method ("b_ok",               &b_ok,              "@brief b_ok constant describing the respective button label") +
  gsi::method ("b_cancel",           &b_cancel,          "@brief b_cancel constant describing the respective button label") +
  gsi::method ("b_yes",              &b_yes,             "@brief b_yes constant describing the respective button label") +
  gsi::method ("b_no",               &b_no,              "@brief b_no constant describing the respective button label") +
  gsi::method ("b_abort",            &b_abort,           "@brief b_abort constant describing the respective button label") +
  gsi::method ("b_retry",            &b_retry,           "@brief b_retry constant describing the respective button label") +
  gsi::method ("b_ignore",           &b_ignore,          "@brief b_ignore constant describing the respective button label") +
  gsi::method ("warning", &warning,
    "@brief Open a warning message box\n"
    "@args title, text, buttons\n"
    "@param title The title of the window\n"
    "@param text The text to show\n"
    "@param buttons A combination (+) of b_... constants describing the buttons to show for the message box\n"
    "@return The b_... constant describing the button that was pressed\n"
  ) +
  gsi::method ("question", &question,
    "@brief Open a question message box\n"
    "@args title, text, buttons\n"
    "@param title The title of the window\n"
    "@param text The text to show\n"
    "@param buttons A combination (+) of b_... constants describing the buttons to show for the message box\n"
    "@return The b_... constant describing the button that was pressed\n"
  ) +
  gsi::method ("info", &info,
    "@brief Open a information message box\n"
    "@args title, text, buttons\n"
    "@param title The title of the window\n"
    "@param text The text to show\n"
    "@param buttons A combination (+) of b_... constants describing the buttons to show for the message box\n"
    "@return The b_... constant describing the button that was pressed\n"
  ) +
  gsi::method ("critical", &critical,
    "@brief Open a critical (error) message box\n"
    "@args title, text, buttons\n"
    "@param title The title of the window\n"
    "@param text The text to show\n"
    "@param buttons A combination (+) of b_... constants describing the buttons to show for the message box\n"
    "@return The b_... constant describing the button that was pressed\n"
  ),
  "@brief Various methods to display message boxes"
);

}

