
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiObject
#define _HDR_gsiObject

#include <memory>

//  For a comprehensive documentation see gsi.h

namespace gsi
{

template <class X> class Class;

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object can be copied
 */
struct CanCopySemanticsTag { 
  CanCopySemanticsTag () { } 
};

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object cannot be copied
 */
struct CantCopySemanticsTag { 
  CantCopySemanticsTag () { } 
};

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object can be default-created
 */
struct CanDefaultCreateSemanticsTag { 
  CanDefaultCreateSemanticsTag () { } 
};

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object cannot be default-created
 */
struct CantDefaultCreateSemanticsTag { 
  CantDefaultCreateSemanticsTag () { }
};

/**
 *  @brief The default properties of a class
 */
template<class X> 
struct ClassTag 
{
  typedef CanCopySemanticsTag copy_semantics;
  typedef CanDefaultCreateSemanticsTag create_semantics;
};

/**
 *  @brief A lower base class for Object<X>
 */
class ObjectBase
{
public:
  ObjectBase () 
    : mp_data (0)
  {
    //  .. nothing yet ..
  }

  ~ObjectBase ()
  {
    mp_data = 0;
  }

  void *data () const
  {
    return mp_data;
  }

protected:
  void *mp_data;
};

template <class X> class Object;

/**
 *  @brief The handler for the client-side data of Object<X>
 *
 *  If the scripting client desires to attach data to the Object<X> objects, it
 *  must reimplement this handler class and register a handler for each ClassBase
 *  object that is registered.
 */
class ClientHandler
{
public:
  ClientHandler () { }
  virtual ~ClientHandler () { }
  virtual void attach (void *&data) = 0;
  virtual void detach (void *&data) = 0;
};

/**
 *  @brief A helper class that handles attachment and detachment of classes to an object
 *
 *  This class avoids having to pull in the whole class framework here.
 */
template <class X>
class ObjectClassAdaptor
{
public:
  ObjectClassAdaptor () 
  {
  }

  ClientHandler *client_handler () const
  {
    return ms_client_handler.get();
  }

  void register_client_handler (ClientHandler *handler) const
  {
    ms_client_handler.reset (handler);
  }

  static void attach (Object<X> *obj) 
  {
    if (ms_client_handler.get()) {
      ms_client_handler->attach (obj->mp_data);
    }
  }

  static void detach (Object<X> *obj) 
  {
    if (ms_client_handler.get() && obj->mp_data) {
      ms_client_handler->detach (obj->mp_data);
    }
  }

private:
  static std::auto_ptr<ClientHandler> ms_client_handler;
};

template<class X>
std::auto_ptr<ClientHandler> ObjectClassAdaptor<X>::ms_client_handler;

/**
 *  @brief A generic base class
 *
 *  Objects that are derived from this template (A being derived from gsi::Object<A>)
 *  offer additional functionality over objects not derived from this base class such as
 *  notification when the C++ object is destroyed and persistency of the client-side object
 *  state.
 */

template <class X>
class Object
  : public ObjectBase
{
public:
  Object ();
  Object (const Object<X> &);
  Object &operator= (const Object<X> &);
  ~Object ();

private:
  friend class ObjectClassAdaptor<X>;
};

//  TODO: currently the client data is not copied. As a consequence, copying the C++
//  objects does not copy the Ruby part of the object but rather leaves it unitialized.
//  Since however, the client-side data feature is supposed to be used to non-copying 
//  objects mainly, this problem is not considered serious for now.

template<class X> inline Object<X>::Object () 
{ 
  ObjectClassAdaptor<X>::attach (this);
}

template<class X> inline Object<X>::Object (const Object<X> &) 
{ 
  ObjectClassAdaptor<X>::attach (this);
}

template<class X> inline Object<X> &Object<X>::operator= (const Object<X> &)
{
  return *this;
}

template<class X> inline Object<X>::~Object ()
{
  ObjectClassAdaptor<X>::detach (this);
}

}

#endif

