
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiSerialisation
#define _HDR_gsiSerialisation

#include <list>

//  For a comprehensive documentation see gsi.h

namespace gsi
{

struct ArglistUnderflowException
  : public tl::Exception
{
  ArglistUnderflowException ()
    : tl::Exception (tl::translate ("To few arguments or no return value supplied"))
  { }
};

struct NilPointerToReference
  : public tl::Exception
{
  NilPointerToReference ()
    : tl::Exception (tl::translate ("nil object passed to a reference"))
  { }
};

/**
 *  @brief Determine the size for one item in the serialization buffer
 */
template<class X>
size_t item_size () 
{
  return (sizeof (X) + (sizeof (void *) - 1)) & ~(sizeof (void *) - 1); 
}

/**
 *  @brief This class provides the basic argument serialization mechanism for the C++/scripting interface
 */
class SerialArgs 
{
public:
  SerialArgs ()
    : mp_buffer (0)
  {
    mp_write = mp_read = mp_buffer;
  }

  SerialArgs (unsigned int len)
    : mp_buffer (len > 0 ? new char [len] : 0)
  {
    mp_write = mp_read = mp_buffer;
  }

  ~SerialArgs ()
  {
    if (mp_buffer) {
      delete [] mp_buffer;
    }
    mp_buffer = 0;
  }

  char *cptr ()
  {
    return mp_buffer;
  }

  const char *cptr () const
  {
    return mp_buffer;
  }

  template <class X> 
  X *set_value (direct_tag, const X &x)
  {
    X *r = (X *)mp_write;
    new (r) X(x);
    mp_write += item_size<X> ();
    return r;
  }

  template <class X> 
  void set_value (ref_tag, X &x) 
  {
    *((X **)mp_write) = &x;
    mp_write += item_size<X *> ();
  }

  template <class X> 
  void set_value (ptr_tag, X *x)
  {
    *((X **)mp_write) = x;
    mp_write += item_size<X *> ();
  }

  template <class X> 
  void set_value (cref_tag, const X &x)
  {
    *((X const **)mp_write) = &x;
    mp_write += item_size<const X *> ();
  }

  template <class X> 
  void set_value (cptr_tag, const X *x)
  {
    *((X const **)mp_write) = x;
    mp_write += item_size<const X *> ();
  }

  void set_value (ccptr_tag, const char *x)
  {
    *((const char **)mp_write) = x;
    mp_write += item_size<const char *> ();
  }

  template <class X>
  X get_value (direct_pod_tag)
  {
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X> 
  X get_value (direct_npod_tag)
  {
    check_data ();
    // :KLUDGE: this is not efficient. We copy an std::string object twice at least:
    X x = *(X *)mp_read;
    ((X *)mp_read)->~X ();
    mp_read += item_size<X> ();
    return x;
  }

  template <class X> 
  X get_value (ccptr_tag)
  {
    const char *r = *((const char **)mp_read);
    mp_read += item_size<const char *> ();
    return r;
  }

  template <class X> 
  X get_value (ref_tag)
  {
    typedef typename type_traits<X>::value_type value_type;
    check_data ();
    value_type *r = *((value_type **)mp_read);
    mp_read += item_size<value_type *> ();
    if (! r) {
      throw NilPointerToReference ();
    }
    return *r;
  }

  template <class X> 
  X get_value (cref_tag)
  {
    typedef typename type_traits<X>::value_type value_type;
    check_data ();
    const value_type *r = *((const value_type **)mp_read);
    mp_read += item_size<const value_type *> ();
    if (! r) {
      throw NilPointerToReference ();
    }
    return *r;
  }

  template <class X> 
  X get_value (ptr_tag)
  {
    //  X is actually an (X *)
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X> 
  X get_value (cptr_tag)
  {
    //  X is actually an (const X *)
    check_data ();
    X r = *((X *)mp_read);
    mp_read += item_size<X> ();
    return r;
  }

  template <class X> 
  X *take_object ()
  {
    check_data ();
    X *x = new X (*((X *)mp_read));
    ((X *)mp_read)->~X ();
    mp_read += item_size<X> ();
    return x;
  }

  void check_data () const
  {
    if (! (mp_read && mp_read < mp_write)) {
      throw ArglistUnderflowException ();
    }
  }

private:
  char *mp_buffer;
  char *mp_read, *mp_write;
  std::list<std::string> m_strbufs;
};

}

#endif

