
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiTypes
#define _HDR_gsiTypes

#include <string>
#include <vector>
#include <stdexcept>

#include "tlInternational.h"
#include "tlException.h"
#include "tlVariant.h"

//  For a comprehensive documentation see gsi.h

namespace gsi
{

// ---------------------------------------------------------------------------------
//  Type system of GSI

class ClassBase;
template<class X> struct ClassTag;
template <class X> class Class;
template <class I, class R, class V> class IterAdaptor;
template <class R, class V> class IterPtrAdaptor;
template <class R, class V> class ConstIterPtrAdaptor;
template <class I, class R, class V> class FreeIterAdaptor;

template<class X>
struct type_tag
{
  // .. intentionally left blank ..
};

/**
 *   @brief Basic GSI type constants
 */
enum BasicType
{
  T_void         = 0,
  T_bool         = 1,
  T_int          = 2,
  T_uint         = 3,
  T_long         = 4,
  T_ulong        = 5,
  T_longlong     = 6,
  T_ulonglong    = 7,
  T_double       = 8,
  T_var          = 9,
  T_string       = 10,
  T_string_ccptr = 11,
  T_object       = 12,
  T_object_ref   = 13,
  T_object_cref  = 14,
  T_object_new   = 15
};

/**
 *  @brief Type tags
 */
struct direct_tag { };
struct direct_npod_tag : public direct_tag { };
struct direct_pod_tag : public direct_tag { };
struct cref_tag { };
struct ref_tag { };
struct cptr_tag { };
struct ptr_tag { };

struct bool_tag : public direct_pod_tag { };
struct int_tag : public direct_pod_tag { };
struct uint_tag : public direct_pod_tag { };
struct long_tag : public direct_pod_tag { };
struct ulong_tag : public direct_pod_tag { };
struct longlong_tag : public direct_pod_tag { };
struct ulonglong_tag : public direct_pod_tag { };
struct double_tag : public direct_pod_tag { };
struct string_tag : public direct_npod_tag { };
struct var_tag : public direct_npod_tag { };
struct ccptr_tag : public direct_pod_tag { };

struct bool_cref_tag : public cref_tag { };
struct int_cref_tag : public cref_tag { };
struct uint_cref_tag : public cref_tag { };
struct long_cref_tag : public cref_tag { };
struct ulong_cref_tag : public cref_tag { };
struct longlong_cref_tag : public cref_tag { };
struct ulonglong_cref_tag : public cref_tag { };
struct double_cref_tag : public cref_tag { };
struct string_cref_tag : public cref_tag { };
struct var_cref_tag : public cref_tag { };

struct bool_ref_tag : public ref_tag { };
struct int_ref_tag : public ref_tag { };
struct uint_ref_tag : public ref_tag { };
struct long_ref_tag : public ref_tag { };
struct ulong_ref_tag : public ref_tag { };
struct longlong_ref_tag : public ref_tag { };
struct ulonglong_ref_tag : public ref_tag { };
struct double_ref_tag : public ref_tag { };
struct string_ref_tag : public ref_tag { };
struct var_ref_tag : public ref_tag { };

struct x_tag : public direct_npod_tag { };
struct x_cref_tag : public cref_tag { };
struct x_cptr_tag : public cptr_tag { };
struct x_cptr_cref_tag : public cptr_tag { };
struct x_cptr_ref_tag : public cptr_tag { };
struct x_ref_tag : public ref_tag { };
struct x_ptr_tag : public ptr_tag { };
struct x_ptr_cref_tag : public ptr_tag { };
struct x_ptr_ref_tag : public ptr_tag { };

struct vector_tag : public direct_npod_tag { };
struct vector_cref_tag : public cref_tag { };
struct vector_ref_tag : public ref_tag { };

/**
 *  @brief Type traits that map a C++ type to a certain tag
 */
template <class X> struct type_traits;

template <>
struct type_traits<bool>
{
  typedef bool_tag tag;
  typedef bool value_type;
  static BasicType code () { return T_bool; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<int>
{
  typedef int_tag tag;
  typedef int value_type;
  static BasicType code () { return T_int; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<unsigned int>
{
  typedef uint_tag tag;
  typedef unsigned int value_type;
  static BasicType code () { return T_uint; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<long>
{
  typedef long_tag tag;
  typedef long value_type;
  static BasicType code () { return T_long; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<long long>
{
  typedef longlong_tag tag;
  typedef long long value_type;
  static BasicType code () { return T_longlong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<unsigned long long>
{
  typedef ulonglong_tag tag;
  typedef unsigned long long value_type;
  static BasicType code () { return T_ulonglong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<unsigned long>
{
  typedef ulong_tag tag;
  typedef unsigned long value_type;
  static BasicType code () { return T_ulong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<double>
{
  typedef double_tag tag;
  typedef double value_type;
  static BasicType code () { return T_double; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<std::string>
{
  typedef string_tag tag;
  typedef std::string value_type;
  static BasicType code () { return T_string; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const char *>
{
  typedef ccptr_tag tag;
  typedef const char *value_type;
  static BasicType code () { return T_string_ccptr; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<tl::Variant>
{
  typedef var_tag tag;
  typedef tl::Variant value_type;
  static BasicType code () { return T_var; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};


template <>
struct type_traits<const bool &>
{
  typedef bool_cref_tag tag;
  typedef bool value_type;
  static BasicType code () { return T_bool; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const int &>
{
  typedef int_cref_tag tag;
  typedef int value_type;
  static BasicType code () { return T_int; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const unsigned int &>
{
  typedef uint_cref_tag tag;
  typedef unsigned int value_type;
  static BasicType code () { return T_uint; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const long &>
{
  typedef long_cref_tag tag;
  typedef long value_type;
  static BasicType code () { return T_long; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const unsigned long &>
{
  typedef ulong_cref_tag tag;
  typedef unsigned long value_type;
  static BasicType code () { return T_ulong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const long long &>
{
  typedef longlong_cref_tag tag;
  typedef long long value_type;
  static BasicType code () { return T_longlong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const unsigned long long &>
{
  typedef ulonglong_cref_tag tag;
  typedef unsigned long long value_type;
  static BasicType code () { return T_ulonglong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const double &>
{
  typedef double_cref_tag tag;
  typedef double value_type;
  static BasicType code () { return T_double; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const std::string &>
{
  typedef string_cref_tag tag;
  typedef std::string value_type;
  static BasicType code () { return T_string; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<const tl::Variant &>
{
  typedef var_cref_tag tag;
  typedef tl::Variant value_type;
  static BasicType code () { return T_var; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};


template <>
struct type_traits<bool &>
{
  typedef bool_ref_tag tag;
  typedef bool value_type;
  static BasicType code () { return T_bool; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<int &>
{
  typedef int_ref_tag tag;
  typedef int value_type;
  static BasicType code () { return T_int; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<unsigned int &>
{
  typedef uint_ref_tag tag;
  typedef unsigned int value_type;
  static BasicType code () { return T_uint; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<long &>
{
  typedef long_ref_tag tag;
  typedef long value_type;
  static BasicType code () { return T_long; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<unsigned long &>
{
  typedef ulong_ref_tag tag;
  typedef unsigned long value_type;
  static BasicType code () { return T_ulong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<long long &>
{
  typedef longlong_ref_tag tag;
  typedef long long value_type;
  static BasicType code () { return T_longlong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<unsigned long long &>
{
  typedef ulonglong_ref_tag tag;
  typedef unsigned long long value_type;
  static BasicType code () { return T_ulonglong; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<double &>
{
  typedef double_ref_tag tag;
  typedef double value_type;
  static BasicType code () { return T_double; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<std::string &>
{
  typedef string_ref_tag tag;
  typedef std::string value_type;
  static BasicType code () { return T_string; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <>
struct type_traits<tl::Variant &>
{
  typedef var_ref_tag tag;
  typedef tl::Variant value_type;
  static BasicType code () { return T_var; }
  static const ClassBase *cls_decl () { return 0; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X>
{
  typedef typename type_traits<X>::tag tag;
  typedef typename type_traits<X>::value_type value_type;
  static BasicType code () { return type_traits<X>::code (); }
  static const ClassBase *cls_decl () { return type_traits<X>::cls_decl (); }
  static bool is_ref () { return type_traits<X>::is_ref (); }
  static bool is_vector () { return type_traits<X>::is_vector (); }
  static bool is_iter () { return type_traits<X>::is_iter (); }
};

template <class X>
struct type_traits
{
  typedef x_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X &>
{
  typedef x_cref_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_cref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X *>
{
  typedef x_cptr_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_cref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X * const &>
{
  typedef x_cptr_cref_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_cref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<const X * &>
{
  typedef x_cptr_ref_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_cref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X &>
{
  typedef x_ref_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_ref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X *>
{
  typedef x_ptr_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_ref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X * const &>
{
  typedef x_ptr_cref_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_ref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits<X * &>
{
  typedef x_ptr_ref_tag tag;
  typedef X value_type;
  static BasicType code () { return T_object_ref; }
  static const ClassBase *cls_decl () { return &Class<X>::decl; }
  static bool is_ref () { return true; }
  static bool is_vector () { return false; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits< std::vector<X> >
{
  typedef vector_tag tag;
  typedef std::vector<X> value_type;
  static BasicType code () { return type_traits<X>::code (); }
  static const ClassBase *cls_decl () { return type_traits<X>::cls_decl (); }
  static bool is_ref () { return false; }
  static bool is_vector () { return true; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits< const std::vector<X> & >
{
  typedef vector_cref_tag tag;
  typedef std::vector<X> value_type;
  static BasicType code () { return type_traits<X>::code (); }
  static const ClassBase *cls_decl () { return type_traits<X>::cls_decl (); }
  static bool is_ref () { return true; }
  static bool is_vector () { return true; }
  static bool is_iter () { return false; }
};

template <class X>
struct type_traits< std::vector<X> & >
{
  typedef vector_ref_tag tag;
  typedef std::vector<X> value_type;
  static BasicType code () { return type_traits<X>::code (); }
  static const ClassBase *cls_decl () { return type_traits<X>::cls_decl (); }
  static bool is_ref () { return true; }
  static bool is_vector () { return true; }
  static bool is_iter () { return false; }
};

template <class I, class R, class V>
struct type_traits< IterAdaptor<I, R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return true; }
};

template <class R, class V>
struct type_traits< IterPtrAdaptor<R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return true; }
};

template <class R, class V>
struct type_traits< ConstIterPtrAdaptor<R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return true; }
};

template <class I, class R, class V>
struct type_traits< FreeIterAdaptor<I, R, V> >
{
  typedef typename type_traits<R>::tag tag;
  typedef typename type_traits<R>::value_type value_type;
  static BasicType code () { return type_traits<R>::code (); }
  static const ClassBase *cls_decl () { return type_traits<R>::cls_decl (); }
  static bool is_ref () { return false; }
  static bool is_vector () { return false; }
  static bool is_iter () { return true; }
};

/**
 *  @brief Generic argument type declaration
 *
 *  The type declaration carries the basic type code and flags, if the argument 
 *  is an iterator delivering the given types of a vector of the given types and
 *  adds class informations if the type is "T_object", "T_object_ref", "T_object_cref" or "T_object_new".
 */
struct ArgType
{
  ArgType (BasicType t, bool v, bool r, bool i, const ClassBase *c) 
    : type (t), vector (v), ref (r), iter (i), cls (c)
  { }

  ArgType () 
    : type (T_void), vector (false), ref (false), iter (false), cls (0) 
  { }

  template <class X>
  void init () 
  { 
    type   = type_traits<X>::code ();
    vector = type_traits<X>::is_vector ();
    ref    = type_traits<X>::is_ref ();
    iter   = type_traits<X>::is_iter ();
    cls    = type_traits<X>::cls_decl ();
  }

  void check_type (const ArgType &a) const;

  bool operator== (const ArgType &b) const
  {
    return type == b.type && vector == b.vector && ref == b.ref && cls == b.cls;
  }

  bool operator!= (const ArgType &b) const
  {
    return !operator== (b);
  }

  std::string to_string () const;

  BasicType type;
  bool vector;
  bool ref;
  bool iter;
  mutable const ClassBase *cls;
};

struct IncompatibleTypeException
  : public tl::Exception
{
  IncompatibleTypeException (const ArgType &got, const ArgType &want)
    : tl::Exception (tl::translate ("Incompatible types: got '") + got.to_string () + tl::translate ("', want '") + want.to_string () + "'")
  { }
};

inline void ArgType::check_type (const ArgType &a) const
{
  if (*this != a) {
    throw IncompatibleTypeException (a, *this);
  }
}

/**
 *  @brief A helper class that provides a base class for wrapped pointers for TempObject
 */
class TempObjectBase 
{
public:
  virtual ~TempObjectBase () { }
};

/**
 *  @brief A helper class that provides a wrapper to a custom type for TempObject
 */
template <class X>
class TempObjectCont : public TempObjectBase 
{
public:
  TempObjectCont (X *x) : mp_x (x) { }
  ~TempObjectCont () { delete mp_x; }
private:
  X *mp_x;
};

/**
 *  @brief A simple autopointer class to hold references to temporary objects
 */
class TempObject
{
public:
  TempObject () : mp_b (0) { }

  template <class X>
  TempObject (X *x) : mp_b (0) 
  { 
    set (x);  
  }

  ~TempObject () {
    if (mp_b) {
      delete mp_b;
    }
    mp_b = 0;
  }

  template <class X>
  void set (X *x) { 
    tl_assert (mp_b == 0);
    mp_b = new TempObjectCont<X> (x); 
  }

private:
  TempObjectBase *mp_b;
};

// ---------------------------------------------------------------------------------
//  Utility: determine non-case "base" of X (nc_x(const X) == X)

template <class X>
struct non_const_x {
  typedef X nc_x;
};

template <class X>
struct non_const_x<const X> {
  typedef X nc_x;
};

// ---------------------------------------------------------------------------------
//  Utility: determine, if X is const 

template <class X>
struct is_const_x {
  static bool value () { return false; }
};

template <class X>
struct is_const_x<const X> {
  static bool value () { return true; }
};

}

#endif

