
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbUserObject.h"
#include "dbClipboard.h"
#include "dbEdgeProcessor.h"
#include "tlString.h"
#include "tlAssert.h"
#include "layPlugin.h"
#include "layRenderer.h"
#include "laySnap.h"
#include "layLayoutView.h"
#include "layConfig.h"
#include "layLayoutCanvas.h"
#include "layProperties.h"
#include "layMainWindow.h"
#include "layExceptions.h"
#include "imgService.h"
#include "AddNewImageDialog.h"

namespace img
{

// -------------------------------------------------------------

class AddNewImageDialog 
  : public QDialog, 
    public Ui::AddNewImageDialog
{
public:
  AddNewImageDialog (QWidget *parent, img::Object *image_object)
    : QDialog (parent), mp_image_object (image_object)
  {
    setupUi (this);
    properties_frame->set_direct_image (image_object);
    properties_frame->update ();
  }

  virtual void accept ()
  {
    BEGIN_PROTECTED 

    properties_frame->set_direct_image (mp_image_object);
    properties_frame->apply ();

    if (mp_image_object->is_empty ()) {
      throw tl::Exception (tl::translate ("No data loaded for that image"));
    }

    QDialog::accept ();

    END_PROTECTED
  }

private:
  img::Object *mp_image_object;
};

// -------------------------------------------------------------

static void 
draw_image (const img::Object &image_object, const lay::Viewport &vp, lay::ViewObjectCanvas &canvas) 
{ 
  // TODO: currently, the images can only be rendered to a bitmap canvas ..
  lay::BitmapViewObjectCanvas *bmp_canvas = dynamic_cast<lay::BitmapViewObjectCanvas *> (&canvas);
  if (! bmp_canvas) {
    return;
  }

  size_t source_width = image_object.width ();
  size_t source_height = image_object.height ();

  QImage &qimage = bmp_canvas->bg_image ();
  db::DBox source_image_box (0.0, 0.0, source_width, source_height);

  //  it the transformation from QImage pixel coordinates (in the "bottom first" orientation) into the
  //  image object's coordinate space.
  db::DCplxTrans it = (vp.trans () * image_object.trans ()).inverted ();

  for (int sl = 0; sl < qimage.height (); ++sl) {

    int y = qimage.height () - sl - 1;

    db::DEdge scanline (db::DPoint (0, y), db::DPoint (qimage.width (), y));
    scanline.transform (it);
    scanline = db::DEdge (db::DPoint (scanline.p1 ().x () / image_object.pixel_width (), scanline.p1 ().y () / image_object.pixel_height ()),
                          db::DPoint (scanline.p2 ().x () / image_object.pixel_width (), scanline.p2 ().y () / image_object.pixel_height ()));

    db::DVector scanline_step (1.0, 0.0);
    scanline_step.transform (it);
    scanline_step.set_x (scanline_step.x () / image_object.pixel_width ());
    scanline_step.set_y (scanline_step.y () / image_object.pixel_height ());

    std::pair<bool, db::DEdge> clipped = scanline.clipped (source_image_box);
    if (clipped.first) {

      double pixel_length = scanline_step.length ();
      int xstart = int (ceil (clipped.second.p1 ().distance (scanline.p1 ()) / pixel_length));
      int xstop = int (floor (clipped.second.p2 ().distance (scanline.p1 ()) / pixel_length));

      xstart = std::max (0, std::min (xstart, qimage.width ()));
      xstop = std::max (0, std::min (xstop + 1, qimage.width ()));

      db::DPoint source_loc (xstart, y);
      source_loc.transform (it);
      source_loc.set_x (source_loc.x () / image_object.pixel_width ());
      source_loc.set_y (source_loc.y () / image_object.pixel_height ());

      QRgb *scanline_data = (QRgb *) qimage.scanLine (sl) + xstart;
      QRgb *pixel_data = (QRgb *) image_object.pixel_data ();

      for (int x = xstart; x < xstop; ++x) {

        size_t sx = size_t (std::max (0.0, floor (source_loc.x ())));
        size_t sy = size_t (std::max (0.0, floor (source_loc.y ())));
        size_t n = sx + sy * source_width;

        *scanline_data = pixel_data [n];

        source_loc += scanline_step;
        ++scanline_data;

      }

    }

  }
}

// -------------------------------------------------------------

View::View (img::Service *service, const img::Object *image_object, img::View::Mode mode)
  : lay::ViewObject (service->widget ()), 
    mp_service (service), m_mode (mode), mp_image_object (image_object)
{
  //  .. nothing else ..
}

View::~View ()
{
  //  .. nothing else ..
}

void 
View::transform_by (const db::DCplxTrans &t)
{
  if (m_trans != t) {
    m_trans = t;
    redraw ();
  }
}

void 
View::image_object (const img::Object *i)
{
  mp_image_object = i;
  redraw ();
}

void 
View::render (const lay::Viewport &vp, lay::ViewObjectCanvas &canvas) 
{ 
  if (! mp_image_object) {
    return;
  }

  db::DCplxTrans t = vp.trans () * m_trans * mp_image_object->trans ();
  db::DCplxTrans it = t.inverted ();

  db::DBox image_box = db::DBox (0.0, 0.0, mp_image_object->width () * mp_image_object->pixel_width (),  mp_image_object->height () * mp_image_object->pixel_height ());

  if (m_mode == mode_normal) {

    //  Limit the box so it does not create overflows in the integer coordinate space. 
    //  The 0.4 factor is purely empirical.
    db::DBox image_box_world = (db::Box::world () * 0.4).transformed (it);
    image_box &= image_box_world;
    if (image_box.empty ()) {
      return;
    }

    std::vector <db::Polygon> frame_p;
    db::DPolygon image_box_poly (image_box);
    image_box_poly.transform (t);
    frame_p.push_back (db::Polygon::from_double (image_box_poly));

    std::vector <db::Polygon> sized_p;
    db::EdgeProcessor ep;
    ep.size (frame_p, -5, sized_p);

    std::vector <db::Polygon> result;
    ep.boolean (frame_p, sized_p, result, db::BooleanOp::ANotB);

    //  obtain bitmap to render on
    lay::CanvasPlane *plane;
    plane = canvas.plane (lay::ViewOp (canvas.foreground_color ().rgb (), lay::ViewOp::Copy, 6));

    for (std::vector <db::Polygon>::const_iterator r = result.begin (); r != result.end (); ++r) {
      canvas.renderer ().draw (*r, db::CplxTrans (), plane, 0, 0, 0);
    }

  } else if (m_mode == mode_transient_move) {

    //  obtain bitmap to render on
    lay::CanvasPlane *plane;
    plane = canvas.plane (lay::ViewOp (canvas.foreground_color ().rgb (), lay::ViewOp::Copy, 0));

    db::DPolygon image_box_poly (image_box);
    image_box_poly.transform (t);
    canvas.renderer ().draw (image_box_poly, db::CplxTrans (), 0, plane, 0, 0);

    db::DCoord enl = it.ctrans (3);
    db::DCoord cl = 0.0;
    db::DCoord cb = 0.0;
    db::DCoord cr = mp_image_object->width () * mp_image_object->pixel_width ();
    db::DCoord ct = mp_image_object->height () * mp_image_object->pixel_height ();

    std::vector <db::DBox> handle_boxes;
    handle_boxes.reserve (8);
    handle_boxes.push_back (db::DBox (cl, cb, cl, cb));
    handle_boxes.push_back (db::DBox (cl, 0.5 * (cb + ct), cl, 0.5 * (cb + ct)));
    handle_boxes.push_back (db::DBox (cl, ct, cl, ct));
    handle_boxes.push_back (db::DBox (cr, cb, cr, cb));
    handle_boxes.push_back (db::DBox (cr, 0.5 * (cb + ct), cr, 0.5 * (cb + ct)));
    handle_boxes.push_back (db::DBox (cr, ct, cr, ct));
    handle_boxes.push_back (db::DBox (0.5 * (cr + cl), ct, 0.5 * (cr + cl), ct));
    handle_boxes.push_back (db::DBox (0.5 * (cr + cl), cb, 0.5 * (cr + cl), cb));

    for (std::vector <db::DBox>::const_iterator hb = handle_boxes.begin (); hb != handle_boxes.end (); ++hb)
    {
      db::DPolygon handle_box_poly (hb->enlarged (db::DPoint (enl, enl)));
      handle_box_poly.transform (t);
      canvas.renderer ().draw (handle_box_poly, db::CplxTrans (), plane, plane, 0, 0);
    }

  } else {

    //  obtain bitmap to render on
    lay::CanvasPlane *plane;
    plane = canvas.plane (lay::ViewOp (canvas.foreground_color ().rgb (), lay::ViewOp::Copy, 0));

    db::DPolygon image_box_poly (image_box);
    image_box_poly.transform (t);
    canvas.renderer ().draw (image_box_poly, db::CplxTrans (), 0, plane, 0, 0);

  }
}

// -------------------------------------------------------------

//  search range for select in pixels
static unsigned int search_range = 5;   // @@@ make variable?

// -------------------------------------------------------------
//  img::Service implementation

Service::Service (db::Manager *manager, lay::LayoutView *view)
  : lay::BackgroundViewObject (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    mp_transient_image (0),
    m_move_mode (Service::move_none)
{ 
  // place images behind the grid
  z_order (-1);
}

Service::~Service ()
{
  for (std::vector<img::View *>::iterator v = m_selected_images.begin (); v != m_selected_images.end (); ++v) {
    delete *v;
  }
  m_selected_images.clear ();
  clear_transient_selection ();
}

bool 
Service::configure (const std::string &name, const std::string &value)
{
  return false;
}

void 
Service::config_finalize ()
{
  // .. nothing yet ..
}

void 
Service::clear_highlights ()
{
  for (std::vector<img::View *>::iterator v = m_selected_images.begin (); v != m_selected_images.end (); ++v) {
    (*v)->visible (false);
  }
}

void 
Service::restore_highlights ()
{
  for (std::vector<img::View *>::iterator v = m_selected_images.begin (); v != m_selected_images.end (); ++v) {
    (*v)->visible (true);
  }
}

void 
Service::highlight (unsigned int n)
{
  for (std::vector<img::View *>::iterator v = m_selected_images.begin (); v != m_selected_images.end (); ++v) {
    (*v)->visible (n-- == 0);
  }
}

img::Object * 
Service::insert_image (const img::Object &image)
{
  //  create the image and insert
  img::Object *new_image = new img::Object (image);
  mp_view->annotation_shapes ().insert (db::DUserObject (new_image));

  //  send a signal
  m_image_list_changed.signal_observers ();

  return new_image;
}

/**
 *  @brief Helper function to determine which move mode to choose given a certain search box and ant::Object
 */
static bool
dragging_what (const img::Object *iobj, const db::DBox &search_dbox, img::Service::MoveMode &mode, db::DPoint &p1)
{
  mode = img::Service::move_all;
  p1 = search_dbox.center ();

  db::DBox ref_box (search_dbox.transformed (iobj->trans ().inverted ()));

  bool lo = ref_box.overlaps (db::DBox (0, 0, 0, iobj->pixel_height () * iobj->height ()));
  bool ro = ref_box.overlaps (db::DBox (iobj->pixel_width () * iobj->width (), 0, iobj->pixel_width () * iobj->width (), iobj->pixel_height () * iobj->height ()));
  bool bo = ref_box.overlaps (db::DBox (0, 0, iobj->pixel_width () * iobj->width (), 0));
  bool to = ref_box.overlaps (db::DBox (0, iobj->pixel_height () * iobj->height (), iobj->pixel_width () * iobj->width (), iobj->pixel_height () * iobj->height ()));
  bool all = ref_box.overlaps (db::DBox (0, 0, iobj->pixel_width () * iobj->width (), iobj->pixel_height () * iobj->height ()));

  if (lo) {
    if (bo) {
      mode = img::Service::move_ll;
      return true;
    } else if (to) {
      mode = img::Service::move_tl;
      return true;
    } else {
      mode = img::Service::move_l;
      return true;
    }
  } else if (ro) {
    if (bo) {
      mode = img::Service::move_lr;
      return true;
    } else if (to) {
      mode = img::Service::move_tr;
      return true;
    } else {
      mode = img::Service::move_r;
      return true;
    }
  } else if (bo) {
    mode = img::Service::move_b;
    return true;
  } else if (to) {
    mode = img::Service::move_t;
    return true;
  } else if (all) {
    mode = img::Service::move_all;
    return true;
  }
  
  return true;
}

bool 
Service::begin_move (bool sel, const db::DPoint &p, lay::angle_constraint_type ac)
{
  //  cancel any pending move or drag operations 
  widget ()->drag_cancel (); // KLUDGE: every service does this to the same service manager

  //  choose move mode
  if (sel) {

    m_move_mode = move_selected;
    m_p1 = p;
    m_trans = db::DTrans ();

    for (std::vector <img::View *>::iterator r = m_selected_images.begin (); r != m_selected_images.end (); ++r) {
      (*r)->thaw ();
    }
    return false;

  } else {
  
    m_move_mode = move_none;

    //  compute search box
    double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
    db::DBox search_dbox = db::DBox (p, p).enlarged (db::DPoint (l, l));

    //  box-selection
    for (lay::AnnotationShapes::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox); ! r.at_end (); ++r) {
      
      const img::Object *iobj = dynamic_cast <const img::Object *> ((*r).ptr ());
      if (iobj && dragging_what (iobj, search_dbox, m_move_mode, m_p1)) {
          
        //  found anything: make the moved image the selection
        clear_selection ();
        m_selected.insert (std::make_pair (mp_view->annotation_shapes ().iterator_from_pointer (&*r), 0));
        m_current = *iobj;
        m_initial = m_current;
        m_selected_images.push_back (new img::View (this, &m_current, img::View::mode_transient_move));
        m_selected_images.back ()->thaw ();
        return true;

      }
      
    }

    //  nothing was found
    return false;

  }
}

void
Service::move_transform (const db::DPoint &p, db::DFTrans tr, lay::angle_constraint_type ac)
{
  if (m_selected_images.empty () || m_selected.empty ()) {
    return;
  }

  if (m_move_mode == move_all) {

    db::DPoint dp = p;

    m_current.transform (db::DTrans (dp) * db::DTrans (tr) * db::DTrans (-dp));

    //  display current images' parameters
    show_message ();

    m_selected_images [0]->redraw ();

  } else if (m_move_mode == move_selected) {

    m_trans *= db::DTrans (m_p1) * db::DTrans (tr) * db::DTrans (-m_p1);

    for (std::vector<img::View *>::iterator r = m_selected_images.begin (); r != m_selected_images.end (); ++r) {
      (*r)->transform_by (m_trans);
    }

  }
}

void 
Service::move (const db::DPoint &p, lay::angle_constraint_type ac)
{
  if (m_selected_images.empty () || m_selected.empty ()) {
    return;
  }

  if (m_move_mode == move_selected) {

    db::DPoint dp = p - m_p1;
    m_p1 = p;

    m_trans = db::DTrans (dp) * m_trans;

    for (std::vector<img::View *>::iterator r = m_selected_images.begin (); r != m_selected_images.end (); ++r) {
      (*r)->transform_by (m_trans);
    }

  } else {

    if (m_move_mode == move_all) {

      db::DPoint dp = p - m_p1;
      m_p1 = p;

      m_current.transform (db::DTrans (dp));

    } else {

      db::DVector dp = p - m_p1;
      dp.transform (m_current.trans ().inverted ());

      m_current = m_initial;

      double w = m_current.pixel_width () * m_current.width ();
      double h = m_current.pixel_height () * m_current.height ();

      db::DVector vv, v;
      
      if (m_move_mode == move_l) {
        vv = db::DVector (-dp.x (), 0.0);
      } else if (m_move_mode == move_r) {
        vv = db::DVector (dp.x (), 0.0);
      } else if (m_move_mode == move_b) {
        vv = db::DVector (0.0, -dp.y ());
      } else if (m_move_mode == move_t) {
        vv = db::DVector (0.0, dp.y ());
      } else if (m_move_mode == move_ll) {
        vv = db::DVector (-dp.x (), -dp.y ());
      } else if (m_move_mode == move_lr) {
        vv = db::DVector (dp.x (), -dp.y ());
      } else if (m_move_mode == move_tl) {
        vv = db::DVector (-dp.x (), dp.y ());
      } else if (m_move_mode == move_tr) {
        vv = db::DVector (dp.x (), dp.y ());
      }

      double min_scale = 0.001;
      vv = db::DVector (std::max (-w * (1.0 - min_scale), vv.x ()), std::max (-h * (1.0 - min_scale), vv.y ()));

      if (m_move_mode == move_ll || m_move_mode == move_lr || m_move_mode == move_tl || m_move_mode == move_tr) {

        double fx = (w + vv.x ()) / w;
        double fy = (h + vv.y ()) / h;
        double f = std::max (fx, fy);

        vv = db::DVector (f * w - w, f * h - h);

      }

      if (m_move_mode == move_l) {
        v = db::DVector (-vv.x (), 0.0);
      } else if (m_move_mode == move_b) {
        v = db::DVector (0.0, -vv.y ());
      } else if (m_move_mode == move_ll) {
        v = db::DVector (-vv.x (), -vv.y ());
      } else if (m_move_mode == move_lr) {
        v = db::DVector (0.0, -vv.y ());
      } else if (m_move_mode == move_tl) {
        v = db::DVector (-vv.x (), 0.0);
      }

      m_current.transform (db::DTrans (v.transformed (m_current.trans ())));

      double pw = (w + vv.x ()) / m_current.width ();
      double ph = (h + vv.y ()) / m_current.height ();

      m_current.set_pixel_width (pw);
      m_current.set_pixel_height (ph);

    }

    //  display current images' parameters
    show_message ();

    m_selected_images [0]->redraw ();

  }

  if (m_move_mode != move_selected) {
    show_message ();
  }
}

void 
Service::show_message ()
{
  //  display current images parameters
  /*@@@
  std::string pos = std::string ("lx: ") + tl::micron_to_string (m_current.p2 ().x () - m_current.p1 ().x ()) 
                      + "  ly: " + tl::micron_to_string (m_current.p2 ().y () - m_current.p1 ().y ()) 
                      + "  l: " + tl::micron_to_string (m_current.p2 ().distance (m_current.p1 ()));
  view ()->message (pos);
  @@@*/
}

void 
Service::end_move (const db::DPoint &, lay::angle_constraint_type)
{
  if (! m_selected_images.empty () && ! m_selected.empty ()) {

    clear_transient_selection ();

    if (m_move_mode == move_selected) {

      //  replace the images that were moved:
      for (std::map<obj_iterator, unsigned int>::const_iterator s = m_selected.begin (); s != m_selected.end (); ++s) {

        const img::Object *iobj = dynamic_cast<const img::Object *> (s->first->ptr ());

        //  compute moved object and replace
        //  KLUDGE: this creates a copy of the data!
        img::Object *inew = new img::Object (*iobj);
        inew->transform (m_trans);
        mp_view->annotation_shapes ().replace (s->first, db::DUserObject (inew));

      }

      //  and make selection "visible"
      selection_to_view ();

      //  send a signal that at least one image has changed
      m_image_changed.signal_observers ();

    } else if (m_move_mode != move_none) {

      //  replace the image that was moved
      mp_view->annotation_shapes ().replace (m_selected.begin ()->first, db::DUserObject (new img::Object (m_current)));

      //  clear the selection (that was artifically created before)
      clear_selection ();

      //  send a signal that at least one image has changed
      m_image_changed.signal_observers ();

    }

  }

  //  termine the operation
  m_move_mode = move_none;
}

void
Service::selection_to_view ()
{
  //  the selection objects need to be recreated since we destroyed the old images
  for (std::vector<img::View *>::iterator v = m_selected_images.begin (); v != m_selected_images.end (); ++v) {
    delete *v;
  }
  m_selected_images.clear ();

  m_selected_images.reserve (m_selected.size ());
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    r->second = m_selected_images.size ();
    const img::Object *iobj = dynamic_cast<const img::Object *> (r->first->ptr ());
    m_selected_images.push_back (new img::View (this, iobj, img::View::mode_normal));
  }
}

db::DBox 
Service::selection_bbox ()
{
  db::DBox box;
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    const img::Object *iobj = dynamic_cast<const img::Object *> (r->first->ptr ());
    box += iobj->box ();
  }
  return box;
}

void 
Service::transform (const db::DCplxTrans &trans)
{
  //  replace the images that were transformed:
  for (std::map<obj_iterator, unsigned int>::const_iterator s = m_selected.begin (); s != m_selected.end (); ++s) {

    const img::Object *iobj = dynamic_cast<const img::Object *> (s->first->ptr ());

    //  compute transformed object and replace
    img::Object *inew = new img::Object (*iobj);
    inew->transform (trans);
    mp_view->annotation_shapes ().replace (s->first, db::DUserObject (inew));

  }

  selection_to_view ();

  //  send a signal that at least one image has changed
  m_image_changed.signal_observers ();
}

void 
Service::edit_cancel () 
{
  clear_selection ();
}

void 
Service::cut ()
{
  if (has_selection ()) {

    //  copy & delete the selected images
    copy_selected ();
    del_selected ();

  }
}

void 
Service::copy ()
{
  //  copy the selected images
  copy_selected ();
}

void
Service::copy_selected ()
{
  //  extract all selected images and paste in "micron" space
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    r->second = m_selected_images.size ();
    const img::Object *iobj = dynamic_cast<const img::Object *> (r->first->ptr ());
    db::Clipboard::instance () += new db::ClipboardValue<img::Object> (*iobj);
  }
}

void 
Service::paste ()
{
  if (db::Clipboard::instance ().begin () != db::Clipboard::instance ().end ()) {

    for (db::Clipboard::iterator c = db::Clipboard::instance ().begin (); c != db::Clipboard::instance ().end (); ++c) {
      const db::ClipboardValue<img::Object> *value = dynamic_cast<const db::ClipboardValue<img::Object> *> (*c);
      if (value) {
        img::Object *image = new img::Object (value->get ());
        mp_view->annotation_shapes ().insert (db::DUserObject (image));
      }
    }

  }

  //  send a signal that the image collection has changed  
  m_image_list_changed.signal_observers ();

}

void 
Service::del ()
{
  if (has_selection ()) {

    //  delete the selected images
    del_selected ();

  }
}

void
Service::del_selected ()
{
  //  positions will hold a set of iterators that are to be erased
  std::vector <lay::AnnotationShapes::iterator> positions;
  positions.reserve (m_selected.size ());
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    positions.push_back (r->first);
  }

  //  clear selection
  clear_selection ();

  //  erase all and insert the ones that we want to keep 
  tl::sort (positions.begin (), positions.end ());  // HINT: must be tl::sort, not std::sort because gcc 3.2.3 has some strange namespace resolution problems ..
  mp_view->annotation_shapes ().erase_positions (positions.begin (), positions.end ());

  //  send a signal that the image collection has changed  
  m_image_list_changed.signal_observers ();
}

bool 
Service::has_selection ()
{
  return ! m_selected.empty ();
}

bool 
Service::select (obj_iterator obj, lay::Editable::SelectionMode mode)
{
  if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
    //  select
    if (m_selected.find (obj) == m_selected.end ()) {
      m_selected.insert (std::make_pair (obj, 0));
      return true;
    }
  } else if (mode == lay::Editable::Reset) {
    //  unselect
    if (m_selected.find (obj) != m_selected.end ()) {
      m_selected.erase (obj);
      return true;
    }
  } else {
    //  invert selection
    if (m_selected.find (obj) != m_selected.end ()) {
      m_selected.erase (obj);
    } else {
      m_selected.insert (std::make_pair (obj, 0));
    }
    return true;
  }
  return false;
}

void
Service::clear_selection ()
{
  select (db::DBox (), lay::Editable::Reset);

  //  clear the transient selection as well so there is no reference to any image left
  clear_transient_selection ();
}

static bool 
is_selected (const img::Object &image, const db::DPoint &pos, double enl, double &distance) 
{
  db::DBox b (image.box ());
  if (! b.enlarged (db::DPoint (enl, enl)).contains (pos)) {
    return false;
  }
  
  db::DEdge edges[4];
  unsigned int nedges = 0;
  edges [nedges++] = db::DEdge (b.p1 (), db::DPoint (b.p2 ().x (), b.p1 ().y ()));
  edges [nedges++] = db::DEdge (db::DPoint (b.p2 ().x (), b.p1 ().y ()), b.p2 ());
  edges [nedges++] = db::DEdge (b.p1 (), db::DPoint (b.p1 ().x (), b.p2 ().y ()));
  edges [nedges++] = db::DEdge (db::DPoint (b.p1 ().x (), b.p2 ().y ()), b.p2 ());
    
  for (unsigned int i = 0; i < nedges; ++i) {
    double d = edges [i].distance_abs (pos);
    if (! i || d < distance) {
      distance = d;
    } 
  }  

  return true;
}

static bool 
is_selected (const img::Object &image, const db::DBox &box)
{
  db::DBox b (image.box ());
  return (box.contains (b.p1 ()) && box.contains (b.p2 ()));
}

double
Service::click_proximity (const db::DPoint &pos)
{
  //  compute search box
  double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
  db::DBox search_dbox = db::DBox (pos, pos).enlarged (db::DPoint (l, l));

  //  point selection: look for the "closest" images
  double dmin = std::numeric_limits <double>::max ();
  bool any_found = false;

  lay::AnnotationShapes::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox);
  while (! r.at_end ()) {
    const img::Object *iobj = dynamic_cast<const img::Object *> ((*r).ptr ());
    if (iobj) {
      double d;
      if (is_selected (*iobj, pos, l, d)) {
        if (! any_found || d < dmin) {
          dmin = d;
        }
        any_found = true;
      }
    }
    ++r;
  }

  //  return the proximity value
  if (any_found) {
    return dmin;
  } else {
    return lay::Editable::click_proximity (pos); 
  } 
}

bool
Service::transient_select (const db::DPoint &pos)
{
  clear_transient_selection ();

  bool any_selected = false;

  //  compute search box
  double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
  db::DBox search_dbox = db::DBox (pos, pos).enlarged (db::DPoint (l, l));

  //  point selection: look for the "closest" image
  double dmin = std::numeric_limits <double>::max ();

  lay::AnnotationShapes::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox);
  lay::AnnotationShapes::touching_iterator rmin (r);
  while (! r.at_end ()) {
    const img::Object *iobj = dynamic_cast<const img::Object *> ((*r).ptr ());
    if (iobj) {
      double d;
      if (is_selected (*iobj, pos, l, d)) {
        if (! any_selected || d < dmin) {
          rmin = r;
          dmin = d;
        }
        any_selected = true;
      }
    }
    ++r;
  }

  //  create the transient marker for the object found 
  if (any_selected) {
    const img::Object *iobj = dynamic_cast<const img::Object *> ((*rmin).ptr ());
    //  HINT: there is no special style for "transient selection on images"
    mp_transient_image = new img::View (this, iobj, mp_view->is_move_mode () && ! mp_view->has_selection () ? img::View::mode_transient_move : img::View::mode_transient);
  }

  return any_selected;
}

void
Service::clear_transient_selection ()
{
  if (mp_transient_image) {
    delete mp_transient_image;
    mp_transient_image = 0;
  }
}

bool
Service::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selected.empty ()) {
      m_selected.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selected.empty ()) {
        m_selected.clear ();
        needs_update = true;
      }
    } else {

      lay::AnnotationShapes::iterator rfrom = mp_view->annotation_shapes ().begin (); 
      lay::AnnotationShapes::iterator rto = mp_view->annotation_shapes ().end ();

      //  extract all images
      for (lay::AnnotationShapes::iterator r = rfrom; r != rto; ++r) {
        const img::Object *iobj = dynamic_cast<const img::Object *> ((*r).ptr ());
        if (iobj) {
          any_selected = true;
          if (select (r, mode)) {
            needs_update = true;
          }
        }
      }
    }

  } else {

    //  compute search box
    double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
    db::DBox search_dbox = box.enlarged (db::DPoint (l, l));

    if (! box.is_point ()) {

      //  box-selection
      lay::AnnotationShapes::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox);
      while (! r.at_end ()) {
        const img::Object *iobj = dynamic_cast<const img::Object *> ((*r).ptr ());
        if (iobj) {
          if (is_selected (*iobj, box)) {
            any_selected = true;
            if (select (mp_view->annotation_shapes ().iterator_from_pointer (&*r), mode)) {
              needs_update = true;
            }
          }
        }
        ++r;
      }

    } else {

      //  point selection: look for the "closest" image
      double dmin = std::numeric_limits <double>::max ();

      lay::AnnotationShapes::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox);
      lay::AnnotationShapes::touching_iterator rmin (r);
      while (! r.at_end ()) {
        const img::Object *iobj = dynamic_cast<const img::Object *> ((*r).ptr ());
        if (iobj) {
          double d;
          if (is_selected (*iobj, box.p1 (), l, d)) {
            if (! any_selected || d < dmin) {
              rmin = r;
              dmin = d;
            }
            any_selected = true;
          }
        }
        ++r;
      }

      //  select the one that was found
      if (any_selected) {
        select (mp_view->annotation_shapes ().iterator_from_pointer (&*rmin), mode);
        needs_update = true;
      }

    }

  }

  //  if required, update the list of image objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  //  return true if at least one element was selected
  return any_selected;
}

lay::PropertiesPage *
Service::properties_page (QWidget *parent)
{
  return new img::PropertiesPage (this, parent);
}

void 
Service::get_selection (std::vector <obj_iterator> &sel) const
{
  sel.clear ();
  sel.reserve (m_selected.size ());

  //  positions will hold a set of iterators that are to be erased
  for (std::map<obj_iterator, unsigned int>::const_iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    sel.push_back (r->first);
  }
}

void
Service::erase_image (obj_iterator pos)
{
  //  clear the selection
  clear_selection ();
    
  //  erase the object
  mp_view->annotation_shapes ().erase (pos);

  //  send a signal that the image list has changed
  m_image_list_changed.signal_observers ();
}

void
Service::erase_image_by_id (size_t id)
{
  obj_iterator img = object_iter_by_id (id);
  if (img != mp_view->annotation_shapes ().end ()) {
    erase_image (img);
  }
}

void
Service::change_image (obj_iterator pos, const img::Object &to)
{
  //  replace the object
  mp_view->annotation_shapes ().replace (pos, db::DUserObject (new img::Object (to)));

  //  and make selection "visible"
  selection_to_view ();
    
  //  send a signal that an image has changed  
  m_image_changed.signal_observers ();
}

void
Service::change_image_by_id (size_t id, const img::Object &to)
{
  obj_iterator img = object_iter_by_id (id);
  if (img != mp_view->annotation_shapes ().end ()) {
    change_image (img, to);
  }
}

void 
Service::render_bg (const lay::Viewport &vp, lay::ViewObjectCanvas &canvas)
{
  lay::AnnotationShapes::touching_iterator user_object = mp_view->annotation_shapes ().begin_touching (vp.box ());
  while (! user_object.at_end ()) {
    const img::Object *image = dynamic_cast <const img::Object *> ((*user_object).ptr ());
    if (image && image->is_visible ()) {
      draw_image (*image, vp, canvas);
    }
    ++user_object;
  }
}

ImageIterator
Service::begin_images () const
{
  return ImageIterator (mp_view->annotation_shapes ().begin (), mp_view->annotation_shapes ().end ());
}

void 
Service::menu_activated (const std::string &symbol)
{
  if (symbol == "img::clear_all_images") {
    manager ()->transaction (tl::translate ("Clear all images")); 
    clear_images ();
    manager ()->commit ();
  } else if (symbol == "img::add_image") {
    add_image ();
  } else {
    lay::Plugin::menu_activated (symbol);
  }
}

void 
Service::add_image ()
{
  img::Object *new_image = new img::Object ();

  AddNewImageDialog dialog (lay::MainWindow::instance (), new_image);
  if (dialog.exec ()) {

    clear_selection ();

    manager ()->transaction (tl::translate ("Add image")); 
    mp_view->annotation_shapes ().insert (db::DUserObject (new_image));
    manager ()->commit ();

    //  send a signal that the image collection has changed  
    m_image_list_changed.signal_observers ();

  } else {
    delete new_image;
  }
}

void
Service::clear_images ()
{
  lay::AnnotationShapes::iterator rfrom = mp_view->annotation_shapes ().begin (); 
  lay::AnnotationShapes::iterator rto = mp_view->annotation_shapes ().end ();

  //  clear selection
  clear_selection ();

  //  extract all images
  std::vector <lay::AnnotationShapes::iterator> positions;
  for (lay::AnnotationShapes::iterator r = rfrom; r != rto; ++r) {
    const img::Object *iobj = dynamic_cast <const img::Object *> (r->ptr ());
    if (iobj) {
      positions.push_back (r);
    }
  }

  //  we can erase these positions after having sorted them
  tl::sort (positions.begin (), positions.end ());  // HINT: must be tl::sort, not std::sort because gcc 3.2.3 has some strange namespace resolution problems ..
  mp_view->annotation_shapes ().erase_positions (positions.begin (), positions.end ());

  //  send a signal that the image collection has changed  
  m_image_list_changed.signal_observers ();
}

const Object *
Service::object_by_id (size_t id) const
{
  obj_iterator i = object_iter_by_id (id);
  if (i == mp_view->annotation_shapes ().end ()) {
    return 0;
  } else {
    return dynamic_cast <const img::Object *> (i->ptr ());
  }
}

Service::obj_iterator 
Service::object_iter_by_id (size_t id) const
{
  //  TODO: this is a O(1) lookup, thus potentially slow.
  //  However, in non-editable mode, maintaining a table is not that straightforward ...
  for (obj_iterator user_object = mp_view->annotation_shapes ().begin (); user_object != mp_view->annotation_shapes ().end (); ++user_object) {
    const img::Object *i = dynamic_cast <const img::Object *> ((*user_object).ptr ());
    if (i->id () == id) {
      return user_object;
    }
  }

  return mp_view->annotation_shapes ().end ();
}

// -------------------------------------------------------------

} // namespace img


