
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layDialogs.h"

#include "dbLayout.h"

#include "tlString.h"
#include "tlException.h"
#include "tlXMLParser.h"

#include "layExceptions.h"
#include "layLayerProperties.h"
#include "layFileDialog.h"
#include "layLayoutView.h"


namespace lay
{

// --------------------------------------------------------------------------------
//  NewLayoutPropertiesDialog implementation

NewLayoutPropertiesDialog::NewLayoutPropertiesDialog (QWidget *parent)
  : QDialog (parent, "new_layout_properties_dialog")
{
  Ui::NewLayoutPropertiesDialog::setupUi (this);
}

NewLayoutPropertiesDialog::~NewLayoutPropertiesDialog ()
{
  //  .. nothing yet ..
}

bool 
NewLayoutPropertiesDialog::exec (std::string &cell_name, double &dbu, double &size, bool &current_panel)
{
  window_le->setText (tl::to_string (size).c_str ());
  dbu_le->setText (tl::to_string (dbu).c_str ());
  topcell_le->setText (cell_name.c_str ());
  current_panel_cb->setChecked (current_panel);

  if (QDialog::exec ()) {

    tl::from_string (window_le->text ().ascii (), size);
    tl::from_string (dbu_le->text ().ascii (), dbu);
    cell_name = topcell_le->text ().ascii ();
    current_panel = current_panel_cb->isChecked ();
    return true;

  } else {
    return false;
  }
}

void 
NewLayoutPropertiesDialog::accept ()
{
BEGIN_PROTECTED;

  double x = 0.0;
  tl::from_string (window_le->text ().ascii (), x);
  tl::from_string (dbu_le->text ().ascii (), x);

  if (topcell_le->text ().isEmpty ()) {
    throw tl::Exception (tl::translate ("The topcell must be specified"));
  }
  
  QDialog::accept ();

END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  NewCellPropertiesDialog implementation

NewCellPropertiesDialog::NewCellPropertiesDialog (QWidget *parent)
  : QDialog (parent, "new_cell_properties_dialog"),
    mp_layout (0)
{
  Ui::NewCellPropertiesDialog::setupUi (this);
}

NewCellPropertiesDialog::~NewCellPropertiesDialog ()
{
  //  .. nothing yet ..
}

bool 
NewCellPropertiesDialog::exec (const db::Layout *layout, std::string &cell_name, double &size)
{
  mp_layout = layout;

  name_le->setText (cell_name.c_str ());
  window_le->setText (tl::to_string (size).c_str ());

  if (QDialog::exec ()) {

    tl::from_string (window_le->text ().ascii (), size);
    cell_name = name_le->text ().ascii ();
    return true;

  } else {
    return false;
  }
}

void 
NewCellPropertiesDialog::accept ()
{
BEGIN_PROTECTED;

  double x = 0.0;
  tl::from_string (window_le->text ().ascii (), x);

  if (mp_layout->cell_by_name (name_le->text ().ascii ()).first) {
    throw tl::Exception (tl::translate ("A cell with that name already exists: %s").c_str (), name_le->text ().ascii ());
  }
  
  QDialog::accept ();

END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  NewLayerPropertiesDialog implementation

NewLayerPropertiesDialog::NewLayerPropertiesDialog (QWidget *parent)
  : QDialog (parent, "new_layer_properties_dialog")
{
  Ui::NewLayerPropertiesDialog::setupUi (this);
}

NewLayerPropertiesDialog::~NewLayerPropertiesDialog ()
{
  //  .. nothing yet ..
}

bool 
NewLayerPropertiesDialog::exec (db::LayerProperties &src)
{
  return exec (lay::CellView (), src);
}

bool 
NewLayerPropertiesDialog::exec (const lay::CellView &cv, db::LayerProperties &src)
{
  if (cv.is_valid ()) {
    layout_lbl->setText ((tl::translate ("Layer for layout: ") + cv->name ()).c_str ());
    layout_lbl->show ();
  } else {
    layout_lbl->hide ();
  }

  if (src.layer >= 0) {
    layer_le->setText (tl::to_string (src.layer).c_str ());
  } else {
    layer_le->setText ("");
  }
  if (src.datatype >= 0) {
    datatype_le->setText (tl::to_string (src.datatype).c_str ());
  } else {
    datatype_le->setText ("");
  }
  name_le->setText (src.name.c_str ());

  if (QDialog::exec ()) {
    get (src);
    return true;
  } else {
    return false;
  }
}

void
NewLayerPropertiesDialog::get (db::LayerProperties &src)
{
  if (layer_le->text () != "") {
    int l = -1;
    tl::from_string (layer_le->text ().ascii (), l);
    src.layer = l;
  } else {
    src.layer = -1;
  }
  if (datatype_le->text () != "") {
    int d = -1;
    tl::from_string (datatype_le->text ().ascii (), d);
    src.datatype = d;
  } else {
    src.datatype = -1;
  }
  src.name = name_le->text ().ascii ();
}

void 
NewLayerPropertiesDialog::accept ()
{
BEGIN_PROTECTED;
  db::LayerProperties lp;
  get (lp);

  if (lp.layer < 0 && lp.datatype < 0) {
    if (lp.name.empty ()) {
      throw tl::Exception (tl::translate ("Either a layer/datatype combination or a name must be specified for a layer"));
    } 
  } else if (lp.layer < 0 || lp.datatype < 0) {
    throw tl::Exception (tl::translate ("Both layer and datatype must be specified for a layer"));
  }
  QDialog::accept ();
END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  MoveOptionsDialog implementation

MoveOptionsDialog::MoveOptionsDialog (QWidget *parent)
  : QDialog (parent, "move_options_dialog")
{
  Ui::MoveOptionsDialog::setupUi (this);
}

MoveOptionsDialog::~MoveOptionsDialog ()
{
  //  .. nothing yet ..
}

bool 
MoveOptionsDialog::exec (db::DPoint &disp)
{
  disp_x_le->setText (tl::to_string (disp.x ()).c_str ());
  disp_y_le->setText (tl::to_string (disp.y ()).c_str ());

  if (QDialog::exec ()) {

    double x = 0.0, y = 0.0;
    tl::from_string (disp_x_le->text ().ascii (), x);
    tl::from_string (disp_y_le->text ().ascii (), y);

    disp = db::DPoint (x, y);

    return true;

  } else {
    return false;
  }
}

void 
MoveOptionsDialog::accept ()
{
BEGIN_PROTECTED;
  double x = 0.0;
  tl::from_string (disp_x_le->text ().ascii (), x);
  tl::from_string (disp_y_le->text ().ascii (), x);
  QDialog::accept ();
END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  RenameCellDialog implementation

RenameCellDialog::RenameCellDialog (QWidget *parent)
  : QDialog (parent, "rename_cell_dialog"), mp_layout (0)
{
  Ui::RenameCellDialog::setupUi (this);
}

RenameCellDialog::~RenameCellDialog ()
{
  //  .. nothing yet ..
}

void 
RenameCellDialog::accept ()
{
BEGIN_PROTECTED;
  if (name_le->text () == "") {
    throw tl::Exception (tl::translate ("A name must be given"));
  }
  if (mp_layout->cell_by_name (name_le->text ().ascii ()).first) {
    throw tl::Exception (tl::translate ("A cell with that name already exists"));
  }
  QDialog::accept ();
END_PROTECTED;
}

bool 
RenameCellDialog::exec (const db::Layout &layout, std::string &name)
{
  mp_layout = &layout;
  name_le->setText (name.c_str ());
  if (QDialog::exec ()) {
    name = name_le->text ().ascii ();
    return true;
  } else {
    return false;
  }
}

// --------------------------------------------------------------------------------
//  DeleteCellModeDialog implementation

DeleteCellModeDialog::DeleteCellModeDialog (QWidget *parent)
  : QDialog (parent, "delete_cell_mode_dialog")
{
  Ui::DeleteCellModeDialog::setupUi (this);
}

DeleteCellModeDialog::~DeleteCellModeDialog ()
{
  //  .. nothing yet ..
}

bool 
DeleteCellModeDialog::exec (int &delete_mode)
{
  QRadioButton *buttons [3] = { shallow_rb, deep_rb, full_rb };

  for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
    buttons [i]->setChecked (delete_mode == i);
  }

  if (QDialog::exec ()) {
    for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
      if (buttons [i]->isChecked ()) {
        delete_mode = i;
      }
    }
    return true;
  } else {
    return false;
  }
}

// --------------------------------------------------------------------------------
//  ClearLayerModeDialog implementation

ClearLayerModeDialog::ClearLayerModeDialog (QWidget *parent)
  : QDialog (parent, "clear_layer_mode_dialog")
{
  Ui::ClearLayerModeDialog::setupUi (this);
}

ClearLayerModeDialog::~ClearLayerModeDialog ()
{
  //  .. nothing yet ..
}

bool 
ClearLayerModeDialog::exec (int &clear_mode, const lay::CellView & /*cv*/)
{
  QRadioButton *buttons [3] = { local_rb, hierarchically_rb, layout_rb };

  for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
    buttons [i]->setChecked (clear_mode == i);
  }

  if (QDialog::exec ()) {
    for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
      if (buttons [i]->isChecked ()) {
        clear_mode = i;
      }
    }
    return true;
  } else {
    return false;
  }
}

// --------------------------------------------------------------------------------
//  OpenLayoutModeDialog implementation

OpenLayoutModeDialog::OpenLayoutModeDialog (QWidget *parent)
  : QDialog (parent, "open_layout_mode_dialog")
{
  Ui::OpenLayoutModeDialog::setupUi (this);
}

OpenLayoutModeDialog::~OpenLayoutModeDialog ()
{
  //  .. nothing yet ..
}

bool 
OpenLayoutModeDialog::exec (int &open_mode)
{
  QRadioButton *buttons [3] = { replace_rb, new_rb, add_rb };

  for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
    buttons [i]->setChecked (open_mode == i);
  }

  if (QDialog::exec ()) {
    for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
      if (buttons [i]->isChecked ()) {
        open_mode = i;
      }
    }
    return true;
  } else {
    return false;
  }
}

// --------------------------------------------------------------------------------
//  BooleanOptionsDialog implementation

BooleanOptionsDialog::BooleanOptionsDialog (QWidget *parent)
  : QDialog (parent, "boolean_options_dialog"),
    mp_view (0)
{
  Ui::BooleanOptionsDialog::setupUi (this);

  connect (cva_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
  connect (cvb_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
  connect (cvr_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
}

BooleanOptionsDialog::~BooleanOptionsDialog ()
{
  //  .. nothing yet ..
}

void
BooleanOptionsDialog::cv_changed (int)
{
  if (! mp_view) {
    return;
  }

  layera_cbx->set_view (mp_view, cva_cbx->currentIndex ());
  layerb_cbx->set_view (mp_view, cvb_cbx->currentIndex ());
  layerr_cbx->set_view (mp_view, cvr_cbx->currentIndex ());
}

bool 
BooleanOptionsDialog::exec (lay::LayoutView *view, int &cv_a, int &layer_a, int &cv_b, int &layer_b, int &cv_r, int &layer_r, int &mode, int &hier_mode, bool &min_coherence)
{
  mp_view = view;

  bool res = false;
  cva_cbx->set_layout_view (view);
  cva_cbx->set_current_cv_index (cv_a);
  cvb_cbx->set_layout_view (view);
  cvb_cbx->set_current_cv_index (cv_b);
  cvb_cbx->setEnabled (true);
  cvr_cbx->set_layout_view (view);
  cvr_cbx->set_current_cv_index (cv_r);

  cv_changed (0 /*dummy*/);

  layera_cbx->set_current_layer (layer_a);
  layerb_cbx->set_current_layer (layer_b);
  layerb_cbx->setEnabled (true);
  layerr_cbx->set_current_layer (layer_r);

  hier_mode_cbx->setCurrentIndex (hier_mode);
  mode_cbx->setCurrentIndex (mode);
  min_coherence_cb->setChecked (min_coherence);

  if (QDialog::exec ()) {

    cv_a = cva_cbx->current_cv_index ();
    cv_b = cvb_cbx->current_cv_index ();
    cv_r = cvr_cbx->current_cv_index ();
    layer_a = layera_cbx->current_layer ();
    layer_b = layerb_cbx->current_layer ();
    layer_r = layerr_cbx->current_layer ();

    hier_mode = hier_mode_cbx->currentIndex ();
    mode = mode_cbx->currentIndex ();
    min_coherence = min_coherence_cb->isChecked ();

    res = true;

  }

  mp_view = 0;
  return res;
}

void 
BooleanOptionsDialog::accept ()
{
BEGIN_PROTECTED;

  int cv_a = cva_cbx->current_cv_index ();
  if (cv_a < 0) {
    throw tl::Exception (tl::translate ("No layout specified for source 'A'"));
  }

  int cv_b = cvb_cbx->current_cv_index ();
  if (cv_b < 0) {
    throw tl::Exception (tl::translate ("No layout specified for source 'B'"));
  }

  int cv_r = cvr_cbx->current_cv_index ();
  if (cv_r < 0) {
    throw tl::Exception (tl::translate ("No layout specified for result"));
  }

  if (fabs (mp_view->cellview (cv_a)->layout ().dbu () - mp_view->cellview (cv_r)->layout ().dbu ()) > db::epsilon ||
      fabs (mp_view->cellview (cv_b)->layout ().dbu () - mp_view->cellview (cv_r)->layout ().dbu ()) > db::epsilon) {
    throw tl::Exception (tl::translate ("All source and result layouts must have the same database unit"));
  }

  if (layera_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for source 'A'"));
  }
  if (layerb_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for source 'B'"));
  }
  if (layerr_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for result layer"));
  }

  if (hier_mode_cbx->currentIndex () == 2 && 
      cva_cbx->current_cv_index () != cvb_cbx->current_cv_index () &&
      cva_cbx->current_cv_index () != cvr_cbx->current_cv_index ()) {
    throw tl::Exception (tl::translate ("All source layouts and result layout must be same in 'cell by cell' mode"));
  }

  QDialog::accept ();
END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  SizingOptionsDialog implementation

SizingOptionsDialog::SizingOptionsDialog (QWidget *parent)
  : QDialog (parent, "sizing_options_dialog"),
    mp_view (0)
{
  Ui::SizingOptionsDialog::setupUi (this);

  connect (cv_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
  connect (cvr_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
}

SizingOptionsDialog::~SizingOptionsDialog ()
{
  //  .. nothing yet ..
}

void
SizingOptionsDialog::cv_changed (int)
{
  if (! mp_view) {
    return;
  }

  layer_cbx->set_view (mp_view, cv_cbx->currentIndex ());
  layerr_cbx->set_view (mp_view, cvr_cbx->currentIndex ());
}

bool 
SizingOptionsDialog::exec (lay::LayoutView *view, int &cv, int &layer, int &cv_r, int &layer_r, double &dx, double &dy, unsigned int &size_mode, int &hier_mode, bool &min_coherence)
{
  mp_view = view;

  bool res = false;
  cv_cbx->set_layout_view (view);
  cv_cbx->set_current_cv_index (cv);
  cvr_cbx->set_layout_view (view);
  cvr_cbx->set_current_cv_index (cv_r);

  cv_changed (0 /*dummy*/);

  layer_cbx->set_current_layer (layer);
  layerr_cbx->set_current_layer (layer_r);

  hier_mode_cbx->setCurrentIndex (hier_mode);
  cutoff_cbx->setCurrentIndex (size_mode);
  if (dx == dy) {
    value_le->setText (tl::sprintf ("%.12g", dx).c_str ());
  } else {
    value_le->setText (tl::sprintf ("%.12g,%.12g", dx, dy).c_str ());
  }
  min_coherence_cb->setChecked (min_coherence);

  if (QDialog::exec ()) {

    cv = cv_cbx->current_cv_index ();
    cv_r = cvr_cbx->current_cv_index ();
    layer = layer_cbx->current_layer ();
    layer_r = layerr_cbx->current_layer ();

    hier_mode = hier_mode_cbx->currentIndex ();
    min_coherence = min_coherence_cb->isChecked ();
    size_mode = cutoff_cbx->currentIndex ();

    QString text (value_le->text ());
    tl::Extractor ex (text.ascii ());
    ex.read (dx);
    if (ex.test (",")) {
      ex.read (dy);
    } else {
      dy = dx;
    }

    res = true;

  }

  mp_view = 0;
  return res;
}

void 
SizingOptionsDialog::accept ()
{
BEGIN_PROTECTED;

  int cv = cv_cbx->current_cv_index ();
  if (cv < 0) {
    throw tl::Exception (tl::translate ("No layout specified for source"));
  }

  int cv_r = cvr_cbx->current_cv_index ();
  if (cv_r < 0) {
    throw tl::Exception (tl::translate ("No layout specified for result"));
  }

  if (fabs (mp_view->cellview (cv)->layout ().dbu () - mp_view->cellview (cv_r)->layout ().dbu ()) > db::epsilon) {
    throw tl::Exception (tl::translate ("Source and result layouts must have the same database unit"));
  }

  if (layer_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for source"));
  }
  if (layerr_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for result layer"));
  }

  if (hier_mode_cbx->currentIndex () == 2 && 
      cv_cbx->current_cv_index () != cvr_cbx->current_cv_index ()) {
    throw tl::Exception (tl::translate ("Source layout and result layout must be same in 'cell by cell' mode"));
  }

  double x, y;
  QString text (value_le->text ());
  tl::Extractor ex (text.ascii ());
  ex.read (x);
  if (ex.test (",")) {
    ex.read (y);
  } 

  QDialog::accept ();
END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  MergeOptionsDialog implementation

MergeOptionsDialog::MergeOptionsDialog (QWidget *parent)
  : QDialog (parent, "merge_options_dialog"),
    mp_view (0)
{
  Ui::MergeOptionsDialog::setupUi (this);

  connect (cv_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
  connect (cvr_cbx, SIGNAL (activated (int)), this, SLOT (cv_changed (int)));
}

MergeOptionsDialog::~MergeOptionsDialog ()
{
  //  .. nothing yet ..
}

void
MergeOptionsDialog::cv_changed (int)
{
  if (! mp_view) {
    return;
  }

  layer_cbx->set_view (mp_view, cv_cbx->currentIndex ());
  layerr_cbx->set_view (mp_view, cvr_cbx->currentIndex ());
}

bool 
MergeOptionsDialog::exec (lay::LayoutView *view, int &cv, int &layer, int &cv_r, int &layer_r, unsigned int &min_wc, int &hier_mode, bool &min_coherence)
{
  mp_view = view;

  bool res = false;
  cv_cbx->set_layout_view (view);
  cv_cbx->set_current_cv_index (cv);
  cvr_cbx->set_layout_view (view);
  cvr_cbx->set_current_cv_index (cv_r);

  cv_changed (0 /*dummy*/);

  layer_cbx->set_current_layer (layer);
  layerr_cbx->set_current_layer (layer_r);

  hier_mode_cbx->setCurrentIndex (hier_mode);
  threshold_le->setText (tl::sprintf ("%u", min_wc).c_str ());
  min_coherence_cb->setChecked (min_coherence);

  if (QDialog::exec ()) {

    cv = cv_cbx->current_cv_index ();
    cv_r = cvr_cbx->current_cv_index ();
    layer = layer_cbx->current_layer ();
    layer_r = layerr_cbx->current_layer ();

    hier_mode = hier_mode_cbx->currentIndex ();
    min_coherence = min_coherence_cb->isChecked ();

    QString text (threshold_le->text ());
    tl::Extractor ex (text.ascii ());
    ex.read (min_wc);

    res = true;

  }

  mp_view = 0;
  return res;
}

void 
MergeOptionsDialog::accept ()
{
BEGIN_PROTECTED;

  int cv = cv_cbx->current_cv_index ();
  if (cv < 0) {
    throw tl::Exception (tl::translate ("No layout specified for source"));
  }

  int cv_r = cvr_cbx->current_cv_index ();
  if (cv_r < 0) {
    throw tl::Exception (tl::translate ("No layout specified for result"));
  }

  if (fabs (mp_view->cellview (cv)->layout ().dbu () - mp_view->cellview (cv_r)->layout ().dbu ()) > db::epsilon) {
    throw tl::Exception (tl::translate ("Source and result layouts must have the same database unit"));
  }

  if (layer_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for source"));
  }
  if (layerr_cbx->current_layer () < 0) {
    throw tl::Exception (tl::translate ("No layer specified for result"));
  }

  if (hier_mode_cbx->currentIndex () == 2 && 
      cv_cbx->current_cv_index () != cvr_cbx->current_cv_index ()) {
    throw tl::Exception (tl::translate ("Source layout and result layout must be same in 'cell by cell' mode"));
  }

  unsigned int min_wc = 0;
  tl::Extractor ex (threshold_le->text ().ascii ());
  ex.read (min_wc);

  QDialog::accept ();
END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  AlignCellOptionsDialog implementation

AlignCellOptionsDialog::AlignCellOptionsDialog (QWidget *parent)
  : QDialog (parent, "align_cell_options_dialog")
{
  Ui::AlignCellOptionsDialog::setupUi (this);

  QToolButton *buttons[3][3] = { { lb, cb, rb }, { lc, cc, rc }, { lt, ct, rt } };

  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      connect (buttons[i][j], SIGNAL (clicked ()), this, SLOT (button_clicked ()));
    }
  }
}

AlignCellOptionsDialog::~AlignCellOptionsDialog ()
{
  //  .. nothing yet ..
}

bool 
AlignCellOptionsDialog::exec (int &mode_x, int &mode_y, bool &visible_only)
{
  vis_only_cbx->setChecked (visible_only);

  QToolButton *buttons[3][3] = { { lb, cb, rb }, { lc, cc, rc }, { lt, ct, rt } };

  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      buttons[i][j]->setChecked (j - 1 == mode_x && i - 1 == mode_y);
    }
  }

  if (QDialog::exec ()) {

    visible_only = vis_only_cbx->isChecked ();

    for (int i = 0; i < 3; ++i) {
      for (int j = 0; j < 3; ++j) {
        if (buttons[i][j]->isChecked ()) {
          mode_x = j - 1;
          mode_y = i - 1;
        }
      }
    }

    return true;

  } else {
    return false;
  }
}

void 
AlignCellOptionsDialog::button_clicked ()
{
  QToolButton *buttons[3][3] = { { lb, cb, rb }, { lc, cc, rc }, { lt, ct, rt } };

  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      if (buttons [i][j] != sender ()) {
        buttons [i][j]->setChecked (false);
      }
    }
  }
}

// --------------------------------------------------------------------------------
//  FlattenInstOptionsDialog implementation

FlattenInstOptionsDialog::FlattenInstOptionsDialog (QWidget *parent, bool enable_pruning)
  : QDialog (parent)
{ 
  setupUi (this);

  if (! enable_pruning) {
    prune_cb->setChecked (false);
    prune_cb->hide ();
  }
}

bool 
FlattenInstOptionsDialog::exec (int &levels, bool &prune) 
{
  first_level_rb->setChecked (false);
  all_levels_rb->setChecked (false);
  spec_levels_rb->setChecked (false);
  levels_sb->setValue ((levels < 0 || levels > levels_sb->maximum ()) ? levels_sb->maximum () : levels);

  if (levels == 1) {
    first_level_rb->setChecked (true);
  } else if (levels < 0 || levels == std::numeric_limits<int>::max ()) {
    all_levels_rb->setChecked (true);
  } else {
    spec_levels_rb->setChecked (true);
  }

  prune_cb->setChecked (prune);

  if (QDialog::exec ()) {

    prune = prune_cb->isChecked ();

    if (first_level_rb->isChecked ()) {
      levels = 1;
      return true;
    } else if (spec_levels_rb->isChecked ()) {
      levels = levels_sb->value ();
      return true;
    } else if (all_levels_rb->isChecked ()) {
      levels = std::numeric_limits<int>::max ();
      return true;
    }

  } 
  return false;
}

}

