
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layLayoutPropertiesForm.h"
#include "layLayoutView.h"
#include "tlInternational.h"
#include "tlString.h"

#include <set>
#include <strstream>

namespace lay
{

// ------------------------------------------------------------

LayoutPropertiesForm::LayoutPropertiesForm (QWidget *parent, lay::LayoutView *view, const char *name, Qt::WFlags fl)
  : QDialog (parent, name, fl), Ui::LayoutPropertiesForm ()
{
  Ui::LayoutPropertiesForm::setupUi (this);

  //  collect the distinct layout handles 
  std::set <lay::LayoutHandle *> handles;
  for (unsigned int n = 0; n < view->cellviews (); ++n) {
    handles.insert (view->cellview (n).operator-> ());
  }
  
  m_handles.reserve (handles.size ());
  for (unsigned int n = 0; n < view->cellviews (); ++n) {
    lay::LayoutHandle *h = view->cellview (n).operator-> ();
    if (handles.find (h) != handles.end ()) {
      m_handles.push_back (h);
      handles.erase (h);
      layout_cbx->addItem (h->name ().c_str ());
    }
  }

  connect (layout_cbx, SIGNAL (activated (int)), this, SLOT (layout_selected (int)));

  layout_selected (0);
}

/**
 *  @brief A comparison operator to sort layers by layer and datatype and then by name
 */
struct CompareLDName 
{
  CompareLDName (const db::Layout &l) 
    : mp_layout (&l)
  {
    //  .. nothing yet ..
  }

  bool operator() (unsigned int a, unsigned int b)
  {
    if (mp_layout->is_valid_layer (a) && mp_layout->is_valid_layer (b)) {
      const db::LayerProperties &lp_a = mp_layout->get_properties (a);
      const db::LayerProperties &lp_b = mp_layout->get_properties (b);
      if (lp_a.layer != lp_b.layer) {
        return lp_a.layer < lp_b.layer;
      }
      if (lp_a.datatype != lp_b.datatype) {
        return lp_a.datatype < lp_b.datatype;
      }
      if (lp_a.name != lp_b.name) {
        return lp_a.name < lp_b.name;
      }
    }
    return false;
  }

private:
  const db::Layout *mp_layout;
};

/**
 *  @brief A comparison operator to sort layers by name and then by layer and datatype 
 */
struct CompareNameLD 
{
  CompareNameLD (const db::Layout &l) 
    : mp_layout (&l)
  {
    //  .. nothing yet ..
  }

  bool operator() (unsigned int a, unsigned int b)
  {
    if (mp_layout->is_valid_layer (a) && mp_layout->is_valid_layer (b)) {
      const db::LayerProperties &lp_a = mp_layout->get_properties (a);
      const db::LayerProperties &lp_b = mp_layout->get_properties (b);
      if (lp_a.name != lp_b.name) {
        return lp_a.name < lp_b.name;
      }
      if (lp_a.layer != lp_b.layer) {
        return lp_a.layer < lp_b.layer;
      }
      if (lp_a.datatype != lp_b.datatype) {
        return lp_a.datatype < lp_b.datatype;
      }
    }
    return false;
  }

private:
  const db::Layout *mp_layout;
};

void
LayoutPropertiesForm::layout_selected (int index)
{
  if (index >= int (m_handles.size ()) || index < 0) {
    return;
  }

  const db::Layout &layout = m_handles [index]->layout ();

  std::ostrstream os;
  size_t num_cells = layout.cells ();

  size_t num_layers = 0;
  for (unsigned int i = 0; i < layout.layers (); ++i) {
    if (layout.is_valid_layer (i)) {
      ++num_layers;
    }
  }

  os << "<html>" << std::endl
     <<   "<body>" << std::endl
     <<     "<h2>" << tl::translate ("Common Properties For '") << m_handles [index]->name () << "'</h2>" << std::endl
     <<     "<table>" << std::endl
     <<       "<tr>"
     <<         "<td>" << tl::translate ("Path") << ":&nbsp;</td><td>" << m_handles [index]->filename () << "</td>"
     <<       "</tr>" << std::endl
     <<       "<tr>"
     <<         "<td>" << tl::translate ("Database unit") << ":&nbsp;</td><td>" << tl::sprintf ("%.12g ", layout.dbu ()) << tl::translate ("micron") << "</td>"
     <<       "</tr>" << std::endl
     <<       "<tr>"
     <<         "<td>" << tl::translate ("Number of cells") << ":&nbsp;</td><td>" << num_cells << "</td>"
     <<       "</tr>" << std::endl
     <<       "<tr>"
     <<         "<td>" << tl::translate ("Number of layers") << ":&nbsp;</td><td>" << num_layers << "</td>"
     <<       "</tr>" << std::endl;
  for (lay::LayoutHandle::meta_info_iterator meta = m_handles [index]->begin_meta (); meta != m_handles [index]->end_meta (); ++meta) {
    os <<     "<tr><td>" << meta->description << "</td><td>" << meta->value << "</td></tr>" << std::endl;
  }
  os <<     "</table>" << std::endl
     <<     "<h2>" << tl::translate ("Top Cells") << "</h2>" << std::endl
     <<     "<table>" << std::endl;
  for (db::Layout::top_down_const_iterator tc = layout.begin_top_down (); tc != layout.end_top_cells (); ++tc) {
    os <<     "<tr><td>" << layout.cell_name (*tc) << "</td></tr>" << std::endl;
  }
  os <<     "</table>" << std::endl;

  std::vector <unsigned int> layers_with_oasis_names;

  std::vector <unsigned int> layers_sorted_by_ld;
  layers_sorted_by_ld.reserve (layout.layers ());
  for (unsigned int i = 0; i < layout.layers (); ++i) {
    layers_sorted_by_ld.push_back (i);
    const db::LayerProperties &lp = layout.get_properties (i);
    if (! lp.name.empty ()) {
      layers_with_oasis_names.push_back (i);
    }
  }

  std::sort (layers_sorted_by_ld.begin (), layers_sorted_by_ld.end (), CompareLDName (layout));
  std::sort (layers_with_oasis_names.begin (), layers_with_oasis_names.end (), CompareNameLD (layout));
   
  os <<     "<h2>" << tl::translate ("Layers (sorted by layer and datatype)") << "</h2>" << std::endl
     <<     "<table>" << std::endl
     <<     "<tr><td><b>" << tl::translate ("Layer/Datatype") << "</b>&nbsp;&nbsp;</td>";
  if (! layers_with_oasis_names.empty ()) {
    os << "<td><b>" << tl::translate("OASIS layer name") << "</b></td>";
  }
  os << "</tr>" << std::endl;

  for (std::vector <unsigned int>::const_iterator i = layers_sorted_by_ld.begin (); i != layers_sorted_by_ld.end (); ++i) {
    if (layout.is_valid_layer (*i)) {
      const db::LayerProperties &lp = layout.get_properties (*i);
      os << "<tr>"
         <<   "<td>" << tl::sprintf ("%d/%d", lp.layer, lp.datatype) << "</td>";
      if (! layers_with_oasis_names.empty ()) {
        os <<   "<td>" << lp.name << "</td>";
      }
      os << "</tr>" << std::endl;
    }
  }

  os <<     "</table>" << std::endl;

  if (! layers_with_oasis_names.empty ()) {

    os <<     "<h2>" << tl::translate ("Layers (sorted by OASIS layer names)") << "</h2>" << std::endl
       <<     "<table>" << std::endl
       <<     "<tr><td><b>" << tl::translate ("OASIS layer name") << "</b>&nbsp;&nbsp;</td><td><b>" << tl::translate("Layer/Datatype") << "</b></td></tr>" << std::endl;
    
    for (std::vector <unsigned int>::const_iterator i = layers_with_oasis_names.begin (); i != layers_with_oasis_names.end (); ++i) {
      if (layout.is_valid_layer (*i)) {
        const db::LayerProperties &lp = layout.get_properties (*i);
        if (! lp.name.empty ()) {
          os << "<tr>"
             <<   "<td>" << lp.name << "</td>"
             <<   "<td>" << tl::sprintf ("%d/%d", lp.layer, lp.datatype) << "</td>"
             << "</tr>" << std::endl;
        }
      }
    }

    os <<     "</table>" << std::endl;

  }

  os <<   "</body>" << std::endl
     << "</html>" << std::endl;
     ;

  os << std::ends;

  text_browser->setHtml (os.str ());
  os.rdbuf ()->freeze (0);
}

}



