
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QWindowsStyle>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QToolBar>
#include <QtGui/QStatusBar>
#include <QtGui/QInputDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QPainter>
#include <QtGui/QDockWidget>
#include <QtGui/QShortcut>
#include <QtGui/QPrintDialog>
#include <QtCore/QFileInfo>
#include <QtCore/QUrl>

#include "tlInternational.h"
#include "tlTimer.h"
#include "tlLog.h"
#include "tlAssert.h"
#include "tlDeferredExecution.h"
#include "dbMemStatistics.h"
#include "dbManager.h"
#include "dbStream.h"
#include "dbSaveLayoutOptions.h"
#include "dbClipboard.h"
#include "edtConfig.h"
#include "laySession.h"
#include "layApplication.h"
#include "layVersion.h"
#include "layConverters.h"
#include "layDialogs.h"
#include "layConfig.h"
#include "layMainWindow.h"
#include "layNavigator.h"
#include "layProgress.h"
#include "layProgressWidget.h"
#include "layStream.h"
#include "layLayerControlPanel.h" // because of LabelWithBackground
#include "layFileDialog.h"
#include "layMainConfigPages.h"
#include "layAbstractMenu.h"
#include "layExceptions.h"
#include "laySaveLayoutOptionsDialog.h"
#include "layLoadLayoutOptionsDialog.h"
#include "layLogViewerDialog.h"
#include "layLayerToolbox.h"
#include "laySettingsForm.h"
#include "layCellSelectionForm.h"
#include "laySelectCellViewForm.h"
#include "layLayoutPropertiesForm.h"
#include "HelpAboutDialog.h"
#include "TipDialog.h"
#include "gsi.h"
#include "gtf.h"
#include "rbaConsoleImpl.h"

namespace lay
{

// -------------------------------------------------------------

class ControlWidgetStack 
  : public QFrame
{
public:
  ControlWidgetStack (QWidget *parent = 0, const char *name = 0)
    : QFrame (parent, name)
  {
    //  Background ist a simple label without a text currently
    mp_bglabel = new LabelWithBackground (this);
    mp_bglabel->setAlignment (Qt::AlignVCenter | Qt::AlignHCenter);
    mp_bglabel->show ();
  }

  void addWidget (QWidget *w)
  {
    m_widgets.push_back (w);
    w->setParent (this);
    resize_children ();
    raiseWidget (m_widgets.size () - 1);

    int mw = 0;
    for (size_t i = 0; i < m_widgets.size (); ++i) {
      mw = std::max (m_widgets [i]->sizeHint ().width (), mw);
      mw = std::max (m_widgets [i]->minimumWidth (), mw);
    }

    if (mw > minimumWidth ()) {
      setMinimumWidth (mw);
      resize (minimumWidth (), height ());
    }
  }

  QSize sizeHint () const
  {
    int w = 0;
    for (size_t i = 0; i < m_widgets.size (); ++i) {
      w = std::max (m_widgets [i]->sizeHint ().width (), w);
    }
    return QSize (w, 0);
  }

  void removeWidget (size_t index)
  {
    if (index < m_widgets.size ()) {
      m_widgets.erase (m_widgets.begin () + index);
    }
    if (m_widgets.size () == 0) {
      mp_bglabel->show ();
    }
  }

  void raiseWidget (size_t index)
  {
    bool any_visible = false;
    for (size_t i = 0; i < m_widgets.size (); ++i) {
      if (m_widgets [i]) {
        if (i == index) {
          m_widgets [i]->show ();
          any_visible = true;
        } else {
          m_widgets [i]->hide ();
        }
      }
    }

    if (! any_visible) {
      mp_bglabel->show ();
    } else {
      mp_bglabel->hide ();
    }
  }
  
  QWidget *widget (size_t index)
  {
    if (index < m_widgets.size ()) {
      return m_widgets [index];
    } else {
      return 0;
    }
  }

  QWidget *background_widget ()
  {
    return mp_bglabel;
  }

protected:
  virtual void resizeEvent (QResizeEvent *)
  {
    resize_children ();
  }

  void resize_children ()
  {
    //  set the geometry of all children 
    for (std::vector <QWidget *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child) {
        (*child)->setGeometry (0, 0, width (), height ());
      }
    }
    mp_bglabel->setGeometry (0, 0, width (), height ());
  }

  std::vector <QWidget *> m_widgets;
  QLabel *mp_bglabel;
};

// -------------------------------------------------------------

class ViewWidgetStack 
  : public QWidget
{
public:
  ViewWidgetStack (QWidget *parent = 0, const char *name = 0)
    : QWidget (parent, name)
  {
    mp_bglabel = new LabelWithBackground (this);
    mp_bglabel->setText (tl::translate ("Use File/Open to open a layout").c_str ());
    mp_bglabel->setAlignment (Qt::AlignVCenter | Qt::AlignHCenter);
    mp_bglabel->show ();
  }

  void addWidget (LayoutView *w)
  {
    m_widgets.push_back (w);
    w->setParent (this);
    resize_children ();
    raiseWidget (m_widgets.size () - 1);

    updateGeometry ();
  }

  void removeWidget (size_t index)
  {
    if (index < m_widgets.size ()) {
      m_widgets.erase (m_widgets.begin () + index);
    }
    if (m_widgets.size () == 0) {
      mp_bglabel->show ();
    }
  }

  void raiseWidget (size_t index)
  {
    size_t i;
    i = 0;
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child, ++i) {
      if (i != index) {
        (*child)->deactivate ();
      }
    }

    if (index < m_widgets.size ()) {
      mp_bglabel->hide ();
      m_widgets [index]->activate ();
      m_widgets [index]->show ();
    } else {
      mp_bglabel->show ();
    }

    i = 0;
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child, ++i) {
      if (i != index) {
        (*child)->hide ();
      }
    }
  }
  
  LayoutView *widget (size_t index)
  {
    if (index < m_widgets.size ()) {
      return m_widgets [index];
    } else {
      return 0;
    }
  }

  QWidget *background_widget ()
  {
    return mp_bglabel;
  }

protected:
  virtual void resizeEvent (QResizeEvent *)
  {
    resize_children ();
  }

  void resize_children ()
  {
    //  set the geometry of all children 
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      (*child)->setGeometry (0, 0, width (), height ());
    }
    mp_bglabel->setGeometry (0, 0, width (), height ());
  }

  std::vector <LayoutView *> m_widgets;
  QLabel *mp_bglabel;
};

// -------------------------------------------------------------

static MainWindow *mw_instance = 0;

MainWindow *
MainWindow::instance ()
{
  return mw_instance; 
}

// -----------------------------------

MainWindow::MainWindow (QApplication *app, const char *name)
    : QMainWindow (0, name), 
      m_mode (std::numeric_limits<unsigned int>::max ()),
      mp_setup_form (new SettingsForm (0, this, "setup_form")),
      m_current_view (std::numeric_limits<size_t>::max ()),
      m_open_mode (0),
      m_edits_enabled (true),
      m_disable_tab_selected (false),
      m_exited (false),
      m_do_update_menu_dm (this, &MainWindow::do_update_menu),
      m_grid_micron (0.001), 
      m_default_grids_updated (true),
      m_new_cell_window_size (2.0),
      m_new_layout_current_panel (false),
      m_default_dbu (0.001),
      m_synchronized_views (false),
      m_synchronous (false),
      m_disabled_edits (0),
      m_work_in_progress (false),
      mp_app (app)
{
  if (mw_instance != 0) {
    throw tl::Exception (tl::translate ("Only one instance of MainWindow may be created"));
  }
  mw_instance = this;

  init_menu ();

  mp_console = new rba::ConsoleImpl ();
  mp_console->setModal (false);

  mp_pr = new lay::ProgressReporter (this, app);

  mp_main_stack_widget = new QStackedWidget (this);
  mp_main_stack_widget->setObjectName ("main_stack");
  setCentralWidget (mp_main_stack_widget);

  mp_main_frame = new QFrame (mp_main_stack_widget, "main_frame");
  mp_main_stack_widget->addWidget (mp_main_frame);

  mp_progress_widget = new ProgressWidget (mp_pr, mp_main_stack_widget);
  mp_progress_widget->setObjectName ("progress");
  mp_main_stack_widget->addWidget (mp_progress_widget);

  mp_main_stack_widget->setCurrentIndex (0);

  QVBoxLayout *vbl = new QVBoxLayout (mp_main_frame);
  vbl->setMargin (0);
  vbl->setSpacing (0);

  mp_tab_bar = new QTabBar (mp_main_frame);
  vbl->addWidget (mp_tab_bar);
  connect (mp_tab_bar, SIGNAL (currentChanged (int)), this, SLOT (view_selected (int)));

  mp_hp_dock_widget = new QDockWidget (tl::translate ("Cells").c_str (), this);
  mp_hp_dock_widget->setObjectName ("hp_dock_widget");
  mp_hp_stack = new ControlWidgetStack (mp_hp_dock_widget, "hp_stack");
  mp_hp_dock_widget->setWidget (mp_hp_stack);
  connect (mp_hp_dock_widget, SIGNAL (visibilityChanged (bool)), this, SLOT (dock_widget_visibility_changed (bool)));
  m_hp_visible = true;

  mp_view_stack = new ViewWidgetStack (mp_main_frame, "view_stack");
  vbl->addWidget (mp_view_stack);

  mp_layer_toolbox_dock_widget = new QDockWidget (tl::translate ("Layer Toolbox").c_str (), this);
  mp_layer_toolbox_dock_widget->setObjectName ("lt_dock_widget");
  mp_layer_toolbox = new LayerToolbox (mp_layer_toolbox_dock_widget, "layer_toolbox");
  mp_layer_toolbox_dock_widget->setWidget (mp_layer_toolbox);
  connect (mp_layer_toolbox_dock_widget, SIGNAL (visibilityChanged (bool)), this, SLOT (dock_widget_visibility_changed (bool)));
  m_layer_toolbox_visible = true;

  mp_lp_dock_widget = new QDockWidget (tl::translate ("Layers").c_str (), this);
  mp_lp_dock_widget->setObjectName ("lp_dock_widget");
  mp_lp_stack = new ControlWidgetStack (mp_lp_dock_widget, "lp_stack");
  mp_lp_dock_widget->setWidget (mp_lp_stack);
  connect (mp_lp_dock_widget, SIGNAL (visibilityChanged (bool)), this, SLOT (dock_widget_visibility_changed (bool)));
  m_lp_visible = true;

  mp_navigator_dock_widget = new QDockWidget (tl::translate ("Navigator").c_str (), this);
  mp_navigator_dock_widget->setObjectName ("navigator_dock_widget");
  mp_navigator = new Navigator (this);
  mp_navigator_dock_widget->setWidget (mp_navigator);
  connect (mp_navigator_dock_widget, SIGNAL (visibilityChanged (bool)), this, SLOT (dock_widget_visibility_changed (bool)));
  m_navigator_visible = true;

  //  Add dock widgets
#if QT_VERSION >= 0x040500
  setTabPosition (Qt::AllDockWidgetAreas, QTabWidget::North);
#endif
  addDockWidget(Qt::LeftDockWidgetArea, mp_navigator_dock_widget);
  addDockWidget(Qt::LeftDockWidgetArea, mp_hp_dock_widget);
  addDockWidget(Qt::RightDockWidgetArea, mp_lp_dock_widget);
  addDockWidget(Qt::RightDockWidgetArea, mp_layer_toolbox_dock_widget);

  mp_tool_bar = new QToolBar (this);
  mp_tool_bar->setCaption (tl::translate ("Toolbar").c_str ());
  mp_tool_bar->setObjectName ("toolbar");
  mp_tool_bar->setMovable (false);
  mp_tool_bar->setToolButtonStyle (Qt::ToolButtonTextUnderIcon);
  addToolBar (Qt::TopToolBarArea, mp_tool_bar);

  QMenuBar *mbar = menuBar ();
  mbar->setObjectName ("menubar");

  mp_menu->build (mbar, mp_tool_bar);

  connect (mp_menu, SIGNAL (changed ()), this, SLOT (menu_changed ()));

  mp_status_bar = statusBar ();
  mp_status_bar->setObjectName ("status_bar");

  mp_msg_label = new QLabel (mp_status_bar, "msg_label");
  mp_status_bar->addWidget (mp_msg_label, 1);

  mp_cp_frame = new QFrame (mp_status_bar);
  mp_status_bar->addWidget (mp_cp_frame);

  QHBoxLayout *cp_frame_ly = new QHBoxLayout (mp_cp_frame);
  cp_frame_ly->setMargin (0);
  cp_frame_ly->setSpacing (0);
  cp_frame_ly->insertSpacing (-1, 6);
  QLabel *xl = new QLabel ("x:", mp_cp_frame, "x");
  cp_frame_ly->addWidget (xl);
  cp_frame_ly->insertSpacing (-1, 6);
  mp_cpx_label = new QLabel (mp_cp_frame, "cpx_label");
  mp_cpx_label->setAlignment (Qt::AlignVCenter | Qt::AlignRight);
  mp_cpx_label->setMinimumSize (100, 0);
  cp_frame_ly->addWidget (mp_cpx_label);
  cp_frame_ly->insertSpacing (-1, 6);
  QLabel *yl = new QLabel ("y:", mp_cp_frame, "y");
  cp_frame_ly->addWidget (yl);
  cp_frame_ly->insertSpacing (-1, 6);
  mp_cpy_label = new QLabel (mp_cp_frame, "cpy_label");
  mp_cpy_label->setAlignment (Qt::AlignVCenter | Qt::AlignRight);
  mp_cpy_label->setMinimumSize (100, 0);
  cp_frame_ly->addWidget (mp_cpy_label);
  cp_frame_ly->insertSpacing (-1, 6);

  //  connect to the menus to provide the dynamic parts
  QMenu *edit_menu = mp_menu->menu ("edit_menu");
  tl_assert (edit_menu != 0);
  connect (edit_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));
  connect (edit_menu, SIGNAL (aboutToHide ()), this, SLOT (popup_menu_hide ()));

  QMenu *zoom_menu = mp_menu->menu ("zoom_menu");
  tl_assert (zoom_menu != 0);
  connect (zoom_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));
  connect (zoom_menu, SIGNAL (aboutToHide ()), this, SLOT (popup_menu_hide ()));

  QMenu *bookmark_menu = mp_menu->menu ("bookmark_menu");
  tl_assert (bookmark_menu != 0);
  connect (bookmark_menu, SIGNAL (aboutToShow ()), this, SLOT (bookmark_menu_show ()));

  QMenu *file_menu = mp_menu->menu ("file_menu");
  tl_assert (file_menu != 0);
  connect (file_menu, SIGNAL (aboutToShow ()), this, SLOT (file_menu_show ()));

  //  select the default mode
  select_mode (lay::LayoutView::default_mode ());

  //  create file dialogs:

  //  session file dialog
  mp_session_fdia = new lay::FileDialog (this, 
                          tl::translate ("Session File"), 
                          tl::translate ("Session files (*.lys);;All files (*)"), 
                          "lys");

  //  bookmarks file dialog
  mp_bookmarks_fdia = new lay::FileDialog (this, 
                            tl::translate ("Bookmarks File"), 
                            tl::translate ("Bookmark files (*.lyb);;All files (*)"), 
                            "lyb");
  //  layer properties
  mp_lprops_fdia = new lay::FileDialog (this, 
                            tl::translate ("Layer Properties File"), 
                            tl::translate ("Layer properties files (*.lyp);;All files (*)"), 
                            "lyp");
  //  screenshots
  mp_screenshot_fdia = new lay::FileDialog (this, 
                            tl::translate ("Screenshot"), 
                            tl::translate ("PNG files (*.png);;All files (*)"), 
                            "png");

  
  //  layout file dialog
  std::string fmts = tl::translate ("All layout files (");
  for (tl::Registrar<db::StreamFormatDeclaration>::iterator rdr = tl::Registrar<db::StreamFormatDeclaration>::begin (); rdr != tl::Registrar<db::StreamFormatDeclaration>::end (); ++rdr) {
    if (rdr != tl::Registrar<db::StreamFormatDeclaration>::begin ()) {
      fmts += " ";
    }
    std::string f = rdr->file_format ();
    const char *fp = f.c_str ();
    while (*fp && *fp != '(') {
      ++fp;
    }
    if (*fp) {
      ++fp;
    }
    while (*fp && *fp != ')') {
      fmts += *fp++;
    }
  }
  fmts += ");;";
  for (tl::Registrar<db::StreamFormatDeclaration>::iterator rdr = tl::Registrar<db::StreamFormatDeclaration>::begin (); rdr != tl::Registrar<db::StreamFormatDeclaration>::end (); ++rdr) {
    fmts += rdr->file_format ();
    fmts += ";;";
  }
  fmts += tl::translate ("All files (*)");
  mp_layout_fdia = new lay::FileDialog (this, tl::translate ("Layout File"), fmts);

  //  save & load layout options
  mp_layout_save_options = new lay::SaveLayoutOptionsDialog (this, tl::translate ("Layout Writer Options"));
  mp_layout_load_options = new lay::LoadLayoutOptionsDialog (this, tl::translate ("Layout Reader Options"));

  //  log viewer dialog
  mp_log_viewer_dialog = new lay::LogViewerDialog (0);

  //  install timer for message timeout
  connect (&m_message_timer, SIGNAL (timeout ()), this, SLOT (message_timer ()));
  m_message_timer.setSingleShot (true);

  //  make the main window accept drops
  setAcceptDrops (true);

  //  initialize the plugins (this should be the last action in the constructor since the
  //  main window should be functional now.
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    lay::PluginDeclaration *pd = const_cast<lay::PluginDeclaration *> (&*cls);
    pd->initialize (this);
  }

  //  establish the configuration
  config_setup ();
}

MainWindow::~MainWindow ()
{
  //  uninitialize the plugins (this should be the first action in the constructor since the
  //  main window should be functional still.
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    lay::PluginDeclaration *pd = const_cast<lay::PluginDeclaration *> (&*cls);
    pd->uninitialize (this);
  }

  //  since the configuration actions unregister themselves, we need to do this before the main
  //  window is gone:
  m_ca_collection.clear ();

  mw_instance = 0;

  //  explicitly delete the views here. Otherwise they
  //  are deleted by ~QWidget, which is too late since then
  //  the manager does not exist any longer.
  m_view_closed_observers.remove_observers (); // don't send events
  close_all ();

  //  delete the Menu after the views because they may want to access them in the destructor

  delete mp_menu;
  mp_menu = 0;

  delete mp_pr;
  mp_pr = 0;

  delete mp_setup_form;
  mp_setup_form = 0;

  delete mp_log_viewer_dialog;
  mp_log_viewer_dialog = 0;
}

void 
MainWindow::init_menu ()
{
  //  default menu layout
  
  MenuLayoutEntry empty_menu [] = {
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry file_menu [] = {
    MenuLayoutEntry ("new_layout:edit:edit_mode", tl::translate ("New Layout"),    SLOT (cm_new_layout ())),
    MenuLayoutEntry ("new_panel:edit:edit_mode", tl::translate ("New Panel"),      SLOT (cm_new_panel ())),
    MenuLayoutEntry::separator ("post_new_group:edit_mode"),
    MenuLayoutEntry ("open:edit",         tl::translate ("Open"),                  SLOT (cm_open ())),
    MenuLayoutEntry ("open_same_panel:edit", tl::translate ("Open In Same Panel"), SLOT (cm_open_too ()),             QKeySequence ("Shift+Ctrl+O")),
    MenuLayoutEntry ("open_new_panel:edit", tl::translate ("Open In New Panel"),   SLOT (cm_open_new_view ()),        QKeySequence ("Ctrl+O")),
    MenuLayoutEntry ("close:edit",        tl::translate ("Close"),                 SLOT (cm_close ()),                QKeySequence ("Ctrl+W")),
    MenuLayoutEntry ("clone",             tl::translate ("Clone Panel"),           SLOT (cm_clone ())),
    MenuLayoutEntry ("reload:edit",       tl::translate ("Reload"),                SLOT (cm_reload ()),               QKeySequence ("Ctrl+R")),
    MenuLayoutEntry ("pull_in:edit",      tl::translate ("Pull In Other Layout"),  SLOT (cm_pull_in ())),
    MenuLayoutEntry ("reader_options",    tl::translate ("Reader Options"),        SLOT (cm_reader_options ())),
    MenuLayoutEntry::separator ("open_recent_group"),
    MenuLayoutEntry ("open_recent_menu:edit",  tl::translate ("Open Recent"),      empty_menu),
    MenuLayoutEntry::separator ("import_group"),
    MenuLayoutEntry ("import_menu:edit",  tl::translate ("Import"),                empty_menu),
    MenuLayoutEntry::separator ("save_group"),
    MenuLayoutEntry ("save",              tl::translate ("Save"),                  SLOT (cm_save ())),
    MenuLayoutEntry ("save_as",           tl::translate ("Save As"),               SLOT (cm_save_as ())),
    MenuLayoutEntry::separator ("setup_group"),
    MenuLayoutEntry ("setup:edit",        tl::translate ("Setup"),                 SLOT (cm_setup ())),
    MenuLayoutEntry::separator ("misc_group"),
    MenuLayoutEntry ("screenshot:edit",   tl::translate ("Screenshot"),            SLOT (cm_screenshot ()),           QKeySequence ("Print")),
    MenuLayoutEntry ("layout_props",      tl::translate ("Layout Properties"),     SLOT (cm_layout_props ())),
    MenuLayoutEntry::separator ("layer_group"),
    MenuLayoutEntry ("load_layer_props:edit", tl::translate ("Load Layer Properties"), SLOT (cm_load_layer_props ())),
    MenuLayoutEntry ("save_layer_props:edit", tl::translate ("Save Layer Properties"), SLOT (cm_save_layer_props ())),
    MenuLayoutEntry::separator ("session_group"),
    MenuLayoutEntry ("restore_session",   tl::translate ("Restore Session"),       SLOT (cm_restore_session ())),
    MenuLayoutEntry ("save_session",      tl::translate ("Save Session"),          SLOT (cm_save_session ())),
    MenuLayoutEntry::separator ("log_group"),
    MenuLayoutEntry ("view_log",          tl::translate ("Log Viewer"),            SLOT (cm_view_log ())),
    MenuLayoutEntry::separator ("print_group"),
    MenuLayoutEntry ("print",             tl::translate ("Print"),                 SLOT (cm_print ()),                QKeySequence ("Ctrl+P")),
    MenuLayoutEntry::separator ("exit_group"),
    MenuLayoutEntry ("exit",              tl::translate ("Exit"),                  SLOT (cm_exit ()),                 QKeySequence ("Ctrl+Q")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_utilities_menu [] = {
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_cell_menu [] = {
    MenuLayoutEntry ("new_cell:edit:edit_mode",     tl::translate ("New Cell"),    SLOT (cm_new_cell ())),
    MenuLayoutEntry ("delete_cell:edit:edit_mode",  tl::translate ("Delete Cell"), SLOT (cm_cell_delete ())),
    MenuLayoutEntry ("rename_cell:edit:edit_mode",  tl::translate ("Rename Cell"), SLOT (cm_cell_rename ())),
    MenuLayoutEntry ("flatten_cell:edit:edit_mode", tl::translate ("Flatten Cell"), SLOT (cm_cell_flatten ())),
    MenuLayoutEntry::separator ("ops_group"),
    MenuLayoutEntry ("adjust_cell_origin:edit:edit_mode", tl::translate ("Adjust Origin"), SLOT (cm_adjust_origin ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_layer_menu [] = {
    MenuLayoutEntry ("new_layer:edit:edit_mode",   tl::translate ("New Layer"),    SLOT (cm_new_layer ())),
    MenuLayoutEntry ("clear_layer:edit:edit_mode", tl::translate ("Clear Layer"),  SLOT (cm_clear_layer ())),
    MenuLayoutEntry ("delete_layer:edit:edit_mode", tl::translate ("Delete Layer"), SLOT (cm_delete_layer ())),
    MenuLayoutEntry ("edit_layer:edit",            tl::translate ("Edit Layer Specification"),    SLOT (cm_edit_layer ())),
    MenuLayoutEntry::separator ("ops_group"),
    MenuLayoutEntry ("edit_boolean:edit:edit_mode", tl::translate ("Boolean Operations"), SLOT (cm_edit_boolean ())),
    MenuLayoutEntry ("edit_merge:edit:edit_mode",  tl::translate ("Merge"),        SLOT (cm_edit_merge ())),
    MenuLayoutEntry ("edit_size:edit:edit_mode",   tl::translate ("Size"),         SLOT (cm_edit_size ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_selection_menu [] = {
    MenuLayoutEntry ("sel_flip_x:edit_mode",       tl::translate ("Flip Horizontally"),  SLOT (cm_sel_flip_x ())),
    MenuLayoutEntry ("sel_flip_y:edit_mode",       tl::translate ("Flip Vertically"),    SLOT (cm_sel_flip_y ())),
    MenuLayoutEntry ("sel_rot_cw:edit_mode",       tl::translate ("Rotate Clockwise"),   SLOT (cm_sel_rot_cw ())),
    MenuLayoutEntry ("sel_rot_ccw:edit_mode",      tl::translate ("Rotate Counterclockwise"),    SLOT (cm_sel_rot_ccw ())),
    MenuLayoutEntry ("sel_free_rot:edit_mode",     tl::translate ("Rotation By Angle"),  SLOT (cm_sel_free_rot ())),
    MenuLayoutEntry ("sel_scale:edit_mode",        tl::translate ("Scale"),        SLOT (cm_sel_scale ())),
    MenuLayoutEntry ("sel_move:edit_mode",         tl::translate ("Move By"),      SLOT (cm_sel_move ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry view_menu [] = {
    MenuLayoutEntry ("show_grid",         tl::translate ("Show Grid"),             std::make_pair (cfg_grid_visible, "?")),
    MenuLayoutEntry ("default_grid:default_grids_group", tl::translate ("Grid"),   empty_menu),
    MenuLayoutEntry::separator ("layout_group"),
    MenuLayoutEntry ("show_texts",        tl::translate ("Show Texts"),            std::make_pair (cfg_text_visible, "?")),
    MenuLayoutEntry ("show_cell_boxes",   tl::translate ("Show Cell Frames"),      std::make_pair (cfg_cell_box_visible, "?")),
    MenuLayoutEntry ("no_stipples",       tl::translate ("Show Layers Without Fill"),    std::make_pair (cfg_no_stipple, "?")),
    MenuLayoutEntry ("synchronized_views", tl::translate ("Synchronized Views"),   std::make_pair (cfg_synchronized_views, "?")),
    MenuLayoutEntry ("edit_top_level_selection:edit_mode", tl::translate ("Select Top Level Objects"),   std::make_pair (edt::cfg_edit_top_level_selection, "?")),
    MenuLayoutEntry::separator ("panels_group"),
    MenuLayoutEntry ("show_toolbar",      tl::translate ("Toolbar"),               std::make_pair (cfg_show_toolbar, "?")),
    MenuLayoutEntry ("show_navigator",    tl::translate ("Navigator"),             std::make_pair (cfg_show_navigator, "?")),
    MenuLayoutEntry ("show_layer_panel",  tl::translate ("Layers"),                std::make_pair (cfg_show_layer_panel, "?")),
    MenuLayoutEntry ("show_layer_toolbox",   tl::translate ("Layer Toolbox"),      std::make_pair (cfg_show_layer_toolbox, "?")),
    MenuLayoutEntry ("show_hierarchy_panel", tl::translate ("Cells"),              std::make_pair (cfg_show_hierarchy_panel, "?")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_menu [] = {
    MenuLayoutEntry ("undo:edit",         tl::translate ("Undo"),                  SLOT (cm_undo ()),                 QKeySequence ("Ctrl+Z")),
    MenuLayoutEntry ("redo:edit",         tl::translate ("Redo"),                  SLOT (cm_redo ()),                 QKeySequence ("Ctrl+Y")),
    MenuLayoutEntry::separator ("basic_group:edit_mode"),
    MenuLayoutEntry ("cell_menu:edit:edit_mode", tl::translate ("Cell"),           edit_cell_menu),
    MenuLayoutEntry ("layer_menu:edit",   tl::translate ("Layer"),         edit_layer_menu),
    MenuLayoutEntry ("selection_menu:edit:edit_mode", tl::translate ("Selection"), edit_selection_menu),
    MenuLayoutEntry::separator ("utils_group:edit_mode"),
    MenuLayoutEntry ("utils_menu:edit:edit_mode", tl::translate ("Utilities"),     edit_utilities_menu),
    MenuLayoutEntry::separator ("misc_group") ,
    MenuLayoutEntry ("delete:edit",       tl::translate ("Delete"),                SLOT (cm_delete ()),               QKeySequence ("Del")),
    MenuLayoutEntry ("show_properties:edit", tl::translate ("Properties"),         SLOT (cm_show_properties ()),      QKeySequence ("Q")),
    MenuLayoutEntry::separator ("cpc_group"),
    MenuLayoutEntry ("copy:edit",         tl::translate ("Copy"),                  SLOT (cm_copy ()),                 QKeySequence ("Ctrl+C")),
    MenuLayoutEntry ("cut:edit",          tl::translate ("Cut"),                   SLOT (cm_cut ()),                  QKeySequence ("Ctrl+X")),
    MenuLayoutEntry ("paste:edit",        tl::translate ("Paste"),                 SLOT (cm_paste ()),                QKeySequence ("Ctrl+V")),
    MenuLayoutEntry::separator ("modes_group"),
    MenuLayoutEntry ("mode_menu",         tl::translate ("Mode"),                  empty_menu),
    MenuLayoutEntry ("select_menu",       tl::translate ("Select"),                empty_menu),
    MenuLayoutEntry::separator ("cancel_group"),
    MenuLayoutEntry ("cancel",            tl::translate ("Cancel"),                SLOT (cm_cancel ()),               QKeySequence ("Esc")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry bookmark_menu [] = {
    MenuLayoutEntry ("goto_bookmark_menu",tl::translate ("Goto Bookmark"),         empty_menu),
    MenuLayoutEntry ("bookmark_view",     tl::translate ("Bookmark This View"),    SLOT (cm_bookmark_view ())),
    MenuLayoutEntry::separator ("bookmark_mgm_group"),
    MenuLayoutEntry ("manage_bookmarks",  tl::translate ("Manage Bookmarks"),      SLOT (cm_manage_bookmarks ())),
    MenuLayoutEntry ("load_bookmarks",    tl::translate ("Load Bookmarks"),        SLOT (cm_load_bookmarks ())),
    MenuLayoutEntry ("save_bookmarks",    tl::translate ("Save Bookmarks"),        SLOT (cm_save_bookmarks ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry zoom_menu [] = {
    MenuLayoutEntry ("max_hier",          tl::translate ("Full Hierarchy"),        SLOT (cm_max_hier ()),             QKeySequence ("*")),
    MenuLayoutEntry ("max_hier_0",        tl::translate ("Box Only"),              SLOT (cm_max_hier_0 ()),           QKeySequence ("0")),
    MenuLayoutEntry ("max_hier_1",        tl::translate ("Top Level Only"),        SLOT (cm_max_hier_1 ()),           QKeySequence ("1")),
    MenuLayoutEntry ("inc_max_hier",      tl::translate ("Increment Hierarchy"),   SLOT (cm_inc_max_hier ()),         QKeySequence ("+")),
    MenuLayoutEntry ("dec_max_hier",      tl::translate ("Decrement Hierarchy"),   SLOT (cm_dec_max_hier ()),         QKeySequence ("-")),
    MenuLayoutEntry::separator ("zoom_group"),
    MenuLayoutEntry ("zoom_fit",          tl::translate ("Zoom Fit"),              SLOT (cm_zoom_fit ()),             QKeySequence ("F2")),
    MenuLayoutEntry ("zoom_fit_sel",      tl::translate ("Zoom Fit Selection"),    SLOT (cm_zoom_fit_sel ()),         QKeySequence ("Shift+F2")),
    MenuLayoutEntry ("zoom_in",           tl::translate ("Zoom In"),               SLOT (cm_zoom_in ()),              QKeySequence ("Return")),
    MenuLayoutEntry ("zoom_out",          tl::translate ("Zoom Out"),              SLOT (cm_zoom_out ()),             QKeySequence ("Shift+Return")),
    MenuLayoutEntry::separator ("pan_group"),
    MenuLayoutEntry ("pan_up",            tl::translate ("Pan Up"),                SLOT (cm_pan_up ()),               QKeySequence ("Up")),
    MenuLayoutEntry ("pan_down",          tl::translate ("Pan Down"),              SLOT (cm_pan_down ()),             QKeySequence ("Down")),
    MenuLayoutEntry ("pan_left",          tl::translate ("Pan Left"),              SLOT (cm_pan_left ()),             QKeySequence ("Left")),
    MenuLayoutEntry ("pan_right",         tl::translate ("Pan Right"),             SLOT (cm_pan_right ()),            QKeySequence ("Right")),
    MenuLayoutEntry::separator ("redraw_group"),
    MenuLayoutEntry ("redraw",            tl::translate ("Redraw"),                SLOT (cm_redraw ())),
    MenuLayoutEntry::separator ("state_group"),
    MenuLayoutEntry ("last_display_state", tl::translate ("Previous State"),       SLOT (cm_last_display_state ()),   QKeySequence ("Shift+Tab")),
    MenuLayoutEntry ("next_display_state", tl::translate ("Next State"),           SLOT (cm_next_display_state ()),   QKeySequence ("Tab")),
    MenuLayoutEntry::separator ("select_group"),
    MenuLayoutEntry ("select_cell:edit",  tl::translate ("Select Cell"),           SLOT (cm_select_cell ())),
    MenuLayoutEntry ("select_current_cell", tl::translate ("Show As New Top"),     SLOT (cm_select_current_cell ()),  QKeySequence ("Ctrl+S")), 
    MenuLayoutEntry ("goto_position",     tl::translate ("Goto Position"),         SLOT (cm_goto_position ()),        QKeySequence ("Ctrl+G")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry help_menu [] = {
    MenuLayoutEntry ("about",             tl::translate ("About"),                 SLOT (cm_help_about ())),
    MenuLayoutEntry ("about_qt",          tl::translate ("About Qt"),              SLOT (cm_help_about_qt ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry tools_menu_with_ruby [] = {
    MenuLayoutEntry ("console",           tl::translate ("RBA Console"),           SLOT (cm_console ())),
    MenuLayoutEntry::separator ("verification_group"),
    MenuLayoutEntry ("verification",      tl::translate ("Verification"),          empty_menu),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry tools_menu_without_ruby [] = {
    MenuLayoutEntry ("verification",      tl::translate ("Verification"),          empty_menu),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry main_menu [] = {
    MenuLayoutEntry ("file_menu",         tl::translate ("&File"),                 file_menu),
    MenuLayoutEntry ("edit_menu",         tl::translate ("&Edit"),                 edit_menu),
    MenuLayoutEntry ("view_menu",         tl::translate ("&View"),                 view_menu),
    MenuLayoutEntry ("bookmark_menu",     tl::translate ("&Bookmarks"),            bookmark_menu),
    MenuLayoutEntry ("zoom_menu",         tl::translate ("&Display"),              zoom_menu),
    MenuLayoutEntry ("tools_menu",        tl::translate ("&Tools"),                lay::Application::instance ()->ruby_interpreter ().ruby_available () ? tools_menu_with_ruby : tools_menu_without_ruby),
    MenuLayoutEntry::separator ("help_group"),
    MenuLayoutEntry ("help_menu",         tl::translate ("&Help"),                 help_menu),
    MenuLayoutEntry ("@toolbar",          "",                                      empty_menu),
    MenuLayoutEntry::last ()
  };

  mp_menu = new AbstractMenu (this);
  mp_menu->init (main_menu);

  lay::LayoutView::init_menu (*mp_menu);
  lay::Navigator::init_menu (*mp_menu);

  //  make all plugins that return true on "implements_editable" into menu entries and
  //  set up the actions accordingly

  if (mp_menu->is_valid ("edit_menu.select_menu")) {

    unsigned int n_plugin = 0;
    int n = 1;
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls, ++n_plugin) {

      std::string title;
      if (cls->implements_editable (title)) {
        
        //  extract first part, which is the name, separated by a tab from the title.
        std::string name = tl::sprintf ("pi_enable_%d", n++);
        std::string t (title);
        const char *tab = strchr (t.c_str (), '\t');
        if (tab) {
          name = std::string (t.c_str (), 0, tab - t.c_str ());
          title = tab + 1;
        } 

        lay::Action action (AbstractMenu::create_action (title));
        gtf::action_connect (action.qaction (), SIGNAL (triggered ()), this, SLOT (enable ()));
        action.qaction ()->setData (QVariant (n_plugin));

        m_plugin_enabled.push_back (true);
        m_plugin_enable_actions.push_back (std::make_pair (n_plugin, action));

        mp_menu->insert_item ("edit_menu.select_menu.end", name, action);

      } else {
        m_plugin_enabled.push_back (false);
      }

    }

  }

  //  add all the custom menus from the plugins
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {

    std::vector<lay::MenuEntry> menu_entries;
    cls->get_menu_entries (menu_entries);

    for (std::vector<lay::MenuEntry>::const_iterator m = menu_entries.begin (); m != menu_entries.end (); ++m) {

      if (m->title.empty ()) {
        mp_menu->insert_separator (m->insert_pos, m->menu_name);
      } else {

        if (m->sub_menu) {
          mp_menu->insert_menu (m->insert_pos, m->menu_name, m->title);
        } else {

          Action action (AbstractMenu::create_action (m->title));
          action.qaction ()->setData (QVariant (QString (m->symbol.c_str ())));
          gtf::action_connect (action.qaction (), SIGNAL (triggered ()), this, SLOT (generic_menu ()));
          mp_menu->insert_item (m->insert_pos, m->menu_name, action);

        }

      }

    }

  }

  //  Fill the mode menu file items from the intrinsic mouse modes followed by the 
  //  modes obtained from the plugins

  int n = 1;

  std::vector <std::string> mode_titles;
  lay::LayoutView::intrinsic_mouse_modes (&mode_titles);

  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    std::string title;
    cls->implements_mouse_mode (title);
    mode_titles.push_back (title);
  }

  QActionGroup *mode_grp = new QActionGroup (this);
  mode_grp->setExclusive (true);

  unsigned int mode_index = 0;
  for (std::vector <std::string>::const_iterator t = mode_titles.begin (); t != mode_titles.end (); ++t, ++mode_index) {

    if (! t->empty ()) {

      //  extract first part, which is the name, separated by a tab from the title.
      std::string name = tl::sprintf ("mode_%d", n++);
      std::string title = *t;
      const char *tab = strchr (t->c_str (), '\t');
      if (tab) {
        name = std::string (t->c_str (), 0, tab - t->c_str ());
        title = tab + 1;
      } 

      Action action (AbstractMenu::create_action (title));
      mode_grp->addAction (action.qaction ());

      action.set_checkable (true);
      action.qaction ()->setData (QVariant (mode_index));

      m_mode_menu_actions.push_back (action);

      mp_menu->insert_item ("edit_menu.mode_menu.end", name, action);
      mp_menu->insert_item ("@toolbar.end", name, action);

      gtf::action_connect (action.qaction (), SIGNAL (triggered ()), this, SLOT (mode ()));

    }

  }

  //  if not in editable mode, hide all entries from "edit_mode" group
  if (! lay::Application::instance ()->is_editable ()) {
    std::vector<std::string> edit_mode_grp = mp_menu->group ("edit_mode");
    for (std::vector<std::string>::const_iterator g = edit_mode_grp.begin (); g != edit_mode_grp.end (); ++g) {
      mp_menu->action (*g).set_visible (false);
    }
  }

  //  get and store the initial (default) key bindings for restore in the setup dialog
  KeyBindingsConfigPage::set_default ();
}

void 
MainWindow::dock_widget_visibility_changed (bool visible)
{
  if (sender () == mp_lp_dock_widget) {
    config_set (cfg_show_layer_panel, tl::to_string (!mp_lp_dock_widget->isHidden ()));
  } else if (sender () == mp_hp_dock_widget) {
    config_set (cfg_show_hierarchy_panel, tl::to_string (!mp_hp_dock_widget->isHidden ()));
  } else if (sender () == mp_navigator_dock_widget) {
    config_set (cfg_show_navigator, tl::to_string (!mp_navigator_dock_widget->isHidden ()));
  } else if (sender () == mp_layer_toolbox_dock_widget) {
    config_set (cfg_show_layer_toolbox, tl::to_string (!mp_layer_toolbox_dock_widget->isHidden ()));
  }
}

void
MainWindow::close_all ()
{
  mp_layer_toolbox->set_view (0);

  //  try a smooth shutdown of the current view
  if (current_view ()) {
    current_view ()->deactivate ();
  }
  m_current_view = mp_views.size ();

  for (unsigned int i = 0; i < mp_views.size (); ++i) {
    mp_views [i]->stop ();
  }

  m_manager.clear ();

  //  Clear the tab bar
  bool f = m_disable_tab_selected;
  m_disable_tab_selected = true;
  while (mp_tab_bar->count () > 0) {
    mp_tab_bar->removeTab (mp_tab_bar->count () - 1);
  }
  m_disable_tab_selected = f;

  //  First pop the mp_views vector and then delete. This way,
  //  any callbacks issued during the deleting of the views do
  //  not find any invalid view pointers but rather nothing.
  while (mp_views.size () > 0) {

    m_view_closed_observers.signal_observers (int (mp_views.size () - 1));

    lay::LayoutView *view = mp_views.back ();
    mp_views.pop_back ();
    mp_lp_stack->removeWidget (mp_views.size ());
    mp_hp_stack->removeWidget (mp_views.size ());
    mp_view_stack->removeWidget (mp_views.size ());

    delete view;

  }

  update_dock_widget_state ();
}

void
MainWindow::about_to_exec ()
{
  bool f;

  if (! lay::Application::instance ()->is_editable ()) {
    TipDialog td (this, 
                  tl::translate ("KLayout has been started in viewer mode. In this mode, editor functions are not available.\n\nTo enable these functions, start KLayout in editor mode by using the \"-e\" command line switch or select it as the default mode in the setup dialog. Choose \"Setup\" in the \"File\" menu and check \"Use editing mode by default\" on the \"Editing Mode\" page in the \"Application\" section."), 
                  "editor-mode");
    if (td.exec_dialog ()) {
      //  Don't bother the user with more dialogs.
      return;
    }
  }

  f = false;
  config_get (cfg_no_stipple, f);
  if (f) {
    TipDialog td (this, 
                  tl::translate ("Layers are shown without fill because fill has been intentionally turned off. This can be confusing since selecting a stipple does not have an effect in this case.\n\nTo turn this feature off, uncheck \"Show Layers Without Fill\" in the \"View\" menu."), 
                  "no-stipple");
    if (td.exec_dialog ()) {
      //  Don't bother the user with more dialogs.
      return;
    }
  }

  f = false;
  config_get (cfg_hide_empty_layers, f);
  if (f) {
    TipDialog td (this, 
                  tl::translate ("The \"Hide Empty Layers\" feature is enabled. This can be confusing, in particular in edit mode, because layers are not shown although they are actually present.\n\nTo disable this feature, uncheck \"Hide Empty Layers\" in the layer panel's context menu."), 
                  "hide-empty-layers");
    if (td.exec_dialog ()) {
      //  Don't bother the user with more dialogs.
      return;
    }
  }

  //  TODO: other tip-of-the-day dialogs?
}

void 
MainWindow::message (const std::string &s, int ms)
{
  mp_msg_label->setText (QString (" ") + QString (s.c_str ()));
  m_message_timer.start (ms);
}

void 
MainWindow::message_timer ()
{
  mp_msg_label->setText ("");
}

void
MainWindow::config_finalize ()
{
  lay::get_from_config (m_reader_options, this);

  // Not set the window state: this ensures we have handled cfg_window_geometry 
  // before we restore the state
  if (! m_config_window_state.empty ()) {
    QByteArray state = QByteArray::fromBase64 (m_config_window_state.c_str ());
    m_config_window_state.clear ();
    restoreState (state);
  }

  // Update the default grids menu if necessary
  if (m_default_grids_updated) {

    m_default_grids_updated = false;

    std::vector<std::string> group = menu ().group ("default_grids_group");

    for (std::vector<std::string>::const_iterator t = group.begin (); t != group.end (); ++t) {
      std::vector<std::string> items = menu ().items (*t);        
      for (std::vector<std::string>::const_iterator i = items.begin (); i != items.end (); ++i) {
        menu ().delete_item (*i);
      }
    }
      
    for (std::vector<lay::Action *>::iterator a = m_default_grid_actions.begin (); a != m_default_grid_actions.end (); ++a) {
      delete *a;
    }
    m_default_grid_actions.clear ();

    int i = 1;
    for (std::vector<double>::const_iterator g = m_default_grids.begin (); g != m_default_grids.end (); ++g, ++i) {
      m_default_grid_actions.push_back (create_config_action (tl::to_string (*g) + tl::translate (" um"), cfg_grid, tl::to_string (*g)));
      m_default_grid_actions.back ()->set_checkable (true);
      m_default_grid_actions.back ()->set_checked (fabs (*g - m_grid_micron) < 1e-10);
      for (std::vector<std::string>::const_iterator t = group.begin (); t != group.end (); ++t) {
        menu ().insert_item (*t + ".end", "default_grid_" + tl::to_string (i), *m_default_grid_actions.back ());
      }
    }

  }

  //  make the changes visible in the setup form if the form is visible 
  mp_setup_form->setup ();
}

bool 
MainWindow::configure (const std::string &name, const std::string &value)
{
  std::map<std::string, std::vector<lay::ConfigureAction *> >::iterator ca = m_configuration_actions.find (name);
  if (ca != m_configuration_actions.end ()) {
    for (std::vector<lay::ConfigureAction *>::const_iterator a = ca->second.begin (); a != ca->second.end (); ++a) {
      (*a)->configure (value);
    }
  }

  if (name == cfg_grid) {

    double g = 0.0;
    tl::from_string (value, g);
    m_grid_micron = g;
    m_default_grids_updated = true;
    return false; // not taken - let others set it too.

  } else if (name == cfg_default_grids) {

    tl::Extractor ex (value.c_str ());
    m_default_grids.clear ();
    m_default_grids_updated = true;

    //  convert the list of grids to a list of doubles 
    while (! ex.at_end ()) {
      double g = 0.0;
      if (! ex.try_read (g)) {
        break;
      }
      m_default_grids.push_back (g);
      ex.test (",");
    }

    return true;

  } else if (name == cfg_stipple_palette) {

    lay::StipplePalette palette = lay::StipplePalette::default_palette ();

    try {
      //  empty string means: default palette
      if (! value.empty ()) {
        palette.from_string (value);
      }
    } catch (...) {
      //  ignore errors: just reset the palette 
      palette = lay::StipplePalette::default_palette ();
    }

    mp_layer_toolbox->set_palette (palette);

    // others need this property too ..
    return false;

  } else if (name == cfg_color_palette) {

    lay::ColorPalette palette = lay::ColorPalette::default_palette ();

    try {
      //  empty string means: default palette
      if (! value.empty ()) {
        palette.from_string (value);
      }
    } catch (...) {
      //  ignore errors: just reset the palette 
      palette = lay::ColorPalette::default_palette ();
    }

    mp_layer_toolbox->set_palette (palette);

    // others need this property too ..
    return false;

  } else if (name == cfg_mru) {

    tl::Extractor ex (value.c_str ());

    m_mru.clear ();
    while (! ex.at_end ()) {
      m_mru.push_back ("");
      ex.read_quoted (m_mru.back ());
    }

    return true;

  } else if (name == cfg_micron_digits) {

    unsigned int d = 5;
    tl::from_string (value, d);
    tl::set_micron_resolution (d);

    return true;

  } else if (name == cfg_dbu_digits) {

    unsigned int d = 2;
    tl::from_string (value, d);
    tl::set_db_resolution (d);

    return true;

  } else if (name == cfg_window_state) {

    //  restore the state on config_finalize to ensure we have handled it after 
    //  restoring the geometry
    m_config_window_state = value;
    return true;

  } else if (name == cfg_window_geometry) {

    if (! value.empty ()) {
      QByteArray state = QByteArray::fromBase64 (value.c_str ());
      restoreGeometry (state);
    }

    return true;

  } else if (name == cfg_show_layer_toolbox) {

    tl::from_string (value, m_layer_toolbox_visible);
    if (m_layer_toolbox_visible) {
      mp_layer_toolbox_dock_widget->show ();
    } else {
      mp_layer_toolbox_dock_widget->hide ();
    }

    return true;

  } else if (name == cfg_show_navigator) {

    tl::from_string (value, m_navigator_visible);
    if (m_navigator_visible) {
      mp_navigator_dock_widget->show ();
    } else {
      mp_navigator_dock_widget->hide ();
    }

    return true;

  } else if (name == cfg_navigator_show_images) {

    bool flag = false;
    tl::from_string (value, flag);
    mp_navigator->show_images (flag);

    return true;

  } else if (name == cfg_navigator_all_hier_levels) {

    bool flag = false;
    tl::from_string (value, flag);
    mp_navigator->all_hier_levels (flag);

    return true;

  } else if (name == cfg_show_toolbar) {

    bool flag = false;
    tl::from_string (value, flag);
    if (flag) {
      mp_tool_bar->show ();
    } else {
      mp_tool_bar->hide ();
    }

    return true;

  } else if (name == cfg_show_hierarchy_panel) {

    tl::from_string (value, m_hp_visible);
    if (m_hp_visible) {
      mp_hp_dock_widget->show ();
    } else {
      mp_hp_dock_widget->hide ();
    }

    return true;

  } else if (name == cfg_show_layer_panel) {

    tl::from_string (value, m_lp_visible);
    if (m_lp_visible) {
      mp_lp_dock_widget->show ();
    } else {
      mp_lp_dock_widget->hide ();
    }

    return true;

  } else if (name == cfg_synchronized_views) {

    bool flag = false;
    tl::from_string (value, flag);
    m_synchronized_views = flag;
    return true;

  } else if (name == cfg_default_dbu) {

    double d = 0.0;
    tl::from_string (value, d);
    m_default_dbu = d;
    return false; // not taken - let others set it too.

  } else if (name == cfg_key_bindings) {

    std::vector<std::pair<std::string, std::string> > key_bindings = unpack_key_binding (value);
    for (std::vector<std::pair<std::string, std::string> >::const_iterator kb = key_bindings.begin (); kb != key_bindings.end (); ++kb) {
      if (menu ().is_valid (kb->first)) {
        menu ().action (kb->first).set_shortcut (kb->second);
      }
    }

    return true;

  } else if (name == cfg_background_color) {

    if (mp_navigator) {
      QColor color;
      ColorConverter ().from_string (value, color);
      mp_navigator->background_color (color);
    }

    //  do not take - let others receive the background color events as well
    return false;

  } else if (name == cfg_technologies) {

    tl::Extractor ex (value.c_str ());
    m_technologies.parse (ex);

    return true;

  } else {
    return false;
  }

}

void 
MainWindow::enable_edits (bool enable)
{
  if (enable) {
    if (m_disabled_edits > 0) {
      --m_disabled_edits;
    }
    enable = (m_disabled_edits == 0);
  } else {
    ++m_disabled_edits;
  }

  std::vector<std::string> edit_grp = mp_menu->group ("edit");
  for (std::vector<std::string>::const_iterator g = edit_grp.begin (); g != edit_grp.end (); ++g) {
    mp_menu->action (*g).set_enabled (enable);
  }

  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->enable_edits (enable);
  }
  
  m_edits_enabled = enable;

  //  call this to establish a consistent set of options on the menu entries
  popup_menu_show ();
}

void
MainWindow::read_dock_widget_state ()
{
  config_set (cfg_show_layer_panel, tl::to_string (!mp_lp_dock_widget->isHidden ()));
  config_set (cfg_show_hierarchy_panel, tl::to_string (!mp_hp_dock_widget->isHidden ()));
  config_set (cfg_show_navigator, tl::to_string (!mp_navigator_dock_widget->isHidden ()));
  config_set (cfg_show_layer_toolbox, tl::to_string (!mp_layer_toolbox_dock_widget->isHidden ()));
}

void 
MainWindow::update_dock_widget_state ()
{
  if (m_hp_visible) {
    mp_hp_dock_widget->show ();
  } else {
    mp_hp_dock_widget->hide ();
  }

  if (m_lp_visible) {
    mp_lp_dock_widget->show ();
  } else {
    mp_lp_dock_widget->hide ();
  }

  if (m_navigator_visible) {
    mp_navigator_dock_widget->show ();
  } else {
    mp_navigator_dock_widget->hide ();
  }

  if (m_layer_toolbox_visible) {
    mp_layer_toolbox_dock_widget->show ();
  } else {
    mp_layer_toolbox_dock_widget->hide ();
  }
}

void
MainWindow::exit ()
{
  m_exited = true;
  do_close ();
  QMainWindow::close ();
}

int 
MainWindow::dirty_files (std::string &dirty_files)
{
  int dirty_layouts = 0;

  if (lay::Application::instance ()->is_editable ()) {
    std::vector <std::string> names;
    lay::LayoutHandle::get_names (names);
    for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {
      lay::LayoutHandle *handle = lay::LayoutHandle::find (*n);
      if (handle && handle->is_dirty ()) {
        ++dirty_layouts;
        if (! dirty_files.empty ()) {
          dirty_files += "\n";
        }
        dirty_files += "  ";
        dirty_files += handle->name ();
      }
    }
  }

  return dirty_layouts;
}

bool
MainWindow::can_close ()
{
  std::string df_list;
  int dirty_layouts = dirty_files (df_list);

  if (dirty_layouts == 0) {
    return true;
  } else {

    QMessageBox mbox (this);
    mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + df_list + "\n\nPress 'Exit Without Saving' to exit anyhow and discard changes.").c_str ());
    mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
    mbox.setIcon (QMessageBox::Warning);
    QAbstractButton *exit_button = mbox.addButton (tl::translate ("Exit Without Saving").c_str (), QMessageBox::YesRole);
    mbox.addButton (QMessageBox::Cancel);

    mbox.exec ();

    return mbox.clickedButton() == exit_button;

  }
}

void
MainWindow::do_close ()
{
  //  close all views
  close_all ();

  //  save the dock widget state with all views closed (that state can be 
  //  used for staring klayout without any layout)
  config_set (cfg_window_geometry, saveGeometry ().toBase64 ().data ());
  config_set (cfg_window_state, saveState ().toBase64 ().data ());
}

void 
MainWindow::closeEvent (QCloseEvent *event)
{
  if (! m_exited) {

    BEGIN_PROTECTED

    if (! can_close ()) {
      event->ignore ();
    } else {
      do_close ();
    }

    END_PROTECTED

  }
}

void
MainWindow::cm_navigator_freeze ()
{
  BEGIN_PROTECTED
  if (mp_navigator) {
    mp_navigator->freeze_clicked ();
  }
  END_PROTECTED
}

void
MainWindow::cm_navigator_close ()
{
  BEGIN_PROTECTED
  if (mp_navigator) {
    mp_navigator->close ();
  }
  END_PROTECTED
}

void
MainWindow::cm_view_log ()
{
  BEGIN_PROTECTED
  mp_log_viewer_dialog->show ();
  END_PROTECTED
}

void
MainWindow::cm_print ()
{
  BEGIN_PROTECTED

  m_printer.setDocName (QString ((tl::translate ("KLayout [") + (current_view () ? current_view ()->title () : std::string ()) + "]").c_str ()));

  QPrintDialog print_dialog (&m_printer, this);
  if (print_dialog.exec() == QDialog::Accepted) {

    if (current_view ()) {

      //  choose a resolution around 300dpi
      double rf = floor (0.5 + 300.0 / m_printer.resolution ());
      m_printer.setResolution (int (floor (0.5 + m_printer.resolution () * rf)));

      QRect page_rect = m_printer.pageRect ();

      QImage img = current_view ()->get_image_with_options (page_rect.width (), page_rect.height (), 2, 1, 1.0 / 3.0, Qt::white, Qt::black, Qt::black, db::DBox ());

      QPainter painter;
      painter.begin (&m_printer);
      painter.drawImage (QPoint (0, 0), img);
      painter.end ();

    } else {
      throw tl::Exception (tl::translate ("No view open to save the layer properties from"));
    }

  }

  END_PROTECTED
}

void
MainWindow::cm_exit ()
{
  BEGIN_PROTECTED
  if (can_close ()) {
    //  actually exit.
    exit ();
  }
  END_PROTECTED
}

lay::LayoutView *
MainWindow::view (int index)
{
  tl_assert (index >= 0 && index < int (mp_views.size ()));
  return mp_views [index];
}

const lay::LayoutView *
MainWindow::view (int index) const
{
  tl_assert (index >= 0 && index < int (mp_views.size ()));
  return mp_views [index];
}

int 
MainWindow::current_view_index () const
{
  if (m_current_view >= mp_views.size ()) {
    return -1;
  } else {
    return int (m_current_view);
  }
}

lay::LayoutView *
MainWindow::current_view ()
{
  if (m_current_view >= mp_views.size ()) {
    return 0;
  } else {
    return mp_views [m_current_view];
  }
}

const lay::LayoutView *
MainWindow::current_view () const
{
  if (m_current_view >= mp_views.size ()) {
    return 0;
  } else {
    return mp_views [m_current_view];
  }
}

void 
MainWindow::cm_show_properties ()
{
  if (current_view ()) {
    current_view ()->show_properties (this);
  }
}

void 
MainWindow::cm_delete ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->del ();
    //  because a "delete" might involve objects currently edited, we cancel the edit after we have deleted the object
    current_view ()->cancel ();
    current_view ()->clear_selection ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_cell_paste ()
{
  BEGIN_PROTECTED
  current_view ()->cm_cell_paste ();
  END_PROTECTED
}

void 
MainWindow::cm_cell_cut ()
{
  BEGIN_PROTECTED
  current_view ()->cm_cell_cut ();
  END_PROTECTED
}

void 
MainWindow::cm_cell_copy ()
{
  BEGIN_PROTECTED
  current_view ()->cm_cell_copy ();
  END_PROTECTED
}

void 
MainWindow::cm_copy ()
{
  BEGIN_PROTECTED

  if (current_view () && current_view ()->has_selection ()) {
    current_view ()->copy ();
    current_view ()->clear_selection ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_paste ()
{
  BEGIN_PROTECTED

  if (current_view () && ! db::Clipboard::instance ().empty ()) {
    current_view ()->cancel ();
    current_view ()->clear_selection ();
    current_view ()->paste ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_cut ()
{
  BEGIN_PROTECTED

  if (current_view () && current_view ()->has_selection ()) {
    current_view ()->cut ();
    current_view ()->cancel (); //  see del() for reason why cancel is after cut
    current_view ()->clear_selection ();
  }

  END_PROTECTED
}

void 
MainWindow::mode ()
{
  QAction *action = dynamic_cast<QAction *> (sender ());
  if (action) {
    select_mode ((unsigned int) action->data ().toInt ());
  }
}

void
MainWindow::select_mode (unsigned int m)
{
  if (m_mode != m) {

    m_mode = m;
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->mode (m);
    }

    //  update mode menu
    for (std::vector<lay::Action>::iterator i = m_mode_menu_actions.begin (); i != m_mode_menu_actions.end (); ++i) {
      i->set_checked (i->qaction ()->data ().toInt () == int (m_mode));
    }

  }
}

void 
MainWindow::enable ()
{
  BEGIN_PROTECTED

  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);

  int m = action->data ().toInt ();
  m_plugin_enabled [m] = ! m_plugin_enabled [m];
  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->enable_editable (m, m_plugin_enabled [m]);
  }

  END_PROTECTED
}

void 
MainWindow::cm_undo ()
{
  BEGIN_PROTECTED

  if (current_view () && m_manager.available_undo ().first) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->clear_selection ();
      (*vp)->cancel ();
    }
    m_manager.undo ();
    current_view ()->update_content ();
  }

  END_PROTECTED
}

void
MainWindow::cm_redo ()
{
  BEGIN_PROTECTED

  if (current_view () && m_manager.available_redo ().first) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->clear_selection ();
      (*vp)->cancel ();
    }
    m_manager.redo ();
    current_view ()->update_content ();
  }

  END_PROTECTED
}

void 
MainWindow::bookmark_menu_show ()
{
  if (mp_menu->is_valid ("bookmark_menu.goto_bookmark_menu")) {

    Action goto_bookmark_action = mp_menu->action ("bookmark_menu.goto_bookmark_menu");
    bool has_bookmarks = current_view () && current_view ()->bookmarks ().size () > 0;

    if (has_bookmarks && m_edits_enabled) {

      goto_bookmark_action.set_enabled (true);

      QMenu *goto_bookmark_menu = goto_bookmark_action.qaction ()->menu ();
      if (goto_bookmark_menu) {

        goto_bookmark_menu->clear ();

        if (current_view ()) {
          const lay::BookmarkList &bookmarks = current_view ()->bookmarks ();
          for (size_t i = 0; i < bookmarks.size (); ++i) {
            QAction *action = goto_bookmark_menu->addAction (bookmarks.name (i).c_str ());
            action->setObjectName (tl::sprintf ("bookmark_%d", i + 1).c_str ());
            gtf::action_connect (action, SIGNAL (triggered ()), this, SLOT (goto_bookmark ()));
            action->setData (QVariant (int (i))); 
          }
        }

      }

    } else {
      goto_bookmark_action.set_enabled (false);
    }

  }
}

void
MainWindow::goto_bookmark ()
{
  BEGIN_PROTECTED

  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);
  size_t id = size_t (action->data ().toInt ());
  if (current_view () && current_view ()->bookmarks ().size () > id) {
    current_view ()->goto_view (current_view ()->bookmarks ().state (id));
  }

  END_PROTECTED
}

void
MainWindow::cm_goto_position ()
{
  BEGIN_PROTECTED

  if (current_view ()) {

    while (true) {

      bool ok = false;

      db::DBox box (current_view ()->box ());
      std::string pos;
      pos += tl::micron_to_string (box.center ().x ()) + "," + tl::micron_to_string (box.center ().y ());
      pos += ",";
      pos += tl::micron_to_string (std::min (box.width (), box.height ()));

      QString text = QInputDialog::getText (tl::translate ("Enter Position").c_str (), tl::translate ("Enter position either as pair (x,y) or with window size (x,y,s)\n(x,y) will be the new window center position, s (if given) the new window size").c_str (),
                                            QLineEdit::Normal, QString (pos.c_str ()), &ok, this);

      if (! ok) {

        break;

      } else if (text.isEmpty ()) {

        throw tl::Exception (tl::translate ("Enter the position"));

      } else {

        double x = 0.0, y = 0.0, s = 0.0;
        tl::Extractor ex (text.ascii ());
        ex >> x >> "," >> y;
        if (! ex.at_end ()) {
          ex >> "," >> s >> tl::Extractor::end ();
          current_view ()->goto_window (db::DPoint (x, y), s);
        } else {
          current_view ()->goto_window (db::DPoint (x, y));
        }
        break;

      }

    }

  }

  END_PROTECTED
}

void
MainWindow::cm_manage_bookmarks ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->manage_bookmarks ();
  }

  END_PROTECTED
}

void
MainWindow::cm_bookmark_view ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    while (true) {
      bool ok = false;
      QString text = QInputDialog::getText (tl::translate ("Enter Bookmark Name").c_str (), tl::translate ("Bookmark name").c_str (),
                                            QLineEdit::Normal, QString::null, &ok, this);
      if (! ok) {
        break;
      } else if (text.isEmpty ()) {
        QMessageBox::critical (this, tl::translate ("Error").c_str (), tl::translate ("Enter a name for the bookmark").c_str ());
      } else {
        current_view ()->bookmark_view (text.ascii ());
        break;
      }
    }
  }

  END_PROTECTED
}

void
MainWindow::update_content ()
{
  BEGIN_PROTECTED

  mp_setup_form->setup ();
  if (current_view ()) {
    current_view ()->update_content ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_fit_sel ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_fit_sel ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_fit ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_fit ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_left ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_left ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_right ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_right ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_up ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_up ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_down ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_down ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_in ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_in ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_out ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_out ();
  }

  END_PROTECTED
}

void 
MainWindow::popup_menu_hide ()
{
  //  Enable all menu items as far as required on closing of the menu. Otherwise,
  //  keyboard shortcuts are not effective when the window is closed and the actions are
  //  left disabled.

  BEGIN_PROTECTED

  if (mp_menu->is_valid ("edit_menu.undo")) {
    Action undo_action = mp_menu->action ("edit_menu.undo");
    undo_action.set_enabled (m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.redo")) {
    Action redo_action = mp_menu->action ("edit_menu.redo");
    redo_action.set_enabled (m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.copy")) {
    Action copy_action = mp_menu->action ("edit_menu.copy");
    copy_action.set_enabled (m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.cut")) {
    Action cut_action = mp_menu->action ("edit_menu.cut");
    cut_action.set_enabled (m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.paste")) {
    Action paste_action = mp_menu->action ("edit_menu.paste");
    paste_action.set_enabled (m_edits_enabled);
  }

  if (mp_menu->is_valid ("zoom_menu.next_display_state")) {
    Action next_display_state_action = mp_menu->action ("zoom_menu.next_display_state");
    next_display_state_action.set_enabled (true);
  } 

  if (mp_menu->is_valid ("zoom_menu.last_display_state")) {
    Action last_display_state_action = mp_menu->action ("zoom_menu.last_display_state");
    last_display_state_action.set_enabled (true);
  }

  END_PROTECTED
}

void 
MainWindow::popup_menu_show ()
{
  BEGIN_PROTECTED

  if (mp_menu->is_valid ("edit_menu.undo")) {

    Action undo_action = mp_menu->action ("edit_menu.undo");

    std::string undo_txt (tl::translate ("&Undo"));
    bool undo_enable = false;
    if (current_view () && m_manager.available_undo ().first) {
      undo_txt += " - " + m_manager.available_undo ().second;
      undo_enable = true;
    }
    undo_action.set_title (undo_txt); 
    undo_action.set_enabled (undo_enable && m_edits_enabled);

  }

  if (mp_menu->is_valid ("edit_menu.redo")) {

    Action redo_action = mp_menu->action ("edit_menu.redo");

    std::string redo_txt (tl::translate ("&Redo"));
    bool redo_enable = false;
    if (current_view () && m_manager.available_redo ().first) {
      redo_txt += " - " + m_manager.available_redo ().second;
      redo_enable = true;
    }
    redo_action.set_title (redo_txt); 
    redo_action.set_enabled (redo_enable && m_edits_enabled);

  }

  if (mp_menu->is_valid ("edit_menu.copy")) {
    Action copy_action = mp_menu->action ("edit_menu.copy");
    copy_action.set_enabled (current_view () && current_view ()->has_selection () && m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.cut")) {
    Action cut_action = mp_menu->action ("edit_menu.cut");
    cut_action.set_enabled (current_view () && current_view ()->has_selection () && m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.paste")) {
    Action paste_action = mp_menu->action ("edit_menu.paste");
    paste_action.set_enabled (! db::Clipboard::instance ().empty () && m_edits_enabled);
  }

  if (mp_menu->is_valid ("zoom_menu.next_display_state")) {
    Action next_display_state_action = mp_menu->action ("zoom_menu.next_display_state");
    next_display_state_action.set_enabled (has_next_display_state ());
  } 

  if (mp_menu->is_valid ("zoom_menu.last_display_state")) {
    Action last_display_state_action = mp_menu->action ("zoom_menu.last_display_state");
    last_display_state_action.set_enabled (has_last_display_state ());
  }

  //  update mode menu
  for (std::vector< std::pair<unsigned int, lay::Action> >::iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i) {
    i->second.set_checkable (true);
    i->second.set_checked (m_plugin_enabled [i->first]);
  }

  END_PROTECTED
}

void
MainWindow::cm_redraw ()
{
  BEGIN_PROTECTED 
  redraw ();
  END_PROTECTED
}

void
MainWindow::redraw ()
{
  if (current_view ()) {
    current_view ()->redraw ();
  }
}

void
MainWindow::cm_cancel ()
{
  BEGIN_PROTECTED 
  cancel ();
  select_mode (lay::LayoutView::default_mode ());
  END_PROTECTED
}

/**
 *  @brief Commits any operations, cancels any pending edit operations and clears the selection.
 */
void
MainWindow::cancel ()
{
  BEGIN_PROTECTED 

  //  if any transaction is pending (this may happen when an operation threw an exception)
  //  close transactions.
  if (m_manager.transacting ()) {
    m_manager.commit ();
  }

  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->cancel ();
  }

  END_PROTECTED
}

void
MainWindow::cm_save_layer_props ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_lprops_fdia->get_save (fn)) {
      current_view ()->save_layer_props (fn);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to save the layer properties from"));
  }

  END_PROTECTED
}

void
MainWindow::load_layer_properties (const std::string &fn, bool all_views, bool add_default)
{
  if (all_views) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->load_layer_props (fn, add_default);
    }
  } else if (current_view ()) {
    current_view ()->load_layer_props (fn, add_default);
  }
}

void
MainWindow::load_layer_properties (const std::string &fn, int cv_index, bool all_views, bool add_default)
{
  if (all_views) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->load_layer_props (fn, cv_index, add_default);
    }
  } else if (current_view ()) {
    current_view ()->load_layer_props (fn, cv_index, add_default);
  }
}

bool 
MainWindow::is_single_cv_layer_properties_file (const std::string &fn)
{
  //  If the file contains information for a single layout but we have multiple ones, 
  //  show the dialog to determine what layout to apply the information to.
  std::vector<lay::LayerPropertiesList> props;
  try {
    tl::XMLFileSource in (fn);
    props.push_back (lay::LayerPropertiesList ());
    props.back ().load (in);
  } catch (...) {
    props.clear ();
    tl::XMLFileSource in (fn);
    lay::LayerPropertiesList::load (in, props);
  }

  //  Collect all cv indices in the layer properties 
  std::set <int> cv;
  for (std::vector<lay::LayerPropertiesList>::const_iterator p = props.begin (); p != props.end (); ++p) {
    for (lay::LayerPropertiesConstIterator lp = p->begin_const_recursive (); ! lp.at_end (); ++lp) {
      if (! lp->has_children ()) {
        cv.insert (lp->source (true).cv_index ());
        if (cv.size () >= 2) {
          break;
        }
      }
    }
  }

  return (cv.size () == 1);
}

void
MainWindow::cm_load_layer_props ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_lprops_fdia->get_open (fn)) {

      int target_cv_index = -2;

      if (current_view ()->cellviews () > 1 && is_single_cv_layer_properties_file (fn)) {

        QStringList items;
        items << QString (tl::translate ("Take it as it is").c_str ());
        items << QString (tl::translate ("Apply to all layouts").c_str ());
        for (unsigned int i = 0; i < current_view ()->cellviews (); ++i) {
          items << QString ((tl::translate ("Apply to ") + current_view ()->cellview (i)->name () + " (@" + tl::to_string (i + 1) + ")").c_str ());
        }

        bool ok;
        QString item = QInputDialog::getItem(this, tl::translate ("Apply Layer Properties File").c_str (),
                                                   tl::translate ("There are multiple layouts in that panel but the layer properties file contains properties for a single one.\nWhat should be done?").c_str (),
                                                   items, 1, false, &ok);
        if (!ok || item.isEmpty()) {
          return;
        }

        target_cv_index = items.indexOf (item) - 2;

      }

      if (target_cv_index > -2) {
        load_layer_properties (fn, target_cv_index, false /*current view only*/, false /*don't add default*/);
      } else {
        load_layer_properties (fn, false /*current view only*/, false /*don't add default*/);
      }

    }
  } else {
    throw tl::Exception (tl::translate ("No view open to load the layer properties for"));
  }

  END_PROTECTED
}

void
MainWindow::save_session (const std::string &fn)
{
  m_current_session = fn;
  lay::Session session;
  session.fetch (*this);
  session.save (fn);
}

void
MainWindow::restore_session (const std::string &fn)
{
  m_current_session = fn;
  lay::Session session;
  session.load (fn);
  session.restore (*this);
  read_dock_widget_state ();
}

void
MainWindow::cm_save_session ()
{
  BEGIN_PROTECTED

  std::string df_list;
  int dirty_layouts = dirty_files (df_list);

  if (dirty_layouts == 0 || 
    QMessageBox::warning (this, 
      tl::translate ("Save Needed For Some Layouts").c_str (),
      (tl::translate ("The following layouts need saving.\nThese layouts must be saved manually:\n\n") + df_list + "\n\nPress 'Ok' to continue.").c_str (),
      QMessageBox::Ok | QMessageBox::Cancel,
      QMessageBox::Cancel) == QMessageBox::Ok) {

    std::string fn = m_current_session;
    if (mp_session_fdia->get_save (fn)) {
      save_session (fn);
    }

  }

  END_PROTECTED
}

void
MainWindow::cm_restore_session ()
{
  BEGIN_PROTECTED

  std::string fn = m_current_session;
  if (mp_session_fdia->get_open (fn)) {

    std::string df_list;
    int dirty_layouts = dirty_files (df_list);

    if (dirty_layouts == 0) {
      restore_session (fn);
    } else {

      QMessageBox mbox (this);
      mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + df_list + "\n\nPress 'Discard Changes' to close them anyhow and discard changes.").c_str ());
      mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
      mbox.setIcon (QMessageBox::Warning);
      QAbstractButton *discard_button = mbox.addButton (tl::translate ("Discard Changes").c_str (), QMessageBox::YesRole);
      mbox.addButton (QMessageBox::Cancel);

      mbox.exec ();

      if (mbox.clickedButton() == discard_button) {
        restore_session (fn);
      }

    }

  }

  END_PROTECTED
}

void
MainWindow::cm_save_bookmarks ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_bookmarks_fdia->get_save (fn)) {
      current_view ()->bookmarks ().save (fn);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to save the bookmarks from"));
  }

  END_PROTECTED
}

void
MainWindow::cm_load_bookmarks ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_bookmarks_fdia->get_open (fn)) {
      BookmarkList bookmarks;
      bookmarks.load (fn);
      current_view ()->bookmarks (bookmarks);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to load the bookmarks for"));
  }

  END_PROTECTED
}

void
MainWindow::cm_select_current_cell ()
{
  BEGIN_PROTECTED

  lay::LayoutView *view = current_view ();
  if (view && view->active_cellview_index () >= 0) {
    lay::LayoutView::cell_path_type path;
    int cvi = view->active_cellview_index ();
    view->current_cell_path (path);
    view->select_cell_fit (path, cvi);
  }

  END_PROTECTED
}

void
MainWindow::cm_open_current_cell ()
{
  BEGIN_PROTECTED

  if (current_view () && current_view ()->active_cellview_index () >= 0) {
    current_view ()->cm_open_current_cell ();
  }

  END_PROTECTED
}

void
MainWindow::cm_select_cell ()
{
  BEGIN_PROTECTED

  if (current_view () && current_view ()->active_cellview_index () >= 0) {

    CellSelectionForm form (0, current_view (), "cell_selection_form");

    if (form.exec () == QDialog::Accepted && 
        form.selected_cellview_index () >= 0) {
      current_view ()->select_cell (form.selected_cellview ().combined_unspecific_path (), form.selected_cellview_index ()); 
      current_view ()->set_current_cell_path (form.selected_cellview_index (), form.selected_cellview ().combined_unspecific_path ());
      current_view ()->zoom_fit ();
    }

  } else {
    throw tl::Exception (tl::translate ("No view open to select a cell for"));
  }

  END_PROTECTED
}

void
MainWindow::cm_screenshot ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_screenshot_fdia->get_save (fn)) {
      current_view ()->save_screenshot (fn);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to create a screenshot from"));
  }

  END_PROTECTED
}

void
MainWindow::cm_save_current_cell_as ()
{
  BEGIN_PROTECTED

  if (current_view ()) {

    int cv_index = current_view ()->active_cellview_index ();
    if (cv_index >= 0 && cv_index < int (current_view ()->cellviews ())) {

      LayoutView::cell_path_type path;
      current_view ()->current_cell_path (path);
      if (! path.empty ()) {

        const lay::CellView &cv = current_view ()->cellview (cv_index);

        QFileInfo file_info (cv->filename ().c_str ());
        std::string suffix = file_info.suffix ().ascii ();

        std::string fn = std::string (cv->layout ().cell_name (path.back ())) + "." + suffix;
        if (mp_layout_fdia->get_save (fn)) {

          db::SaveLayoutOptions options (current_view ()->cellview (cv_index)->save_options ());
          options.set_dbu (current_view ()->cellview (cv_index)->layout ().dbu ());
          bool gzip = false;
          options.set_format_from_filename (fn, gzip);

          bool save_permanently = true; // TODO: save this permanently?
          if (mp_layout_save_options->get_options (current_view (), cv_index, fn, gzip, options, this, save_permanently)) {
            options.select_cell (path.back ());
            cv->save_as (fn, gzip, options, false /*don't update*/);
            add_mru (fn);
          }

        }

      }

    }

  }

  END_PROTECTED
}

void
MainWindow::cm_save ()
{
  do_save (false);
}

void
MainWindow::cm_save_as ()
{
  do_save (true);
}

void
MainWindow::do_save (bool as)
{
  BEGIN_PROTECTED

  if (current_view ()) {

    int cv_index = -1;
    if (current_view ()->cellviews () > 1) {
      SelectCellViewForm form (0, current_view (), tl::translate ("Select Layout To Save"), true /*single selection*/);
      form.set_selection (current_view ()->active_cellview_index ());
      if (form.exec () == QDialog::Accepted) {
        cv_index = form.selected_cellview ();
      }
    } else if (current_view ()->cellviews () == 1) {
      cv_index = 0;
    }

    std::string fn;
    if (cv_index >= 0) {

      fn = current_view ()->cellview (cv_index)->filename ();
      if ((!as && !fn.empty ()) || mp_layout_fdia->get_save (fn)) {

        db::SaveLayoutOptions options (current_view ()->cellview (cv_index)->save_options ());
        options.set_dbu (current_view ()->cellview (cv_index)->layout ().dbu ());
        bool gzip = false;
        options.set_format_from_filename (fn, gzip);
        bool save_permanently = true; // TODO: save this permanently?
        if (mp_layout_save_options->get_options (current_view (), cv_index, fn, gzip, options, this, save_permanently)) {
          current_view ()->save_as ((unsigned int) cv_index, fn, gzip, options);
          add_mru (fn);
        }

      }

    }

  } else {
    throw tl::Exception (tl::translate ("No view open to save"));
  }

  END_PROTECTED
}

void
MainWindow::cm_setup ()
{
  mp_setup_form->show ();
  mp_setup_form->setup ();
}

void
MainWindow::view_selected (int index)
{
  if (index >= 0 && index < int (views ())) {

    //  Hint: setting the focus to the tab bar avoids problem with dangling keyboard focus.
    //  Sometimes, the focus was set to the hierarchy level spin buttons which caught Copy&Paste
    //  events in effect.
    mp_tab_bar->setFocus ();

    if (! m_disable_tab_selected) {
      select_view (index);
    }

  }
}

void
MainWindow::select_view (int index)
{
  bool dis = m_disable_tab_selected;
  m_disable_tab_selected = true; // prevent recursion

  tl_assert (index >= 0 && index < int (views ()));

  int view_index_org = current_view_index ();

  mp_tab_bar->setCurrentIndex (index);

  bool box_set = (m_synchronized_views && current_view () != 0);
  db::DBox box;
  if (box_set) {
    box = current_view ()->viewport ().box ();
  }

  m_current_view = index;

  mp_layer_toolbox->set_view (current_view ());

  if (current_view ()) {

    if (box_set) {
      current_view ()->zoom_box (box);
    }

    mp_view_stack->raiseWidget (m_current_view);
    mp_hp_stack->raiseWidget (m_current_view);
    mp_lp_stack->raiseWidget (m_current_view);
    mp_setup_form->setup ();

  }

  m_current_view_observers.signal_observers ();
  m_current_view_observers.signal_observers (int (view_index_org));

  m_disable_tab_selected = dis;
}

void 
MainWindow::cm_open_too ()
{
  open (2);
}

void 
MainWindow::cm_open_new_view ()
{
  open (1);
}

void 
MainWindow::cm_open ()
{
  open (0);
}

void 
MainWindow::cm_pull_in ()
{
  BEGIN_PROTECTED

  std::vector <std::string> names;
  lay::LayoutHandle::get_names (names);

  QStringList layouts;
  for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {
    layouts << n->c_str ();
  }

  if (layouts.size () == 0) {
    throw tl::Exception (tl::translate ("No layouts loaded"));
  }

  bool ok = false;
  QString item = QInputDialog::getItem (this, tl::translate ("Choose Layout").c_str (),
                                        tl::translate ("Choose an existing layout for being opened in the current view\nadditionally to the layouts already shown").c_str (),
                                        layouts, 0, false, &ok);
  if (ok) {
    lay::LayoutHandle *layout_handle = lay::LayoutHandle::find (std::string (item.ascii ()));
    if (layout_handle) {
      if (! current_view ()) {
        create_view ();
      }
      if (current_view ()) {
        current_view ()->add_layout (layout_handle, true);
      }
    }
  }

  END_PROTECTED
}

void
MainWindow::cm_reader_options ()
{
  bool save_permanently = true; // TODO: save permanently?
  mp_layout_load_options->get_options (m_reader_options, this, save_permanently);
}

void 
MainWindow::cm_new_panel ()
{
  create_view ();
}

void 
MainWindow::cm_new_layout ()
{
  BEGIN_PROTECTED

  lay::NewLayoutPropertiesDialog dialog (this);
  if (dialog.exec (m_new_cell_cell_name, m_default_dbu, m_new_cell_window_size, m_new_layout_current_panel)) {

    const lay::CellView &cellview = create_or_load_layout (0, m_new_layout_current_panel ? 2 : 1 /*= new view*/);

    cellview->layout ().dbu (m_default_dbu);
    db::cell_index_type new_ci = cellview->layout ().add_cell (m_new_cell_cell_name.empty () ? 0 : m_new_cell_cell_name.c_str ());

    current_view ()->select_cell (new_ci, current_view ()->active_cellview_index ());
    current_view ()->zoom_box (db::DBox (-0.5 * m_new_cell_window_size, -0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size));
    current_view ()->set_hier_levels (std::make_pair (0, 1));

    //  make the current settings default for new calls of the dialog (and maybe for other
    //  components)
    config_set (cfg_default_dbu, m_default_dbu);
    config_finalize ();

  }

  END_PROTECTED
}

void 
MainWindow::cm_adjust_origin ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr || curr->active_cellview_index () < 0) {
    throw tl::Exception (tl::translate ("No view open to create a new cell inside"));
  }

  curr->cm_align_cell_origin ();

  END_PROTECTED
}

void 
MainWindow::cm_new_cell ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr || curr->active_cellview_index () < 0) {
    throw tl::Exception (tl::translate ("No view open to create a new cell inside"));
  }

  NewCellPropertiesDialog cell_prop_dia (this); 
  if (cell_prop_dia.exec (& curr->cellview (curr->active_cellview_index ())->layout (), m_new_cell_cell_name, m_new_cell_window_size)) {

    db::cell_index_type new_ci = curr->new_cell (curr->active_cellview_index (), m_new_cell_cell_name.c_str ());
    curr->select_cell (new_ci, curr->active_cellview_index ());
    curr->zoom_box (db::DBox (-0.5 * m_new_cell_window_size, -0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size));

    if (curr->get_max_hier_levels () < 1 || curr->get_min_hier_levels () > 0) {
      curr->set_hier_levels (std::make_pair (0, 1));
    }

  }

  END_PROTECTED
}

void 
MainWindow::cm_sel_move ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_move ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_scale ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_scale ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_free_rot ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_free_rot ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_rot_ccw ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_rot_ccw ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_rot_cw ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_rot_cw ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_flip_y ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_flip_y ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_flip_x ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_flip_x ();

  END_PROTECTED
}

void 
MainWindow::cm_edit_layer ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to edit a layer specification"));
  }

  curr->cm_edit_layer ();

  END_PROTECTED
}

void 
MainWindow::cm_delete_layer ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to create a delete layer from"));
  }

  curr->cm_delete_layer ();

  END_PROTECTED
}

void 
MainWindow::cm_clear_layer ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to create a clear layer of"));
  }

  curr->cm_clear_layer ();

  END_PROTECTED
}

void 
MainWindow::cm_edit_boolean ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to perform boolean operations on"));
  }

  curr->cm_edit_boolean ();

  END_PROTECTED
}

void 
MainWindow::cm_edit_size ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to size a layer"));
  }

  curr->cm_edit_size ();

  END_PROTECTED
}

void 
MainWindow::cm_edit_merge ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to merge a layer"));
  }

  curr->cm_edit_merge ();

  END_PROTECTED
}

void 
MainWindow::cm_new_layer ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to create a new layer inside"));
  }

  curr->cm_new_layer ();

  END_PROTECTED
}

void 
MainWindow::cm_layout_props ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to show layout properties for"));
  }

  LayoutPropertiesForm lp_form (this, curr, "layout_props_form");
  lp_form.exec ();

  END_PROTECTED
}

void 
MainWindow::cm_clone ()
{
  BEGIN_PROTECTED
  clone_current_view ();
  END_PROTECTED
}

void
MainWindow::clone_current_view ()
{
  lay::LayoutView *view = 0;
  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to clone"));
  }

  //  create a new view
  view = new lay::LayoutView (current_view (), &m_manager, this, mp_view_stack);
  connect (view, SIGNAL (title_changed ()), this, SLOT (view_title_changed ()));
  mp_views.push_back (view);

  //  we must resize the widget here to set the geometry properly. 
  //  This is required to make zoom_fit work.
  view->setGeometry (0, 0, mp_view_stack->width (), mp_view_stack->height ());
  view->show ();

  //  set initial attributes
  view->set_hier_levels (curr->get_hier_levels ());

  //  select the current mode and select the enabled editables
  view->mode (m_mode);
  for (std::vector< std::pair<unsigned int, lay::Action> >::const_iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i) {
    view->enable_editable (i->first, m_plugin_enabled [i->first]);
  }

  lay::DisplayState state;
  current_view ()->save_view (state);
  view->goto_view (state);

  view->update_content ();

  m_current_view = mp_views.size () - 1;

  mp_layer_toolbox->set_view (current_view ());

  mp_view_stack->addWidget (view);
  mp_lp_stack->addWidget (view->layer_control_frame ());
  mp_hp_stack->addWidget (view->hierarchy_control_frame ());

  bool f = m_disable_tab_selected;
  m_disable_tab_selected = true;
  int index = mp_tab_bar->insertTab (-1, view->title ().c_str ());
  m_disable_tab_selected = f;

  m_new_view_observers.signal_observers ();
  select_view (index);

  update_dock_widget_state ();
}

void 
MainWindow::cm_close ()
{
  if (current_view ()) {

    std::vector <int> selected;

    if (current_view ()->cellviews () > 1) {

      SelectCellViewForm form (0, current_view (), tl::translate ("Select Layouts To Close"));
      form.set_selection (current_view ()->active_cellview_index ());
    
      if (form.exec () == QDialog::Accepted) {
        selected = form.selected_cellviews ();
      }
        
    } else if (current_view ()->cellviews () > 0) {
      selected.push_back (0);
    }

    if (selected.size () > 0) {

      int dirty_layouts = 0;
      std::string dirty_files;

      if (lay::Application::instance ()->is_editable ()) {
        for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
          if (current_view ()->cellview (*i)->is_dirty ()) {

            std::string name = current_view ()->cellview (*i)->name ();

            int count = 0;

            for (std::vector <lay::LayoutView *>::const_iterator v = mp_views.begin (); v != mp_views.end (); ++v) {
              for (unsigned int cvi = 0; cvi < (*v)->cellviews (); ++cvi) {
                if ((*v)->cellview (cvi)->name () == name) {
                  ++count;
                }
              }
            }

            for (std::vector <int>::const_iterator ii = selected.begin (); ii != selected.end (); ++ii) {
              if (current_view ()->cellview (*ii)->name () == name) {
                --count;
              }
            }

            //  only report layouts as dirty which will vanish if we would close all layouts 
            if (count <= 0) { 
              ++dirty_layouts;
              if (! dirty_files.empty ()) {
                dirty_files += "\n";
              }
              dirty_files += name;
            }

          }
        }
      }

      bool can_close = true;
      if (dirty_layouts != 0) {

        QMessageBox mbox (this);
        mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + dirty_files + tl::translate ("\n\nPress 'Close Without Saving' to close them anyway and discard changes\n")).c_str ()),
        mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
        mbox.setIcon (QMessageBox::Warning);
        QAbstractButton *can_close_button = mbox.addButton (tl::translate ("Close Without Saving").c_str (), QMessageBox::YesRole);
        mbox.addButton (QMessageBox::Cancel);

        mbox.exec ();

        can_close = (mbox.clickedButton() == can_close_button);

      }

      if (can_close) {

        BEGIN_PROTECTED

        //  Actually erase the selected cellviews
        if (current_view ()->cellviews () == selected.size ()) {
          //  all cellviews selected - simply close
          close_current_view ();
        } else {
          std::sort (selected.begin (), selected.end ());
          int offset = 0;
          for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
            current_view ()->erase_cellview ((unsigned int)(*i - offset));
            ++offset;
          }
        }

        END_PROTECTED

      }

    } else {
      close_current_view ();
    }

  }

}

void 
MainWindow::close_current_view ()
{
  if (current_view ()) {

    BEGIN_PROTECTED

    //  this suppresses view_selected events that would otherwise be created
    bool f = m_disable_tab_selected;
    m_disable_tab_selected = true;

    bool box_set = false;
    db::DBox box;
    if (m_synchronized_views) {
      box = current_view ()->viewport ().box ();
      box_set = true;
    }

    mp_tab_bar->removeTab (m_current_view);
    mp_view_stack->removeWidget (m_current_view);
    mp_lp_stack->removeWidget (m_current_view);
    mp_hp_stack->removeWidget (m_current_view);

    m_view_closed_observers.signal_observers (int (m_current_view));

    delete current_view ();
    mp_views.erase (mp_views.begin () + m_current_view, mp_views.begin () + m_current_view + 1);

    int view_index_org = current_view_index ();

    if (m_current_view >= mp_views.size () && m_current_view > 0) {
      --m_current_view;
    }

    mp_layer_toolbox->set_view (current_view ());

    if (current_view ()) {
      if (box_set) {
        current_view ()->zoom_box (box);
      }
      mp_tab_bar->setCurrentTab (m_current_view);
      mp_view_stack->raiseWidget (m_current_view);
      mp_lp_stack->raiseWidget (m_current_view);
      mp_hp_stack->raiseWidget (m_current_view);
    }

    m_disable_tab_selected = f;

    m_current_view_observers.signal_observers ();
    m_current_view_observers.signal_observers (int (view_index_org));

    update_dock_widget_state ();

    END_PROTECTED

  }
}

void
MainWindow::cm_reload ()
{    
  BEGIN_PROTECTED 

  if (current_view ()) {

    std::vector <int> selected;

    if (current_view ()->cellviews () > 1) {

      SelectCellViewForm form (0, current_view (), tl::translate ("Select Layouts To Reload"));
      form.select_all ();
    
      if (form.exec () == QDialog::Accepted) {
        selected = form.selected_cellviews ();
      }
        
    } else if (current_view ()->cellviews () > 0) {
      selected.push_back (0);
    }

    if (selected.size () > 0) {

      int dirty_layouts = 0;
      std::string dirty_files;

      if (lay::Application::instance ()->is_editable ()) {
        for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
          if (current_view ()->cellview (*i)->is_dirty ()) {
            ++dirty_layouts;
            if (! dirty_files.empty ()) {
              dirty_files += "\n";
            }
            dirty_files += current_view ()->cellview (*i)->name ();
          }
        }
      }

      bool can_reload = true;
      if (dirty_layouts != 0) {

        QMessageBox mbox (this);
        mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + dirty_files + "\n\nPress 'Reload Without Saving' to reload anyhow and discard changes.").c_str ());
        mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
        mbox.setIcon (QMessageBox::Warning);
        QAbstractButton *yes_button = mbox.addButton (tl::translate ("Reload Without Saving").c_str (), QMessageBox::YesRole);
        mbox.addButton (QMessageBox::Cancel);

        mbox.exec ();

        can_reload = (mbox.clickedButton() == yes_button);

      }

      if (can_reload) {

        //  Actually reload
        for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
          reload_layout (*i);
        }

      }

    }

  }

  END_PROTECTED
}

void 
MainWindow::set_technologies (const Technologies &technologies)
{
  m_technologies = technologies;
  config_set (cfg_technologies, m_technologies.to_string ());
}

void 
MainWindow::set_technology (const std::string &name, const Technology &technology)
{
  if (m_technologies.has_technology (name)) {
    *m_technologies.technology_by_name (name) = technology;
  } else {
    m_technologies.add (new Technology (technology));
  }

  config_set (cfg_technologies, m_technologies.to_string ());
}

void
MainWindow::add_mru (const std::string &fn_rel)
{
  std::vector <std::string> new_mru (m_mru);
  std::string fn (QFileInfo (fn_rel.c_str ()).absoluteFilePath ());

  for (std::vector<std::string>::iterator mru = new_mru.begin (); mru != new_mru.end (); ++mru) {
    if (*mru == fn) {
      new_mru.erase (mru);
      break;
    }
  }

  new_mru.push_back (fn);

  if (new_mru.size () > 10) {
    new_mru.erase (new_mru.begin ());
  }

  std::string config_str;
  for (std::vector<std::string>::const_iterator mru = new_mru.begin (); mru != new_mru.end (); ++mru) {
    if (! config_str.empty ()) {
      config_str += " ";
    }
    config_str += tl::to_quoted_string (*mru);
  }

  config_set (cfg_mru, config_str);
}

void 
MainWindow::file_menu_show ()
{
  if (mp_menu->is_valid ("file_menu.open_recent_menu")) {

    Action open_recent_action = mp_menu->action ("file_menu.open_recent_menu");

    if (m_mru.size () > 0 && m_edits_enabled) {

      open_recent_action.set_enabled (true);

      QMenu *open_recent_menu = open_recent_action.qaction ()->menu ();
      if (open_recent_menu) {

        open_recent_menu->clear ();

        for (std::vector<std::string>::iterator mru = m_mru.end (); mru != m_mru.begin (); ) {
          --mru;
          unsigned int i = std::distance (m_mru.begin (), mru);
          QAction *action = open_recent_menu->addAction (mru->c_str ());
          action->setObjectName (tl::sprintf ("open_recent_%d", i + 1).c_str ());
          gtf::action_connect (action, SIGNAL (triggered ()), this, SLOT (open_recent ()));
          action->setData (QVariant (int (i))); 
        }

      }

    } else {
      open_recent_action.set_enabled (false);
    }

  }
}

void
MainWindow::open_recent ()
{
  BEGIN_PROTECTED

  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);
  size_t n = size_t (action->data ().toInt ());
  if (n < m_mru.size ()) {

    OpenLayoutModeDialog open_mode_dialog (this);
    if (views () == 0 || open_mode_dialog.exec (m_open_mode)) {

      std::string fn (m_mru [n]);  //  create a copy since we change m_mru later

      if (m_open_mode == 0) {

        std::string df_list;
        int dirty_layouts = dirty_files (df_list);

        bool can_open = true;
        if (dirty_layouts != 0) {

          QMessageBox mbox (this);
          mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + df_list + "\n\nPress 'Close Without Saving' to open the layout and discard changes.").c_str ());
          mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
          mbox.setIcon (QMessageBox::Warning);
          QAbstractButton *yes_button = mbox.addButton (tl::translate ("Close Without Saving").c_str (), QMessageBox::YesRole);
          mbox.addButton (QMessageBox::Cancel);

          mbox.exec ();

          can_open = (mbox.clickedButton() == yes_button);

        }

        if (! can_open) {
          return;
        }

      }

      load_layout (fn, m_open_mode);

      add_mru (fn);  //  make it the latest

    }

  }

  END_PROTECTED
}

void 
MainWindow::open (int mode)
{
  BEGIN_PROTECTED 

  static std::string fn;
  if (mp_layout_fdia->get_open (fn)) {

    if (mode == 0) {

      std::string df_list;
      int dirty_layouts = dirty_files (df_list);

      bool can_open = true;
      if (dirty_layouts != 0) {

        QMessageBox mbox (this);
        mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + df_list + "\n\nPress 'Close Without Saving' to open the layout and discard changes.").c_str ());
        mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
        mbox.setIcon (QMessageBox::Warning);
        QAbstractButton *yes_button = mbox.addButton (tl::translate ("Close Without Saving").c_str (), QMessageBox::YesRole);
        mbox.addButton (QMessageBox::Cancel);

        mbox.exec ();

        can_open = (mbox.clickedButton() == yes_button);

      }

      if (! can_open) {
        return;
      }

    }

    load_layout (fn, mode);
    add_mru (fn);

  }

  END_PROTECTED
}

void
MainWindow::reload_layout (unsigned int cv_index)
{
  lay::LayoutView *view = current_view ();

  if (view && view->cellviews () > cv_index) {
    view->reload_layout (cv_index);
  }
}

const lay::CellView &
MainWindow::load_layout (const std::string &filename, int mode)
{
  return create_or_load_layout (&filename, mode);
}

const lay::CellView &
MainWindow::create_layout (int mode)
{
  return create_or_load_layout (0, mode);
}

int 
MainWindow::do_create_view ()
{
  //  create a new view
  lay::LayoutView *view = new lay::LayoutView (&m_manager, this, mp_view_stack);

  connect (view, SIGNAL (title_changed ()), this, SLOT (view_title_changed ()));

  mp_views.push_back (view);

  //  we must resize the widget here to set the geometry properly. 
  //  This is required to make zoom_fit work.
  view->setGeometry (0, 0, mp_view_stack->width (), mp_view_stack->height ());
  view->show ();

  //  set initial attributes
  view->set_synchronous (synchronous ());

  int tl = 0;
  config_get (cfg_initial_hier_depth, tl);
  view->set_hier_levels (std::make_pair (0, tl));

  //  select the current mode and select the enabled editables
  view->mode (m_mode);
  for (std::vector< std::pair<unsigned int, lay::Action> >::const_iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i) {
    view->enable_editable (i->first, m_plugin_enabled [i->first]);
  }

  return int (mp_views.size () - 1);
}

int
MainWindow::create_view ()
{
  //  create a new view
  int view_index = do_create_view ();

  //  add a new tab and make the new view the current one
  m_current_view = mp_views.size () - 1;

  mp_layer_toolbox->set_view (current_view ());

  mp_view_stack->addWidget (mp_views.back ());
  mp_lp_stack->addWidget (mp_views.back ()->layer_control_frame ());
  mp_hp_stack->addWidget (mp_views.back ()->hierarchy_control_frame ());

  bool f = m_disable_tab_selected;
  m_disable_tab_selected = true;
  int index = mp_tab_bar->insertTab (-1, current_view ()->title ().c_str ());
  m_disable_tab_selected = f;

  m_new_view_observers.signal_observers ();
  select_view (index);

  update_dock_widget_state ();

  return view_index;
}

const lay::CellView & 
MainWindow::create_or_load_layout (const std::string *filename, int mode)
{
  lay::LayoutView *vw = 0;

  int view_index_org = current_view_index ();

  if (! current_view ()) {
    mode = 1;
  }

  if (mode == 1) {
    //  create a new view
    vw = view (do_create_view ());
  } else {
    //  take the current view
    vw = current_view ();
    if (mode == 0) {
      //  reset the hierarchy depth in the "replace" case
      int tl = 0;
      config_get (cfg_initial_hier_depth, tl);
      vw->set_hier_levels (std::make_pair (0, tl));
    }
  }

  unsigned int cv_index = 0; 

  try {

    //  load or create the layout
    if (filename != 0) {
      cv_index = vw->load_layout (*filename, m_reader_options, mode == 2);
    } else {
      cv_index = vw->create_layout (mode == 2);
    }

    //  make the new view the current one
    if (mode == 1) {

      m_current_view = mp_views.size () - 1;

      mp_layer_toolbox->set_view (current_view ());

      mp_view_stack->addWidget (mp_views.back ());
      mp_lp_stack->addWidget (mp_views.back ()->layer_control_frame ());
      mp_hp_stack->addWidget (mp_views.back ()->hierarchy_control_frame ());

      bool f = m_disable_tab_selected;
      m_disable_tab_selected = true;
      int index = mp_tab_bar->insertTab (-1, current_view ()->title ().c_str ());
      m_disable_tab_selected = f;
      select_view (index);

    } else if (mode == 0) {
      mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());
    } else if (mode == 2) {
      mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());
    }

    m_new_view_observers.signal_observers ();
    m_current_view_observers.signal_observers ();
    m_current_view_observers.signal_observers (int (view_index_org));

    update_dock_widget_state ();

  } catch (...) {

    //  clean up in case of an error ..
    if (mode == 1) {
      delete mp_views.back ();
      mp_views.pop_back ();
    }

    throw;

  }

  return vw->cellview (cv_index);
}

void 
MainWindow::view_title_changed ()
{
  //  transfer the current view's text into the tab bar
  unsigned int tabs = mp_tab_bar->count ();
  for (unsigned int i = 0; i < tabs && i < views (); ++i) {
    mp_tab_bar->setTabText (i, view (i)->title ().c_str ());
  }
}

double
MainWindow::grid_micron () const
{
  return m_grid_micron;
}

void 
MainWindow::cm_inc_max_hier ()
{
  BEGIN_PROTECTED 

  int new_to = get_max_hier_levels () + 1;
  set_hier_levels (std::make_pair (get_min_hier_levels (), new_to));

  END_PROTECTED
}

void 
MainWindow::cm_dec_max_hier ()
{
  BEGIN_PROTECTED 

  int new_to = get_max_hier_levels () > 0 ? get_max_hier_levels () - 1 : 0;
  set_hier_levels (std::make_pair (std::min (get_min_hier_levels (), new_to), new_to));

  END_PROTECTED
}

void 
MainWindow::cm_max_hier ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->max_hier ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_max_hier_0 ()
{
  BEGIN_PROTECTED 
  set_hier_levels (std::make_pair (std::min (get_min_hier_levels (), 0), 0));
  END_PROTECTED
}

void 
MainWindow::cm_max_hier_1 ()
{
  BEGIN_PROTECTED 
  set_hier_levels (std::make_pair (std::min (get_min_hier_levels (), 0), 1));
  END_PROTECTED
}

void 
MainWindow::set_hier_levels (std::pair<int, int> l)
{
  if (l != get_hier_levels ()) {
    if (current_view ()) {
      current_view ()->set_hier_levels (l);
    }
  } 
}

std::pair<int, int> 
MainWindow::get_hier_levels () const
{
  if (current_view ()) {
    return current_view ()->get_hier_levels ();
  } else {
    int tl = 0;
    config_get (cfg_initial_hier_depth, tl);
    return std::make_pair (0, tl);
  }
}

void 
MainWindow::cm_last_display_state ()
{
  BEGIN_PROTECTED 

  if (has_last_display_state ()) {
    current_view ()->last_display_state ();
  }

  END_PROTECTED
}

bool 
MainWindow::has_last_display_state ()
{
  if (current_view ()) {
    return current_view ()->has_last_display_state ();
  } else {
    return false;
  }
}

void 
MainWindow::cm_next_display_state ()
{
  BEGIN_PROTECTED 

  if (has_next_display_state ()) {
    current_view ()->next_display_state ();
  }

  END_PROTECTED
}

bool 
MainWindow::has_next_display_state ()
{
  if (current_view ()) {
    return current_view ()->has_next_display_state ();
  } else {
    return false;
  }
}

void 
MainWindow::set_synchronous (bool sync_mode)
{
  m_synchronous = sync_mode;
  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->set_synchronous (sync_mode);
  }
}

void
MainWindow::current_pos (double x, double y, bool dbu_units) 
{
  mp_cpx_label->setText (dbu_units ? tl::db_to_string (x).c_str () : tl::micron_to_string (x).c_str ());
  mp_cpy_label->setText (dbu_units ? tl::db_to_string (y).c_str () : tl::micron_to_string (y).c_str ());
}

void
MainWindow::clear_current_pos ()
{
  mp_cpx_label->setText ("");
  mp_cpy_label->setText ("");
}

void 
MainWindow::set_progress_text (const std::string &text)
{
  if (mp_progress_widget) {
    mp_progress_widget->set_text (text);
  }
}

void 
MainWindow::set_progress_value (double v, const std::string &value)
{
  if (mp_progress_widget) {
    mp_progress_widget->set_value (v, value);
  }
}

bool 
MainWindow::eventFilter (QObject *watched, QEvent *event)
{
  if (event->type () == QEvent::MouseButtonDblClick ||
      event->type () == QEvent::MouseButtonPress ||
      event->type () == QEvent::MouseButtonRelease ||
      event->type () == QEvent::MouseMove ||
      event->type () == QEvent::KeyPress ||
      event->type () == QEvent::KeyRelease) {

    QObject *o = watched;
    while (o) {
      //  if the watched object is a child of the progress widget, pass the event on to this ..
      if (o == mp_progress_widget) {
        return QObject::eventFilter(watched, event);
      }
      o = o->parent ();
      //  stop at the level of a MessageBox
      if (dynamic_cast<QMessageBox *> (o)) {
        return QObject::eventFilter(watched, event);
      }
    }
    return true;

  } else {
    // standard event processing
    return QObject::eventFilter(watched, event);
  }
}

void 
MainWindow::show_progress_bar (bool show)
{
  if (m_work_in_progress != show) {

    m_work_in_progress = show;
    if (show) {
      //  to avoid recursions of any kind, disallow any user interaction except 
      //  cancelling the operation
      mp_app->installEventFilter (this);
      QApplication::setOverrideCursor (QCursor (Qt::WaitCursor));
    } else {
      mp_app->removeEventFilter (this);
      QApplication::restoreOverrideCursor ();
    }

    //  HINT: enabling the scheduler is accumulative - make sure that this method is only called when necessary
    tl::DeferredMethodScheduler::instance ()->enable (! show);

    mp_main_stack_widget->setCurrentIndex (show ? 1 : 0);

    if (show) {
      clear_current_pos ();
    }

  }
}

void
MainWindow::cm_console ()
{
  mp_console->show ();
}

void
MainWindow::cm_help_about ()
{
  HelpAboutDialog help_about_dialog (this);
  help_about_dialog.exec ();
}

void
MainWindow::cm_help_about_qt ()
{
  QApplication::aboutQt ();
}

lay::Action &
MainWindow::action_for_slot (const char *slot)
{
  std::map<std::string, lay::Action>::iterator a = m_actions_for_slot.find (std::string (slot));
  if (a != m_actions_for_slot.end ()) {
    return a->second;
  } else {
    Action a (new ActionHandle (this));
    gtf::action_connect (a.qaction (), SIGNAL (triggered ()), this, slot);
    return m_actions_for_slot.insert (std::make_pair (std::string (slot), a)).first->second;
  }
}

lay::Action *
MainWindow::create_config_action (const std::string &title, const std::string &cname, const std::string &cvalue)
{
  lay::ConfigureAction *ca = new lay::ConfigureAction(this, title, cname, cvalue);
  m_ca_collection.insert (ca);
  return ca;
}

lay::Action *
MainWindow::create_config_action (const std::string &cname, const std::string &cvalue)
{
  lay::ConfigureAction *ca = new lay::ConfigureAction(this, std::string (), cname, cvalue);
  m_ca_collection.insert (ca);
  return ca;
}

void 
MainWindow::register_config_action (const std::string &name, lay::ConfigureAction *action)
{
  std::map<std::string, std::vector<lay::ConfigureAction *> >::iterator ca = m_configuration_actions.insert (std::make_pair (name, std::vector<lay::ConfigureAction *> ())).first;
  for (std::vector<lay::ConfigureAction *>::iterator a = ca->second.begin (); a != ca->second.end (); ++a) {
    if (*a == action) {
      return; // already registered
    }
  }

  ca->second.push_back (action);
}
  
void 
MainWindow::unregister_config_action (const std::string &name, lay::ConfigureAction *action)
{
  std::map<std::string, std::vector<lay::ConfigureAction *> >::iterator ca = m_configuration_actions.find (name);
  if (ca != m_configuration_actions.end ()) {
    for (std::vector<lay::ConfigureAction *>::iterator a = ca->second.begin (); a != ca->second.end (); ++a) {
      if (*a == action) {
        ca->second.erase (a);
        return;
      }
    }
  }
}

void 
MainWindow::generic_menu ()
{
  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);
  generic_menu (std::string (action->data ().toString ().toAscii ()));
}

void 
MainWindow::generic_menu (const std::string &symbol)
{
  BEGIN_PROTECTED 

  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    if (cls->menu_activated (symbol)) {
      return;
    }
  }

  if (current_view ()) {
    current_view ()->menu_activated (symbol);
  } else {
    throw tl::Exception (tl::translate ("No view is active"));
  }

  END_PROTECTED
}

void
MainWindow::menu_changed ()
{
  //  delay actual rebuilding of the menu to collect multiple change events.
  m_do_update_menu_dm ();
}

void
MainWindow::do_update_menu ()
{
  mp_menu->build (menuBar (), mp_tool_bar);
}

void
MainWindow::cm_cell_delete ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_delete ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_rename ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_rename ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_flatten ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_flatten ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_select ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_select ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_hide ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_hide ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_show ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_show ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_show_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_show_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_select_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_select_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_new_tab ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_new_tab ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_rename_tab ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_rename_tab ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_remove_tab ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_remove_tab ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_hide_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_hide_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_hide ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_hide ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_show_only ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_show_only ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_show_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_show_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_show ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_show ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_rename ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_rename ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_delete ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_delete ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_insert ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_insert ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_group ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_group ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_ungroup ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_ungroup ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_source ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_source ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_name ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_name ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_ild ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_ild ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_idl ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_idl ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_ldi ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_ldi ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_dli ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_dli ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_by_index ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_by_index ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_by_datatype ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_by_datatype ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_by_layer ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_by_layer ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_flatten ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_flatten ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_expand_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_expand_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_add_missing ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_add_missing ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_remove_unused ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_remove_unused ();
  }

  END_PROTECTED
}

void 
MainWindow::dragEnterEvent(QDragEnterEvent *event)
{
  if (event->mimeData () && event->mimeData ()->urls ().size () >= 1 && event->mimeData ()->urls ().begin ()->scheme () == "file") {
    QString path = event->mimeData ()->urls ().begin ()->path ();
    event->acceptProposedAction ();
  }
}

void 
MainWindow::dropEvent(QDropEvent *event)
{
  BEGIN_PROTECTED

  if (event->mimeData ()) {

    QList<QUrl> urls = event->mimeData ()->urls ();
    for (QList<QUrl>::const_iterator url = urls.begin (); url != urls.end (); ++url) {

      if (url->scheme () != "file") {
        continue;
      }

      QString path = url->toLocalFile ();
      QFileInfo file_info (path);
      QString suffix = file_info.suffix ().toLower ();

      if (suffix == "rb" || suffix == "rba") {
        
        //  load ruby file
        lay::Application::instance ()->ruby_interpreter ().load_file (path.ascii ());

      } else if (suffix == "rbm") {

        //  install ruby module permanently
        if (QMessageBox::question (this, 
                                   tl::translate ("Install Ruby Module").c_str (), 
                                   tl::sprintf (tl::translate ("Install ruby module '%s' permanently?\n\nPress 'Yes' to install the ruby module in the installation directory permanently.").c_str (), file_info.fileName ().ascii ()).c_str (),
                                   QMessageBox::Yes | QMessageBox::No, 
                                   QMessageBox::No) == QMessageBox::Yes) {
          QFileInfo target (QDir (lay::Application::instance ()->inst_path ().c_str ()), file_info.fileName ());
          if (! target.exists () || QMessageBox::question (this,
                                                           tl::translate ("Overwrite Module").c_str (), 
                                                           tl::translate ("Overwrite existing module?").c_str (),
                                                           QMessageBox::Yes | QMessageBox::No, 
                                                           QMessageBox::No) == QMessageBox::Yes) {
            if (target.exists ()) {
              QFile target_file (target.filePath ());
              if (! target_file.remove ()) {
                throw tl::Exception (tl::translate ("Unable to remove file '%s'").c_str (), target.filePath ().ascii ());
              }
              if (! QFile (path).copy (target.filePath ())) {
                throw tl::Exception (tl::translate ("Unable to copy file to '%s'").c_str (), target.filePath ().ascii ());
              }
            }
          }
        }

      } else if (suffix == "lyp") {

        load_layer_properties (path.ascii (), false /*current view only*/, false /*don't add a default*/);

      } else if (suffix == "lys") {

        restore_session (path.ascii ());

      } else if (suffix == "lyb") {

        if (current_view ()) {
          BookmarkList bookmarks;
          bookmarks.load (path.ascii ());
          current_view ()->bookmarks (bookmarks);
        }

      } else {

        OpenLayoutModeDialog open_mode_dialog (this);
        if (views () == 0 || open_mode_dialog.exec (m_open_mode)) {

          if (m_open_mode == 0) {

            std::string df_list;
            int dirty_layouts = dirty_files (df_list);

            bool can_open = true;
            if (dirty_layouts != 0) {

              QMessageBox mbox (this);
              mbox.setText ((tl::translate ("The following layouts need saving:\n\n") + df_list + "\n\nPress 'Close Without Saving' to open the layout and discard changes.").c_str ());
              mbox.setWindowTitle (tl::translate ("Save Needed").c_str ());
              mbox.setIcon (QMessageBox::Warning);
              QAbstractButton *yes_button = mbox.addButton (tl::translate ("Close Without Saving").c_str (), QMessageBox::YesRole);
              mbox.addButton (QMessageBox::Cancel);

              mbox.exec ();

              can_open = (mbox.clickedButton() == yes_button);

            }

            if (! can_open) {
              return;
            }

          }

          load_layout (path.ascii (), m_open_mode);

          add_mru (path.ascii ());

        }

      }

    }

  }

  END_PROTECTED
}

// ------------------------------------------------------------
//  Implementation of the "tip" dialog

TipDialog::TipDialog (lay::MainWindow *parent, const std::string &text, const std::string &key)
  : QDialog (parent), mp_mw (parent), m_key (key)
{
  mp_ui = new Ui::TipDialog ();
  mp_ui->setupUi (this);

  mp_ui->dont_show_cbx->setChecked (false);
  mp_ui->tip_text->setText (text.c_str ());
}

bool
TipDialog::exec_dialog ()
{
  bool must_show = true;

  std::string th;
  mp_mw->config_get (cfg_tip_window_hidden, th);

  //  test if we need to show this window
  tl::Extractor ex (th.c_str ());
  while (! ex.at_end ()) {
    std::string k;
    ex.read_word (k, "_-.");
    if (k == m_key) {
      must_show = false;
      break;
    }
    ex.test (",");
  }

  if (must_show) {
    exec ();
    return true;
  } else {
    return false;
  }
}

void
TipDialog::accept ()
{
  //  register this dialog as dont-show on next startup
  if (mp_ui->dont_show_cbx->isChecked ()) {
    std::string th;
    mp_mw->config_get (cfg_tip_window_hidden, th);
    if (! th.empty ()) {
      th += ",";
    } 
    th += m_key;
    mp_mw->config_set (cfg_tip_window_hidden, th);
  }

  QDialog::accept ();
}

TipDialog::~TipDialog ()
{
  delete mp_ui;
  mp_ui = 0;
}

// ------------------------------------------------------------
//  Implementation of the "help about" dialog

HelpAboutDialog::HelpAboutDialog (QWidget *parent)
  : QDialog (parent)
{
  mp_ui = new Ui::HelpAboutDialog ();
  mp_ui->setupUi (this);

  std::string s = std::string (lay::Version::name ()) + " " + lay::Version::version () +
                  "\n\n" + lay::Version::about_text ();
  std::string t = tl::translate ("About ") + lay::Version::name ();

  setWindowTitle (t.c_str ());
  mp_ui->main->setText (s.c_str ());
}

HelpAboutDialog::~HelpAboutDialog ()
{
  delete mp_ui;
  mp_ui = 0;
}

} // namespace lay

