
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layRedrawThreadWorker.h"
#include "layRedrawThread.h"

namespace lay
{

//  time delay until the first snapshot is taken
const int first_snapshot_delay = 20;

// -------------------------------------------------------------
//  RedrawThreadWorker implementation 

RedrawThreadWorker::RedrawThreadWorker (RedrawThread *redraw_thread)
  : mp_redraw_thread (redraw_thread)
{
  mp_layout = 0;
  mp_cell_var_cache = 0;
  m_cache_hits = 0;
  m_cache_misses = 0;
  m_cv_index = -1;
  mp_canvas = 0;
  m_test_count = 0;
  m_from_level = 0;
  m_to_level = 0;
  m_from_level_default = 0;
  m_to_level_default = 0;
  m_drawing_context = false;
  m_nlayers = 0;
  m_box_text_transform = false;
  m_box_font = 0;
  m_min_size_for_label = 1;
  m_text_font = 0;
  m_text_visible = false;
  m_text_lazy_rendering = false;
  m_show_properties = false;
  m_apply_text_trans = false;
  m_default_text_size = 0.0;
  m_drop_small_cells = false;
  m_drop_small_cells_value = 0;
  m_drop_small_cells_cond = lay::LayoutView::DSC_Min;
  m_draw_array_border_instances = false;
  m_layer = 0;
  m_level = 0;
  m_custom_draw_boxes = false;
  m_inv_prop_sel = false;
  m_ci = 0;
  m_prop_id = 0;
  m_clock = tl::Clock::current ();

  for (unsigned int i = 0; i < sizeof (m_planes) / sizeof (m_planes[0]); ++i) {
    m_planes[i] = 0;
  }
}

RedrawThreadWorker::~RedrawThreadWorker ()
{
  for (unsigned int i = 0; i < sizeof (m_planes) / sizeof (m_planes[0]); ++i) {
    if (m_planes[i]) {
      delete m_planes[i];
      m_planes[i] = 0;
    }
  }
}

void 
RedrawThreadWorker::perform_task (tl::Task *task)
{
  RedrawThreadTask *redraw_thread_task = dynamic_cast <RedrawThreadTask *> (task);
  if (! redraw_thread_task) {
    return;
  }

  m_level = 0;
  m_from_level = m_from_level_default;
  m_to_level = m_to_level_default;

  int task_id = redraw_thread_task->id ();

  if (task_id >= 0) {

    //  draw a layer
    for (unsigned int i = 0; i < (unsigned int) planes_per_layer; ++i) {
      m_planes[i]->clear ();
    }

    //  HINT: the order in which the planes are delivered (the index stored in the first member of the pair below)
    //  must correspond with the order by which the ViewOp's are created inside LayoutView::set_view_ops
    m_buffers.clear ();
    for (unsigned int i = 0; i < (unsigned int) planes_per_layer / 3; ++i) {

      //  context level planes
      unsigned int i1 = task_id * (planes_per_layer / 3) + special_planes_before + i;
      mp_canvas->initialize_plane (m_planes[i], i1); 
      m_buffers.push_back (std::make_pair (i1, m_planes [i]));

      //  child level planes (if used)
      unsigned int i2 = (task_id + m_nlayers) * (planes_per_layer / 3) + special_planes_before + i;
      mp_canvas->initialize_plane (m_planes [i + planes_per_layer / 3], i2); 
      m_buffers.push_back (std::make_pair (i2, m_planes [i + planes_per_layer / 3]));

      //  current level planes
      unsigned int i3 = (task_id + m_nlayers * 2) * (planes_per_layer / 3) + special_planes_before + i;
      mp_canvas->initialize_plane (m_planes [i + 2 * (planes_per_layer / 3)], i3); 
      m_buffers.push_back (std::make_pair (i3, m_planes [i + 2 * (planes_per_layer / 3)]));

    }

    const RedrawLayerInfo &li = mp_redraw_thread->get_layer_info (task_id);

    if (li.cellview_index >= 0) {

      //  determine layout and cell associated with this layer ..
      const lay::CellView &cv = m_cellviews [li.cellview_index];
      if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {

        mp_layout = &cv->layout ();
        m_cv_index = li.cellview_index;
        m_ci = cv.cell_index ();

        int ctx_path_length = int (m_cellviews [m_cv_index].specific_path ().size ());

        if (li.hier_levels.has_from_level ()) {
          m_from_level = li.hier_levels.from_level (ctx_path_length, m_from_level);
        }
        if (li.hier_levels.has_to_level ()) {
          m_to_level = li.hier_levels.to_level (ctx_path_length, m_to_level);
        }

        mp_prop_sel = &li.prop_sel;
        m_inv_prop_sel = li.inverse_prop_sel;
        if (mp_prop_sel->empty () && m_inv_prop_sel) {
          //  no property selection
          mp_prop_sel = 0;
        }

        if (li.layer_index >= 0) {

          m_layer = li.layer_index;
       
          if (tl::verbosity () >= 30) {
            tl::info << tl::translate ("Drawing layer: ") << mp_layout->get_properties (m_layer).name;
          }
          tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing layer"));

          for (std::vector<db::CplxTrans>::const_iterator t = li.trans.begin (); t != li.trans.end (); ++t) {
            m_trans0 = m_trans = (m_vp_trans * mp_layout->dbu ()) * db::DCplxTrans (*t);
            m_mi_cache.clear ();
            iterate_variants (&RedrawThreadWorker::draw_layer);
          }

        } else if (li.cell_frame) {

          //  if no specific layer is assigned, draw cell boxes with the style given
          if (tl::verbosity () >= 30) {
            tl::info << tl::translate ("Drawing custom frames");
          }
          tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing frames"));

          for (std::set< std::pair<db::CplxTrans, int> >::const_iterator b = m_box_variants.begin (); b != m_box_variants.end (); ++b) {
            if (b->second == li.cellview_index) {
              m_trans0 = m_trans = (m_vp_trans * mp_layout->dbu ()) * db::DCplxTrans (b->first);
              m_custom_draw_boxes = true;
              iterate_variants (&RedrawThreadWorker::draw_boxes);
            }
          }

        }

        mp_prop_sel = 0;
        m_inv_prop_sel = false;

      }

    }

  } else if (task_id == draw_boxes_queue_entry) {

    //  draw the bounding boxes 
    if (tl::verbosity () >= 30) {
      tl::info << tl::translate ("Drawing frames");
    }
    tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing frames"));

    m_planes [0]->clear ();
    m_planes [1]->clear ();
    m_planes [2]->clear ();

    m_buffers.clear ();

    //  current level box planes
    mp_canvas->initialize_plane (m_planes[0], 0); 
    m_buffers.push_back (std::make_pair (0, m_planes[0]));

    //  child level box planes (if used)
    mp_canvas->initialize_plane (m_planes[1], 1); 
    m_buffers.push_back (std::make_pair (1, m_planes[1]));

    //  context level box planes
    mp_canvas->initialize_plane (m_planes[2], 2); 
    m_buffers.push_back (std::make_pair (2, m_planes[2]));

    for (std::set< std::pair<db::CplxTrans, int> >::const_iterator b = m_box_variants.begin (); b != m_box_variants.end (); ++b) {

      const lay::CellView &cv = m_cellviews [b->second];
      if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {

        mp_layout = &cv->layout ();
        m_ci = cv.cell_index ();
        m_cv_index = b->second;

        m_trans0 = m_trans = (m_vp_trans * mp_layout->dbu ()) * db::DCplxTrans (b->first);
        m_custom_draw_boxes = false;

        iterate_variants (&RedrawThreadWorker::draw_boxes);

      }

    }

  } else if (task_id == draw_custom_queue_entry) {

    //  draw the decorations
    if (tl::verbosity () >= 30) {
      tl::info << tl::translate ("Drawing decorations");
    }
    tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing decorations"));

    m_planes [0]->clear ();

    m_buffers.clear ();
    mp_canvas->initialize_plane (m_planes[0], m_nlayers * planes_per_layer + special_planes_before); 
    m_buffers.push_back (std::make_pair (m_nlayers * planes_per_layer + special_planes_before, m_planes [0]));

    unsigned int nd = 0;
    for (std::vector <lay::Drawing *>::iterator d = mp_drawings.begin (); d != mp_drawings.end (); ++d, ++nd) {

      //  temporarily create bitmap objects, paint on them,
      //  transfer them to the canvas and clear them again.
      //  This operation is not interrupted by any "test_snapshot".

      std::vector <lay::CanvasPlane *> tmp_planes;
      tmp_planes.reserve ((*d)->num_planes ());
      for (unsigned int i = 0; i < (*d)->num_planes (); ++i) {
        tmp_planes.push_back (mp_canvas->create_drawing_plane ());
        mp_canvas->initialize_plane (tmp_planes.back (), nd, i); 
      }

      //  currently, all cellviews are painted over each other ..
      for (unsigned int i = 0; i < m_cellviews.size (); ++i) {
        test_snapshot (); 
        const lay::CellView &cv = m_cellviews [i];
        if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {
          db::CplxTrans trans (m_vp_trans * cv->layout ().dbu ());
          (*d)->paint_cv_on_planes (cv, trans, tmp_planes);
        }
      }

      //  do the non-cv-related painting
      test_snapshot (); 
      (*d)->paint_on_planes (m_vp_trans, tmp_planes, *mp_renderer);

      for (unsigned int i = 0; i < (*d)->num_planes (); ++i) {
        mp_canvas->set_drawing_plane (nd, i, tmp_planes [i]); 
      }

      while (! tmp_planes.empty ()) {
        delete tmp_planes.back ();
        tmp_planes.pop_back ();
      }

    }

  }

  transfer ();
  m_buffers.clear ();

  mp_redraw_thread->task_finished (task_id);
}

void 
RedrawThreadWorker::finish ()
{
  //  release all cellview references here.
  m_cellviews.clear ();

  //  free the planes
  for (unsigned int i = 0; i < sizeof (m_planes) / sizeof (m_planes[0]); ++i) {
    if (m_planes[i]) {
      delete m_planes[i];
      m_planes[i] = 0;
    }
  }
}

void
RedrawThreadWorker::setup (LayoutView *view, RedrawThreadCanvas *canvas, int width, int height, const db::DCplxTrans &vp_trans)
{
  m_width = width;
  m_height = height;
  m_vp_trans = vp_trans;

  mp_canvas = canvas;

  mp_drawings.clear ();
  for (lay::Drawings::iterator d = view->drawings ()->begin (); d != view->drawings ()->end (); ++d) {
    mp_drawings.push_back (&*d);
  }

  //  allow a very short time to pass before we issue the
  //  first update event.
  m_clock = tl::Clock::current () - tl::Clock ((update_interval - first_snapshot_delay) * 0.001);

  //  initialize the drawing planes
  for (unsigned int i = 0; i < (unsigned int) planes_per_layer; ++i) {
    if (m_planes[i]) {
      delete m_planes[i];
    }
    m_planes[i] = mp_canvas->create_drawing_plane ();
  }

  //  give the renderer some initial capacity.
  mp_renderer.reset (mp_canvas->create_renderer ());
  mp_renderer->reserve_edges (100);

  //  copy everything that we need so there is no need to access 
  //  members of lay::LayoutView in the drawing thread.
  //  Note: copying the cellviews will create new references to the
  //  layout objects. These are not automatically freed when the 
  //  drawing ends but rather on "stop". The advantage of this is
  //  that, since "stop" is called from the main thread like "start",
  //  we don't challenge lay::CellView's MT compliance.
  m_from_level_default = view->get_hier_levels ().first;
  m_to_level_default = view->get_hier_levels ().second;
  m_min_size_for_label = view->min_inst_label_size ();
  m_box_text_transform = view->cell_box_text_transform ();
  m_box_font = view->cell_box_text_font ();
  m_text_font = view->text_font ();
  m_text_visible = view->text_visible ();
  m_text_lazy_rendering = view->text_lazy_rendering ();
  m_show_properties = view->show_properties_as_text ();
  m_apply_text_trans = view->apply_text_trans ();
  m_default_text_size = view->default_text_size ();
  m_drop_small_cells = view->drop_small_cells ();
  m_drop_small_cells_value = view->drop_small_cells_value ();
  m_drop_small_cells_cond = view->drop_small_cells_cond ();
  m_draw_array_border_instances = view->draw_array_border_instances ();
  m_abstract_mode_width = view->abstract_mode_width ();
  m_child_context_enabled = view->child_context_enabled ();
  m_test_count = 0;

  m_trans = db::CplxTrans ();
  mp_prop_sel = 0;
  m_inv_prop_sel = false;
  m_custom_draw_boxes = false;
  m_drawing_context = false;

  m_hidden_cells = view->hidden_cells ();

  m_cellviews.clear ();
  m_cellviews.reserve (view->cellviews ());
  for (unsigned int i = 0; i < view->cellviews (); ++i) {
    m_cellviews.push_back (view->cellview (i));
  }

  m_nlayers = mp_redraw_thread->num_layers (); 

  m_box_variants = view->cv_transform_variants ();

  m_ci = 0;
  m_prop_id = 0;
}

void
RedrawThreadWorker::transfer ()
{
  for (std::vector<std::pair<unsigned int, lay::CanvasPlane *> >::iterator b = m_buffers.begin (); b != m_buffers.end (); ++b) {
    mp_canvas->set_plane (b->first, b->second);
  }
}

void 
RedrawThreadWorker::test_snapshot ()
{
  checkpoint ();

  if (mp_redraw_thread->num_workers () > 0) {
    if (m_test_count == 0) {

      m_test_count = 100; // TODO: make variable?

      tl::Clock c = tl::Clock::current ();
      if ((c - m_clock).seconds () > update_interval * 0.001) {
        transfer ();
        mp_redraw_thread->wakeup_checked ();
        m_clock = c;
      }

    } else {
      --m_test_count;
    }
  }
}

void 
RedrawThreadWorker::draw_cell (const db::Box &box, const std::string &txt, db::properties_id_type prop_id)
{
  lay::Renderer &r = *mp_renderer;
  r.clear ();

  db::DBox dbox = m_trans * box;
  if (m_trans.is_ortho ()) {
    r.insert (dbox);
  } else {
    db::Polygon poly (box);
    for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      r.insert (m_trans * *e);
    }
  }

  if (! txt.empty (), dbox.width () > m_min_size_for_label && dbox.height () > m_min_size_for_label) {
    r.insert (dbox, txt, 
                db::Font (m_box_font), 
                db::HAlignCenter, 
                db::VAlignCenter, 
                //  TODO: apply "real" transformation?
                db::DFTrans (m_box_text_transform ? m_trans.fp_trans ().rot () : db::DFTrans::r0));
  }

  unsigned int plane_group = 2;
  if (m_drawing_context) {
    plane_group = 0;
  } else if (m_child_context_enabled && m_level > 0) {
    plane_group = 1;
  }

  //  TODO: there should not be two methods of drawing instances.
  //  In addition, instance drawing should be a method of the renderer
  if (m_custom_draw_boxes) {

    /* no vertices for instances
    r.render_vertices (*m_planes[3 + plane_group * (planes_per_layer / 3)], 1);
    */
    r.render_fill     (*m_planes[0 + plane_group * (planes_per_layer / 3)]);
    r.render_contour  (*m_planes[1 + plane_group * (planes_per_layer / 3)]);
    r.render_texts    (*m_planes[2 + plane_group * (planes_per_layer / 3)]);

    if (m_prop_id != 0 && m_show_properties) {
      r.draw_propstring (m_prop_id, &mp_layout->properties_repository (), dbox.p1 (), m_planes [2 + plane_group * (planes_per_layer / 3)], m_trans);
    }

  } else {

    r.render_texts   (*m_planes [plane_group]);
    r.render_contour (*m_planes [plane_group]);

    if (m_prop_id != 0 && m_show_properties) {
      r.draw_propstring (m_prop_id, &mp_layout->properties_repository (), dbox.p1 (), m_planes [plane_group], m_trans);
    }

  }

}

static bool
cells_in (const db::Layout *layout, const db::Cell &cell, 
          const std::set <db::cell_index_type> &selected,
          int levels,
          std::set <std::pair <int, db::cell_index_type> > &cache)
{
  if (selected.find (cell.cell_index ()) != selected.end ()) {
    return true;
  }
  if (levels > 0) {
    for (db::Cell::child_cell_iterator c = cell.begin_child_cells (); ! c.at_end (); ++c) {
      if (cache.find (std::make_pair (levels, *c)) == cache.end ()) {
        if (cells_in (layout, layout->cell (*c), selected, levels - 1, cache)) {
          return true;
        }
        cache.insert (std::make_pair (levels, *c));
      }
    }
  } 
  return false;
}

static bool 
need_draw_box (const db::Layout *layout, const db::Cell &cell, 
               int level, int to_level, 
               const std::vector <std::set <db::cell_index_type> > &hidden_cells, unsigned int cv_index) 
{
  if (level > to_level) {
    return false;
  }
  if (hidden_cells.size () > cv_index && ! hidden_cells [cv_index].empty ()) {
    std::set <std::pair <int, db::cell_index_type> > cache;
    if (cells_in (layout, cell, hidden_cells [cv_index], to_level - level, cache)) {
      return true;
    }
  }
  return int (cell.hierarchy_levels ()) + level >= to_level;
}

void 
RedrawThreadWorker::draw_boxes ()
{
  lay::Renderer &r = *mp_renderer;

  //  do not draw, if there is nothing to draw
  if (mp_layout->cells () <= m_ci || m_vp.empty ()) {
    return;
  }

  const db::Cell &cell = mp_layout->cell (m_ci);

  //  we will never come to a valid level ..
  if (! need_draw_box (mp_layout, cell, m_level, m_to_level, m_hidden_cells, m_cv_index)) {
    return;
  }
  if (cell_var_cached ()) {
    return;
  }

  //  For small bboxes, the cell outline can be reduced ..
  db::Box bbox = cell.bbox ();

  if (bbox.empty ()) {

    //  no shapes there and below ..

  } else if (m_drop_small_cells && drop_cell (cell)) {

    //  small cell dropped

  } else if (m_level == m_to_level || (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ())) {

    //  paint the box on this level
    draw_cell (bbox, mp_layout->cell_name (m_ci), m_prop_id);

  } else if (m_level < m_to_level) {

    db::DBox dbbox = m_trans * bbox;
    if (dbbox.width () < 1.5 && dbbox.height () < 1.5) {

      //  the cell is a very small box and we know there must be 
      //  some level at which to draw the boundary: just draw it 
      //  here and stop looking further down ..
      draw_cell (bbox, std::string (), 0);

    } else {

      db::box_convert <db::CellInst> bc (*mp_layout);
     
      db::CplxTrans trans = m_trans;
      db::cell_index_type ci = m_ci;
      db::properties_id_type prop_id = m_prop_id;

      db::Box vp = m_vp;

      //  create a set of boxes to look into
      db::Coord aw = db::coord_traits<db::Coord>::rounded (m_abstract_mode_width / mp_layout->dbu ());
      std::vector<db::Box> vv;
      if (m_level == 1 && m_abstract_mode_width > 0 && bbox.width () > aw * 2 && bbox.height () > aw * 2) {
        vv.reserve (4);
        vv.push_back (m_vp & db::Box (bbox.left (), bbox.bottom (), bbox.left () + aw, bbox.top ()));
        vv.push_back (m_vp & db::Box (bbox.right () - aw, bbox.bottom (), bbox.right (), bbox.top ()));
        vv.push_back (m_vp & db::Box (bbox.left () + aw, bbox.bottom (), bbox.right () - aw, bbox.bottom () + aw));
        vv.push_back (m_vp & db::Box (bbox.left () + aw, bbox.top () - aw, bbox.right () - aw, bbox.top ()));
      } else {
        vv.reserve (1);
        vv.push_back (m_vp);
      }

      ++m_level;

      //  dive down into the hierarchy ..
      for (std::vector<db::Box>::const_iterator v = vv.begin (); v != vv.end (); ++v) {

        if (! v->empty ()) {

          db::Cell::touching_iterator inst = cell.begin_touching (*v); 
          while (! inst.at_end ()) {

            const db::CellInstArray &cell_inst = inst->cell_inst ();
            db::properties_id_type cell_inst_prop = inst->prop_id ();

            db::Point a, b;
            unsigned long amax, bmax; 
            bool simplify = false;
            if (cell_inst.is_regular_array (a, b, amax, bmax)) {

              db::DBox inst_box;
              if (cell_inst.is_complex ()) {
                inst_box = trans * (cell_inst.complex_trans () * mp_layout->cell (cell_inst.object ().cell_index ()).bbox ());
              } else {
                inst_box = trans * mp_layout->cell (cell_inst.object ().cell_index ()).bbox ();
              }
              if (((a.x () == 0 && b.y () == 0) || (a.y () == 0 && b.x () == 0)) && 
                  inst_box.width () < 1.5 && inst_box.height () < 1.5 && 
                  (amax <= 1 || trans.ctrans (a.distance ()) < 1.5) && 
                  (bmax <= 1 || trans.ctrans (b.distance ()) < 1.5)) {
                simplify = true;
              }

            }

            if (simplify) {

              //  The array can be simplified if there are levels below to draw
              if (need_draw_box (mp_layout, mp_layout->cell (cell_inst.object ().cell_index ()), m_level, m_to_level, m_hidden_cells, m_cv_index)) {
                
                db::box_convert <db::CellInst> bc (*mp_layout);

                unsigned int plane_group = 2;
                if (m_drawing_context) {
                  plane_group = 0;
                } else if (m_child_context_enabled && m_level > 0) {
                  plane_group = 1;
                }

                r.draw (cell_inst.bbox (bc), trans, m_planes[plane_group], 0, 0, 0);

              }

            } else {

              //  The array (or single instance) must be iterated instance
              //  by instance
              for (db::CellInstArray::iterator p = cell_inst.begin_touching (*v, bc); ! p.at_end (); ++p) {
            
                test_snapshot (); 
                db::CplxTrans t (cell_inst.complex_trans (*p));
                m_trans = trans * t;
                m_vp = db::Box::from_double (t.inverted () * *v);
                m_ci = cell_inst.object ().cell_index ();
                m_prop_id = cell_inst_prop;
                draw_boxes ();

              }

            }

            ++inst;
          
          }

        }

      }

      --m_level;
      m_trans = trans;
      m_vp = vp;
      m_ci = ci;
      m_prop_id = prop_id;

    }

  }

}

/**
 *  @brief A helper function to determine if there are any shapes on the cell below to a certain hierarchy level
 *
 *  @return 0, if there are no shapes
 *  @return 1, if there are only contour-only shapes
 *  @return 2, if there are texts or properties to draw anywhere
 */
static int 
any_shapes (RedrawThreadWorker::MicroInstanceCache &cache, 
            const db::Layout &layout, db::cell_index_type cell_index, unsigned int layer, 
            unsigned long levels, const std::set<db::cell_index_type> *hidden_cells,
            const std::set<db::properties_id_type> *prop_sel, bool inv_prop_sel)
{
  //  if the cell is "hidden", it does not need to be drawn
  if (hidden_cells && hidden_cells->find (cell_index) != hidden_cells->end ()) {
    return 0;
  }

  //  the cache contains all cells that do not have any shapes
  RedrawThreadWorker::MicroInstanceCache::const_iterator c = cache.find (std::make_pair (cell_index, levels));
  if (c != cache.end ()) {
    return c->second;
  }

  int ret = 0;

  const db::Cell &cell = layout.cell (cell_index);

  if (! cell.shapes (layer).begin (db::ShapeIterator::Texts).at_end ()) {

    ret = 2;

  } else if (! cell.shapes (layer).begin (db::ShapeIterator::AllWithProperties, prop_sel, inv_prop_sel).at_end ()) {

    //  shapes with properties count like text objects ..
    ret = 2;

  } else if (! cell.shapes (layer).begin (db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Boxes, prop_sel, inv_prop_sel).at_end ()) {

    ret = 1;

  } else if (levels > 1) {
    
    for (db::Cell::child_cell_iterator cc = cell.begin_child_cells (); ! cc.at_end (); ++cc) {
      int r = any_shapes (cache, layout, *cc, layer, levels - 1, hidden_cells, prop_sel, inv_prop_sel);
      if (r > 0) {
        ret = r;
        break;
      }
    }

  }

  cache.insert (std::make_pair (std::make_pair (cell_index, levels), ret));

  return ret;

}

void 
RedrawThreadWorker::draw_layer ()
{
  lay::Renderer &r = *mp_renderer;

  unsigned int plane_group = 2;
  if (m_drawing_context) {
    plane_group = 0;
  } else if (m_child_context_enabled && m_level > 0) {
    plane_group = 1;
  }

  lay::CanvasPlane *fill   = m_planes[0 + plane_group * (planes_per_layer / 3)];
  lay::CanvasPlane *frame  = m_planes[1 + plane_group * (planes_per_layer / 3)];
  lay::CanvasPlane *text   = m_planes[2 + plane_group * (planes_per_layer / 3)];
  lay::CanvasPlane *vertex = m_planes[3 + plane_group * (planes_per_layer / 3)];

  test_snapshot (); 

  //  do not draw, if there is nothing to draw
  if (mp_layout->cells () <= m_ci || m_vp.empty ()) {
    return;
  }
  if (cell_var_cached ()) {
    return;
  }

  const db::Cell &cell = mp_layout->cell (m_ci);

  //  For small bboxes, the cell outline can be reduced ..
  db::Box bbox = cell.bbox (m_layer);

  if (m_drop_small_cells && drop_cell (cell)) {

    //  small cell dropped

  } else if (! bbox.empty ()) {

    bool hidden = (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ());
    bool need_to_dive = (m_level + 1 < m_to_level) && ! hidden;

    db::Box cell_bbox = cell.bbox ();

    //  draw this level
    if (m_level >= m_from_level && m_level < m_to_level && ! hidden) {

      //  configure renderer to text drawing ..
      r.draw_texts (m_text_visible);
      r.draw_properties (m_show_properties);
      r.default_text_size (db::Coord (m_default_text_size / mp_layout->dbu ()));
      r.set_font (db::Font (m_text_font));
      r.apply_text_trans (m_apply_text_trans);

      db::DBox dbbox_tot = m_trans * cell_bbox;
      bool text_simplified = m_text_lazy_rendering && (dbbox_tot.width () <= 8.0 || dbbox_tot.height () <= 8.0);
          
      const db::Shapes &shapes = cell.shapes (m_layer);

      db::DBox dbbox = m_trans * bbox;
      if ((dbbox.width () < 2.5 && dbbox.height () < 1.5) || 
          (dbbox.width () < 1.5 && dbbox.height () < 2.5)) {

        //  Here we first analyze the cell to find out 
        //  how far we need to go while maintaining the precision:

        int a;
        if (int (m_hidden_cells.size ()) <= m_cv_index || m_hidden_cells [m_cv_index].empty ()) {
          if (! mp_prop_sel && (text_simplified || ! m_text_visible)) {
            //  if there are no cells hiddden and there is no need to determine whether there are texts
            //  to draw (since these are switched off etc.), we do not need to employ the complicated
            //  "any_shapes" function. If a property selection is applied however, we need "any_shapes".
            a = 1;
          } else {
            a = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, 0, mp_prop_sel, m_inv_prop_sel);
          }
        } else {
          a = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, & m_hidden_cells [m_cv_index], mp_prop_sel, m_inv_prop_sel);
        }

        if (a > 0) {
          //  any shapes here: paint bbox for simplification
          r.clear ();
          r.insert (dbbox);
          r.render_vertices (*vertex, 1);
          r.render_contour (*frame);
        }

        //  texts not visible or no texts below - do not dive further into hierarchy
        if (text_simplified || ! m_text_visible || a < 2) {
          need_to_dive = false;
        }

      } else {

        //  create a set of boxes to look into
        db::Coord aw = db::coord_traits<db::Coord>::rounded (m_abstract_mode_width / mp_layout->dbu ());
        std::vector<db::Box> vv;
        if (m_level == 1 && m_abstract_mode_width > 0 && cell_bbox.width () > aw * 2 && cell_bbox.height () > aw * 2) {
          vv.reserve (4);
          vv.push_back (m_vp & db::Box (cell_bbox.left (), cell_bbox.bottom (), cell_bbox.left () + aw, cell_bbox.top ()));
          vv.push_back (m_vp & db::Box (cell_bbox.right () - aw, cell_bbox.bottom (), cell_bbox.right (), cell_bbox.top ()));
          vv.push_back (m_vp & db::Box (cell_bbox.left () + aw, cell_bbox.bottom (), cell_bbox.right () - aw, cell_bbox.bottom () + aw));
          vv.push_back (m_vp & db::Box (cell_bbox.left () + aw, cell_bbox.top () - aw, cell_bbox.right () - aw, cell_bbox.top ()));
        } else {
          vv.reserve (1);
          vv.push_back (m_vp);
        }

        //  dive down into the hierarchy ..
        for (std::vector<db::Box>::const_iterator v = vv.begin (); v != vv.end (); ++v) {

          if (! v->empty ()) {

            db::ShapeIterator shape (shapes.begin_touching (*v, db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths, mp_prop_sel, m_inv_prop_sel));
            while (! shape.at_end ()) {
              test_snapshot (); 
              r.draw (*shape, &mp_layout->properties_repository (), m_trans, fill, frame, vertex, text);
              ++shape;
            }

          }

        }

      }

      //  In lazy text rendering mode, all texts are only rendered if the cell is "reasonable large". 
      //  Empirically, a reasonable minimum dimension of 8x8 pixels was determined.
      //  otherwise just a few texts are rendered.
      if (m_text_visible) {

        //  this is the number of texts to draw in lazy text rendering mode
        size_t ntexts = 2;
        if (! text_simplified) {
          ntexts = std::numeric_limits <size_t>::max ();
        }

        //  create a set of boxes to look into
        db::Coord aw = db::coord_traits<db::Coord>::rounded (m_abstract_mode_width / mp_layout->dbu ());
        std::vector<db::Box> vv;
        if (m_level == 1 && m_abstract_mode_width > 0 && cell_bbox.width () > aw * 2 && cell_bbox.height () > aw * 2) {
          vv.reserve (4);
          vv.push_back (m_vp & db::Box (cell_bbox.left (), cell_bbox.bottom (), cell_bbox.left () + aw, cell_bbox.top ()));
          vv.push_back (m_vp & db::Box (cell_bbox.right () - aw, cell_bbox.bottom (), cell_bbox.right (), cell_bbox.top ()));
          vv.push_back (m_vp & db::Box (cell_bbox.left () + aw, cell_bbox.bottom (), cell_bbox.right () - aw, cell_bbox.bottom () + aw));
          vv.push_back (m_vp & db::Box (cell_bbox.left () + aw, cell_bbox.top () - aw, cell_bbox.right () - aw, cell_bbox.top ()));
        } else {
          vv.reserve (1);
          vv.push_back (m_vp);
        }

        //  dive down into the hierarchy ..
        for (std::vector<db::Box>::const_iterator v = vv.begin (); v != vv.end (); ++v) {

          if (! v->empty ()) {

            db::ShapeIterator shape = shapes.begin_touching (*v, db::ShapeIterator::Texts, mp_prop_sel, m_inv_prop_sel);
            while (! shape.at_end () && ntexts > 0) {

              test_snapshot (); 

              r.draw (*shape, &mp_layout->properties_repository (), m_trans, fill, frame, vertex, text);
              ++shape;

              --ntexts;

            }

          }

        }

      }

    }

    db::box_convert <db::CellInst> bc (*mp_layout, m_layer);

    //  dive down into the hierarchy ..
    if (need_to_dive) {

      db::CplxTrans trans = m_trans;
      db::Box vp = m_vp;
      db::cell_index_type ci = m_ci;

      //  create a set of boxes to look into
      db::Coord aw = db::coord_traits<db::Coord>::rounded (m_abstract_mode_width / mp_layout->dbu ());
      std::vector<db::Box> vv;
      if (m_level == 1 && m_abstract_mode_width > 0 && cell_bbox.width () > aw * 2 && cell_bbox.height () > aw * 2) {
        vv.reserve (4);
        vv.push_back (m_vp & db::Box (cell_bbox.left (), cell_bbox.bottom (), cell_bbox.left () + aw, cell_bbox.top ()));
        vv.push_back (m_vp & db::Box (cell_bbox.right () - aw, cell_bbox.bottom (), cell_bbox.right (), cell_bbox.top ()));
        vv.push_back (m_vp & db::Box (cell_bbox.left () + aw, cell_bbox.bottom (), cell_bbox.right () - aw, cell_bbox.bottom () + aw));
        vv.push_back (m_vp & db::Box (cell_bbox.left () + aw, cell_bbox.top () - aw, cell_bbox.right () - aw, cell_bbox.top ()));
      } else {
        vv.reserve (1);
        vv.push_back (m_vp);
      }

      ++m_level;

      //  dive down into the hierarchy ..
      for (std::vector<db::Box>::const_iterator v = vv.begin (); v != vv.end (); ++v) {

        if (! v->empty ()) {

          db::Cell::touching_iterator inst = cell.begin_touching (*v); 
          while (! inst.at_end ()) {

            const db::CellInstArray &cell_inst = inst->cell_inst ();
            ++inst;

            m_ci = cell_inst.object ().cell_index ();
            bool hidden = (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ());

            db::Box cell_box = mp_layout->cell (m_ci).bbox (m_layer);
            if (! cell_box.empty () && ! hidden) {

              db::Point a, b;
              unsigned long amax = 0, bmax = 0; 
              bool simplify = false;
              bool anything = true;

              if (cell_inst.is_regular_array (a, b, amax, bmax)) {

                db::DBox inst_box;
                if (cell_inst.is_complex ()) {
                  inst_box = trans * (cell_inst.complex_trans () * cell_box);
                } else {
                  inst_box = trans * cell_box;
                }

                if (((a.x () == 0 && b.y () == 0) || (a.y () == 0 && b.x () == 0)) && 
                    inst_box.width () < 1.5 && inst_box.height () < 1.5 && 
                    (amax <= 1 || trans.ctrans (a.distance ()) < 1.5) && 
                    (bmax <= 1 || trans.ctrans (b.distance ()) < 1.5)) {

                  int as;
                  if (int (m_hidden_cells.size ()) <= m_cv_index) {
                    as = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, 0, mp_prop_sel, m_inv_prop_sel);
                  } else {
                    as = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, & m_hidden_cells [m_cv_index], mp_prop_sel, m_inv_prop_sel);
                  }
                  if (as == 0) {
                    anything = false;
                  } else if ((m_text_visible && as == 1) || (! m_text_visible && as >= 1)) {
                    //  only for simple shapes this array instance can be simplified ...
                    simplify = true;
                  }

                }

              }

              if (simplify) {

                //  The array can be simplified ..
                r.draw (cell_inst.bbox (bc), trans, 0, 0, 0, 0);

                r.render_contour (*frame);
                r.render_fill (*frame);
                r.render_contour (*vertex);
                r.render_fill (*vertex);

              } else if (anything) {

                for (db::CellInstArray::iterator p = cell_inst.begin_touching (*v, bc); ! p.at_end (); ++p) {

                  if (! m_draw_array_border_instances || 
                      p.index_a () <= 0 || (unsigned long)p.index_a () == amax - 1 || p.index_b () <= 0 || (unsigned long)p.index_b () == bmax - 1) {

                    db::CplxTrans t (cell_inst.complex_trans (*p));
                    m_trans = trans * t;
                    m_vp = db::Box::from_double (t.inverted () * *v);
                    draw_layer ();

                  } 

                }

              }

            }

          }

        }

      }

      --m_level;
      m_trans = trans;
      m_vp = vp;
      m_ci = ci;

    }

  }

}

bool
RedrawThreadWorker::drop_cell (const db::Cell &cell)
{
  db::DBox bbox = m_trans * cell.bbox ();

  double value = 0;
  if (m_drop_small_cells_cond == lay::LayoutView::DSC_Min) {
    value = std::min (bbox.width (), bbox.height ());
  } else if (m_drop_small_cells_cond == lay::LayoutView::DSC_Max) {
    value = std::max (bbox.width (), bbox.height ());
  } else {
    value = bbox.width () + bbox.height ();
  }

  return (value < double (m_drop_small_cells_value));
}

bool 
RedrawThreadWorker::cell_var_cached () 
{
  if (mp_cell_var_cache) {
    //  Use the native transformation (just including cell instantiation components) to enable
    //  fuzzy comparison of floating-point coordinates: this requires a well-defined unit system to
    //  allow the definition of an uncertainty value.
    db::CplxTrans db_trans ((m_vp_trans * mp_layout->dbu ()).inverted () * db::DCplxTrans (m_trans));
    if (mp_cell_var_cache->find (std::make_pair (db_trans, m_ci)) != mp_cell_var_cache->end ()) {
      ++m_cache_hits;
      return true;
    }
    ++m_cache_misses;
    mp_cell_var_cache->insert (std::make_pair (db_trans, m_ci));
  }
  return false;
}

void 
RedrawThreadWorker::iterate_variants (void (RedrawThreadWorker::*what) ())
{
  m_level = 0;

  //  save current state
  db::CplxTrans trans = m_trans;
  int from_level = m_from_level;
  int to_level = m_to_level;
  
  //  if a context path is given, we adjust the levels such that the target (not the context 
  //  cell) is drawn and the context cell is visible through addressing negative levels. The
  //  iterate_variants_rec methods takes care of converting the negative hierarchy levels into
  //  traversals along the context path bottom up.
  size_t ctx_path_length = m_cellviews [m_cv_index].specific_path ().size ();
  if (ctx_path_length > 0) {
    m_from_level -= ctx_path_length;
    m_to_level -= ctx_path_length;
    m_trans = m_trans * m_cellviews [m_cv_index].context_trans ();
  }

  if (m_from_level_default < 0 || ctx_path_length > 0) {

    //  if we start from above the hierarchy, we need to establish a 
    //  cell variant cache to at least avoid painting the current cell
    //  multiple times.
    std::set <std::pair <db::CplxTrans, db::cell_index_type>, lay::CellVariantCacheCompare> cell_var_cache;
    mp_cell_var_cache = &cell_var_cache;

    //  Use the cell variant cache to exclude the basic instance from the 
    //  drawing in the first pass.
    cell_var_cache.insert (std::make_pair (m_cellviews [m_cv_index].context_trans (), m_ci));

    m_cache_hits = m_cache_misses = 0;

    //  draw the context for the current instance
    m_drawing_context = true;
    iterate_variants_rec (what, true);

    cell_var_cache.clear ();

    //  draw the current instance without context (using a minimum of from_level=0 for this)
    mp_cell_var_cache = 0;
    m_drawing_context = false;
    int fl = m_from_level;
    if (m_from_level < 0) {
      m_from_level = 0;
    }
    iterate_variants_rec (what, false);
    m_from_level = fl;

    if (tl::verbosity () >= 40) {
      tl::info << tl::translate ("Cell variant cache hits/misses: ") << m_cache_hits << "/" << m_cache_misses;
    }

  } else {
    mp_cell_var_cache = 0;
    m_drawing_context = false;
    iterate_variants_rec (what, false);
  }

  //  restore state
  m_trans = trans;
  m_from_level = from_level;
  m_to_level = to_level;
}

void 
RedrawThreadWorker::iterate_variants_rec (void (RedrawThreadWorker::*what) (), bool spread)
{
  db::Cell::parent_inst_iterator p = mp_layout->cell (m_ci).begin_parent_insts ();
  int context_path_length = int (m_cellviews [m_cv_index].specific_path ().size ());

  if ((spread || m_level > m_from_level) && m_level + context_path_length > 0) {

    //  pull an specific instance from the instance stack and move this one up
    --m_level;

    const db::InstElement &ie = m_cellviews [m_cv_index].specific_path ().end () [m_level];

    db::cell_index_type ci = m_ci;
    db::CplxTrans trans = m_trans;
    db::Box vp = m_vp;

    if (m_level + context_path_length > 0) {
      m_ci = m_cellviews [m_cv_index].specific_path ().end () [m_level - 1].inst_ptr.cell_index ();
    } else {
      m_ci = m_cellviews [m_cv_index].ctx_cell_index ();
    }

    db::CplxTrans t (ie.complex_trans ());
    m_trans = trans * t.inverted ();
    m_vp = db::Box::from_double (t * vp);
    iterate_variants_rec (what, spread);

    m_trans = trans;
    m_vp = vp;
    m_ci = ci;

    ++m_level;

  } else if (m_level > (spread ? (m_from_level_default - context_path_length) : m_from_level) && ! p.at_end ()) {

    //  one level up ..
    --m_level;

    db::CplxTrans trans = m_trans;
    db::Box vp = m_vp;
    db::cell_index_type ci = m_ci;

    while (! p.at_end ()) {

      db::Cell::cell_inst_array_type pi = (*p).inst ();

      m_ci = pi.object ().cell_index ();

      for (db::Cell::cell_inst_array_type::iterator pp = pi.begin (); ! pp.at_end (); ++pp) {

        db::CplxTrans t (pi.complex_trans (*pp));
        m_trans = trans * t;
        m_vp = db::Box::from_double (t.inverted () * vp);
        iterate_variants_rec (what, spread);

      }

      ++p;

    }

    m_trans = trans;
    m_vp = vp;
    m_ci = ci;

    ++m_level;

  } else {

    db::Coord lim = std::numeric_limits<db::Coord>::max ();
    db::DBox world (m_trans * db::Box (db::Point (-lim, -lim), db::Point (lim, lim)));
    m_vp = db::Box::from_double (m_trans.inverted () * (world & db::DBox (db::DPoint (0.0, 0.0), db::DPoint (m_width, m_height))));
    m_vp &= mp_layout->cell (m_ci).bbox (); // this avoids problems with accessing designs through very large viewports

    (this->*what) ();

  }

}

}

