
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "layRenderer.h"
#include "layCanvasPlane.h"
#include "laySnap.h"

#include "tlAlgorithm.h"

#include <algorithm>

namespace lay 
{

// ----------------------------------------------------------------------------------------------
//  Renderer implementation

Renderer::Renderer (unsigned int width, unsigned int height, double resolution)
  : m_draw_texts (true),
    m_draw_properties (false),
    m_default_text_size (16),
    m_apply_text_trans (true),
    m_precise (false),
    m_font (db::DefaultFont),
    m_width (width), m_height (height),
    m_resolution (resolution)
{
  // .. nothing else ..
}

void 
Renderer::draw (const db::Shape &shape, const db::PropertiesRepository *prep, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  if (shape.is_text ()) {
    
    db::Point p = shape.text_trans ().disp ();
    db::DPoint dp = trans * p;

    if ((vertices || frame) &&
        dp.x () < m_width - 0.5 && dp.x () > -0.5 && 
        dp.y () < m_height - 0.5 && dp.y () > -0.5) {

      clear ();

      db::Point pp (db::Point::from_double (dp));

      //  generate a dot (cross, if marked) at the text's position
      if (vertices) {
        vertices->pixel (pp.x (), pp.y ());
      }
      if (frame) {
        frame->pixel (pp.x (), pp.y ());
      }

    }

    if (m_draw_texts && text) {

      db::DFTrans fp (db::DFTrans::r0);
      db::DCoord h = trans.ctrans (m_default_text_size);
      db::Font font = shape.text_font () == db::NoFont ? m_font : shape.text_font ();

      if (m_apply_text_trans && font != db::NoFont && font != db::DefaultFont) {
        fp = db::DFTrans (trans.fp_trans () * shape.text_trans ());
        h = trans.ctrans (shape.text_size () > 0 ? shape.text_size () : m_default_text_size);
      }

      db::HAlign halign = shape.text_halign ();
      db::VAlign valign = shape.text_valign ();

      double fy = 0.0;
      if (valign == db::VAlignBottom || valign == db::NoVAlign) {
        fy = 1.0;
      } else if (valign == db::VAlignTop) {
        fy = -1.0;
      }

      double fx = 0.0;
      if (halign == db::HAlignLeft || halign == db::NoHAlign) {
        fx = 1.0;
      } else if (halign == db::HAlignRight) {
        fx = -1.0;
      }

      db::DPoint tp1 (db::DPoint (fx * 2.0, fy * 2.0 + (fy - 1) * 0.5 * h));
      db::DPoint tp2 (db::DPoint (fx * 2.0, fy * 2.0 + (fy + 1) * 0.5 * h));

      clear ();

      insert (db::DBox (fp (tp1) + dp, fp (tp2) + dp), shape.text_string (), font, halign, valign, fp);

      render_texts (*text);

    }

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      draw_propstring (shape.prop_id (), prep, dp, text, trans);
    }

  } else if (shape.is_polygon ()) {

    //  simplify to a rectangle if possible
    db::Box b (shape.bbox ());
    if (simplify_to_box (b, trans)) {
      draw (b, trans, fill, frame, vertices, text);
    } else {

      clear ();
      db::Shape::polygon_edge_iterator e = shape.begin_edge (); 
      for ( ; ! e.at_end (); ++e) {
        insert (trans * *e);
      }

      if (vertices) {
        render_vertices (*vertices, 1);
      }
      if (fill) {
        render_fill (*fill);
      }
      if (frame) {
        render_contour (*frame);
      }

    }

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      db::Shape::polygon_edge_iterator e = shape.begin_edge (); 
      db::DPoint pref = trans ((*e).p1 ());
      draw_propstring (shape.prop_id (), prep, pref, text, trans);
    }

  } else if (shape.is_edge ()) {

    draw (shape.edge (), trans, fill, frame, vertices, text);

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      draw_propstring (shape.prop_id (), prep, trans (shape.edge ().p1 ()), text, trans);
    }

  } else if (shape.is_path ()) {

    //  simplify to a rectangle if possible
    db::Box b (shape.bbox ());
    if (simplify_to_box (b, trans)) {
      draw (b, trans, fill, frame, vertices, text);
    } else {

      //  instantiate the path and draw
      db::Path path;
      shape.path (path);
      draw (path, trans, fill, frame, vertices, text);

    }

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      db::DPoint pref = trans (*shape.begin_point ());
      draw_propstring (shape.prop_id (), prep, pref, text, trans);
    }

  } else if (shape.is_box ()) {

    draw (shape.box (), trans, fill, frame, vertices, text);

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      draw_propstring (shape.prop_id (), prep, trans (shape.box ().p1 ()), text, trans);
    }

  }
}

void 
Renderer::draw (const db::Polygon &poly, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::Box b (poly.box ());
  if (simplify_to_box (b, trans)) {
    insert (trans * b);
  } else {
    for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      insert (trans * *e);
    }
  }

  if (vertices) {
    render_vertices (*vertices, 1);
  }
  if (fill) {
    render_fill (*fill);
  }
  if (frame) {
    render_contour (*frame);
  }

}

void 
Renderer::draw (const db::DPolygon &poly, const db::DCplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::DBox b (poly.box ());
  if (simplify_to_box (b, trans)) {
    insert (trans * b);
  } else {
    for (db::DPolygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      insert (trans * *e);
    }
  }

  if (vertices) {
    render_vertices (*vertices, 1);
  }
  if (fill) {
    render_fill (*fill);
  }
  if (frame) {
    render_contour (*frame);
  }
}

void 
Renderer::draw (const db::ShortBox &box, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  if (! box.empty ()) {

    clear ();
    insert (db::Box (box), trans);

    if (vertices) {
      render_vertices (*vertices, 2);
    }
    if (fill) {
      render_fill (*fill);
    }
    if (frame) {
      render_contour (*frame);
    }

  }
}

void 
Renderer::draw (const db::Box &box, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  if (! box.empty ()) {
  
    clear ();
    insert (box, trans);

    if (vertices) {
      render_vertices (*vertices, 2);
    }
    if (fill) {
      render_fill (*fill);
    }
    if (frame) {
      render_contour (*frame);
    }

  }
}

void 
Renderer::draw (const db::DBox &box, const db::DCplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  if (! box.empty ()) {
  
    clear ();
    insert (box, trans);

    if (vertices) {
      render_vertices (*vertices, 2);
    }
    if (fill) {
      render_fill (*fill);
    }
    if (frame) {
      render_contour (*frame);
    }

  }
}

void 
Renderer::draw (const db::Path &path, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::Box b (path.box ());
  if (simplify_to_box (b, trans)) {
    draw (b, trans, fill, frame, vertices, text);
    return;
  } 

  //  generate the hull and produce the edges from this only if the path is considerably wide
  //  otherwise just render the spine
  
  double w = trans.ctrans (path.width ());

  bool thin = (w < 0.5);
  bool quite_thin = (w < 3.0);

  //  render the border
  if (! thin) {

    db::DPath::pointlist_type pts;
    path.transformed (trans).hull (pts);

    db::DPath::pointlist_type::const_iterator p = pts.begin ();
    if (p != pts.end ()) {

      db::DPath::pointlist_type::const_iterator pp = p;
      ++pp;

      while (pp != pts.end ()) {
        insert (db::DEdge (*p, *pp));
        p = pp;
        ++pp;
      }

      insert (db::DEdge (*p, pts [0]));
    }

    if (fill) {
      render_fill (*fill);
    }
    if (frame) {
      render_contour (*frame);
    }

  }

  //  render the spine edges
  if (! quite_thin || thin) {

    clear ();

    db::Path::iterator q = path.begin ();
    if (q != path.end ()) {

      db::Path::iterator qq = q;
      ++qq;

      bool first = true;
      while (qq != path.end ()) {

        db::Edge seg (*q, *qq);
        q = qq;
        ++qq;

        //  If the path is simplified, the spine is drawn instead of the contour.
        //  We must apply the begin/end extensions then.
        
        if (thin && qq == path.end ()) {
          if (path.extensions ().second != 0 && (seg.dx () != 0 || seg.dy () != 0)) {
            db::DPoint ed (seg.p2 () - seg.p1 ());
            ed *= 1.0 / ed.double_distance ();
            seg = db::Edge (seg.p1 (), seg.p2 () + db::Point::from_double (ed * double (path.extensions ().second)));
          }
        } 

        if (first) {
          first = false;
          if (thin && path.extensions ().first != 0 && (seg.dx () != 0 || seg.dy () != 0)) {
            db::DPoint ed (seg.p2 () - seg.p1 ());
            ed *= 1.0 / ed.double_distance ();
            seg = db::Edge (seg.p1 () - db::Point::from_double (ed * double (path.extensions ().first)), seg.p2 ());
          }
        }
          
        insert (trans * seg);

      }

    }

    if (vertices) {
      render_vertices (*vertices, 0);
    }
    if (frame) {
      render_contour (*frame);
    }

  }
}

void 
Renderer::draw (const db::DPath &path, const db::DCplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::DBox b (path.box ());
  if (simplify_to_box (b, trans)) {
    draw (b, trans, fill, frame, vertices, text);
    return;
  } 

  //  generate the hull and produce the edges from this 

  db::DPath::pointlist_type pts;
  path.hull (pts);

  db::DPath::pointlist_type::const_iterator p = pts.begin ();
  if (p != pts.end ()) {

    db::DPath::pointlist_type::const_iterator pp = p;
    ++pp;

    while (pp != pts.end ()) {
      insert (trans * db::DEdge (*p, *pp));
      p = pp;
      ++pp;
    }

    insert (trans * db::DEdge (*p, pts [0]));
  }

  if (fill) {
    render_fill (*fill);
  }
  if (frame) {
    render_contour (*frame);
  }

  //  render the spine edges

  clear ();

  db::DPath::iterator q = path.begin ();
  if (q != path.end ()) {

    db::DPath::iterator qq = q;
    ++qq;

    while (qq != path.end ()) {
      insert (trans * db::DEdge (*q, *qq));
      q = qq;
      ++qq;
    }

  }

  if (vertices) {
    render_vertices (*vertices, 0);
  }
  if (frame) {
    render_contour (*frame);
  }
}

void 
Renderer::draw (const db::Text &txt, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  db::Point p = txt.trans ().disp ();
  db::DPoint dp = trans * p;

  if ((vertices || frame) &&
      dp.x () < m_width - 0.5 && dp.x () > -0.5 && 
      dp.y () < m_height - 0.5 && dp.y () > -0.5) {

    clear ();

    db::Point pp (db::Point::from_double (dp));

    //  generate a dot (cross, if marked) at the text's position
    if (vertices) {
      vertices->pixel (pp.x (), pp.y ());
    }
    if (frame) {
      frame->pixel (pp.x (), pp.y ());
    }

  }

  if (m_draw_texts && text) {

    db::DFTrans fp (db::DFTrans::r0);
    db::DCoord h = trans.ctrans (m_default_text_size);
    db::Font font = txt.font () == db::NoFont ? m_font : txt.font ();

    if (m_apply_text_trans && font != db::NoFont && font != db::DefaultFont) {
      fp = db::DFTrans (trans.fp_trans () * txt.trans ());
      h = trans.ctrans (txt.size () > 0 ? txt.size () : m_default_text_size);
    }

    double fy = 0.0;
    if (txt.valign () == db::VAlignBottom || txt.valign () == db::NoVAlign) {
      fy = 1.0;
    } else if (txt.valign () == db::VAlignTop) {
      fy = -1.0;
    }

    double fx = 0.0;
    if (txt.halign () == db::HAlignLeft || txt.halign () == db::NoHAlign) {
      fx = 1.0;
    } else if (txt.halign () == db::HAlignRight) {
      fx = -1.0;
    }

    //  impose a 2 pixel offset to separate the text from the origin
    db::DPoint tp1 (db::DPoint (fx * 2.0, fy * 2.0 + (fy - 1) * 0.5 * h));
    db::DPoint tp2 (db::DPoint (fx * 2.0, fy * 2.0 + (fy + 1) * 0.5 * h));

    clear ();

    insert (db::DBox (fp (tp1) + dp, fp (tp2) + dp), txt.string (), font, txt.halign (), txt.valign (), fp);
    
    render_texts (*text);

  }

}

void 
Renderer::draw (const db::DText &txt, const db::DCplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  db::DPoint p = txt.trans ().disp ();
  db::DPoint dp = trans * p;

  if ((vertices || frame) &&
      dp.x () < m_width - 0.5 && dp.x () > -0.5 && 
      dp.y () < m_height - 0.5 && dp.y () > -0.5) {

    clear ();

    db::Point pp (db::Point::from_double (dp));

    //  generate a dot (cross, if marked) at the text's position
    if (vertices) {
      vertices->pixel (pp.x (), pp.y ());
    }
    if (frame) {
      frame->pixel (pp.x (), pp.y ());
    }

  }

  if (m_draw_texts && text) {

    db::DFTrans fp (db::DFTrans::r0);
    db::DCoord h = trans.ctrans (m_default_text_size_dbl);
    db::Font font = txt.font () == db::NoFont ? m_font : txt.font ();

    if (m_apply_text_trans && font != db::NoFont && font != db::DefaultFont) {
      fp = trans.fp_trans () * db::DFTrans (txt.trans ());
      h = trans.ctrans (txt.size () > 0 ? txt.size () : m_default_text_size_dbl);
    }

    double fy = 0.0;
    if (txt.valign () == db::VAlignBottom || txt.valign () == db::NoVAlign) {
      fy = 1.0;
    } else if (txt.valign () == db::VAlignTop) {
      fy = -1.0;
    }

    double fx = 0.0;
    if (txt.halign () == db::HAlignLeft || txt.halign () == db::NoHAlign) {
      fx = 1.0;
    } else if (txt.halign () == db::HAlignRight) {
      fx = -1.0;
    }

    db::DPoint tp1 (db::DPoint (fx * 2.0, fy * 2.0 + (fy - 1) * 0.5 * h));
    db::DPoint tp2 (db::DPoint (fx * 2.0, fy * 2.0 + (fy + 1) * 0.5 * h));

    clear ();

    insert (db::DBox (fp (tp1) + dp, fp (tp2) + dp), txt.string (), font, txt.halign (), txt.valign (), fp);
    
    render_texts (*text);

  }

}

void 
Renderer::draw (const db::Edge &edge, const db::CplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  clear ();
  insert (trans * edge);

  if (vertices) {
    render_vertices (*vertices, 0);
  }
  if (frame) {
    render_contour (*frame);
  }
}

void 
Renderer::draw (const db::DEdge &edge, const db::DCplxTrans &trans,
                lay::CanvasPlane *fill, lay::CanvasPlane *frame, lay::CanvasPlane *vertices, lay::CanvasPlane *text)
{
  clear ();
  insert (trans * edge);

  if (vertices) {
    render_vertices (*vertices, 0);
  }
  if (frame) {
    render_contour (*frame);
  }
}

void 
Renderer::draw_propstring (db::properties_id_type id, 
                           const db::PropertiesRepository *prep, const db::DPoint &pref, 
                           lay::CanvasPlane *text, const db::CplxTrans &trans)
{
  db::DPoint tp1 (pref + db::DPoint (2.0, -2.0));
  db::DPoint tp2 (pref + db::DPoint (2.0, -2.0 - trans.ctrans (m_default_text_size)));

  clear ();

  std::string ptext;

  const char *sep = "";
  const db::PropertiesRepository::properties_set &props = prep->properties (id);
  for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
    ptext += sep;
    sep = "\n";
    ptext += prep->prop_name (p->first).to_string ();
    ptext += ": ";
    ptext += p->second.to_string ();
  }

  insert (db::DBox (tp1, tp2), ptext, m_font,
          db::HAlignLeft, db::VAlignTop, 
          db::DFTrans (db::DFTrans::r0));
  
  render_texts (*text);
}

template <class Box, class Trans>
bool 
Renderer::simplify_to_box (Box &b, const Trans &trans)
{
  const double small_size_threshold = 1.0;

  bool ortho = trans.is_ortho ();
  if (! m_precise && ((ortho && trans.ctrans (std::min (b.width (), b.height ())) < small_size_threshold) || 
                      (! ortho && trans.ctrans (std::max (b.width (), b.height ())) < small_size_threshold))) {

    if (trans.ctrans (b.width ()) < small_size_threshold) {
      typename Box::coord_type c = b.center ().x ();
      b.set_left (c);
      b.set_right (c);
    } 
    if (trans.ctrans (b.height ()) < small_size_threshold) {
      typename Box::coord_type c = b.center ().y ();
      b.set_top (c);
      b.set_bottom (c);
    } 

    return true;

  } else {
    return false;
  }
}

}


