
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "laySvgRendering.h"
#include "layCanvasPlane.h"
#include "layViewObject.h"
#include "layRedrawThread.h"
#include "layRedrawThreadCanvas.h"
#include "layRenderer.h"
#include "layLayoutView.h"
#include "tlAssert.h"

#include <fstream>
#include <iostream>

namespace lay {


// -----------------------------------------------------------------------------------
//  CanvasPlane specialization for SVG output

class SvgCanvasPlane 
  : public CanvasPlane
{
public:
  /**
   *  @brief Constructor
   */
  SvgCanvasPlane ()
  {
    // .. nothing yet ..
  }
  
  /**
   *  @brief Constructor
   */
  SvgCanvasPlane (const lay::ViewOp &view_op)
  {
    set_view_op (view_op);
  }
  
  /**
   *  @brief Constructor
   */
  SvgCanvasPlane (const std::vector<lay::ViewOp> &view_ops)
  {
    set_view_ops (view_ops);
  }
  
  /**
   *  @brief Destructor
   */
  virtual ~SvgCanvasPlane ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Clear the plane
   */
  virtual void clear () 
  {
    m_fill_paths.clear ();
    m_contour_paths.clear ();
  }

  /**
   *  @brief Assignment of content
   */
  void copy_from (const SvgCanvasPlane &plane)
  {
    if (this != &plane) {
      m_fill_paths = plane.m_fill_paths;
      m_contour_paths = plane.m_contour_paths;
    }
  }
  
  /**
   *  @brief Fill method 
   *
   *  This method is supposed to draw a small rectangle of 1 pixel
   *
   *  @param x,y The coordinate where to put the pixel (lower left corner)
   */
  virtual void pixel (unsigned int y, unsigned int x)
  {
    
    // @@@

  }

  /**
   *  @brief Render a set of edges in fill mode
   */
  void render_fill (const std::vector<db::DEdge> &edges)
  {
    m_fill_paths.push_back (path_d (edges));
  }

  /**
   *  @brief Render a set of edges in contour mode
   */
  void render_contour (const std::vector<db::DEdge> &edges)
  {
    m_contour_paths.push_back (path_d (edges));
  }

  /**
   *  @brief Render a set of edges as vertices
   */
  void render_vertices (const std::vector<db::DEdge> &edges, int mode)
  {

    // @@@

  }

  /**
   *  @brief Set the view op to the plane
   */
  void set_view_op (const lay::ViewOp &view_op) 
  {
    m_view_ops.clear ();
    unsigned int transparent = 0xffffff & (view_op.andmask () & ~view_op.ormask ());
    if (transparent != 0xffffff) {
      m_view_ops.push_back (view_op);
    }
  }

  /**
   *  @brief Set the view ops to the plane
   */
  void set_view_ops (const std::vector<lay::ViewOp> &view_ops) 
  {
    m_view_ops.reserve (view_ops.size ());
    m_view_ops.clear ();
    
    for (std::vector<lay::ViewOp>::const_iterator v = view_ops.begin (); v != view_ops.end (); ++v) {
      unsigned int transparent = 0xffffff & (v->andmask () & ~v->ormask ());
      if (transparent != 0xffffff) {
        m_view_ops.push_back (*v);
      }
    }
  }

  void produce (std::ostream &output, lay::LayoutView *mp_view)
  {
    for (std::vector <lay::ViewOp>::const_iterator v = m_view_ops.begin (); v != m_view_ops.end (); ++v) {

      size_t id = (size_t)this; // @@@
      unsigned int color = 0xffffff & v->ormask ();

      if (! m_fill_paths.empty ()) {

        output << "<defs>" << std::endl;
        output << "  <pattern patternUnits=\"userSpaceOnUse\" id=\"Fill" << id << "\" x=\"0\" y=\"0\" width=\"32\" height=\"32\" style=\"fill:#" << tl::sprintf("%06x", color) << "\">" << std::endl;

        const uint32_t *pattern = mp_view->dither_pattern ().pattern (v->dither_index ());
        for (unsigned int y = 0; y < 32; ++y) {
          for (unsigned int x = 0; x < 32; ++x) {
            if ((pattern[y] & (1 << x)) != 0) {
              output << "    <rect x=\"" << x << "\" y=\"" << y << "\" width=\"1\" height=\"1\"/>" << std::endl;
            }
          }
        }

        output << "  </pattern>" << std::endl;
        output << "</defs>" << std::endl;
        output << "<g style=\"fill:url(#Fill" << id << ");fill-rule:nonzero;stroke:none\">" << std::endl;

        for (std::list<std::string>::const_iterator s = m_fill_paths.begin (); s != m_fill_paths.end (); ++s) {
          output << "<path d=\"" << *s << "\"/>" << std::endl;
        }

        output << "</g>" << std::endl;

      }

      if (! m_contour_paths.empty ()) {

        unsigned int color = 0xffffff & v->ormask ();
        // @@@ unsigned int transparent = 0xffffff & (view_op.andmask () & ~view_op.ormask ());

        output << "<g style=\"fill:none;stroke:#" <<  tl::sprintf("%06x", color) << ";stroke-width:" << tl::to_string (v->width ()) << "\">" << std::endl;
        for (std::list<std::string>::const_iterator s = m_contour_paths.begin (); s != m_contour_paths.end (); ++s) {
          output << "<path d=\"" << *s << "\"/>" << std::endl;
        }
        output << "</g>" << std::endl;
      }

    }

    m_fill_paths.clear ();
    m_contour_paths.clear ();
  }

public:
// private: @@@
  std::vector<lay::ViewOp> m_view_ops;
  std::list<std::string> m_fill_paths;
  std::list<std::string> m_contour_paths;

  std::string path_d (const std::vector<db::DEdge> &edges) 
  {
    db::DPoint last_pt;
    bool last_pt_valid = false;
    bool any_before = false;

    std::string d;
    d.reserve (1000);

    for (std::vector<db::DEdge>::const_iterator e = edges.begin (); e != edges.end (); ++e) {

      if (any_before) {
        d += " ";
      }
      any_before = true;

      if (! last_pt_valid || fabs (e->p1 ().x () - last_pt.x ()) > 1e-3 || fabs (e->p1 ().y () - last_pt.y ()) > 1e-3) {
        d += tl::sprintf ("M%.1f,%.1f ", e->p1 ().x (), e->p1 ().y ());
      }
      d += tl::sprintf ("L%.1f,%.1f", e->p2 ().x (), e->p2 ().y ());

      last_pt = e->p2 ();
      last_pt_valid = true;

    }

    return d;
  }
};

// -----------------------------------------------------------------------------------
//  Renderer specialization for SVG output

class SvgRenderer 
  : public Renderer
{
public:
  /**
   *  @brief The ctor
   */
  SvgRenderer (unsigned int width, unsigned int height, double resolution)
    : Renderer (width, height, resolution)
  { }

  /**
   *  @brief The destructor
   */
  virtual ~SvgRenderer () { }

  /**
   *  @brief Reserve space for n edges 
   *
   *  It is strongly recommended to use this method to
   *  reserve space in advance before any insert operation for edges or boxes.
   */
  virtual void reserve_edges (size_t n) 
  {
    m_edges.reserve (n);
  }

  /**
   *  @brief Reserve space for n texts 
   *
   *  It is strongly recommended to use this method to
   *  reserve space in advance before any insert operation for texts.
   */
  virtual void reserve_texts (size_t n) 
  {
    // does nothing.
  }

  /** 
   *  @brief Clear the content
   */
  virtual void clear () 
  {
    m_edges.clear ();
    // m_texts.clear (); @@@
  }

  /** 
   *  @brief Insert a box with a transformation (because it may not be orthogonal)
   */
  virtual void insert (const db::Box &b, const db::CplxTrans &t) 
  {
    db::DPoint p1 = t.trans (db::Point (b.left (), b.bottom ()));
    db::DPoint p2 = t.trans (db::Point (b.left (), b.top ()));
    db::DPoint p3 = t.trans (db::Point (b.right (), b.top ()));
    db::DPoint p4 = t.trans (db::Point (b.right (), b.bottom ()));

    insert (db::DEdge (p1, p2));
    insert (db::DEdge (p2, p3));
    insert (db::DEdge (p3, p4));
    insert (db::DEdge (p4, p1));
  }

  /** 
   *  @brief Insert a box with a transformation (because it may not be orthogonal)
   */
  virtual void insert (const db::DBox &b, const db::DCplxTrans &t) 
  {
    db::DPoint p1 = t.trans (db::DPoint (b.left (), b.bottom ()));
    db::DPoint p2 = t.trans (db::DPoint (b.left (), b.top ()));
    db::DPoint p3 = t.trans (db::DPoint (b.right (), b.top ()));
    db::DPoint p4 = t.trans (db::DPoint (b.right (), b.bottom ()));

    insert (db::DEdge (p1, p2));
    insert (db::DEdge (p2, p3));
    insert (db::DEdge (p3, p4));
    insert (db::DEdge (p4, p1));
  }

  /** 
   *  @brief Insert a box 
   */
  virtual void insert (const db::DBox &b) 
  {
    db::DPoint p1 = db::DPoint (b.left (), b.bottom ());
    db::DPoint p2 = db::DPoint (b.left (), b.top ());
    db::DPoint p3 = db::DPoint (b.right (), b.top ());
    db::DPoint p4 = db::DPoint (b.right (), b.bottom ());

    insert (db::DEdge (p1, p2));
    insert (db::DEdge (p2, p3));
    insert (db::DEdge (p3, p4));
    insert (db::DEdge (p4, p1));
  }

  /** 
   *  @brief Insert an edge 
   *
   *  The edges must be a db::DEdge type.
   */
  virtual void insert (const db::DEdge &e) 
  {
    m_edges.push_back (e);
  }

  /**
   *  @brief Insert a text
   *
   *  @box The box in which to inscribe the text (or the point at which to align if the width and height is zero)
   *  @text The text to render
   *  @font The font to use
   *  @halign The horizontal alignment
   *  @valign The vertical alignment
   *  @trans The orientation of the text
   */
  virtual void insert (const db::DBox &box, const std::string &text, db::Font font, db::HAlign halign, db::VAlign valign, db::DFTrans trans)
  {
    // @@@
  }

  /**
   *  @brief Render the interior of the object to the bitmap 
   */
  virtual void render_fill (lay::CanvasPlane &plane) 
  {
    lay::SvgCanvasPlane *svg_plane = dynamic_cast<lay::SvgCanvasPlane *> (&plane);
    if (svg_plane && ! m_edges.empty ()) {
      svg_plane->render_fill (m_edges);
    }
  }

  /**
   *  @brief Render the contour of the object to the bitmap 
   */
  virtual void render_contour (lay::CanvasPlane &plane) 
  {
    lay::SvgCanvasPlane *svg_plane = dynamic_cast<lay::SvgCanvasPlane *> (&plane);
    if (svg_plane && ! m_edges.empty ()) {
      svg_plane->render_contour (m_edges);
    }
  }

  /**
   *  @brief Render the vertices of the object to the bitmap 
   */
  virtual void render_vertices (lay::CanvasPlane &plane, int mode)
  {
    lay::SvgCanvasPlane *svg_plane = dynamic_cast<lay::SvgCanvasPlane *> (&plane);
    if (svg_plane && ! m_edges.empty ()) {
      svg_plane->render_vertices (m_edges, mode);
    }
  }

  /**
   *  @brief Render the texts of the object to the bitmap
   */
  virtual void render_texts (lay::CanvasPlane &bitmap) 
  {
    // does nothing.
  }

private:
  std::vector<db::DEdge> m_edges;
};

// -----------------------------------------------------------------------------------
//  RedrawThreadCanvas specialization for SVG output

class SvgRedrawThreadCanvas
  : public RedrawThreadCanvas
{
public:
  /**
   *  @brief Constructor
   */
  SvgRedrawThreadCanvas (lay::LayoutView *view, unsigned int width, unsigned int height, double resolution, std::ostream &output)
    : mp_view (view), m_width (width), m_height (height), m_resolution (resolution), m_output (output)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Destructor
   */
  virtual ~SvgRedrawThreadCanvas ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Prepare the given number of planes
   *
   *  This method is called from RedrawThread::start (), not from the
   *  redraw thread.
   */
  virtual void prepare (unsigned int nlayers, unsigned int width, unsigned int height, double resolution, const lay::Drawings *drawings)
  {
    RedrawThreadCanvas::prepare(nlayers, width, height, resolution, drawings);

    lock ();

    //  delete the buffers and create new ones 
    m_planes.clear ();
    m_planes.reserve (nlayers);

    const std::vector<lay::ViewOp> &view_ops = mp_view->get_view_ops ();
    // @@@ why not? tl_assert (nlayers == view_ops.size ());

    for (unsigned int i = 0; i < nlayers; ++i) {
      if (i < view_ops.size ()) {
        m_planes.push_back (lay::SvgCanvasPlane (view_ops[i]));
      } else {
        m_planes.push_back (lay::SvgCanvasPlane ());
      }
    }

    for (lay::Drawings::const_iterator d = drawings->begin (); d != drawings->end (); ++d) {
      m_drawing_planes.push_back (std::vector <lay::SvgCanvasPlane> ());
      m_drawing_planes.back ().reserve (d->num_planes ());
      for (unsigned int i = 0; i < d->num_planes (); ++i) {
        // @@@ why not? tl_assert (d->num_planes () == d->get_view_ops ().size ());
        std::vector<lay::ViewOp> view_ops = d->get_view_ops (*this, mp_view->background_color (), mp_view->foreground_color (), mp_view->active_color ());
        if (i < view_ops.size ()) {
          m_drawing_planes.back ().push_back (lay::SvgCanvasPlane (view_ops[i]));
        } else {
          m_drawing_planes.back ().push_back (lay::SvgCanvasPlane ());
        }
      }
    }

    unlock ();
  }

  /**
   *  @brief Set a plane
   *
   *  This method is called from the redraw thread to transfer data for a certain plane.
   *  The planes have been reserved before with prepare ().
   */
  virtual void set_plane (unsigned int n, const lay::CanvasPlane *plane)
  {
    lock ();
    if (n < m_planes.size ()) {
      const lay::SvgCanvasPlane *svg_plane = dynamic_cast<const lay::SvgCanvasPlane *> (plane);
      tl_assert (svg_plane != 0);
      m_planes [n].copy_from (*svg_plane); 
    }
    unlock ();
  }

  /**
   *  @brief Set a plane for the drawing number d and index n within the drawing.
   *
   *  This method is called from the redraw thread.
   */
  virtual void set_drawing_plane (unsigned int d, unsigned int n, const lay::CanvasPlane *plane)
  {
    lock ();
    if (d < m_drawing_planes.size () && n < m_drawing_planes [d].size ()) {
      const lay::SvgCanvasPlane *svg_plane = dynamic_cast<const lay::SvgCanvasPlane *> (plane);
      tl_assert (svg_plane != 0);
      m_drawing_planes [d][n].copy_from (*svg_plane); 
    }
    unlock ();
  }
  
  /** 
   *  @brief Create a new, unassociated drawing plane
   */
  virtual lay::CanvasPlane *create_drawing_plane ()
  {
    return new SvgCanvasPlane (lay::ViewOp ());
  }

  /** 
   *  @brief Initialize a drawing plane for drawing on plane number n
   */
  // @@@ drop this!
  virtual void initialize_plane (lay::CanvasPlane *plane, unsigned int n)
  {
    // .. nothing yet ..
  }

  /** 
   *  @brief Initialize a drawing plane for drawing on drawing d and plane number n
   */
  // @@@ drop this!
  virtual void initialize_plane (lay::CanvasPlane *plane, unsigned int d, unsigned int n)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Provide the renderer
   */
  virtual lay::Renderer *create_renderer () 
  { 
    return new lay::SvgRenderer (m_width, m_height, m_resolution); 
  }

  /**
   *  @brief Produce the code
   */
  void produce ()
  {
    for (std::vector<lay::SvgCanvasPlane>::iterator p = m_planes.begin (); p != m_planes.end (); ++p) {
      p->produce (m_output, mp_view);
    }
    for (std::vector<std::vector<lay::SvgCanvasPlane> >::iterator pp = m_drawing_planes.begin (); pp != m_drawing_planes.end (); ++pp) {
      for (std::vector<lay::SvgCanvasPlane>::iterator p = pp->begin (); p != pp->end (); ++p) {
        p->produce (m_output, mp_view);
      }
    }
    m_planes.clear ();
    m_drawing_planes.clear ();
  }

private:
  std::vector<lay::SvgCanvasPlane> m_planes;
  std::vector<std::vector<lay::SvgCanvasPlane> > m_drawing_planes;
  lay::LayoutView *mp_view;
  int m_width, m_height;
  double m_resolution;
  std::ostream &m_output;
};

// -----------------------------------------------------------------------------------
//  ViewObjectCanvas specialization for SVG output

class SvgViewObjectCanvas
  : public ViewObjectCanvas
{
public:
  /**
   *  The constructor 
   */
  SvgViewObjectCanvas (lay::LayoutView *view, unsigned int width, unsigned int height, double resolution, std::ostream &output)
    : mp_view (view), m_renderer (width, height, resolution), m_width (width), m_height (height), m_resolution (resolution), m_output (output)
  {
    //  .. nothing yet ..
  }

  /**
   *  The destructor 
   */
  ~SvgViewObjectCanvas ()
  {
    clear_alloc_planes ();
  }

  /**
   *  @brief CanvasPlane provider
   *
   *  This method is used by the view objects to obtain the planes
   *  they should paint themselves on. 
   *  A style can be specified that determines how the plane will
   *  be displayed. CanvasPlanes may be shared if the style is identical.
   *  The plane index of the ViewOp is not used. 
   */
  virtual lay::CanvasPlane *plane (const lay::ViewOp &style)
  {
    std::map <lay::ViewOp, unsigned int>::iterator b = m_fg_plane_table.find (style);
    if (b == m_fg_plane_table.end ()) {

      //  we need to create a new plane
      m_fg_plane_table.insert (std::make_pair (style, mp_alloc_planes.size ()));
      lay::SvgCanvasPlane *plane = new lay::SvgCanvasPlane (style);
      mp_alloc_planes.push_back (plane);
      return plane;
      
    } else {
      //  we can recycle a current one
      return mp_alloc_planes [b->second];
    }
  }

  /**
   *  @brief CanvasPlane provider for a sequence of operations on the same plane
   *
   *  This method is used by the view objects to obtain the planes
   *  they should paint themselves on. 
   *  A operator vector can be specified that determines how the plane will
   *  be displayed. The operators are executed in the order given on the same plane.
   *  CanvasPlanes may be shared if the styles are identical.
   *  The plane index of the ViewOp is not used. 
   */
  virtual lay::CanvasPlane *plane (const std::vector<lay::ViewOp> &style)
  {
    std::map <std::vector<lay::ViewOp>, unsigned int>::iterator b = m_fgv_plane_table.find (style);
    if (b == m_fgv_plane_table.end ()) {

      //  we need to create a new plane
      m_fgv_plane_table.insert (std::make_pair (style, mp_alloc_planes.size ()));
      lay::SvgCanvasPlane *plane = new lay::SvgCanvasPlane (style);
      mp_alloc_planes.push_back (plane);
      return plane;
      
    } else {
      //  we can recycle a current one
      return mp_alloc_planes [b->second];
    }
  }

  /**
   *  @brief Sort the planes in the painting order
   */
  virtual void sort_planes ()
  {
    // .. not used currently ..
  }

  /**
   *  @brief Provide the renderer
   */
  virtual lay::Renderer &renderer () 
  { 
    return m_renderer; 
  }

  virtual QColor background_color () const
  {
    return mp_view->background_color ();
  }

  virtual QColor foreground_color () const
  {
    return mp_view->foreground_color ();
  }

  virtual QColor active_color () const
  {
    return mp_view->active_color ();
  }

  /**
   *  @brief Get the resolution
   */
  virtual double resolution () const
  {
    return m_resolution;
  }

  /**
   *  @brief Produce the code
   */
  void produce ()
  {
    for (std::vector<lay::SvgCanvasPlane *>::iterator p = mp_alloc_planes.begin (); p != mp_alloc_planes.end (); ++p) {
      (*p)->produce (m_output, mp_view);
    }
    clear_alloc_planes ();
  }

private:
  lay::LayoutView *mp_view;
  lay::SvgRenderer m_renderer;
  unsigned int m_width, m_height;
  double m_resolution;
  std::ostream &m_output;
  std::map <lay::ViewOp, unsigned int> m_fg_plane_table;
  std::map <std::vector <lay::ViewOp>, unsigned int> m_fgv_plane_table;
  std::vector <lay::SvgCanvasPlane *> mp_alloc_planes;

  void 
  clear_alloc_planes ()
  {
    for (std::vector <lay::SvgCanvasPlane *>::iterator i = mp_alloc_planes.begin (); i != mp_alloc_planes.end (); ++i) {
      if (*i) {
        delete *i;
      }
    }
    mp_alloc_planes.clear ();
  }
};

// -----------------------------------------------------------------------------------
//  Main function implementation

void write_svg_image (lay::LayoutView *view, const std::string &fn, double dwidth, double dheight, double resolution)
{
  std::ofstream output (fn.c_str (), std::ios::out);
  if (! output.good ()) {
    throw tl::Exception (tl::translate ("Unable to open file %s for writing").c_str (), fn.c_str ());
  }

  tl_assert (resolution > 1e-6);
  tl_assert (dwidth > resolution);
  tl_assert (dheight > resolution);

  // @@@ needed?
  unsigned int width = (unsigned int)(dwidth / resolution + 0.5);
  unsigned int height = (unsigned int)(dheight / resolution + 0.5);

  //  write preamble
  output << "<?xml version=\"1.0\" standalone=\"no\"?>" << std::endl
         << "<!-- Created with KLayout (http://www.klayout.de/) -->" << std::endl
         << "<svg" << std::endl
         << "  xmlns:svg=\"http://www.w3.org/2000/svg\"" << std::endl
         << "  xmlns=\"http://www.w3.org/2000/svg\"" << std::endl
         << "  xmlns:xlink=\"http://www.w3.org/1999/xlink\"" << std::endl
         << "  version=\"1.0\"" << std::endl
         << "  width=\"" << dwidth << "\"" << std::endl
         << "  height=\"" << dheight << "\"" << std::endl
         << "  id=\"svg\">" << std::endl;

  double ma = resolution;
  double mb = 0.0;
  double mc = 0.0;
  double md = -resolution;
  double me = 0.0;
  double mf = dheight;

  output << "<g transform=\"matrix(" << ma << "," << mb << "," << mc << "," << md << "," << me << "," << mf << ")\">" << std::endl;

  unsigned int bg_color = view->background_color ().rgb () & 0xffffff;
  output << "<rect x=\"0\" y=\"0\" width=\"" << width << "\" height=\"" << height << "\" style=\"fill:#" << tl::sprintf("%06x", bg_color) << "\"/>" << std::endl;

  //  provide canvas objects for the layout bitmaps and the foreground/background objects
  SvgRedrawThreadCanvas rd_canvas (view, width, height, resolution * 4.0, output); // @@@ TODO: this assumes that the resolution is given in mm
  SvgViewObjectCanvas vo_canvas (view, width, height, resolution * 4.0, output); // @@@ TODO: this assumes that the resolution is given in mm

  //  compute the new viewport 
  Viewport vp (width, height, view->viewport ().target_box ());

  //  paint the background objects.
  view->view_object_widget ()->do_render_bg (vp, vo_canvas);

  //  render the layout
  lay::RedrawThread redraw_thread (&rd_canvas, view);
  redraw_thread.start (0 /*synchronous*/, view->get_redraw_layers (), vp, resolution * 4.0); // @@@ TODO: this assumes that the resolution is given in mm
  redraw_thread.stop (); // safety
  rd_canvas.produce ();

  //  render the foreground parts ..
  view->view_object_widget ()->do_render (vp, vo_canvas, true);
  view->view_object_widget ()->do_render (vp, vo_canvas, false);
  vo_canvas.produce ();

  //  write tail
  output << "</g>" << std::endl;
  output << "</svg>" << std::endl;
}

}

