
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layTechnology.h"
#include "layPlugin.h"

#include <stdio.h>

namespace lay
{

// -----------------------------------------------------------------------------

Technologies::Technologies ()
{
  m_technologies.push_back (new Technology (std::string (""), "(Default)"));
}

Technologies::~Technologies ()
{
  // .. nothing yet ..
}

std::string 
Technologies::to_string () const
{
  std::string result;
  for (tl::stable_vector <Technology>::const_iterator t = m_technologies.begin (); t != m_technologies.end (); ++t) {
    result += tl::to_word_or_quoted_string (t->name ());
    result += "(";
    result += tl::to_word_or_quoted_string (t->description ());
    result += "):";
    result += t->to_string ();
    result += ";";
  }
  return result;
}

void 
Technologies::parse (tl::Extractor &ex)
{
  m_technologies.clear ();
  m_technologies.push_back (new Technology (std::string (""), "(Default)"));

  while (! ex.at_end ()) {

    std::string name;
    std::string description;
    ex.read_word_or_quoted (name);
    if (ex.test ("(")) {
      ex.read_word_or_quoted (description);
      ex.expect (")");
    }
    ex.expect (":");

    Technology *technology = technology_by_name (name);
    if (! technology) {
      technology = new Technology (name, description);
      throw tl::Exception (tl::translate ("Invalid technology name: ") + name);
    }
    technology->parse (ex);

    ex.test (";");

  }
}

void 
Technologies::add (Technology *technology)
{
  m_technologies.push_back (technology);
}

void 
Technologies::remove (iterator position)
{
  if (position != m_technologies.begin ()) {
    m_technologies.erase (position);
  }
}

bool 
Technologies::has_technology (const std::string &name) const
{
  for (tl::stable_vector<Technology>::const_iterator t = m_technologies.begin (); t != m_technologies.end (); ++t) {
    if (t->name () == name) {
      return true;
    }
  }

  return false;
}

Technology *
Technologies::technology_by_name (const std::string &name)
{
  for (tl::stable_vector<Technology>::iterator t = m_technologies.begin (); t != m_technologies.end (); ++t) {
    if (t->name () == name) {
      return &*t;
    }
  }

  return &*m_technologies.begin ();
}

// -----------------------------------------------------------------------------
//  Technology implementation

Technology::Technology (const std::string &name, const std::string &description)
  : m_name (name), m_description (description)
{
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    if (cls->technology_component_provider () != 0) {
      m_components.push_back (cls->technology_component_provider ()->create_component ());
    }
  }
}

Technology::~Technology ()
{
  for (std::vector <TechnologyComponent *>::const_iterator c = m_components.begin (); c != m_components.end (); ++c) {
    delete *c;
  }
  m_components.clear ();
}

Technology::Technology (const Technology &d)
{
  for (std::vector <TechnologyComponent *>::const_iterator c = d.m_components.begin (); c != d.m_components.end (); ++c) {
    m_components.push_back ((*c)->clone ());
  }
}

Technology &Technology::operator= (const Technology &d)
{
  if (this != &d) {

    for (std::vector <TechnologyComponent *>::const_iterator c = m_components.begin (); c != m_components.end (); ++c) {
      delete *c;
    }
    m_components.clear ();

    for (std::vector <TechnologyComponent *>::const_iterator c = d.m_components.begin (); c != d.m_components.end (); ++c) {
      m_components.push_back ((*c)->clone ());
    }

  }

  return *this;
}

std::string 
Technology::to_string () const
{
  std::string result;
  result += "(";
  for (std::vector <TechnologyComponent *>::const_iterator c = m_components.begin (); c != m_components.end (); ++c) {
    result += tl::to_word_or_quoted_string ((*c)->name ());
    result += ":";
    result += (*c)->to_string ();
    result += ";";
  }
  result += ")";
  return result;
}

void 
Technology::parse (tl::Extractor &ex)
{
  ex.test ("(");

  while (! ex.at_end () && ! ex.test (")")) {

    std::string name;
    ex.read_word_or_quoted (name);
    ex.expect (":");

    TechnologyComponent *component = component_by_name (name);
    if (! component) {
      throw tl::Exception (tl::translate ("Invalid technology component name: ") + name);
    }
    component->parse (ex);

    ex.test (";");

  }
}

TechnologyComponent *
Technology::component_by_name (const std::string &component_name)
{
  for (tl::stable_vector <TechnologyComponent>::iterator c = m_components.begin (); c != m_components.end (); ++c) {
    if (c->name () == component_name) {
      return &*c;
    }
  }

  return 0;
}

}

