
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_layTechnology
#define HDR_layTechnology

#include "config.h"

#include "tlStableVector.h"
#include "tlString.h"

#include <QtGui/QFrame>

namespace lay
{

class Technology;
class TechnologyComponent;
class TechnologyComponentEditor;

/**
 *  @brief A container for the technology settings
 *
 *  The container associates a technology with a name and provides an
 *  iterator for the technologies.
 *  The container features at least one technology (the default) which is
 *  present in any case. If a technology with an unknown name is requested,
 *  this default technology is returned.
 */
class KLAYOUT_DLL Technologies
{
public:
  typedef tl::stable_vector<Technology>::iterator iterator;
  typedef tl::stable_vector<Technology>::const_iterator const_iterator;

  /**
   *  @brief The constructor
   */
  Technologies ();

  /**
   *  @brief The destructor
   */
  ~Technologies ();

  /**
   *  @brief Non-const iterator - begin 
   */
  iterator begin ()
  {
    return m_technologies.begin ();
  }

  /**
   *  @brief Non-const iterator - end 
   */
  iterator end ()
  {
    return m_technologies.end ();
  }

  /**
   *  @brief Const iterator - begin 
   */
  const_iterator begin () const
  {
    return m_technologies.begin ();
  }

  /**
   *  @brief Const iterator - end 
   */
  const_iterator end () const
  {
    return m_technologies.end ();
  }

  /**
   *  @brief Converts the technology setup to a string
   *
   *  The string is intended for serialization purposes.
   */
  std::string to_string () const;

  /**
   *  @brief Parses the technoloy setup from a string extractor
   */
  void parse (tl::Extractor &ex);

  /**
   *  @brief Adds a technology to the setup
   *
   *  The container becomes owner of the technology object.
   */
  void add (Technology *technology);

  /**
   *  @brief Remove a technology from the setup
   */
  void remove (iterator position);

  /**
   *  @brief Checks, if a technology with the given name exists
   */
  bool has_technology (const std::string &name) const;

  /**
   *  @brief Returns the technology with the given name
   *
   *  If no technology with that name exists, the default technology is returned.
   */
  Technology *technology_by_name (const std::string &name);

  /**
   *  @brief Returns the technology with the given name (const version)
   *
   *  If no technology with that name exists, the default technology is returned.
   */
  const Technology *technology_by_name (const std::string &name) const
  {
    return const_cast<Technologies *> (this)->technology_by_name (name);
  }
  
private:
  tl::stable_vector<Technology> m_technologies;
};

/**
 *  @brief A technology 
 *
 *  This class represents one technology.
 *  A technology has a name and a description.
 */
class KLAYOUT_DLL Technology
{
public:
  /**
   *  @brief The constructor 
   */
  Technology (const std::string &name, const std::string &description);

  /**
   *  @brief The copy constructor 
   */
  Technology (const Technology &tech);

  /**
   *  @brief The destructor
   */
  ~Technology ();

  /**
   *  @brief Assignment
   */
  Technology &operator= (const Technology &tech);

  /**
   *  @brief Gets the name 
   */
  const std::string &name () const
  {
    return m_name;
  }

  /**
   *  @brief Gets the description 
   */
  const std::string &description () const
  {
    return m_description;
  }

  /**
   *  @brief Converts the technology to a string
   */
  std::string to_string () const;

  /**
   *  @brief Parses the technology from a string
   */
  void parse (tl::Extractor &ex);

  /**
   *  @brief Gets the technology component by the component name
   *
   *  If no component with that name exists, 0 is returned.
   */
  TechnologyComponent *component_by_name (const std::string &component_name);

private:
  std::string m_name, m_description;
  std::vector <TechnologyComponent *> m_components;
};

/**
 *  @brief A technology component
 *
 *  A technology component is a part of the data for one technology. 
 *  Plugins may register technology components in every technology and
 *  use those components to store their specific data.
 *  A technology component has a name and a description. The name is used
 *  to identify a component within a technology. The description is shown
 *  in the setup dialogs.
 *  This class is the base class for all technology components.
 */
class KLAYOUT_DLL TechnologyComponent
{
public:
  /**
   *  @brief The constructor
   *
   *  @param name The name of the technology component
   *  @param descriptor The description of the technology component
   */
  TechnologyComponent (const std::string &name, const std::string &description)
    : m_name (name), m_description (description)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief The destructor
   */
  virtual ~TechnologyComponent ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Gets the name 
   */
  const std::string &name () const
  {
    return m_name;
  }

  /**
   *  @brief Gets the description 
   */
  const std::string &description () const
  {
    return m_description;
  }

  /**
   *  @brief Converts the technology component to a string
   */
  virtual std::string to_string () const
  {
    return std::string ();
  }

  /**
   *  @brief Parses the technology component from a string
   */
  virtual void parse (tl::Extractor &ex)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Creates the editor 
   *
   *  The implementation of this method is supposed to create an editor for
   *  the given component. It is guaranteed that the component that is given
   *  to the editor is one that has been delivered by create_component ().
   */
  virtual TechnologyComponentEditor *create_editor (QWidget *parent) 
  {
    return 0;
  }

  /**
   *  @brief Clone this instance
   */
  virtual TechnologyComponent *clone () const = 0;

private:
  std::string m_name, m_description;
};

/**
 *  @brief A base class for an editor for a technology component
 *
 *  A technology component provider can provide an editor for the component
 *  To do so, it must implement a TechnologyComponentEditor that provides an 
 *  editor for it's component.
 */
class KLAYOUT_DLL TechnologyComponentEditor
  : public QFrame
{
public:
  /**
   *  @brief The constructor
   */
  TechnologyComponentEditor (QWidget *parent)
    : QFrame (parent)
  {
    // .. nothing yet ..
  }

  /**
   *  @brief The destructor
   */
  virtual ~TechnologyComponentEditor ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Set the technology and component for the editor
   */
  void set_technology (Technology *tech, TechnologyComponent *tech_component)
  {
    mp_tech = tech;
    mp_tech_component = tech_component;
  }

  /**
   *  @brief Sets up the editor with the given technology and component
   */
  virtual void setup () 
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Commits the edits to the technology and component
   */
  virtual void commit () 
  {
    // .. nothing yet ..
  }

protected:
  Technology *tech () 
  {
    return mp_tech;
  }

  TechnologyComponent *tech_component () 
  {
    return mp_tech_component;
  }

private:
  Technology *mp_tech;
  TechnologyComponent *mp_tech_component;
};

/**
 *  @brief A base class for a technology component provider
 */
class KLAYOUT_DLL TechnologyComponentProvider
{
public:
  /**
   *  @brief The constructor
   */
  TechnologyComponentProvider ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief The destructor
   */
  virtual ~TechnologyComponentProvider ()
  {
    // .. nothing yet ..
  }

  /**
   *  @brief Creates the technology component
   */
  virtual TechnologyComponent *create_component () const = 0;
};

}

#endif

