
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QKeyEvent>

#include "layExceptions.h"
#include "layApplication.h"
#include "tlString.h"
#include "rbaConsoleImpl.h"
#include "Console.h"

namespace rba 
{

ConsoleImpl::ConsoleImpl ()
  : QDialog (0 /*parent*/, "rba_console"), 
    m_in_event_handler (false),
    m_os (OS_none),
    m_new_line (true)
{
  mp_ui = new Ui::ConsoleForm ();
  mp_ui->setupUi (this);

  mp_ui->input_field->installEventFilter (this);
  mp_ui->input_field->setCompleter (0);

  m_def_color = mp_ui->text_widget->textColor ();
  m_def_font = mp_ui->text_widget->currentFont ();

  connect (mp_ui->close_button, SIGNAL (clicked ()), this, SLOT (hide ()));
  connect (mp_ui->clear_button, SIGNAL (clicked ()), this, SLOT (clearLog ()));
  connect (mp_ui->input_field, SIGNAL (editTextChanged (const QString &)), this, SLOT (editTextChanged (const QString &)));

  m_history_index = -1;
}

ConsoleImpl::~ConsoleImpl ()
{
  delete mp_ui;
  mp_ui = 0;
}

void
ConsoleImpl::clearLog ()
{
  mp_ui->text_widget->clear ();
  m_new_line = true;
  m_os = OS_none;
}

void 
ConsoleImpl::showEvent (QShowEvent *)
{
  lay::Application::instance ()->ruby_interpreter ().install_console (this);
}

void 
ConsoleImpl::hideEvent (QHideEvent *)
{
  lay::Application::instance ()->ruby_interpreter ().install_console (0);
}

void 
ConsoleImpl::editTextChanged (const QString &text)
{
  m_history_index = -1;
  if (! m_in_event_handler) {
    m_edit_text = text;
  }
}

void
ConsoleImpl::execute (const QString &cmd)
{
  BEGIN_PROTECTED
  try {
    write_str ("> ", OS_echo);
    write_str (cmd.ascii (), OS_echo);
    write_str ("\n", OS_echo);
    lay::Application::instance ()->ruby_interpreter ().eval_string_and_print (cmd.ascii ());
  } catch (RubyError &ex) {
    if (ex.line () > 0) {
      write_str (ex.sourcefile ().c_str (), OS_stderr);
      write_str (":", OS_stderr);
      write_str (tl::to_string (ex.line ()).c_str (), OS_stderr);
      write_str (": ", OS_stderr);
    }
    write_str (ex.msg ().c_str (), OS_stderr);
    write_str ("\n", OS_stderr);
  } catch (tl::Exception &ex) {
    write_str (ex.msg ().c_str (), OS_stderr);
    write_str ("\n", OS_stderr);
  } catch (std::runtime_error &ex) {
    write_str (ex.what (), OS_stderr);
    write_str ("\n", OS_stderr);
  }
  END_PROTECTED
}

bool 
ConsoleImpl::eventFilter (QObject *obj, QEvent *event)
{
  if (obj == mp_ui->input_field && event->type() == QEvent::KeyPress) {

    QKeyEvent *key_event = dynamic_cast<QKeyEvent *> (event);
    if (key_event->key () == Qt::Key_Return) {

      QString cmd = mp_ui->input_field->currentText ();
      if (cmd != "") {

        if (! (m_history_index >= 0 && m_history_index < mp_ui->input_field->count () && cmd == mp_ui->input_field->itemText (m_history_index))) {
          mp_ui->input_field->addItem (cmd);
        }

        execute (cmd);

        mp_ui->input_field->clearEditText ();
        m_edit_text = "";
        m_history_index = -1;

      }

      return true;

    } else if (key_event->key () == Qt::Key_Up) {

      m_in_event_handler = true; // prevent setting of m_edit_text

      int hi = m_history_index;
      if (hi < 0) {
        if (mp_ui->input_field->count () > 0) {
          hi = mp_ui->input_field->count () - 1;
          mp_ui->input_field->setCurrentItem (hi);
        }
      } else if (hi > 0 && hi <= mp_ui->input_field->count ()) {
        --hi;
        mp_ui->input_field->setCurrentItem (hi);
      }
        
      m_in_event_handler = false;
      m_history_index = hi;

      return true;

    } else if (key_event->key () == Qt::Key_Down) {

      m_in_event_handler = true; // prevent setting of m_edit_text

      int hi = m_history_index;
      if (hi < 0) {
        if (mp_ui->input_field->count () > 0) {
          hi = mp_ui->input_field->count () - 1;
          mp_ui->input_field->setCurrentItem (hi);
        }
      } else if (hi < mp_ui->input_field->count () - 1) {
        ++hi;
        mp_ui->input_field->setCurrentItem (hi);
      } else {
        hi = mp_ui->input_field->count ();
        mp_ui->input_field->setEditText (m_edit_text);
      }
        
      m_in_event_handler = false;
      m_history_index = hi;

      return true;

    }

  } 
  return QDialog::eventFilter(obj, event);

}

void 
ConsoleImpl::write_str (const char *text, output_stream os)
{
  if (! mp_ui->text_widget->textCursor ().atEnd ()) {
    QTextCursor c = mp_ui->text_widget->textCursor ();
    c.movePosition (QTextCursor::End);
    mp_ui->text_widget->setTextCursor (c);
  }

  if (m_os != OS_none && os != m_os && ! m_new_line) {
    //  insert a new line if the stream changes ..
    write_str ("\n", m_os);
  }

  if (m_os != os) {
    if (os == OS_stdout) {
      mp_ui->text_widget->setCurrentFont (m_def_font);
      mp_ui->text_widget->setTextColor (m_def_color);
    } else if (os == OS_echo) {
      mp_ui->text_widget->setCurrentFont (m_def_font);
      mp_ui->text_widget->setTextColor (QColor (0, 0, 255));
    } else if (os == OS_stderr) {
      QFont f (m_def_font);
      f.setBold (true);
      mp_ui->text_widget->setCurrentFont (f);
      mp_ui->text_widget->setTextColor (QColor (255, 0, 0));
    } 
  }

  m_os = os;

  for (const char *t = text; *t; ) {

    const char *t0 = t;
    for ( ; *t && *t != '\n'; ++t)
      ;

    mp_ui->text_widget->insertPlainText (QString::fromAscii (t0, t - t0));

    if (*t == '\n') {
      ++t;
      //  new line: terminate line
      mp_ui->text_widget->insertPlainText ("\n");
      m_new_line = true;
    } else {
      m_new_line = false;
    }

  }

  mp_ui->text_widget->ensureCursorVisible ();
}

}

