
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "tlExpression.h"
#include "tlString.h"

#include <map>
#include <vector>
#include <sstream>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <QtCore/QFileInfo>
#include <QtCore/QDir>

//  Suggestions for further functions:
//  - provide date/time function
//  - file functions: (drive), combine, split_path
//  - file tests: exists, locate (in path), readable, writeable 
//  - glob: match, expand (files) - needs glob code

namespace tl
{

// ----------------------------------------------------------------------------
//  Implementation of EvalError

EvalError::EvalError (const std::string &what, const char *where)
  : tl::Exception (what + " at '.." + std::string (where) + "'")
{
  // .. nothing yet ..
}

// ----------------------------------------------------------------------------
//  Utilities for evaluation

static double to_double (tl::Extractor &ex, const tl::Variant &v)
{
  if (v.can_convert_to_double ()) {
    return v.to_double ();
  } else if (v.is_list ()) {
    return v.get_list ().size ();
  } else {
    throw EvalError (tl::translate ("Float value expected"), ex.skip ());
  }
}

static double to_double (tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  if (v.size () != 1) {
    throw EvalError (tl::translate ("Function expects a single numeric argument"), ex.skip ());
  }

  return to_double (ex, v [0]);
}

static long to_long (tl::Extractor &ex, const tl::Variant &v)
{
  if (v.can_convert_to_long ()) {
    return v.to_long ();
  } else if (v.is_list ()) {
    return long (v.get_list ().size ());
  } else {
    throw EvalError (tl::translate ("Integer value expected"), ex.skip ());
  }
}

// ----------------------------------------------------------------------------
//  Implementation of functions

static tl::Variant
sin_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (sin (to_double (ex, v)));
}

static tl::Variant
sinh_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (sinh (to_double (ex, v)));
}

static tl::Variant
cos_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (cos (to_double (ex, v)));
}

static tl::Variant
cosh_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (cosh (to_double (ex, v)));
}

static tl::Variant
tan_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (tan (to_double (ex, v)));
}

static tl::Variant
tanh_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (tanh (to_double (ex, v)));
}

static tl::Variant
log_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (::log (to_double (ex, v)));
}

static tl::Variant
log10_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (log10 (to_double (ex, v)));
}

static tl::Variant
exp_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (exp (to_double (ex, v)));
}

static tl::Variant
floor_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (floor (to_double (ex, v)));
}

static tl::Variant
ceil_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (ceil (to_double (ex, v)));
}

static tl::Variant
round_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  //  VC++ does not have "round"
  //  return tl::Variant (round (to_double (ex, v)));
  return tl::Variant (floor (0.5 + to_double (ex, v)));
}

static tl::Variant
sqrt_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (sqrt (to_double (ex, v)));
}

static tl::Variant
acos_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (acos (to_double (ex, v)));
}

#ifndef _MSC_VER // not available on MS VC++
static tl::Variant
acosh_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (acosh (to_double (ex, v)));
}
#endif

static tl::Variant
asin_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (asin (to_double (ex, v)));
}

#ifndef _MSC_VER // not available on MS VC++
static tl::Variant
asinh_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (asinh (to_double (ex, v)));
}
#endif

static tl::Variant
atan_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (atan (to_double (ex, v)));
}

#ifndef _MSC_VER // not available on MS VC++
static tl::Variant
atanh_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &v)
{
  return tl::Variant (atanh (to_double (ex, v)));
}
#endif

static tl::Variant
min_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  tl::Variant r;
  for (std::vector <tl::Variant>::const_iterator v = vv.begin (); v != vv.end (); ++v) {
    if (! v->is_nil () && (r.is_nil () || *v < r)) {
      r = *v;
    }
  }

  return r;
}

static tl::Variant
max_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  tl::Variant r;
  for (std::vector <tl::Variant>::const_iterator v = vv.begin (); v != vv.end (); ++v) {
    if (! v->is_nil () && (r.is_nil () || r < *v)) {
      r = *v;
    }
  }

  return r;
}

static tl::Variant
pow_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'pow' function expects exactly two arguments"), ex.skip ());
  }

  return tl::Variant (pow (to_double (ex, vv [0]), to_double (ex, vv [1])));
}

static tl::Variant
atan2_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'atan2' function expects exactly two arguments"), ex.skip ());
  }

  return tl::Variant (atan2 (to_double (ex, vv [0]), to_double (ex, vv [1])));
}

static tl::Variant
to_f_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'to_f' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].to_double ());
}

static tl::Variant
to_s_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'to_s' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].to_string ());
}

static tl::Variant
to_i_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'to_i' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].to_long ());
}

static tl::Variant
is_string_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'is_string' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].is_string ());
}

static tl::Variant
is_numeric_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'is_numeric' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].is_long () || vv [0].is_double ());
}

static tl::Variant
is_array_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'is_array' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].is_list ());
}

static tl::Variant
is_nil_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'is_nil' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (vv [0].is_nil ());
}

static tl::Variant
gsub_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 3) {
    throw EvalError (tl::translate ("'gsub' function expects exactly three arguments"), ex.skip ());
  }

  std::string s (vv [0].to_string ());
  std::string x (vv [1].to_string ());
  std::string y (vv [2].to_string ());

  std::string r;
  r.reserve (s.size ());

  size_t p = 0;
  for (size_t pp = 0; (pp = s.find (x, p)) != std::string::npos; p = pp + x.size ()) {
    r += std::string (s, p, pp - p);
    r += y;
  }

  r += std::string (s, p, std::string::npos);

  return tl::Variant (r);
}

static tl::Variant
sub_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 3) {
    throw EvalError (tl::translate ("'sub' function expects exactly three arguments"), ex.skip ());
  }

  std::string s (vv [0].to_string ());
  std::string x (vv [1].to_string ());
  std::string y (vv [2].to_string ());

  std::string r;

  size_t p = s.find (x);
  if (p != std::string::npos) {

    r.reserve (s.size () + y.size () - x.size ());
    r += std::string (s, 0, p);
    r += y;
    r += std::string (s, p + x.size ());
    return tl::Variant (r);

  } else {
    return tl::Variant (s);
  }

}

static tl::Variant
find_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'find' function expects exactly two arguments"), ex.skip ());
  }

  std::string s (vv [0].to_string ());
  std::string x (vv [1].to_string ());

  size_t p = s.find (x);
  if (p != std::string::npos) {
    return tl::Variant (long (p));
  } else {
    return tl::Variant ();
  }
}

static tl::Variant
rfind_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'rfind' function expects exactly two arguments"), ex.skip ());
  }

  std::string s (vv [0].to_string ());
  std::string x (vv [1].to_string ());

  size_t p = s.rfind (x);
  if (p != std::string::npos) {
    return tl::Variant (long (s.size () - (p + x.size ())));
  } else {
    return tl::Variant ();
  }
}

static tl::Variant
len_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'len' function expects exactly one argument"), ex.skip ());
  }

  if (vv [0].is_list ()) {
    return tl::Variant (long (vv [0].end () - vv [0].begin ()));
  } else {
    return tl::Variant (long (strlen (vv [0].to_string ())));
  }
}

static tl::Variant
substr_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 3 && vv.size () != 2) {
    throw EvalError (tl::translate ("'substr' function expects two or three arguments"), ex.skip ());
  }

  std::string s (vv [0].to_string ());

  long len = std::string::npos;
  if (vv.size () > 2) {
    len = std::max (long (0), to_long (ex, vv [2]));
  }

  long l = to_long (ex, vv [1]);
  if (l < 0) {
    l = long (s.size ()) + l;
    if (l < 0) {
      len += l;
      l = 0;
    }
  } 

  size_t from = size_t (l);

  if (len == 0 || from >= s.size ()) {
    return tl::Variant ("");
  } else if (from + len >= s.size ()) {
    return tl::Variant (std::string (s, from));
  } else {
    return tl::Variant (std::string (s, from, len));
  }
}

static tl::Variant
join_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'join' function expects exactly two arguments"), ex.skip ());
  }

  if (! vv[0].is_list ()) {
    throw EvalError (tl::translate ("First argument of 'join' function must be a list"), ex.skip ());
  }

  std::ostringstream r;

  std::string s (vv [1].to_string ());

  bool first = true;
  for (tl::Variant::iterator i = vv [0].begin (); i != vv [0].end (); ++i) {
    if (first) {
      r << s;
      first = false;
    }
    r << i->to_string ();
  }

  return tl::Variant (r.str ());
}

static tl::Variant
item_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'item' function expects exactly two arguments"), ex.skip ());
  }

  if (! vv[0].is_list ()) {
    throw EvalError (tl::translate ("First argument of 'item' function must be a list"), ex.skip ());
  }

  long index = to_long (ex, vv [1]);
  if (index < 0 || index >= long (vv [0].end () - vv [0].begin ())) {
    return tl::Variant ();
  } else {
    return *(vv [0].begin () + index);
  }
}

static tl::Variant
split_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'split' function expects exactly two arguments"), ex.skip ());
  }

  tl::Variant r = tl::Variant::empty_list ();
  std::string t (vv [0].to_string ()); 
  std::string s (vv [1].to_string ());

  size_t p = 0;
  for (size_t pp = 0; (pp = t.find (s, p)) != std::string::npos; p = pp + s.size ()) {
    r.push (tl::Variant (std::string (t, p, pp - p)));
  }

  r.push (tl::Variant (std::string (t, p)));

  return r;
}

static tl::Variant
true_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 0) {
    throw EvalError (tl::translate ("'true' function must not have arguments"), ex.skip ());
  }

  return tl::Variant (true);
}

static tl::Variant
false_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 0) {
    throw EvalError (tl::translate ("'false' function must not have arguments"), ex.skip ());
  }

  return tl::Variant (false);
}

static tl::Variant
nil_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 0) {
    throw EvalError (tl::translate ("'nil' function must not have arguments"), ex.skip ());
  }

  return tl::Variant ();
}

static tl::Variant
env_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'env' function expects exactly two arguments"), ex.skip ());
  }

  const char *env = getenv (vv [0].to_string ());
  if (env) {
    return tl::Variant (env);
  } else {
    return tl::Variant ();
  }
}

static tl::Variant
error_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'error' function expects exactly one argument"), ex.skip ());
  }

  throw tl::Exception (vv [0].to_string ());
}

static tl::Variant
absolute_file_path_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'absolute_file_path' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).absoluteFilePath ().ascii ());
}

static tl::Variant
absolute_path_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'absolute_path' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).absolutePath ().ascii ());
}

static tl::Variant
path_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'path' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).path ().ascii ());
}

static tl::Variant
basename_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'basename' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).baseName ().ascii ());
}

static tl::Variant
extension_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'extension' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).completeSuffix ().ascii ());
}

static tl::Variant
file_exists_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'file_exists' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).exists ());
}

static tl::Variant
is_dir_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 1) {
    throw EvalError (tl::translate ("'is_dir' function expects exactly one argument"), ex.skip ());
  }

  return tl::Variant (QFileInfo (vv [0].to_string ()).isDir ());
}

static tl::Variant
combine_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () != 2) {
    throw EvalError (tl::translate ("'combine' function expects two arguments"), ex.skip ());
  }

  return tl::Variant (QFileInfo (QDir (vv [0].to_string ()), vv [1].to_string ()).filePath ().ascii ());
}

static tl::Variant
sprintf_f (const tl::Eval *, tl::Extractor &ex, const std::vector <tl::Variant> &vv)
{
  if (vv.size () < 1) {
    throw EvalError (tl::translate ("'sprintf' function expects at least one argument"), ex.skip ());
  }

  std::ostringstream os;
  int def_prec = os.precision();

  std::string f (vv [0].to_string ()); 

  unsigned int a = 1;

  for (const char *cp = f.c_str (); *cp; ) {

    if (*cp == '%' && cp[1] == '%') {
      os << '%';
      cp += 2;
    } else if (*cp == '%') {

      ++cp;
      if (*cp == '-') {
        ++cp;
        os << std::left;
      } else {
        os << std::right;
      }

      if (*cp == '0') {
        ++cp;
        os.fill('0');
      } else {
        os.fill(' ');
      }

      unsigned int width = 0;
      while (isdigit (*cp) && *cp) {
        width = (width * 10) + (unsigned int)(*cp - '0');
        ++cp;
      }
      os.width(width);

      if (*cp == '.') {
        ++cp;
        unsigned int prec = 0;
        while (isdigit (*cp) && *cp) {
          prec = (prec * 10) + (unsigned int)(*cp - '0');
          ++cp;
        }
        os.precision(prec);
      } else {
        os.precision(def_prec);
      }

      if (*cp == 'x') {
        os.setf (std::ios::hex, std::ios::basefield);
        if (a < vv.size ()) {
          os << vv [a].to_ulong ();
        }
      } if (*cp == 'u') {
        os.setf (std::ios_base::fmtflags (0), std::ios::basefield);
        if (a < vv.size ()) {
          os << vv [a].to_ulong ();
        }
      } if (*cp == 'd') {
        os.setf (std::ios_base::fmtflags (0), std::ios::basefield);
        if (a < vv.size ()) {
          os << vv [a].to_long ();
        }
      } if (*cp == 's') {
        os.setf (std::ios_base::fmtflags (0), std::ios::basefield);
        if (a < vv.size ()) {
          os << vv [a].to_string ();
        }
      } if (*cp == 'g') {
        os.setf (std::ios_base::fmtflags (0), std::ios::basefield);
        os.setf (std::ios_base::fmtflags (0), std::ios::floatfield);
        if (a < vv.size ()) {
          os << vv [a].to_double ();
        }
      } if (*cp == 'e') {
        os.setf (std::ios_base::fmtflags (0), std::ios::basefield);
        os.setf (std::ios::scientific, std::ios::floatfield);
        if (a < vv.size ()) {
          os << vv [a].to_double ();
        }
      } if (*cp == 'f') {
        os.setf (std::ios_base::fmtflags (0), std::ios::basefield);
        os.setf (std::ios::fixed, std::ios::floatfield);
        if (a < vv.size ()) {
          os << vv [a].to_double ();
        }
      }

      if (*cp) {
        ++cp;
      }

      ++a;

    } else {
      os << *cp;
      ++cp;
    }

  }

  os << std::ends;

  return tl::Variant (os.str ());
}

// ----------------------------------------------------------------------------
//  Implementation of Eval

std::map <std::string, tl::Variant> Eval::m_global_vars;
std::map <std::string, function_t> Eval::m_global_functions;

static std::map <std::string, function_t> ms_functions;

struct FunctionInitializer
{
  FunctionInitializer ()
  {
    std::pair <std::string, function_t> function_defs[] = {
      std::pair <std::string, function_t> ("sin", &sin_f),
      std::pair <std::string, function_t> ("sinh", &sinh_f),
      std::pair <std::string, function_t> ("cos", &cos_f),
      std::pair <std::string, function_t> ("cosh", &cosh_f),
      std::pair <std::string, function_t> ("tan", &tan_f),
      std::pair <std::string, function_t> ("tanh", &tanh_f),
      std::pair <std::string, function_t> ("log", &log_f),
      std::pair <std::string, function_t> ("log10", &log10_f),
      std::pair <std::string, function_t> ("exp", &exp_f),
      std::pair <std::string, function_t> ("floor", &floor_f),
      std::pair <std::string, function_t> ("ceil", &ceil_f),
      std::pair <std::string, function_t> ("round", &round_f),
      std::pair <std::string, function_t> ("sqrt", &sqrt_f),
      std::pair <std::string, function_t> ("max", &max_f),
      std::pair <std::string, function_t> ("min", &min_f),
      std::pair <std::string, function_t> ("pow", &pow_f),
      std::pair <std::string, function_t> ("acos", &acos_f),
#ifndef _MSC_VER // not available on MS VC++
      std::pair <std::string, function_t> ("acosh", &acosh_f),
      std::pair <std::string, function_t> ("asinh", &asinh_f),
      std::pair <std::string, function_t> ("atanh", &atanh_f),
#endif
      std::pair <std::string, function_t> ("asin", &asin_f),
      std::pair <std::string, function_t> ("atan", &atan_f),
      std::pair <std::string, function_t> ("atan2", &atan2_f),
      std::pair <std::string, function_t> ("to_f", &to_f_f),
      std::pair <std::string, function_t> ("to_s", &to_s_f),
      std::pair <std::string, function_t> ("to_i", &to_i_f),
      std::pair <std::string, function_t> ("is_string", &is_string_f),
      std::pair <std::string, function_t> ("is_numeric", &is_numeric_f),
      std::pair <std::string, function_t> ("is_array", &is_array_f),
      std::pair <std::string, function_t> ("is_nil", &is_nil_f),
      std::pair <std::string, function_t> ("join", &join_f),
      std::pair <std::string, function_t> ("split", &split_f),
      std::pair <std::string, function_t> ("item", &item_f),
      std::pair <std::string, function_t> ("sub", &sub_f),
      std::pair <std::string, function_t> ("gsub", &gsub_f),
      std::pair <std::string, function_t> ("find", &find_f),
      std::pair <std::string, function_t> ("rfind", &rfind_f),
      std::pair <std::string, function_t> ("len", &len_f),
      std::pair <std::string, function_t> ("substr", &substr_f),
      std::pair <std::string, function_t> ("env", &env_f),
      std::pair <std::string, function_t> ("error", &error_f),
      std::pair <std::string, function_t> ("sprintf", &sprintf_f),
      std::pair <std::string, function_t> ("false", &false_f),
      std::pair <std::string, function_t> ("true", &true_f),
      std::pair <std::string, function_t> ("nil", &nil_f),
      std::pair <std::string, function_t> ("absolute_file_path", &absolute_file_path_f),
      std::pair <std::string, function_t> ("absolute_path", &absolute_path_f),
      std::pair <std::string, function_t> ("path", &path_f),
      std::pair <std::string, function_t> ("basename", &basename_f),
      std::pair <std::string, function_t> ("extension", &extension_f),
      std::pair <std::string, function_t> ("file_exists", &file_exists_f),
      std::pair <std::string, function_t> ("is_dir", &is_dir_f),
      std::pair <std::string, function_t> ("combine", &combine_f)
      // TODO: other functions
    };

    for (size_t i = 0; i < sizeof (function_defs) / sizeof (function_defs [0]); ++i) {
      ms_functions.insert (function_defs [i]);
    }
  }
};

//  perform the initialization:
static FunctionInitializer function_initializer;

static std::map <std::string, tl::Variant> ms_constants;

struct ConstantInitializer
{
  ConstantInitializer ()
  {
    std::pair <std::string, tl::Variant> consts_defs[] = {
      std::pair <std::string, tl::Variant> ("M_PI", tl::Variant (M_PI)),
      std::pair <std::string, tl::Variant> ("M_E", tl::Variant (M_E)),
    };

    for (size_t i = 0; i < sizeof (consts_defs) / sizeof (consts_defs [0]); ++i) {
      ms_constants.insert (consts_defs [i]);
    }
  }
};

//  perform the initialization:
static ConstantInitializer constant_initializer;

Eval::Eval ()
{
  // .. nothing yet ..
}

void 
Eval::set_var (const std::string &name, const tl::Variant &var)
{
  m_local_vars.insert (std::make_pair (name, tl::Variant ())).first->second = var;
}

void 
Eval::set_global_var (const std::string &name, const tl::Variant &var)
{
  m_global_vars.insert (std::make_pair (name, tl::Variant ())).first->second = var;
}

void 
Eval::define_function (const std::string &name, function_t function)
{
  m_local_functions.insert (std::make_pair (name, (function_t) 0)).first->second = function;
}

void 
Eval::define_global_function (const std::string &name, function_t function)
{
  m_global_functions.insert (std::make_pair (name, (function_t) 0)).first->second = function;
}

tl::Variant 
Eval::eval (const std::string &expr)
{
  tl::Extractor ex (expr.c_str ());

  tl::Variant a;
  eval_top (ex, &a);

  if (! ex.at_end ()) {
    throw EvalError (tl::translate ("Unexpected text after expression"), ex.skip ());
  }

  return a;
}

void
Eval::eval_top (tl::Extractor &ex, tl::Variant *v)
{
  do {
    eval_if (ex, v);
  } while (ex.test (";"));
}

void
Eval::eval_if (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor ex1 (ex);
  eval_boolean (ex, v);

  if (ex.test ("?")) {

    bool f = (v != 0 && v->to_bool ());

    eval_if (ex, f ? v : 0);

    if (! ex.test (":")) {
      throw tl::EvalError (tl::translate ("Expected ':'"), ex.skip ());
    }

    eval_if (ex, f ? 0 : v);
    
  }
}

void
Eval::eval_boolean (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor exa (ex);
  eval_conditional (ex, v);

  while (true) {

    tl::Extractor exb (ex);

    if (ex.test("||")) {

      eval_conditional (ex, (v != 0 && v->to_bool ()) ? 0 : v);

    } else if (ex.test ("&&")) {

      eval_conditional (ex, (v != 0 && v->to_bool ()) ? v : 0);

    } else {
      break;
    }

    exa = exb;

  }
}

void
Eval::eval_conditional (tl::Extractor &ex, tl::Variant *v)
{
  eval_shift (ex, v);

  while (true) {

    tl::Variant a;

    if (ex.test("<=")) {
      eval_shift (ex, v ? &a : 0);
      if (v) {
        *v = tl::Variant (*v < a || a == *v);
      }
    } else if (ex.test("<")) {
      eval_shift (ex, v ? &a : 0);
      if (v) {
        *v = tl::Variant (*v < a);
      }
    } else if (ex.test (">=")) {
      eval_shift (ex, v ? &a : 0);
      if (v) {
        *v = tl::Variant (a < *v || a == *v);
      }
    } else if (ex.test (">")) {
      eval_shift (ex, v ? &a : 0);
      if (v) {
        *v = tl::Variant (a < *v);
      }
    } else if (ex.test ("==")) {
      eval_shift (ex, v ? &a : 0);
      if (v) {
        *v = tl::Variant (a == *v);
      }
    } else if (ex.test ("!=")) {
      eval_shift (ex, v ? &a : 0);
      if (v) {
        *v = tl::Variant (! (a == *v));
      }
    } else {
      break;
    }

  }
}

void
Eval::eval_shift (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor exa (ex);
  eval_addsub (ex, v);

  while (true) {

    tl::Variant u;
    tl::Extractor exb (ex);

    if (ex.test("<<")) {

      eval_addsub (ex, v ? &u : 0);
      if (v) {
        *v = tl::Variant (to_long (exa, *v) << to_long (exb, u));
      }

    } else if (ex.test (">>")) {

      eval_addsub (ex, v ? &u : 0);
      if (v) {
        *v = tl::Variant (to_long (exa, *v) >> to_long (exb, u));
      }

    } else {
      break;
    }

  }
}

void
Eval::eval_addsub (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor exa (ex);
  eval_product (ex, v);

  while (true) {

    tl::Extractor exb (ex);
    tl::Variant u;

    if (ex.test("+")) {

      eval_product (ex, v ? &u : 0);

      if (v) {
        if (v->is_string () || u.is_string ()) {
          *v = tl::Variant (std::string (v->to_string ()) + u.to_string ());
        } else if (v->is_long () && u.is_long ()) {
          *v = tl::Variant (to_long (exa, *v) + to_long (exb, u));
        } else {
          *v = tl::Variant (to_double (exa, *v) + to_double (exb, u));
        }
      }

    } else if (ex.test ("-")) {

      eval_product (ex, v ? &u : 0);

      if (v) {
        if (v->is_long () && u.is_long ()) {
          *v = tl::Variant (to_long (exa, *v) - to_long (exb, u));
        } else {
          *v = tl::Variant (to_double (exa, *v) - to_double (exb, u));
        }
      }

    } else {
      break;
    }

  }
}

void
Eval::eval_product (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor exa (ex);
  eval_bitwise (ex, v);

  while (true) {

    tl::Extractor exb (ex);
    tl::Variant u;

    if (ex.test("*")) {

      eval_bitwise (ex, v ? &u : 0);

      if (v) {

        if (v->is_string ()) {

          long x = to_long (exb, u);
          if (x < 0) {
            throw EvalError (tl::translate ("Numeric argument of '*' operator with string must be positive"), ex.skip ());
          }

          std::string s;
          s.reserve (strlen (v->to_string ()) * size_t (x));
          while (x-- > 0) {
            s += v->to_string ();
          }

          *v = tl::Variant (s);

        } else if (u.is_string ()) {

          long x = to_long (exa, *v);
          if (x < 0) {
            throw EvalError (tl::translate ("Numeric argument of '*' operator with string must be positive"), ex.skip ());
          }

          std::string s;
          s.reserve (strlen (u.to_string ()) * size_t (x));
          while (x-- > 0) {
            s += u.to_string ();
          }

          *v = tl::Variant (s);

        } else if (v->is_long () && u.is_long ()) {
          *v = tl::Variant (to_long (exa, *v) * to_long (exb, u));
        } else {
          *v = tl::Variant (to_double (exa, *v) * to_double (exb, u));
        }

      }


    } else if (ex.test ("/")) {

      eval_bitwise (ex, v ? &u : 0);

      if (v) {
        if (v->is_long () && u.is_long ()) {
          *v = tl::Variant (to_long (exa, *v) / to_long (exb, u));
        } else {
          *v = tl::Variant (to_double (exa, *v) / to_double (exb, u));
        }
      }

    } else if (ex.test ("%")) {

      eval_bitwise (ex, v ? &u : 0);

      if (v) {
        *v = tl::Variant (to_long (exa, *v) % to_long (exb, u));
      }

    } else {
      break;
    }

  }
}

void
Eval::eval_bitwise (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor exa (ex);
  eval_unary (ex, v);

  while (true) {

    tl::Extractor exb (ex);
    tl::Variant u;

    if (exb.test("||")) {
      break; // not handled here
    } else if (exb.test("&&")) {
      break; // not handled here
    } else if (ex.test("|")) {

      eval_unary (ex, v ? &u : 0);
      if (v) {
        *v = tl::Variant (to_long (exa, *v) | to_long (exb, u));
      }

    } else if (ex.test ("&")) {

      eval_unary (ex, v ? &u : 0);
      if (v) {
        *v = tl::Variant (to_long (exa, *v) & to_long (exb, u));
      }

    } else if (ex.test ("^")) {

      eval_unary (ex, v ? &u : 0);
      if (v) {
        *v = tl::Variant (to_long (exa, *v) ^ to_long (exb, u));
      }

    } else {
      break;
    }

  }
}

void
Eval::eval_unary (tl::Extractor &ex, tl::Variant *v)
{
  tl::Extractor exa (ex);

  if (ex.test ("!")) {

    eval_unary (ex, v);
    if (v) {
      *v = tl::Variant (! v->to_bool ());
    }

  } else if (ex.test ("-")) {

    eval_unary (ex, v);
    if (v) {
      if (v->is_double ()) {
        *v = tl::Variant (-v->to_double ());
      } else if (v->is_long ()) {
        *v = tl::Variant (-v->to_long ());
      } else {
        throw EvalError (tl::translate ("Expected numeric value for '-' operator"), ex.skip ());
      }
    }
    
  } else if (ex.test ("~")) {

    eval_unary (ex, v);
    if (v) {
      *v = tl::Variant (~to_long (exa, *v));
    }
    
  } else {
    eval_atomic (ex, v);
  }
}

void
Eval::eval_atomic (tl::Extractor &ex, tl::Variant *v)
{
  double g = 0.0;
  std::string t;

  if (ex.test ("(")) {

    eval_top (ex, v);
    if (! ex.test (")")) {
      throw EvalError (tl::translate ("Expected closing bracket ')'"), ex.skip ());
    }

  } else if (ex.test ("[")) {

    std::vector <tl::Variant> vv;

    if (! ex.test ("]")) {

      do {

        if (v) {
          vv.push_back (tl::Variant ());
          eval_top (ex, &vv.back ());
        } else {
          eval_top (ex, 0);
        }

        if (ex.test ("]")) {
          break;
        } else if (! ex.test (",")) {
          throw EvalError (tl::translate ("Expected closing bracket ']'"), ex.skip ());
        }

      } while (true);

    }

    if (v) {
      *v = tl::Variant (vv.begin (), vv.end ());
    }

  } else if (ex.test ("0x")) {

    long n = 0;
    while (! ex.at_end ()) {
      if (isdigit (*ex) || (tolower (*ex) <= 'f' && tolower (*ex) >= 'a')) {
        if ((n * 16) / 16 != n) {
          throw EvalError (tl::translate ("Hexadecimal number overflow"), ex.skip ());
        }
        n *= 16;
        if (isdigit (*ex)) {
          n += long (*ex - '0');
        } else {
          n += long (tolower (*ex) - 'a' + 10);
        }
        ++ex;
      } else {
        break;
      }
    }

    if (v) {
      *v = tl::Variant (n);
    }

  } else if (ex.try_read (g)) {

    if (v) {
      *v = tl::Variant (g);
    }

  } else if (ex.try_read_quoted (t)) {

    if (v) {
      *v = tl::Variant (t);
    }

  } else if (ex.try_read_word (t)) {

    function_t function = 0;

    std::map <std::string, function_t>::const_iterator f;

    f = m_local_functions.find (t);
    if (f != m_local_functions.end ()) {
      function = f->second;
    }

    if (!function) {
      f = m_global_functions.find (t);
      if (f != m_global_functions.end ()) {
        function = f->second;
      }
    }

    if (!function) {
      f = ms_functions.find (t);
      if (f != ms_functions.end ()) {
        function = f->second;
      }
    }

    if (function) {

      //  a function: collect the parameters
      std::vector <tl::Variant> vv;

      tl::Extractor exb (ex);

      if (ex.test ("(")) {

        do {

          if (v) {
            vv.push_back (tl::Variant ());
            eval_top (ex, &vv.back ());
          } else {
            eval_top (ex, 0);
          }

          if (ex.test (")")) {
            break;
          } else if (! ex.test (",")) {
            throw EvalError (tl::translate ("Expected closing bracket ')'"), ex.skip ());
          }

        } while (true);

      }

      if (v) {
        *v = (*f->second) (this, exb, vv);
      }

    } else {

      std::map <std::string, tl::Variant>::const_iterator f;

      f = ms_constants.find (t);
      if (f != ms_constants.end ()) {

        if (v) {
          *v = f->second;
        }

      } else {

        f = m_local_vars.find (t);
        if (f != m_local_vars.end ()) {

          if (v) {
            *v = f->second;
          }

        } else {

          f = m_global_vars.find (t);

          if (f != m_global_vars.end ()) {
            if (v) {
              *v = f->second;
            }
          } else {
            throw EvalError (tl::translate ("Unknown variable") + " '" + t + "'", ex.skip ());
          }

        }

      }

    }

  } else {
    throw EvalError (tl::translate ("Expected constant, function or bracket expression"), ex.skip ());
  }
}

std::string 
Eval::interpolate (const std::string &str)
{
  std::ostringstream os;
  os.precision(8);

  tl::Extractor ex (str.c_str ());

  while (*ex) {
    if (*ex == '$') {
      ++ex;
      if (*ex == '$') {
        os << '$';
        ++ex;
      } else {
        tl::Variant v;
        eval_atomic (ex, &v);
        //  use default precision instead of full precision of to_string ..
        if (v.is_double ()) {
          os << v.to_double();
        } else {
          os << v.to_string ();
        }
      }
    } else {
      os << *ex;
      ++ex;
    }
  }

  return os.str ();
}

}

