
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_tlExpression
#define HDR_tlExpression

#include "config.h"

#include "tlException.h"
#include "tlVariant.h"

#include <map>
#include <vector>

namespace tl 
{

class Eval;

typedef tl::Variant (*function_t) (const tl::Eval *eval, tl::Extractor &ex, const std::vector <tl::Variant> &args);

/**
 *  @brief An exception thrown by the evaluation
 */
class KLAYOUT_DLL EvalError 
  : public tl::Exception
{
public:
  EvalError (const std::string &what, const char *where);
};

/**
 *  @brief Provides a simple formula parser and evaluator.
 */
class KLAYOUT_DLL Eval
{
public:
  /**
   *  @brief Create a new object for expression evaluation
   */
  Eval ();

  /**
   *  @brief virtual dtor to enable dynamic_cast on derived classes.
   */
  virtual ~Eval () { }

  /**
   *  @brief Define a global function for use within an expression
   */
  static void define_global_function (const std::string &name, function_t function);

  /**
   *  @brief Define a function for use within an expression
   */
  void define_function (const std::string &name, function_t function);

  /**
   *  @brief Define a global variable for use within an expression
   */
  static void set_global_var (const std::string &name, const tl::Variant &var);

  /**
   *  @brief Define a variable for use within an expression
   */
  void set_var (const std::string &name, const tl::Variant &var);

  /**
   *  @brief Evaluate the expression and return the result
   */
  tl::Variant eval (const std::string &expr);

  /**
   *  @brief Interpolate the string and return the result
   *
   *  Interpolation will replace all expressions of the form
   *  '$<atomic>' by their string value.
   */
  std::string interpolate (const std::string &str);

private:
  std::map <std::string, tl::Variant> m_local_vars;
  static std::map <std::string, tl::Variant> m_global_vars;
  std::map <std::string, function_t> m_local_functions;
  static std::map <std::string, function_t> m_global_functions;

  void eval_top (tl::Extractor &ex, tl::Variant *v);
  void eval_if (tl::Extractor &ex, tl::Variant *v);
  void eval_boolean (tl::Extractor &ex, tl::Variant *v);
  void eval_conditional (tl::Extractor &ex, tl::Variant *v);
  void eval_shift (tl::Extractor &ex, tl::Variant *v);
  void eval_addsub (tl::Extractor &ex, tl::Variant *v);
  void eval_product (tl::Extractor &ex, tl::Variant *v);
  void eval_bitwise (tl::Extractor &ex, tl::Variant *v);
  void eval_unary (tl::Extractor &ex, tl::Variant *v);
  void eval_atomic (tl::Extractor &ex, tl::Variant *v);
};

}

#endif

