
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <math.h>
#include <stdio.h>

#include "tlProgress.h"
#include "tlString.h"
#include "tlAssert.h"

namespace tl
{

static ProgressAdaptor *sp_adaptor = 0;

Progress::Progress (const std::string &desc, size_t max_count, size_t yield_interval)
  : m_desc (desc), m_title (desc),
    m_interval_count (0), 
    m_yield_interval (yield_interval), 
    m_last_value (-1.0), 
    m_cancelled (false)
{
  if (max_count) {
    m_format = "%.0f%%";
    m_res = 1.0;
    m_unit = double (max_count) / 100.0;
  } else {
    m_format = "%.0f";
    m_res = 1.0;
    m_unit = 1.0;
  }

  m_count = 0;

  if (sp_adaptor) {
    sp_adaptor->register_object (this);
  }
}

Progress::~Progress ()
{
  if (sp_adaptor) {
    sp_adaptor->unregister_object (this);
  }
}

double
Progress::value () const
{
  return double (m_count) / m_unit; 
}

std::string 
Progress::formatted_value () const
{
  return tl::sprintf (m_format.c_str (), value ());
}

void 
Progress::register_adaptor (ProgressAdaptor *pa)
{
  tl_assert (sp_adaptor == 0 || pa == 0);
  sp_adaptor = pa;
}

void 
Progress::signal_break ()
{
  m_cancelled = true;
}

void
Progress::set_desc (const std::string &d)
{
  if (sp_adaptor && d != m_desc) {

    m_desc = d;
    sp_adaptor->trigger (this);
    sp_adaptor->yield (this);

    if (m_cancelled) {
      m_cancelled = false;
      throw tl::BreakException ();
    }

  }

}

Progress &
Progress::set (size_t count) throw (tl::BreakException)
{
  m_count = count;
  if (++m_interval_count >= m_yield_interval) {

    double value = floor (double (m_count) / (m_unit * m_res) + 1e-6);
    if (value != m_last_value) {
      m_last_value = value;
      if (sp_adaptor) {
        sp_adaptor->trigger (this);
      }
    }

    m_interval_count = 0;
    if (sp_adaptor) {
      tl::Clock now = tl::Clock::current ();
      if ((now - m_last_yield).seconds () > 0.1) {
        m_last_yield = now;
        sp_adaptor->yield (this);
      }
    }

    if (m_cancelled) {
      m_cancelled = false;
      throw tl::BreakException ();
    }

  }

  return *this;
}

} // namespace tl


