
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

 
#ifndef HDR_utHead
#define HDR_utHead

#include <vector>
#include <iostream>
#include <sstream>
#include <string.h>

#include "tlString.h"
#include "tlException.h"
#include "dbStatic.h"
#include "rba.h"

#if defined(_MSC_VER)

extern void debug_print (const std::string &s);

#else

inline void debug_print (const std::string &s) 
{
  std::cerr << s << std::endl;
}

#endif

namespace ut {

struct Exception 
  : public tl::Exception
{
  Exception (const std::string &msg)
    : tl::Exception (msg)
  { }
};

struct TestBase 
{
  TestBase (const std::string &test);
  virtual ~TestBase () { }

  bool do_test () const
  {
    debug_print (std::string ("Running ") + m_test + " ..");
    try {
      execute ();
      return true;
    } catch (tl::Exception &ex) {
      debug_print (std::string ("*** Test ") + m_test + " failed:");
      debug_print (ex.msg ());
      return false;
    }
  }

private:
  virtual void execute () const throw (tl::Exception) = 0;
  std::string m_test;
};

struct Registrar
{
  static void reg (const ut::TestBase *t) 
  {
    if (! ms_instance) {
      ms_instance = new Registrar ();
    }
    ms_instance->m_tests.push_back (t);
  }

  static int do_tests () 
  {
    int failed = 0;
    if (ms_instance) {
      for (std::vector <const ut::TestBase *>::const_iterator 
           t = ms_instance->m_tests.begin (); 
           t != ms_instance->m_tests.end (); ++t) {
        if (! (*t)->do_test ()) {
          ++failed;
        }
      }
    }
    return failed;
  }
  
private:
  static Registrar *ms_instance;

  Registrar () : m_tests () { }

  std::vector <const ut::TestBase *> m_tests;
};

inline TestBase::TestBase (const std::string &test)
  : m_test (test)
{
  ut::Registrar::reg (this);
}

extern bool verbose ();

} // namespace ut

#define TEST(NAME) \
  static void TestImpl_##NAME##_execute () throw (tl::Exception); \
  namespace {\
struct TestImpl##NAME \
      : public ut::TestBase \
    { \
      TestImpl##NAME () : TestBase (__FILE__ ":" #NAME) { } \
      virtual void execute () const throw (tl::Exception) { TestImpl_##NAME##_execute (); }; \
    }; \
    static TestImpl##NAME TestImpl_Inst##NAME; \
  } \
  static void TestImpl_##NAME##_execute () throw (tl::Exception)

#define EXPECT_EQ(WHAT,EQUALS) \
  if (!((WHAT) == (EQUALS))) { \
    std::stringstream sstr; \
    sstr << __LINE__ << ": " << #WHAT << " does not equal " << #EQUALS << " (actual value is " << (WHAT) << ")" << std::ends; \
    throw ut::Exception (sstr.str ()); \
  } 

#define EXPECT_NE(WHAT,EQUALS) \
  if (!((WHAT) != (EQUALS))) { \
    std::stringstream sstr; \
    sstr << __LINE__ << ": " << #WHAT << " does equal " << #EQUALS << " (actual value is " << (WHAT) << ")" << std::ends; \
    throw ut::Exception (sstr.str ()); \
  } 

#define EXPECT(WHAT) \
  if (!(WHAT)) { throw ut::Exception (__LINE__ << ": " << #WHAT " is not true"); } 

#define FAIL_ARG(MSG,WHAT) \
  { \
    std::stringstream sstr; \
    sstr << __LINE__ << ": " << MSG << ", value is " << (WHAT) << std::ends; \
    throw ut::Exception (sstr.str ()); \
  } 

#if !defined(_MSC_VER)

//  On Linux/MinGW every test is an inidividual executable.

namespace ut 
{

ut::Registrar *ut::Registrar::ms_instance = 0;

static bool _verbose_flag = false;

bool verbose ()
{
  return _verbose_flag;
}

} // namespace ut

rba::RubyInterpreter *ruby_interp_instance = 0;

int 
main (int argc, const char *argv[]) 
{
  //  With Ruby, "fclose" is redirected to the Ruby-DLL. Because of that, Ruby needs to be initialized before fclose can be used ...
  rba::RubyInterpreter::initialize ();
  rba::RubyInterpreter rbinterp;
  ruby_interp_instance = &rbinterp;

  for (int a = 1; a < argc; ++a) {
    if (! strcmp (argv [a], "-v")) {
      ut::_verbose_flag = true;
    } else if (! strcmp (argv [a], "-e")) {
      db::set_editable_mode (true);
    }
  }
  int failed = ut::Registrar::do_tests ();
  if (failed > 0) {
    debug_print (std::string ("Summary: ") + tl::to_string (failed) + " test(s) failed");
  } else {
    debug_print ("All tests passed.");
  }
  return failed;
}

#endif

#endif

