
use_minitest = (RUBY_VERSION =~ /^1\.9/)

# in MSVC environment:
if ENV["RUBY"] 
  $:.push("#{ENV["RUBY"]}/lib/ruby/#{RUBY_VERSION}")
end

if use_minitest
  require "minitest/unit"
  test_base = MiniTest::Unit::TestCase
else
  require "test/unit"
  test_base = Test::Unit::TestCase
end

# Set this to true to disable some tests involving exceptions
$leak_check = false

# extend A
class RBA::A 

  def initialize
    @offset = nil
  end
  def s( o )
    @offset = o
  end
  def g
    return @offset
  end
  def m
    return @offset+a1
  end

private
  @offset

end

class BasicTest < test_base

  def test_00

    GC.start 
    
    # all references of A are released now:
    ac0 = RBA::A::a0
    assert_equal( ac0, 0 )

    a = RBA::A.new_a(100)
    assert_equal( RBA::A::a0, ac0 + 1 )

    a = nil
    GC.start
    assert_equal( RBA::A::a0, ac0 )

    a = RBA::A.new
    assert_equal( RBA::A::a0, ac0 )  # delayed instantiation of detached objects - A is actually created if it is used first
    a.a2   # just check, if it can be called
    assert_equal( RBA::A::a0, ac0 + 1 )

    # open question: with ruby 1.8, aa is not deleted if the assert_equal is missing. Why?
    # maybe the GC does not like to be called that frequently?
    aa = a.dup
    assert_equal( RBA::A::a0, ac0 + 2 )

    aa = nil
    GC.start 

    if RUBY_VERSION == "1.9.1"
      # GC works differently in 1.9.1(?) - but no leak (verified with valgrind)
      ac0 = RBA::A::a0
    else
      assert_equal( RBA::A::a0, ac0 + 1 )
    end

    a = nil
    GC.start

    if RUBY_VERSION == "1.9.1"
      # GC works differently in 1.9.1(?) - but no leak (verified with valgrind)
      ac0 = RBA::A::a0
    else
      assert_equal( RBA::A::a0, ac0 )
    end

    a = RBA::A.new
    assert_equal( RBA::A::a0, ac0 )  # delayed instantiation of detached objects - A is actually created if it is used first
    a.a2   # just check, if it can be called
    assert_equal( RBA::A::a0, ac0 + 1 )

    assert_equal( a.a1, 17 )
    a.a5 -5
    assert_equal( a.a1, -5 )
    a.a5 0x7fffffff
    assert_equal( a.a1, 0x7fffffff )
    a.a5 -0x80000000
    assert_equal( a.a1, -0x80000000 )

    assert_equal( a.a3("a"), 1 )
    assert_equal( a.a3("ab"), 2 )
    assert_equal( a.a3(""), 0 )

    assert_equal( a.a4([1]), 1.0 )
    assert_equal( a.a4([1, 125e-3]), 0.125 )
    assert_equal( a.a4([5, 1, -1.25]), -1.25 )

    arr = []
    a.a6 { |d| arr.push(d) } 
    assert_equal(arr, [5, 1, -1.25])

    arr = []
    a.a7 { |d| arr.push(d) } 
    assert_equal(arr, [5, 1, -1.25])

    arr = []
    a.a8 { |d| arr.push(d) } 
    assert_equal(arr, [5, 1, -1.25])

    a.destroy
    assert_equal( RBA::A::a0, ac0 )

    if !$leak_check 

      error_caught = false
      begin 
        a.a2  # object has been destroyed already
      rescue
        error_caught = true
      end
      assert_equal( error_caught, true )

      error_caught = false
      begin 
        a.destroy  # object has been destroyed already
      rescue
        error_caught = true
      end
      assert_equal( error_caught, true )

    end

    assert_equal( RBA::A::a0, ac0 )
    a = RBA::A::new_a( 55 )
    assert_equal( RBA::A::a0, ac0 + 1 )
    assert_equal( a.a1, 55 )
    a.destroy
    assert_equal( RBA::A::a0, ac0 )

  end

  def test_12

    a1 = RBA::A.new
    a1.a5( -15 )
    a2 = a1
    a3 = a2.dup

    assert_equal( a1.a1, -15 )
    assert_equal( a2.a1, -15 )
    assert_equal( a3.a1, -15 )

    a1.a5( 11 )
    a3.a5( -11 )
    
    assert_equal( a1.a1, 11 )
    assert_equal( a2.a1, 11 )
    assert_equal( a3.a1, -11 )

  end

  def test_13

    b = RBA::B.new

    if !$leak_check 

      err_caught = false
      begin
        b.b7.a1 # cannot call non-const method on const reference
      rescue 
        err_caught = true
      end
      assert_equal( err_caught, true )

    end
      
    b.b7.a2 

    assert_equal( b.b1, 5 )
    assert_equal( b.b2, "" )
    b.b5( "xyz" )
    assert_equal( b.b2, "xyz" )
    assert_equal( b.b5a, "xyz" )
    b.b5b( "yx", "zz" )
    assert_equal( b.b2, "yxzz" )
    assert_equal( b.b5a, "yxzz" )

    arr = []

    err_caught = false

    if !$leak_check 

      begin
        b.b10 { |a| arr.push(a.a1) }  # b10 is a const iterator - cannot call a1 on it
      rescue 
        err_caught = true
      end
      assert_equal( err_caught, true )
      assert_equal(arr, [])

    end
   
    err_caught = false

    if !$leak_check 

      begin
        b.b10p { |a| arr.push(a.a1) }  # b10p is a const iterator - cannot call a1 on it
      rescue 
        err_caught = true
      end
      assert_equal( err_caught, true )
      assert_equal(arr, [])

    end
   
    arr = []
    b.b10 { |a| arr.push(a.dup.a1) } 
    assert_equal(arr, [100, 121, 144])

    arr = []
    b.b10 { |a| arr.push(a.a1c) } 
    assert_equal(arr, [100, 121, 144])

    arr = []
    b.b10p { |a| arr.push(a.dup.a1) } 
    assert_equal(arr, [100, 121, 144])

    arr = []
    b.b10p { |a| arr.push(a.a1c) } 
    assert_equal(arr, [100, 121, 144])

    arr = []
    b.b11 { |a| arr.push(a.a1) } 
    assert_equal(arr, [100, 121, 144])

    arr = []
    b.b12 { |a| arr.push(a.a1) } 
    assert_equal(arr, [7100, 7121, 7144, 7169])

    aarr = b.b16a 
    arr = []
    aarr.each { |a| arr.push(a.a1) } 
    assert_equal(arr, [100, 121, 144])

    aarr = b.b16b 
    arr = []
    aarr.each { |a| arr.push(a.a1) } 
    assert_equal(arr, [100, 121, 144])

    aarr = b.b16c 
    arr = []
    aarr.each { |a| arr.push(a.a1) } 
    assert_equal(arr, [100, 121, 144])

    b.b17a( [ RBA::A.new_a( 101 ), RBA::A.new_a( -122 ) ] )
    arr = []
    b.b11 { |a| arr.push(a.a1) } 
    assert_equal(arr, [101, -122])

    b.b17a( [] )
    arr = []
    b.b11 { |a| arr.push(a.a1) } 
    assert_equal(arr, [])

    b.b17b( [ RBA::A.new_a( 102 ), RBA::A.new_a( -123 ) ] )
    arr = []
    b.b11 { |a| arr.push(a.a1) } 
    assert_equal(arr, [102, -123])

    b.b17c( [ RBA::A.new_a( 100 ), RBA::A.new_a( 121 ), RBA::A.new_a( 144 ) ] )
    arr = []
    b.b11 { |a| arr.push(a.a1) } 
    assert_equal(arr, [100, 121, 144])

    if !$leak_check 

      arr = []
      begin
        b.b13 { |a| arr.push(a.a1) } 
      rescue 
        err_caught = true
      end
      assert_equal( err_caught, true )
      assert_equal(arr, [])

    end

    arr = []
    b.b13 { |a| arr.push(a.a1c) } 
    assert_equal(arr, [-3100, -3121])

    arr = []
    b.b18 { |a| arr.push(a.a1) } 
    assert_equal(arr, [100, 121, 144])

  end

  def test_14

    a = RBA::A.new
    a.a5( 22 )

    b = RBA::B.new
    assert_equal( b.b3( a ), 22 )
    assert_equal( b.b4( a ), "b4_result: 22" )
    a.a5( -6 )
    assert_equal( b.b3( a ), -6 )
    assert_equal( b.b4( a ), "b4_result: -6" )
    assert_equal( b.b4( RBA::A.new ), "b4_result: 17" )

  end

  def test_15

    a = RBA::A_NC.new
    a.a5( 22 )

    b = RBA::B.new
    assert_equal( b.b3( a ), 22 )
    assert_equal( b.b4( a ), "b4_result: 22" )
    a.a5( -6 )
    assert_equal( b.b3( a ), -6 )
    assert_equal( b.b4( a ), "b4_result: -6" )
    assert_equal( b.b4( RBA::A_NC.new ), "b4_result: 17" )

  end

  class X < Data
  end
  class Y < Object
  end

  def test_16

    if $leak_check 
      return
    end

    # Test, if this throws an error (object of class X passed to A argument):
    begin
      b = RBA::B.new
      assert_equal( b.b4( X.new ), "b4_result: -6" )
      assert_equal( false, true )  # this must never hit
    rescue
      assert_equal( $!.to_s(), "allocator undefined for BasicTest::X" );
    end
  
    # Test, if this throws an error (object of class X passed to A argument):
    begin
      b = RBA::B.new
      assert_equal( b.b4( Y.new ), "b4_result: -6" )
      assert_equal( false, true )  # this must never hit
    rescue
      assert_equal( $!.to_s(), "wrong argument type BasicTest::Y (expected Data)" );
    end
  
    # Test, if this throws an error (object of class X passed to A argument):
    begin
      b = RBA::B.new
      bb = RBA::B.new
      assert_equal( b.b4( bb ), "b4_result: -6" )
      assert_equal( false, true )  # this must never hit
    rescue
      assert_equal( $!.to_s(), "Unexpected object type (expected argument of class A)" );
    end
  
  end

  def test_17

    # test copies of objects being returned

    b = RBA::B.new

    a = b.b6( 1971 );
    assert_equal( a.a1, 1971 );
    assert_equal( b.b9( a ), 1971 );

    aa = b.b6( -61 );
    assert_equal( b.b9cref( aa ), -61 );
    assert_equal( a.a1, 1971 );
    assert_equal( b.b9( a ), 1971 );
    assert_equal( aa.a1, -61 );
    assert_equal( b.b9( aa ), -61 );

    aa.a5 98;
    a.a5 100;
    
    assert_equal( a.a1, 100 );
    assert_equal( b.b9( a ), 100 );
    assert_equal( aa.a1, 98 );
    assert_equal( b.b9( aa ), 98 );

  end

  def test_18

    # Test references to objects (returned by b.b7)

    b = RBA::B.new
    b.b8( 77 )
    assert_equal( b.b7.a1c, 77 );

    b.b8cref( 79 )
    assert_equal( b.b7.a1c, 79 );

    aref = b.b7
    err_caught = false

    if !$leak_check 

      begin 
        x = aref.a1 # cannot call non-const method on const reference (as delivered by b7)
      rescue
        err_caught = true
      end
      assert_equal( err_caught, true )
      assert_equal( aref.a1c, 79 );

    end

    b.b8( -1 )
    assert_equal( aref.a1c, -1 );

  end

  class C_IMP1 < RBA::C 
    def f(s)
      return 615
    end
  end

  class C_IMP2 < RBA::C 
    def f(s)
      return s.size
    end
  end

  class C_IMP3 < RBA::C 
  end

  def test_19

    c1 = C_IMP1.new
    assert_equal( c1.g("x"), 615 );

    c2 = C_IMP2.new
    assert_equal( c2.g("x"), 1 );
    assert_equal( c2.g(""), 0 );
    assert_equal( c2.g("abc"), 3 );
    assert_equal( c1.g("x"), 615 );

    c3 = C_IMP3.new
    assert_equal( c3.g("x"), 1977 );

    assert_equal( RBA::C.s1, 4451 );
    RBA::C.s2clr
    RBA::C.s2( 7.0 )
    assert_equal( RBA::C.s3( 5.5 ), "5.500" );

    arr = []
    RBA::C.each { |i| arr.push i }
    assert_equal( arr, [ 0, 1, 2, 3, 4, 5, 6 ] )

    assert_equal( C_IMP1.s1, 4451 );
    C_IMP1.s2( 1.0 )
    assert_equal( C_IMP1.s3( 1.5 ), "1.500" );

    arr = []
    C_IMP1.each { |i| arr.push i }
    assert_equal( arr, [ 0, 1, 2, 3, 4, 5, 6, 0 ] )

    assert_equal( C_IMP2.s1, 4451 );
    C_IMP2.s2( 2.0 )
    assert_equal( C_IMP2.s3( -1.5 ), "-1.500" );

    arr = []
    C_IMP2.each { |i| arr.push i }
    assert_equal( arr, [ 0, 1, 2, 3, 4, 5, 6, 0, 0, 1 ] )

  end

  def test_20

    b = RBA::B.new

    a1 = b.b14a( true )
    a2 = b.b14b
    assert_equal( a1.a1, 17 )
    assert_equal( a2.a1, 17 )
    a1.a5( 761 )
    assert_equal( a1.a1, 761 )
    assert_equal( a2.a1, 761 )

    a1 = b.b14a( false )
    assert_equal( a1, nil )
    
    assert_equal( b.b15( b.b14b ), true )
    assert_equal( b.b15( b.b14a( false ) ), false )
    assert_equal( b.b15( nil ), false )

  end

  def test_21

    # test client data binding to C++ objects 
    
    b = RBA::B.new
    
    b.b14b.s( 117 )
    assert_equal( b.b14b.g, 117 )

    n = 0
    b.b10_nc { |a| a.s( n ); n += 1 } 

    arr = []
    b.b10 { |a| arr.push( a.g ) }
    assert_equal( arr, [ 0, 1, 2 ] )

    arr = []
    b.b10p { |a| arr.push( a.g ) }
    assert_equal( arr, [ 0, 1, 2 ] )

  end

  def test_22

    # test client data binding to C++ objects 
    
    b = RBA::B.new
    
    assert_equal( b.b20a( 5.0 ), false )
    assert_equal( b.b20a( nil ), true )
    assert_equal( b.b20a( 1 ), false )
    assert_equal( b.b20a( "hallo" ), false )
    assert_equal( b.b20a( false ), false )
    assert_equal( b.b20a( true ), false )
    assert_equal( b.b20a( 10000000000000000 ), false )
    assert_equal( b.b20b( 5.0 ), true )
    assert_equal( b.b20b( nil ), false )
    assert_equal( b.b20b( 1 ), false )
    assert_equal( b.b20b( "hallo" ), false )
    assert_equal( b.b20b( false ), false )
    assert_equal( b.b20b( true ), false )
    # on 64 bit platforms this fits into a long value, therefore this test returns false:
    # assert_equal( b.b20b( 10000000000000000 ), false )
    assert_equal( b.b20c( 5.0 ), false )
    assert_equal( b.b20c( nil ), false )
    assert_equal( b.b20c( 1 ), true )
    assert_equal( b.b20c( "hallo" ), false )
    assert_equal( b.b20c( false ), true )
    assert_equal( b.b20c( true ), true )
    # on 64 bit platforms this fits into a long value, therefore this test returns true:
    # assert_equal( b.b20c( 10000000000000000 ), false )
    assert_equal( b.b20d( 5.0 ), false )
    assert_equal( b.b20d( nil ), false )
    assert_equal( b.b20d( 1 ), false )
    assert_equal( b.b20d( "hallo" ), true )
    assert_equal( b.b20d( false ), false )
    assert_equal( b.b20d( true ), false )
    assert_equal( b.b20d( 10000000000000000 ), false )

    assert_equal( b.b21a( 50 ), "50" )
    assert_equal( b.b21a( true ), "1" )
    assert_equal( b.b21a( false ), "0" )
    assert_equal( b.b21a( "hallo" ), "hallo" )
    assert_equal( b.b21a( 5.5 ), "5.5" )
    assert_equal( b.b21a( nil ), "nil" )

    assert_equal( b.b21b( 50 ), 50.0 )
    assert_equal( b.b21b( true ), 1.0 )
    assert_equal( b.b21b( false ), 0.0 )
    assert_equal( b.b21b( 5.5 ), 5.5 )
    assert_equal( b.b21b( nil ), 0.0 )

    assert_equal( b.b21c( 50 ), 50 )
    assert_equal( b.b21c( true ), 1 )
    assert_equal( b.b21c( false ), 0 )
    assert_equal( b.b21c( 5.5 ), 5 )
    assert_equal( b.b21c( nil ), 0 )

    assert_equal( b.b22a( [ 1, "hallo", 5.5 ] ), 3 ) 
    assert_equal( b.b23a, [ 1, "hallo", 5.5 ] ) 
    a = [] 
    b.b24 { |x| a.push( x ) } 
    assert_equal( a, [ 1, "hallo", 5.5 ] ) 
    assert_equal( b.b22c, 5.5 )
    assert_equal( b.b22d, 5.5 )
    assert_equal( b.b22a( [ 1, "hallo" ] ), 2 ) 
    assert_equal( b.b23b, [ 1, "hallo" ] ) 
    assert_equal( b.b23d, [ 1, "hallo" ] ) 
    assert_equal( b.b22c, "hallo" )
    assert_equal( b.b22d, "hallo" )
    assert_equal( b.b22a( [ ] ), 0 ) 
    assert_equal( b.b23c, [ ] ) 
    a = [] 
    b.b24 { |x| a.push( x ) } 
    assert_equal( a, [ ] ) 
    assert_equal( b.b22b, nil )
    assert_equal( b.b22c, nil )
    assert_equal( b.b22d, nil )
    assert_equal( b.b22a( [ [ 1, "hallo" ], [ 10, 17, 20 ] ] ), 2 ) 
    assert_equal( b.b23a, [ [ 1, "hallo" ], [ 10, 17, 20 ] ] ) 
    a = [] 
    b.b24 { |x| a.push( x ) } 
    assert_equal( a, [ [ 1, "hallo" ], [ 10, 17, 20 ] ] ) 

  end

  def test_23

    b = RBA::B::new
    a = RBA::A::new

    assert_equal( b.bx, 17 )
    assert_equal( b.b30, 17 )
    assert_equal( b.bx( 5 ), "xz" )
    assert_equal( b.by( 5 ), "xz" )
    assert_equal( b.b31( 6 ), "xz" )
    assert_equal( b.b33( a ), "aref" )
    assert_equal( b.bx( a ), "aref" )
    assert_equal( b.bx( "a", 15 ), 20.5 )
    assert_equal( b.b32( "b", 25 ), 20.5 )

    err_caught = false
    begin 
      # invalid number of arguments
      assert_equal( b.by, "xz" )
    rescue
      err_caught = true
    end
    assert_equal( err_caught, true )

    err_caught = false
    begin 
      # invalid number of arguments
      assert_equal( b.bx( 1, 5, 7 ), "xz" )
    rescue
      err_caught = true
    end
    assert_equal( err_caught, true )

    b.destroy
    a.destroy

  end

    class RBA::E
      def m 
        @m
      end
      def m=(x)
        @m = x
      end
      @m = nil
    end

  def test_24

    n0 = 0
    n1 = 0
    n2 = "" 

    # Events
    e = RBA::E.new
    e.m = 100

    e.s1 # no event installed
    assert_equal( 0, n0 )
    e.s2
    assert_equal( 0, n1 )
    e.s3
    assert_equal( "", n2 )

    assert_equal( 100, e.m )
    e.e0 { n0 += 1 }
    e.e1 { |x| n1 += x.m }
    e.e2 { |i,s| n2 += i.to_s + s; } 

    e.s1
    assert_equal( 1, n0 )
    e.s1
    assert_equal( 2, n0 )

    # using lambda
    p = lambda { n0 += 2 }
    e.e0(&p)
    e.s1
    assert_equal( 4, n0 )

    e.s2
    assert_equal( 100, n1 )
    e.m = 1
    e.s2
    assert_equal( 101, n1 )

    e.s3
    assert_equal( "18hallo", n2 )
    e.s3
    assert_equal( "18hallo18hallo", n2 )

    e.e0 { raise "X" }
    error_caught = false
    begin 
      e.s1
    rescue
      error_caught = true
    end
    assert_equal( error_caught, true )

  end

end

# In the test environment, we cannot make sure that we destroy the ruby interpreter before the RBA
# environment is shut down. Therefore we must release all RBA objects by explicitly calling the GC
# and start the test suite manually.

if !use_minitest

  # Ruby 1.8.7 with test/unit:
  require 'test/unit/ui/console/testrunner'
  result = Test::Unit::UI::Console::TestRunner.run(BasicTest)    

  GC.start

  if (result.failure_count + result.error_count) > 0
    raise "Tests failed: #{result.to_s}"
  end

else

  mt = MiniTest::Unit.new
  
  # Ruby 1.9.1 with minitest/unit:
  # Make sure that _test_FIRST is run as the first one - to have a zero instance count of "A"
  mt.run(["-n", "/^test_00$/"])
  mt.run(["-n", "/^test_[1-9]\\d/"])

  GC.start

  if mt.errors + mt.failures > 0
    raise "Tests failed (#{mt.errors} Errors, #{mt.failures} Failures)"
  end

end

